************************************************************************
*                                                                      *
*                           This is PARNI                              *
*                                                                      *
*          for importance sampling and/or denstiy estimation.          *
*                                                                      *
* author: Andreas van Hameren <Andre.HamerenREMOVETHIS@ifj.edu.pl>     *
*   date: 08-10-2009                                                   *
*                                                                      *
* Please cite                                                          *
*    A. van Hameren, 
*    Acta Phys.Polon.B40:259-272(2009), arXiv:0710.2448 [hep-ph]       *
* in publications with results obtained with the help of this program. *
************************************************************************
*
* For full examples, have a look at the file examples.f
*
* Rough example density estimation:
*
*      call avh_parni_init(ID ,itask ,ndimen,nbatch,nchmax)
*      do iev=1,nev
*        call your_data_generator(wdat,xdat)
*        call avh_parni_adapt(ID ,wdat,xdat)
*      enddo
*      call avh_parni_marg(ID ,nunit,0)
*
*
* Rough example importance sampling:
*
*      call avh_parni_init(ID ,itask ,ndimen,nbatch,nchmax)
*      do iev=1,nev
*        call avh_parni_generate(ID ,xx)
*        value = your_integrand(xx)
*        call avh_parni_calcwght(ID ,ww ,xx)
*        value = value*ww
*        call avh_parni_adaptone(ID ,value)
*        call your_weight_analysis(value)
*      enddo
*      call avh_parni_marg(ID ,nunit,0)
*      call avh_parni_result(ID)
*
* The user needs to write the contents of subroutine avh_parni_random
*                                                                      *
************************************************************************
*
* Include-files:
*
**** avh_parni.h *******************************************************
** The maximum size of an array containing all channels of all copies:
*      integer   avh_parni_size
*      parameter(avh_parni_size=10000)
** The maximal number of copies:
*      integer   avh_parni_ncopy
*      parameter(avh_parni_ncopy=10)
** The maximal number of dimension:
*      integer   avh_parni_ndim
*      parameter(avh_parni_ndim=10)
************************************************************************
**** avh_parni_priv.h **************************************************
*      include 'avh_parni.h'
*      integer   ISIZE
*      parameter(ISIZE=avh_parni_size)
*      integer   ISIZE_ncopy
*      parameter(ISIZE_ncopy=avh_parni_ncopy)
*      integer   ISIZE_ndim
*      parameter(ISIZE_ndim=avh_parni_ndim)
*      integer   ISIZE_nstp
*      parameter(ISIZE_nstp=100)
*      double precision weight(ISIZE)
*     &                ,sumwt0(ISIZE)
*     &                ,sumwt1(ISIZE)
*     &                ,sumwt2(ISIZE)
*     &                ,totwt0(ISIZE_ncopy)
*     &                ,totwt1(ISIZE_ncopy)
*     &                ,totwt2(ISIZE_ncopy)
*      common/avh_parni_dbl/ weight,sumwt0,sumwt1,sumwt2
*     &                            ,totwt0,totwt1,totwt2
*      integer mother(ISIZE)
*     &       ,daught(ISIZE)
*     &       ,cutdim(ISIZE)
*     &       ,channl(ISIZE)
*     &       ,nodech(ISIZE)
*     &       ,pointr(ISIZE)
*     &       ,offset(ISIZE_ncopy)
*     &       ,copies(ISIZE_ncopy)
*     &       ,nchann(ISIZE_ncopy)
*     &       ,resrvd(ISIZE_ncopy)
*     &       ,nchmax(ISIZE_ncopy)
*     &       ,ndimen(ISIZE_ncopy)
*     &       ,nbatch(ISIZE_ncopy)
*     &       ,nsofar(ISIZE_ncopy)
*     &       ,iitask(ISIZE_ncopy)
*     &       ,ncopy
*      common/avh_parni_int/ mother,daught,cutdim,channl,nodech,pointr
*     &,offset,copies,nchann,resrvd,nchmax,ndimen,nbatch,nsofar,iitask
*     &,ncopy
*      logical start_sort(ISIZE_ncopy),do_adapt(ISIZE_ncopy)
*      common/avh_parni_log/ start_sort,do_adapt
*      double precision xstr(ISIZE_ncopy,ISIZE_ndim),wstr(ISIZE_ncopy)
*      common/avh_parni_str/ xstr,wstr
************************************************************************


      subroutine avh_parni_random(rho,nn)
*  ********************************************************************
*  * return a double precision array "rho" of "nn" random numbers
*  * between 0 and 1
*  ********************************************************************
      implicit none
      integer nn
      double precision rho(nn)
ccostas [
      integer I
      double precision rnmy
ccostas ]
c      write(6,*) 'ERROR in avh_parni_random: '
c     &          ,'please implement random number generator.'
c      stop
* Out-comment the 3 lines above and put a line like for example:
c     call rangen(rho,nn)
ccostas [
      do I=1,nn
       rho(I)=rnmy(0)
      enddo
ccostas ]
      end




      subroutine avh_parni_init(ID ,task ,ndimen_in,nbatch_in,nchmax_in)
*  *********************************************************************
*  * All arguments are input.
*  * "ID" is the label of the 'copy' of the algorithm. It should be
*  *    a number between 0 and avh_parni_ncopy+1.
*  * "task" sets the kind of optimizations used internally. For the
*  *    moment the options are 1,2,3 ,6,7,8 ,11,12,13 ,16,17,18.
*  *    See extended explanation below. In case of importance sampling
*  *    task=11 seems to be the best choice. In case of density
*  *    estimation, task MUST be set to 11 or 16.
*  * "ndimen_in" is the number of dimensions of the hypercube.
*  * "nbatch_in" is the size of the batches of data after which an
*  *    optimization step is performed. The square-root of the number
*  *    of data points that is going to be used for the optimization
*  *    seems to be a good choice. The larger this number, the more
*  *    conservative the algorithm, i.e. the smaller the number of
*  *    channels.
*  * "nchmax_in" is the maximal number of channels. If put to 0, the
*  *    number of channels keeps growing. If larger than 0, boxes
*  *    originating from the same parent with the smallest weight are
*  *    merged if nchmax is exeeded after splitting boxes. This process
*  *    continues: boxes are splitted and others merged and the number 
*  *    of channels is kept around nchmax.
*  *
*  * Let f be the integrand and denote
*  *
*  *   < f > = integral( f(x), whole hypercube )
*  *
*  * Furthermore, the density built by parni is given by
*  *
*  *   p(x) = sum_{i=1}^{nchannel} w_i*t_i(x)/v_i
*  *
*  * where t_i is the characteristic function of the hyperbox
*  * corresponding to channel i, and v_i is the volume of that box.
*  * For the different tasks, the weights w_i are optimized as follows:
*  *
*  * task=1,6,11,16: w_i = < f*t_i >                  (histogram-like)
*  * task=2,7,12,17: w_i = sqrt( < f^2*t_i > * v_i )  (minimal variance)
*  * task=3,8,13,18: w_i = < f*t_i >^a * sqrt(< f^2*t_i > * v_i)^(1-a)
*  *
*  * divided by the normalisation such that their sum is equal to 1.
*  * The number  a  for task=3,8,13,18 is set to a=0.5 (look for the
*  * string par0 in this file).
*  * Channels are dissected if they have the largest w_i or w_i*v_i
*  *
*  * task=1,2,3,11,12,13: dissect channel with largest w_i
*  * task=6,7,8,16,17,18: dissect channel with largest w_i*v_i
*  *
*  * The integrals are estimated as follows: (up to an overall
*  * normalization)
*  *
*  * task<10:  < f*t_i > = sum( f(xj)*t_i(xj) )/sum( t_i(xj) )
*  * task>10:  < f*t_i > = sum( f(xj)*t_i(xj) )*v_i/w_i
*  *
*  * and the same for f^2. The sum is over all events. For the
*  * adaptation, this means the following. Let W(xj) be the numbers
*  * going into  avh_parni_adapt( ID, W(xj), xj ) , then
*  * 
*  * task<1 :  w_i <-- sum( W(xj)*t_i(xj)*w_i/v_i )/sum( t_i(xj) )
*  * task>11:  w_i <-- sum( W(xj)*t_i(xj) )
*  *
*  * summary
*  *
*  * task=1  : histogram-like  , largest w_i    , sum( t_i(xj) )
*  * task=2  : minimal variance, largest w_i    , sum( t_i(xj) )
*  * task=3  : mix             , largest w_i    , sum( t_i(xj) )
*  * task=6  : histogram-like  , largest w_i*v_i, sum( t_i(xj) )
*  * task=7  : minimal variance, largest w_i*v_i, sum( t_i(xj) )
*  * task=8  : mix             , largest w_i*v_i, sum( t_i(xj) )
*  * task=11 : histogram-like  , largest w_i    , w_i/v_i
*  * task=12 : minimal variance, largest w_i    , w_i/v_i
*  * task=13 : mix             , largest w_i    , w_i/v_i
*  * task=16 : histogram-like  , largest w_i*v_i, w_i/v_i
*  * task=17 : minimal variance, largest w_i*v_i, w_i/v_i
*  * task=18 : mix             , largest w_i*v_i, w_i/v_i
*  *
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID,task,ndimen_in,nbatch_in,nchmax_in
     &,obj,oo,node
      logical firstcall
      data firstcall/.true./
      save firstcall
*
      if (firstcall) then
        firstcall = .false.
        write(*,'(a36,a36)') '####################################'
     &                      ,'####################################'
        write(*,'(a36,a36)') '#                                   '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '#                         You are us'
     &                      ,'ing PARNI                          #'
        write(*,'(a36,a36)') '#                                   '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '#          for importance sampling a'
     &                      ,'nd/or density estimation.          #'
        write(*,'(a36,a36)') '#                                   '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '# author: Andreas van Hameren <hamer'
     &                      ,'enREMOVETHIS@ifj.edu.pl>           #'
        write(*,'(a36,a36)') '#   date: 08-10-2009                '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '#                                   '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '# Please cite                       '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '#    A. van Hameren,                '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '#    Acta Phys.Polon.B40:259-272(200'
     &                      ,'9), arXiv:0710.2448 [hep-ph]       #'
        write(*,'(a36,a36)') '# in publications with results obtai'
     &                      ,'ned with the help of this program. #'
        write(*,'(a36,a36)') '#                                   '
     &                      ,'                                   #'
        write(*,'(a36,a36)') '####################################'
     &                      ,'####################################'
      endif
*
      if (nchmax_in.eq.0) then
        call avh_parni_reserv(ID ,max(20,nbatch_in))
      else
        call avh_parni_reserv(ID ,max(20,nchmax_in))
      endif
*
      obj = copies(ID)
      oo = offset(obj)
*
      ndimen(obj) = ndimen_in
      nbatch(obj) = nbatch_in
      nchmax(obj) = nchmax_in
*
      do node=1,2*resrvd(obj)
        cutdim(oo+node) = 0
        daught(oo+node) = 0
        sumwt0(oo+node) = 0d0
        sumwt1(oo+node) = 0d0
        sumwt2(oo+node) = 0d0
        pointr(oo+node) = node
      enddo
*
      nchann(obj) = 1
      mother(oo+1) = 0
      daught(oo+1) = 0
      channl(oo+1) = 1
      nodech(oo+1) = 1
      pointr(oo+1) = 1
      weight(oo+1) = 1d0
*
      totwt1(obj) = 0d0
      totwt2(obj) = 0d0
      totwt0(obj) = 0d0
*
      nsofar(obj) = 0
      iitask(obj) = task
      start_sort(obj) = .true.
      do_adapt(obj) = .true.
      end


      subroutine avh_parni_set_nsteps( ID ,nsteps_in )
*  ********************************************************************
*  ********************************************************************
      implicit none
      integer ID,nsteps_in
      end


      subroutine avh_parni_reserv(ID ,resrvd_in)
*  ********************************************************************
*  * Update the value of "resrvd"
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID ,resrvd_in
     &,obj,ii,dif,stp,low,upp,jj,jjpdif
      double precision spillover
      parameter(spillover=1.0d0)
      logical firstcall
      data firstcall/.true./
      save firstcall
*
      if (firstcall) then
        firstcall = .false.
        ncopy = 0
        offset(1) = 0
        do ii=1,ISIZE_ncopy
          copies(ii) = 0
          seipoc(ii) = 0
          resrvd(ii) = 0
        enddo
      endif
*
      obj = copies(ID)
      if (obj.ne.0) then
        if (resrvd_in.ne.resrvd(obj)) then
          dif = 2*( resrvd_in - resrvd(obj) )
          if     (dif.gt.0) then
            stp = -1
            low = offset(ncopy+1)
            upp = offset(obj+1) + 1
          elseif (dif.lt.0) then
            stp = 1
            low = offset(obj+1) + 1
            upp = offset(ncopy+1)
          endif
          do jj=low,upp,stp
            jjpdif = jj+dif
            mother(jjpdif) = mother(jj)
            daught(jjpdif) = daught(jj)
            cutdim(jjpdif) = cutdim(jj)
            channl(jjpdif) = channl(jj)
            nodech(jjpdif) = nodech(jj)
            pointr(jjpdif) = pointr(jj)
            weight(jjpdif) = weight(jj)
            sumwt1(jjpdif) = sumwt1(jj)
            sumwt2(jjpdif) = sumwt2(jj)
            sumwt0(jjpdif) = sumwt0(jj)
             mother(jj) = 0
             daught(jj) = 0
             cutdim(jj) = 0
             channl(jj) = 0
             nodech(jj) = 0
             pointr(jj) = 0
             sumwt1(jj) = 0d0
             sumwt2(jj) = 0d0
             sumwt0(jj) = 0d0
          enddo
          do jj=2*resrvd(obj)+1,2*resrvd_in
            pointr(offset(obj)+jj) = jj
          enddo
          do jj=obj+1,ncopy+1
            offset(jj) = offset(jj) + dif
          enddo
          resrvd(obj) = resrvd_in
        endif
      else
        ncopy = ncopy + 1
        copies(ID) = ncopy
        seipoc(ncopy) = ID
        resrvd(ncopy) = resrvd_in
        offset(ncopy+1) = offset(ncopy) + 2*resrvd(ncopy)
      endif
c      write(6,*) 'MESSAGE from avh_parni: '
c     &          ,'total reserved size updated to',offset(ncopy+1)
      end


      subroutine avh_parni_weight( ID ,ww,xx )
*  ********************************************************************
*  * Return the latest generated point "xx" and corresponding weight
*  * "ww".
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision xx(ISIZE_ndim),ww
      integer ID
     &,obj,idm
*
      obj = copies(ID)
*
      do idm=1,ndimen(obj)
        xx(idm) = xstr(obj,idm)
      enddo
      ww = wstr(obj)
      end


      subroutine avh_parni_calcwght( ID ,ww ,xx )
*  *********************************************************************
*  * Returns the inverse value of of the density built by the
*  * algorithm at the point "xx" in the hypercube [0,1]^ndimen
*  * This routine also stores (overwrites) both "ww" and "xx"
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision ww,xx(ISIZE_ndim)
     &,avh_parni_halfpwr
      integer ID
     &,obj,oo,ich,nstp,idm
*
      obj = copies(ID)
      oo = offset(obj)
*
      call avh_parni_channel(obj ,ich,nstp ,xx)
*
      do idm=1,ndimen(obj)
        xstr(obj,idm) = xx(idm)
      enddo
      wstr(obj) = avh_parni_halfpwr(nstp)/weight(oo+pointr(oo+ich))
*
      ww = wstr(obj)
      end


      subroutine avh_parni_wght( ID ,ww ,xx )
*  *********************************************************************
*  * The same as calcwght
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision ww,xx(ISIZE_ndim)
      integer ID
      call avh_parni_calcwght( ID ,ww ,xx )
      end


      double precision function avh_parni_density( ID ,xx )
*  *********************************************************************
*  * Returns the value of the density built by the algorithm at
*  * the point "xx" in the hypercube [0,1]^ndimen
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision xx(ISIZE_ndim)
     &,avh_parni_halfpwr
      integer ID
     &,obj,oo,ich,nstp
*
      obj = copies(ID)
      oo = offset(obj)
*
      call avh_parni_channel(obj ,ich,nstp ,xx)
      avh_parni_density = weight(oo+pointr(oo+ich))
     &                  / avh_parni_halfpwr(nstp)
      end

      double precision function avh_parni_halfpwr(nstp)
*  ********************************************************************
*  * Powers of 0.5
*  ********************************************************************
      implicit none
      integer nstp ,maxpwr,ii
      parameter(maxpwr=100)
      double precision halfpwr(0:maxpwr)
      logical firstcall
      data firstcall/.true./
      save firstcall,halfpwr
      if (firstcall) then
        firstcall = .false.
        halfpwr(0) = 1d0
        do ii=1,maxpwr
          halfpwr(ii) = halfpwr(ii-1)/2d0
        enddo
      endif
      avh_parni_halfpwr = halfpwr(nstp)
      end


      subroutine avh_parni_channel(obj ,ich,nstp ,xx)
*  *********************************************************************
*  * Returns the number "ich" of the channel containing "xx".
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision xx(ISIZE_ndim)
     &,x0(ISIZE_ndim),x1(ISIZE_ndim),xh
      integer obj,ich,nstp
     &,oo,node,dght,idm
*
      oo = offset(obj)
*
      node = 1
      dght = daught(oo+node)
      do idm=1,ndimen(obj)
        x0(idm) = 0d0
        x1(idm) = 1d0
      enddo
      nstp = 0
      do while(dght.ne.0)
        idm = cutdim(oo+node)
        xh  = (x0(idm)+x1(idm))/2
        if (xx(idm).lt.xh) then
          node = dght
          x1(idm) = xh
        else
          node = dght+1
          x0(idm) = xh
        endif
        nstp = nstp+1
        dght = daught(oo+node)
      enddo
      ich = channl(oo+node)
      end


      subroutine avh_parni_generate(ID, xx)
*  *********************************************************************
*  *  "xx" is a data point, distributed following the density
*  *  built by the algorithm.
*  * The value of xx and the corresponding weight are stored.
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision xx(ISIZE_ndim)
     &,x0(ISIZE_ndim),x1(ISIZE_ndim),rho(ISIZE_ndim),rch,wght
     &,avh_parni_halfpwr
      integer ID
     &,obj,oo,ich,idm,nstp,node,jch,nch
*
      obj = copies(ID)
      oo = offset(obj)
      nch = nchann(obj)
*
      call avh_parni_random(rch,1)
      ich = 2*nch-1
      do while(ich.gt.nch)
        jch = 2*(ich-nch)-1
        wght = weight(oo+pointr(oo+jch))
        if (rch.ge.wght) then
          rch = rch - wght
          ich = jch+1
        else
          ich = jch
        endif
      enddo
*
      node = nodech(oo+ich)
      call avh_parni_box(obj ,x0,x1,nstp ,node)
*
      call avh_parni_random(rho,ndimen(obj))
      do idm=1,ndimen(obj)
        xx(idm) = x0(idm) + (x1(idm)-x0(idm))*rho(idm)
      enddo
*
      do idm=1,ndimen(obj)
        xstr(obj,idm) = xx(idm)
      enddo
      wstr(obj) = avh_parni_halfpwr(nstp)/weight(oo+pointr(oo+ich))
      end

      subroutine avh_parni_box(obj ,x0,x1,nstp ,node)
*  ********************************************************************
*  * Return the box that corresponds to "node"
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision x0(ISIZE_ndim),x1(ISIZE_ndim)
      integer obj ,nstp,node
     &,idm,istp,dimnsn(ISIZE_nstp)
      logical upperh(ISIZE_nstp)
*
      call avh_parni_history(obj ,dimnsn,upperh,nstp ,node)
      do idm=1,ndimen(obj)
        x0(idm) = 0d0
        x1(idm) = 1d0
      enddo
      do istp=nstp,1,-1
        idm = dimnsn(istp)
        if (upperh(istp)) then
          x0(idm) = (x0(idm)+x1(idm))/2
        else
          x1(idm) = (x0(idm)+x1(idm))/2
        endif
      enddo
      end

      subroutine avh_parni_history(obj ,dimnsn,upperh,nstp ,node_in)
*  ********************************************************************
*  * Return the history of cutting dimensions and choosing halfs
*  * that led to "node_in"
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer obj,dimnsn(ISIZE_nstp),nstp,node_in
     &,oo,mthr,node
      logical upperh(ISIZE_nstp)
*
      oo = offset(obj)
*
      nstp = 0
      node = node_in
      mthr = mother(oo+node)
      do while (mthr.ne.0)
        nstp = nstp+1
        dimnsn(nstp) = cutdim(oo+mthr)
        upperh(nstp) = (node.ne.2*(node/2))
        node = mthr
        mthr = mother(oo+node)
      enddo
      end

      double precision function avh_parni_volume(obj ,node_in)
*  ********************************************************************
*  * Volume of box at node_in
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer obj,node_in
     &,oo,nstp,mthr,node
      double precision avh_parni_halfpwr
*
      oo = offset(obj)
*
      nstp = 0
      node = node_in
      mthr = mother(oo+node)
      do while (mthr.ne.0)
        nstp = nstp+1
        node = mthr
        mthr = mother(oo+node)
      enddo
      avh_parni_volume = avh_parni_halfpwr(nstp)
      end


      subroutine avh_parni_adaptall( wdat )
*  ********************************************************************
*  * Collect "wdat" for adaptation of all instances of PARNI
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision wdat
     &,xx(ISIZE_ndim)
      integer
     & obj,idm,ID
*
      do obj=1,ncopy
*
        do idm=1,ndimen(obj)
          xx(idm) = xstr(obj,idm)
        enddo
        ID = seipoc(obj)
        call avh_parni_adapt( ID ,wdat ,xx )
*
      enddo
      end


      subroutine avh_parni_adaptone( ID ,wdat )
*  ********************************************************************
*  * Collect "wdat" for adaptation of the instance ID of PARNI
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision wdat
     &,xx(ISIZE_ndim)
      integer ID
     &,obj,idm
*
      obj = copies(ID)
*
      do idm=1,ndimen(obj)
        xx(idm) = xstr(obj,idm)
      enddo
      call avh_parni_adapt( ID ,wdat ,xx )
      end


      subroutine avh_parni_adapt(ID ,wdat,xdat)
*  *********************************************************************
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision wdat,xdat(ISIZE_ndim)
     &,sum1,sum2,tmp,par0,par1,par2,avh_parni_halfpwr,avh_parni_volume
      parameter(par0=0.5d0) ! should be between 0 and 1
      parameter(par1=par0,par2=1d0-par0)
      integer ID
     &,obj,oo,node,ich,jch,kch,nstp
      logical cont
*
      obj = copies(ID)
      oo = offset(obj)
*
      call avh_parni_channel(obj ,ich,nstp ,xdat)
      if (iitask(obj).le.10) then
        tmp = wdat*weight(oo+pointr(oo+ich))/avh_parni_halfpwr(nstp)
      else
        tmp = wdat
      endif
      kch = oo+pointr(oo+nchann(obj)+ich)
      sumwt0(kch) = sumwt0(kch) + 1d0
      sumwt1(kch) = sumwt1(kch) + tmp
      sumwt2(kch) = sumwt2(kch) + tmp*tmp
      nsofar(obj) = nsofar(obj) + 1
*
      if (nsofar(obj).eq.nbatch(obj)) then
        sum1 = 0d0
        sum2 = 0d0
        do ich=1,nchann(obj)
          jch = oo+pointr(oo+ich)
          kch = oo+pointr(oo+nchann(obj)+ich)
          sumwt0(jch) = sumwt0(jch) + sumwt0(kch)
          sumwt1(jch) = sumwt1(jch) + sumwt1(kch)
          sumwt2(jch) = sumwt2(jch) + sumwt2(kch)
          if (iitask(obj).le.10) then
            node = nodech(oo+ich)
            tmp = weight(jch)/avh_parni_volume(obj,node)
            sum1 = sum1 + sumwt1(kch)/tmp
            sum2 = sum2 + sumwt2(kch)/(tmp*tmp)
          else
            sum1 = sum1 + sumwt1(kch)
            sum2 = sum2 + sumwt2(kch)
          endif
        enddo
        totwt1(obj) = totwt1(obj) + sum1
        totwt2(obj) = totwt2(obj) + sum2
        totwt0(obj) = totwt0(obj) + dble(nbatch(obj))
        nsofar(obj) = 0
*
        if     (iitask(obj).eq.1.or.iitask(obj).eq.6) then
          do ich=1,nchann(obj)
            jch = oo+pointr(oo+ich)
            weight(jch) = sumwt1(jch)/sumwt0(jch)
     &                  * avh_parni_volume(obj,nodech(oo+ich))
          enddo
        elseif (iitask(obj).eq.2.or.iitask(obj).eq.7) then
          do ich=1,nchann(obj)
            jch = oo+pointr(oo+ich)
            weight(jch) = dsqrt( sumwt2(jch)/sumwt0(jch) )
     &                  * avh_parni_volume(obj,nodech(oo+ich))
          enddo
        elseif (iitask(obj).eq.3.or.iitask(obj).eq.8) then
          do ich=1,nchann(obj)
            jch  = oo+pointr(oo+ich)
            weight(jch) = ( sumwt1(jch)/sumwt0(jch) )**par1
     &                  * dsqrt( sumwt2(jch)/sumwt0(jch) )**par2
     &                  * avh_parni_volume(obj,nodech(oo+ich))
          enddo
        elseif (iitask(obj).eq.11.or.iitask(obj).eq.16) then
          do ich=1,nchann(obj)
            jch = oo+pointr(oo+ich)
            weight(jch) = sumwt1(jch)
          enddo
        elseif (iitask(obj).eq.12.or.iitask(obj).eq.17) then
          do ich=1,nchann(obj)
            jch = oo+pointr(oo+ich)
            weight(jch) = dsqrt( sumwt2(jch) )
          enddo
        elseif (iitask(obj).eq.13.or.iitask(obj).eq.18) then
          do ich=1,nchann(obj)
            jch = oo+pointr(oo+ich)
            weight(jch) = sumwt1(jch)**par1 * dsqrt( sumwt2(jch) )**par2
          enddo
        endif
*
        if (do_adapt(obj)) then
          cont = .true.
          do while (cont)
            call avh_parni_findmax(obj ,cont)
            if (cont) call avh_parni_dissect(obj)
            if (nchann(obj).ge.resrvd(obj)) then
              if (nchmax(obj).eq.0) then
                jch = int(1.5d0*resrvd(obj))
              else
                jch = int(1.1d0*resrvd(obj))
              endif
              call avh_parni_reserv(ID, jch)
            endif
          enddo
*
          if (nchmax(obj).gt.0) then
          do while (nchann(obj).gt.nchmax(obj))
            call avh_parni_merge(obj)
          enddo
          endif
        endif
*
        do ich=1,nchann(obj)
          kch = oo+pointr(oo+nchann(obj)+ich)
          sumwt0(kch) = 0d0
          sumwt1(kch) = 0d0
          sumwt2(kch) = 0d0
        enddo
*
        call avh_parni_add(obj)
*
        do ich=1,nchann(obj)
          daught(oo+nodech(oo+ich)) = 0
        enddo
      endif
      end


      subroutine avh_parni_findmax(obj ,cont)
*  *********************************************************************
*  * determine the channel with the largest weight and put it last
*  *********************************************************************
      implicit none 
      include 'avh_parni_priv.h'
      integer obj 
     &,oo,ichmax,node,ich,lch(ISIZE/ISIZE_ncopy),nsame,nch
      double precision
     & wchmax,ineff,ineff_old(ISIZE_ncopy),rho,par1,par2,vlm,wch
     &,weightmax,avh_parni_large,avh_parni_volume
      parameter(par1=0.5d0)
      parameter(par2=1d0-par1)
      logical cont,firstcall,case1
      data firstcall/.true./
      save ineff_old,firstcall
*
      oo = offset(obj)
      nch = nchann(obj)
*
      if (start_sort(obj)) then
        start_sort(obj) = .false.
        ineff_old(obj) = avh_parni_large()
      endif
*
      nsame = 1
      wchmax = -1d0
      weightmax = -1d0
      case1 = (                          iitask(obj).le.5
     &         .or.iitask(obj).gt.10.and.iitask(obj).le.15)
!
      if (case1) then
        do ich=1,nch
          wch = weight(oo+pointr(oo+ich))
          if     (wch.gt.wchmax) then
            wchmax = wch
            lch(1) = ich
            nsame = 1
          elseif (wch.eq.wchmax) then
            nsame = nsame+1
            lch(nsame) = ich
          endif
          weightmax = wchmax
        enddo
      else
        do ich=1,nch
          wch = weight(oo+pointr(oo+ich))
          vlm = avh_parni_volume(obj,nodech(oo+ich))
          wch = wch**par1 * vlm**par2
          if     (wch.gt.wchmax) then
            wchmax = wch
            lch(1) = ich
            nsame = 1
          elseif (wch.eq.wchmax) then
            nsame = nsame+1
            lch(nsame) = ich
          endif
          weightmax = wchmax
        enddo
      endif
*
      ineff = weightmax*nch
*
      if (ineff.ge.ineff_old(obj)) then
        ineff_old(obj) = avh_parni_large()
        cont = .false.
      else
        ineff_old(obj) = ineff
        cont = .true.
      endif
*
      if (nsame.gt.1) then
        call avh_parni_random(rho,1)
        ichmax = lch(1+int(nsame*rho))
      else
        ichmax = lch(1)
      endif
*
      call avh_parni_switch(obj ,ichmax,nch)
      end


      subroutine avh_parni_dissect(obj)
*  *********************************************************************
*  * dissect the last channel into two new ones.
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision
     & rho,stdev,w1old,w2old,w0old,weold
      integer obj
     &,oo,node,idm,istp,nstp,ncut(ISIZE_ndim),ncutmin,maxidm ,nch
     &,dimnsn(ISIZE_nstp),nsame,ldm(ISIZE_ndim),node1,node2,jch1,jch2
      logical init,upperh(ISIZE_nstp)
*
      oo = offset(obj)
      nch = nchann(obj)
      node = nodech(oo+nch)
*
* determine the dimension in which to dissect box 1:
* determine largerst dimension in box 1,
* which is the one with the fewest number of cuts
      call avh_parni_history(obj ,dimnsn,upperh,nstp ,node)
      if (nstp.ge.ISIZE_nstp) return ! restrict minimal bin-size
      do idm=1,ndimen(obj)
        ncut(idm) = 0
      enddo
      do istp=1,nstp
        ncut(dimnsn(istp)) = ncut(dimnsn(istp)) + 1
      enddo
      maxidm = 1
      do idm=1,ndimen(obj)
        if (ncut(idm).lt.nstp) then
          nstp = ncut(idm)
          maxidm = idm
        endif
      enddo
* if several dimensions have this size, choose one at random
      nsame = 0
      do idm=1,ndimen(obj)
        if (ncut(idm).eq.nstp) then
          nsame = nsame+1
          ldm(nsame) = idm
        endif
      enddo
      if (nsame.gt.1) then
        call avh_parni_random(rho,1)
        maxidm = ldm( 1+int(nsame*rho) )
      endif
*
      jch1 = oo+pointr(oo+nch)
      weold = weight(jch1)
      w0old = sumwt0(jch1)
      w1old = sumwt1(jch1)
      w2old = sumwt2(jch1)
*
      node1 = 2*nch
      node2 = 2*nch+1 
      cutdim(oo+node ) = maxidm
      daught(oo+node ) = node1
      mother(oo+node1) = node
      daught(oo+node1) = 0
      cutdim(oo+node1) = 0
      mother(oo+node2) = node
      daught(oo+node2) = 0
      cutdim(oo+node2) = 0
*
      nodech(oo+ 2*nch)   = node1
      nodech(oo+ 2*nch+1) = node2
      channl(oo+ node1 )  = 2*nch
      channl(oo+ node2 )  = 2*nch+1
      call avh_parni_switch(obj ,nch  ,2*nch  )
      call avh_parni_switch(obj ,nch+1,2*nch+1)
*
      jch1 = oo+pointr(oo+nch)
      jch2 = oo+pointr(oo+nch+1)
      sumwt0(jch1) = w0old/2d0
      sumwt0(jch2) = w0old/2d0
      sumwt1(jch1) = w1old/2d0
      sumwt1(jch2) = w1old/2d0
      sumwt2(jch1) = w2old/2d0
      sumwt2(jch2) = w2old/2d0
      weight(jch1) = weold/2d0
      weight(jch2) = weold/2d0
*
      nchann(obj) = nchann(obj)+1
      end


      subroutine avh_parni_merge(obj)
*  *********************************************************************
*  * determine the pair of nodes with the smallest weights
*  * and merge the boxes to one box
*  *********************************************************************
      implicit none 
      include 'avh_parni_priv.h'
      double precision
     & wchmin,rho,wchtmp,avh_parni_large
      integer obj 
     &,oo,ich1,ich2,node,ich,nsame,jch,nch,sstr,mthr
     &,jch1,jch2,lch(ISIZE/ISIZE_ncopy)
*
      oo = offset(obj)
      nch = nchann(obj)
*
      nsame = 1
      lch(1) = 1
      wchmin = avh_parni_large()
      do ich=1,nch
        node = nodech(oo+ich)
        sstr = (node/2)*2
        if (node.eq.sstr) sstr = sstr+1
        if (daught(oo+sstr).eq.0) then
          jch = channl(oo+sstr)
          wchtmp = weight(oo+pointr(oo+ich)) + weight(oo+pointr(oo+jch))
          if     (wchtmp.lt.wchmin) then
            wchmin = wchtmp
            lch(1) = ich
            nsame = 1
          elseif (wchtmp.eq.wchmin) then
            nsame = nsame+1
            lch(nsame) = ich
          endif
        endif
      enddo
*
      if (nsame.gt.2) then
        call avh_parni_random(rho,1)
        ich1 = lch(1+int(nsame*rho))
      else
        ich1 = lch(1)
      endif
      node = nodech(oo+ich1)
      sstr = (node/2)*2
      if (node.ne.sstr) node = sstr
      sstr = sstr+1
*
      ich1 = channl(oo+node)
      ich2 = channl(oo+sstr)
*
      mthr = mother(oo+node)
      ich  = channl(oo+mthr)
      jch  = oo+pointr(oo+ich)
      jch1 = oo+pointr(oo+ich1)
      jch2 = oo+pointr(oo+ich2)
      weight(jch) = weight(jch1) + weight(jch2)
      sumwt0(jch) = sumwt0(jch1) + sumwt0(jch2)
      sumwt1(jch) = sumwt1(jch1) + sumwt1(jch2)
      sumwt2(jch) = sumwt2(jch1) + sumwt2(jch2)
      cutdim(oo+mthr) = 0
      daught(oo+mthr) = 0
*
      call avh_parni_switch(obj ,ich ,2*nch-1)
      call avh_parni_switch(obj ,ich1,2*nch-1)
      call avh_parni_switch(obj ,ich2,  nch  )
      call avh_parni_switch(obj ,nch ,2*nch-2)
*
      call avh_parni_rename(obj ,node,2*nch-2)
*
      nchann(obj) = nchann(obj)-1
      end


      subroutine avh_parni_switch(obj ,ich,jch)
*  ********************************************************************
*  * Switch number for channel ich and jch
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer obj ,ich,jch
     &,oo,itmp,ii,jj
*
      oo = offset(obj)
      ii = oo+ich
      jj = oo+jch
*
      itmp       = pointr(ii)
      pointr(ii) = pointr(jj)
      pointr(jj) = itmp
      itmp       = nodech(ii)
      nodech(ii) = nodech(jj)
      nodech(jj) = itmp
      channl(oo+nodech(ii)) = ich
      channl(oo+nodech(jj)) = jch
      end


      subroutine avh_parni_rename(obj ,node1_in,node2_in)
*  ********************************************************************
*  * Switch the numbers of node1 and her sister with those of node2
*  * and her sister
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer obj ,node1_in,node2_in
     &,oo,node1,node2,sstr1,sstr2,n1,n2,s1,s2,itmp
*
      oo = offset(obj)
      node1 = node1_in
      node2 = node2_in
*
      sstr1 = (node1/2)*2
      if (node1.ne.sstr1) node1 = sstr1
      sstr1 = sstr1+1
      sstr2 = (node2/2)*2
      if (node2.ne.sstr2) node2 = sstr2
      sstr2 = sstr2+1
*
      if (node1.eq.node2) return
      n1 = oo+node1
      s1 = oo+sstr1
      n2 = oo+node2
      s2 = oo+sstr2
      itmp       = mother(n1)
      mother(n1) = mother(n2)
      mother(n2) = itmp
      itmp       = daught(n1)
      daught(n1) = daught(n2)
      daught(n2) = itmp
      itmp       = cutdim(n1)
      cutdim(n1) = cutdim(n2)
      cutdim(n2) = itmp
      itmp       = channl(n1)
      channl(n1) = channl(n2)
      channl(n2) = itmp
      nodech(oo+channl(n1)) = node1
      if (daught(n1).ne.0) then
        mother(oo+daught(n1)  ) = node1
        mother(oo+daught(n1)+1) = node1
      endif
      if (mother(n1).ne.0) daught(oo+mother(n1)) = node1
      nodech(oo+channl(n2)) = node2
      if (daught(n2).ne.0) then
        mother(oo+daught(n2)  ) = node2
        mother(oo+daught(n2)+1) = node2
      endif
      if (mother(n2).ne.0) daught(oo+mother(n2)) = node2
      itmp       = mother(s1)
      mother(s1) = mother(s2)
      mother(s2) = itmp
      itmp       = daught(s1)
      daught(s1) = daught(s2)
      daught(s2) = itmp
      itmp       = cutdim(s1)
      cutdim(s1) = cutdim(s2)
      cutdim(s2) = itmp
      itmp       = channl(s1)
      channl(s1) = channl(s2)
      channl(s2) = itmp
      nodech(oo+channl(s1)) = sstr1
      if (daught(s1).ne.0) then
        mother(oo+daught(s1)  ) = sstr1
        mother(oo+daught(s1)+1) = sstr1
      endif
      nodech(oo+channl(s2)) = sstr2
      if (daught(s2).ne.0) then
        mother(oo+daught(s2)  ) = sstr2
        mother(oo+daught(s2)+1) = sstr2
      endif
      end


      subroutine avh_parni_add(obj)
*  ********************************************************************
*  * Add the weights together in a tree-structure, and normalize them.
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer obj
     &,oo,nnew,ich,jch
      double precision
     & cnst
*
      oo = offset(obj)
*
      nnew = nchann(obj)
      do ich=1,2*nchann(obj)-3,2
        nnew = nnew+1
        weight(oo+pointr(oo+nnew))
     &  = weight(oo+pointr(oo+ich)) + weight(oo+pointr(oo+ich+1))
      enddo
      cnst = 1d0/weight(oo+pointr(oo+nnew))
      do ich=1,2*nchann(obj)-1
        jch = oo+pointr(oo+ich)
        weight(jch) = weight(jch) * cnst
      enddo
      end


      subroutine avh_parni_clean(ID)
*  ********************************************************************
*  * Put all weights to zero but keep cell-structure. Stop adaptation.
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID
     &,obj,oo,ich,jch
*
      obj = copies(ID)
      oo = offset(obj)
*
      do ich=1,2*nchann(obj)
        jch = oo+pointr(oo+ich)
        sumwt0(jch) = 0d0
        sumwt1(jch) = 0d0
        sumwt2(jch) = 0d0
        weight(jch) = 0d0
      enddo
      totwt1(obj) = 0d0
      totwt2(obj) = 0d0
      totwt0(obj) = 0d0
      do_adapt(obj) = .false.
      end


      subroutine avh_parni_restartall
*  ********************************************************************
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer
     & obj,oo,ich,jch
*
      do obj=1,ncopy
*
        oo = offset(obj)
*
        do ich=1,2*nchann(obj)
          jch = oo+pointr(oo+ich)
          sumwt0(jch) = 0d0 !sumwt0(jch)/2d0
          sumwt1(jch) = 0d0 !sumwt1(jch)/2d0
          sumwt2(jch) = 0d0 !sumwt2(jch)/2d0
        enddo
        totwt1(obj) = 0d0 !totwt1(obj)/2d0
        totwt2(obj) = 0d0 !totwt2(obj)/2d0
        totwt0(obj) = 0d0 !totwt0(obj)/2d0
*
      enddo
      end


      subroutine avh_parni_updateall( itask_in ,nbatch_in ,nchmax_in )
*  ********************************************************************
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer itask_in,nbatch_in,nchmax_in
     &,obj
*
      do obj=1,ncopy
        iitask(obj) = itask_in
        nbatch(obj) = nbatch_in
        nchmax(obj) = nchmax_in
      enddo
      end


      subroutine avh_parni_free( ID )
*  ********************************************************************
*  * Returns the smallest value for ID that is still free
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID
     &,ii
      ii = 1
      do while(copies(ii).gt.0)
        ii = ii+1
      enddo
      ID = ii
      end


      subroutine avh_parni_marg(ID ,nunit ,label)
*  ********************************************************************
*  * The marginal distribution for each dimension, that is the
*  * distribution obtained by integrating over all other dimensions,
*  * is written into files. The input "nunit" is the unit to write to,
*  * and "label" is an extra label for those files. To view, for
*  * example, the distribution for ID=37, label=0 and dimension=2, use
*  *   gnuplot> plot 'PARNI37d2' w l
*  * For non-zero label, say label=900, use
*  *   gnuplot> plot 'PARNI37d2_900' w l
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID,nunit,label
     &,obj,oo,nch,idm,length,ndig,ich,nstp,node,nx,ix,jx
      double precision
     & x0(ISIZE_ndim),x1(ISIZE_ndim),xx(ISIZE),height(ISIZE),rr
      character(24) filename,chii,spaces
      parameter(spaces='                        ')
*
      obj = copies(ID)
      oo = offset(obj)
      nch = nchann(obj)
*
      call avh_parni_charnum(chii,ndig ,ID)
      write(6,'(a28,a24)') 'Writing marginals for PARNI ',chii
*
      do idm=1,ndimen(obj)
*
      filename = spaces
      filename(1:5) = 'PARNI'
      length = 5
      call avh_parni_charnum(chii,ndig ,ID)
      filename(length+1:length+ndig) = chii(1:ndig)
        length = length+ndig
      filename(length+1:length+1) = 'd'
        length = length+1
      call avh_parni_charnum(chii,ndig ,idm)
      filename(length+1:length+ndig) = chii(1:ndig)
        length = length+ndig
      if (label.gt.0) then
        filename(length+1:length+1) = '_'
          length = length+1
        call avh_parni_charnum(chii,ndig ,label)
        filename(length+1:length+ndig) = chii(1:ndig)
          length = length+ndig
      endif
      open(nunit,file=filename)
*
      do ich=1,nch
        node = nodech(oo+ich)
        call avh_parni_box(obj ,x0,x1,nstp ,node)
        xx(2*ich-1) = x0(idm)
        xx(2*ich  ) = x1(idm)
      enddo
      nx = 2*nch
      call avh_parni_sort(xx,nx)
      nx = nx+1
      ix = 0
      do while (ix.lt.nx)
        ix = ix+1
        if (xx(ix).eq.xx(ix+1)) then
          do jx=ix+1,nx-1
            xx(jx) = xx(jx+1)
          enddo
          ix = ix-1
          nx = nx-1
        endif
      enddo
      do ix=1,nx-1
        height(ix) = 0d0
        do ich=1,nch
          node = nodech(oo+ich)
          call avh_parni_box(obj ,x0,x1,nstp ,node)
          rr = min(xx(ix+1),x1(idm)) - max(xx(ix),x0(idm))
          if (rr.lt.0) rr = 0d0
          rr = rr/(x1(idm)-x0(idm))
          height(ix) = height(ix) + rr * weight(oo+pointr(oo+ich))
        enddo
        height(ix) = height(ix)/(xx(ix+1)-xx(ix))
      enddo
      do ix=1,nx-1
        write(nunit,101) xx(ix  ),'0d0'
        write(nunit,102) xx(ix  ),height(ix)
        write(nunit,102) xx(ix+1),height(ix)
        write(nunit,101) xx(ix+1),'0d0'
      enddo
  101 format(e16.8,1x,a3)
  102 format(2e16.8)
*
      close(nunit)
      call avh_parni_charnum(chii,ndig ,idm)
      write(6,*) 'To view marginal for dim ',chii(1:ndig)
     &          ,', use gnuplot> plot "',filename(1:length),'" w l'
      enddo
      end

      subroutine avh_parni_sort(xx,nn)
*  ********************************************************************
*  ********************************************************************
      implicit none
      integer           nn     ,ll,ii,jj,ir
      double precision  xx(nn) ,xxb 
      if (nn.le.1) goto 30
*     
      ll = nn/2 + 1
      ir = nn
  10  continue
        if (ll.gt.1) then 
          ll = ll - 1
          xxb = xx(ll)
        else
          xxb = xx(ir)
          xx(ir) = xx(1)
          ir = ir - 1
          if (ir.eq.1) then
            xx(1) = xxb
            goto 30
          endif
        endif
        ii = ll
        jj = ll + ll
  20    continue
        if (jj.le.ir) then
          if (jj.lt.ir) then
            if (xx(jj).lt.xx(jj+1)) jj = jj + 1
          endif
          if (xxb.lt.xx(jj)) then
            xx(ii) = xx(jj)
            ii = jj
            jj = jj + jj
          else
            jj = ir + 1
          endif
        goto 20
        endif
        xx(ii) = xxb
      goto 10
  30  continue
*
      end


      subroutine avh_parni_plot(ID ,nunit ,label)
*  *********************************************************************
*  * Only works for dimension=1 and dimension=2.
*  * Consider the example that ID=37 and label=0. A plot of the built
*  * density is written in the file with the name PARNI37p. To view,
*  * open gnuplot and put 
*  *   gnuplot> plot 'PARNIp37' w l
*  * for the 1-dim case. For the the 2-dim case, the cell structure
*  * will be depicted this way. To get the full 3-dim plot for the
*  * 2-dim case, use
*  *   gnuplot> splot 'PARNIp37' w l
*  * Input: "nunit" is a number for the unit to write to.
*  *        "label" is, if label>0, an extra label to the filename
*  *********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      double precision height,x0(ISIZE_ndim),x1(ISIZE_ndim),y0,y1,y2
     &,avh_parni_halfpwr
      integer ID,nunit,label ,obj,oo,ich,node,nstp,ndig,length
      character(24) filename,chii,spaces
      parameter(spaces='                        ')
*
      obj = copies(ID)
      oo = offset(obj)
*
      call avh_parni_charnum(chii,ndig ,ID)
c      write(6,'(a31,a24)') 'Writing distribution for PARNI ',chii
*
      filename = spaces
      filename(1:5) = 'PARNI'
        length = 5
      call avh_parni_charnum(chii,ndig ,ID)
      filename(length+1:length+ndig) = chii(1:ndig)
        length = length+ndig
      filename(length+1:length+1) = 'p'
        length = length+1
      if (label.gt.0) then
        filename(length+1:length+1) = '_'
          length = length+1
        call avh_parni_charnum(chii,ndig ,label)
        filename(length+1:length+ndig) = chii(1:ndig)
          length = length+ndig
      endif
      open(nunit,file=filename)
*
      if (ndimen(obj).eq.1) then
*
  101 format(e16.8,1x,a3,1x,a3,1x,a3)
  102 format(e16.8,e16.8,e16.8,e16.8)
      do ich=1,nchann(obj)
        node = nodech(oo+ich)
        call avh_parni_box(obj ,x0,x1,nstp ,node)
        height = weight(oo+pointr(oo+ich))/avh_parni_halfpwr(nstp)
        y0 = sumwt0(oo+pointr(oo+ich))/totwt0(obj)
     &     / avh_parni_halfpwr(nstp)
        y1 = dabs(height-y0)*weight(oo+pointr(oo+ich))!avh_parni_halfpwr(nstp)
        write(nunit,101) x0(1),'0d0' ,'0d0','0d0'
        write(nunit,102) x0(1),height,   y0,   y1
        write(nunit,102) x1(1),height,   y0,   y1
        write(nunit,101) x1(1),'0d0' ,'0d0','0d0'
      enddo
c      write(6,*) 'To view distribution, please use '
c     &          ,'gnuplot> plot "',filename(1:length),'" w l'
*
      elseif (ndimen(obj).eq.2) then
*
  201 format(3e16.8)
  202 format()
      do ich=1,nchann(obj)
        node = nodech(oo+ich)
        call avh_parni_box(obj ,x0,x1,nstp ,node)
        height = 0d0
        write(nunit,201) x0(1),x0(2),height
        write(nunit,201) x0(1),x1(2),height
        write(nunit,201) x1(1),x1(2),height
        write(nunit,201) x1(1),x0(2),height
        write(nunit,201) x0(1),x0(2),height
        write(nunit,202)
        height = weight(oo+pointr(oo+ich))/avh_parni_halfpwr(nstp)
        write(nunit,201) x0(1),x0(2),height
        write(nunit,201) x0(1),x1(2),height
        write(nunit,201) x1(1),x1(2),height
        write(nunit,201) x1(1),x0(2),height
        write(nunit,201) x0(1),x0(2),height
        write(nunit,202)
        write(nunit,202)
      enddo
      write(6,*) 'To view cell-structure, please use '
     &          ,'gnuplot> plot "',filename(1:length),'" w l'
      write(6,*) 'To view distribution, please use '
     &          ,'gnuplot> splot "',filename(1:length),'" w l'
*
      endif
*
      close(nunit)
*
      end

  
      subroutine avh_parni_result(ID)
*  ********************************************************************
*  ********************************************************************
      implicit none
      include 'avh_parni_priv.h'
      integer ID ,nunit
     &,obj,oo,length,ndig ,ich,jch,node,nstp,dimnsn(ISIZE_nstp)
      logical upperh(ISIZE_nstp)
      parameter(nunit=6)
      double precision ave,err
      character(24) filename,chii
*
      obj = copies(ID)
      oo = offset(obj)
*
      filename = '                        '
      filename(1:6) = 'PARNIr'
      length = 6
      call avh_parni_charnum(chii,ndig ,ID)
      filename(length+1:length+ndig) = chii(1:ndig)
      length = length+ndig
*
      ave = totwt1(obj)/totwt0(obj)
      err = (totwt2(obj)/totwt0(obj)-ave*ave)/(totwt0(obj)-1d0)
      err = dsqrt(err)
*
c      open(nunit,file=filename)
      write(nunit,'(a36,a36)') '####################################'
     &                        ,'####################################'
      write(nunit,'(a1,a42,a1,i2,a1,a25)')
     &  '#','Diagnostics for the optimization of PARNI',' ',ID,':','#'
      write(nunit,'(a1,a7,a1,i3 ,a1,a15,a1,i7 ,a1,a20,a1,i6 ,a8)')
     &  '#','task',':',iitask(obj)
     & ,',','batch-size',':',nbatch(obj)
     & ,',','number of channels',':',nchann(obj),'#'
      write(nunit,'(a1,a10,a1,d24.16,a4,d23.16 ,a9)')
     &  '#','average',':',ave,'+/-',err,'(crude) #'
      write(nunit,'(a1,a19,a1,f13.0 ,a1,a16,a1,f11.8 ,a9)')
     &  '#','number of events',':',totwt0(obj)
     & ,',','relative error',':',err/dabs(ave),'(crude) #'
      write(nunit,'(a36,a36)') '####################################'
     &                        ,'####################################'
c      close(nunit)
      end
      

      double precision function avh_parni_large()
*  ********************************************************************
*  * Rediculously large positive double precision number
*  ********************************************************************
      implicit none
      double precision xx,yy,large,prec
      logical firstcall
      data firstcall /.true./
      save firstcall,large
      if (firstcall) then
        firstcall = .false.
        xx = 1d0
        yy = xx
        do while (xx.eq.yy)
          prec = xx
          xx = xx/2
          yy = -1d0+dexp(dlog(1d0+xx))
        enddo
        large = 1d300 ! double precision
        if (prec.lt.1d-24) large = large**16 ! quadruple precision
        write(6,*) 'MESSAGE from avh_parni: '
     &            ,'avh_parni_large set to ',large
      endif
      avh_parni_large = large
      end


      subroutine avh_parni_charnum(chii,ndig ,ii)
*  ********************************************************************
*  * Determines the character representing "ii".
*  ********************************************************************
      implicit none
      integer ndig,ii ,istp,iden,ndigmax,jj,ipos,idig
*
      parameter(ndigmax=9)
*
      character(24) chii ,nmbr
      character(1) arab(0:9)
      data arab/'0','1','2','3','4','5','6','7','8','9'/
      chii = '                        '
      istp = ii
      iden = 10**(ndigmax-1)
      if (ii.ge.10*iden) then
        write(6,*) 'WARNING from avh_parni_charnum: ii =',ii
     &            ,' is too large, returning "_x".'
        chii(1:2) = '_x'
        ndig = 2
        return
      endif
      do jj=ndigmax,1,-1
        ipos = ndigmax-jj+1
        idig = istp/iden
        nmbr(ipos:ipos) = arab(idig)
        istp = istp - iden*idig
        iden = iden/10
      enddo
      ipos=1
      do while (nmbr(ipos:ipos).eq.arab(0))
        ipos = ipos+1
      enddo
      ndig = ndigmax-ipos+1
      chii(1:ndig) = nmbr(ipos:ndigmax)
      end
