      function colcor(icol1,icol2,k,l,n)

c-----------------------------------------------------------------------
c...  calculates the color correlator between lines k and l for 
c...  given color flows icol1 and icol2 
c...  l - emitter ij     
c...  k - spectator
c...  color correlator symmetric l <-> k
c-----------------------------------------------------------------------

      include 'declare.h'      
      include 'dimensions.h'

      double precision colcor
      dimension icol1(mxpart,2),icol2(mxpart,2)

      double precision singlet,delta
      parameter (NCOL=3)

      colcor=0d0

      nc=ncycles(icol1,icol2,-1,1,-1,1,n)
      singlet=NCOL**(nc-1)
      
      do kdir=1,2
         if (icol1(k,kdir).eq.0) cycle

         do ldir=1,2
            if (icol1(l,ldir).eq.0) cycle

            nckl=ncycles(icol1,icol2,k,kdir,l,ldir,n)
            delta=NCOL**nckl-singlet

            if (kdir.ne.ldir) delta=-delta

            colcor=colcor+delta
         enddo
      enddo

      colcor=0.5d0*colcor

      end

c-----------------------------------------------------------------------

      function ncycles(icol1,icol2,k,kdir,l,ldir,n)

c-----------------------------------------------------------------------
c...  count the number of cycles in a graph obtained by 
c...  connecting two color flow graphs                                     
c...  icol1, icol2 - color flow graphs in color flow notation   
c...  k,l - lines to interchange
c...  kdir,ldir - directions of the lines to interchange
c...  besides lines k and l, every line from the first graph is 
c...  connected to the line from the second graph corresponding 
c...  to the same particle and having the same direction. it is 
c...  assumed that the lines of the second graph have reversed 
c...  color flow
c-----------------------------------------------------------------------
      
      include 'declare.h'   
      include 'dimensions.h'

      dimension icol1(mxpart,2),icol2(mxpart,2)
      logical first,used(2,mxpart,2)
      integer dir

c...  initialization
      first=.true.

      do i=1,n
         do j=1,2
            used(1,i,j)=.false.
            used(2,i,j)=.false.
         enddo
      enddo

      ncycles=0
      
c...  find an element of an unused cycle in one of the graphs
 10   do j=1,2
         do i=1,n
            do dir=1,2
               if (icol1(i,dir).ne.0.and..not.used(j,i,dir)) goto 20
            enddo
         enddo
      enddo

c...  all elements have been used
      goto 100

c...  explore the cycle
 20   istart=i

c...  if the first graph has been explored, run only through the second
      if (j.eq.2) then
         j=i
         dir=3-dir
         first=.false.
         goto 40
      endif

c...  move from one end of a line to another in the first graph
 30   used(1,i,dir)=.true.
      do j=1,n
         if (icol1(j,3-dir).eq.icol1(i,dir)) exit
      enddo
      used(1,j,3-dir)=.true.

c...  jump between lines
      if (j.eq.k.and.3-dir.eq.kdir) then
         if (kdir.ne.ldir) then
            i=l
            goto 50
         endif
         j=l
      else if (j.eq.l.and.3-dir.eq.ldir) then
         if (kdir.ne.ldir) then
            i=k
            goto 50
         endif
         j=k
      endif

c...  move from one end of a line to another in the second graph
 40   used(2,j,3-dir)=.true.
      do i=1,n
         if (icol2(i,dir).eq.icol2(j,3-dir)) exit
      enddo
      used(2,i,dir)=.true.

c...  jump between lines
      if (i.eq.k.and.dir.eq.kdir) then
         if (kdir.ne.ldir) then
            j=l
            if (first) goto 40
         else
            i=l
         endif
      else if (i.eq.l.and.dir.eq.ldir) then
         if (kdir.ne.ldir) then
            j=k
            if (first) goto 40
         else
            i=k
         endif
      endif

c...  check whether the cycle has been closed
 50   if (first) then
         if (i.ne.istart) goto 30
      else
         if (j.ne.istart) goto 40
      endif

c...  next cycle
      ncycles=ncycles+1
      goto 10

c...  no further cycles
 100  continue
      end

c-----------------------------------------------------------------------

      function nclrcfns(icolor,n)

c-----------------------------------------------------------------------
c...  count the number of possible color configurations
c...  icolor - any color configuration to determine the parameters
c...  n - number of particles
c-----------------------------------------------------------------------
      
      include 'declare.h'      
      include 'dimensions.h'

      integer nclrcfns,icolor(mxpart,2),factorial

      nclrs=0
      do i=1,n
         if (icolor(i,1).ne.0) nclrs=nclrs+1
      enddo
      
      nclrcfns=0
      nfac=factorial(nclrs)
      do i=0,nclrs
         do j=0,nclrs
            do k=0,nclrs
               if (i+j+k.eq.nclrs) then
                  ifac=factorial(i)
                  jfac=factorial(j)
                  kfac=factorial(k)
                  nclrcfns=nclrcfns+(nfac/(ifac*jfac*kfac))**2
               endif
            enddo
         enddo
      enddo
      
      end

c-----------------------------------------------------------------------

      function factorial(i)

c-----------------------------------------------------------------------
c...  
c-----------------------------------------------------------------------
      
      integer factorial
      integer i,ifac(0:12)
      data ifac/1,1,2,6,24,120,720,5040,40320,362880,3628800,39916800,
     .          479001600/

      if (i.gt.12) then
         print*,'too large factorial argument',i
         stop
      endif

      factorial=ifac(i)

      end
