c-----------------------------------------------------------------------
c...  Copyright (C) 2009 M. Czakon, C. Papadopoulos and M. Worek
c...  Copyright (C) 2013 G. Bevilacqua, M. Czakon, M. Kubocz, M. Worek
c...  For details see: e-Print: arXiv:0905.0883 [hep-ph] 
c...                   e-Print: arXiv:1308.5605 [hep-ph]
c-----------------------------------------------------------------------
#define SINGLEPOINT 0
#define DEBUG1 0
#define DEBUG2 0
#define DEBUG3 0


      subroutine dipoles(key,weight,wmeasure)

c-----------------------------------------------------------------------
c...  Real matrix element squared minus the sum of all dipole
c...  contributions. The phase space point is specified by the phegas
c...  common block.
c...  
c...  key - if 1, only the initialization phase is run
c...        if 2, the weight is computed
c...        if 3, same as 2, but helicity sampling is on
c...
c...  weight - difference of the real radiation matrix element squared
c...           and the sum over dipoles
c...
c...  wmeasure - phase space measure, necessary for correct helicity
c...             sampling and histograms
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles.h'
c...  already included in common_dipoles.h
c      include 'dimensions.h'
      include 'common_kt.h'

c...  antiparticles array
      dimension an(-12:44)
      data an/12,11,10,9,8,7,6,5,4,3,2,1,0,
     .     -1,-2,-3,-4,-5,-6,-7,-8,-9,-10,-11,-12,
     .     13,14,15,16,17,18,19,20,
     .     21,22,23,24,25,26,27,28,29,30,
     .     31,32,34,33,35,36,37,38,39,40,41,42,44,43/

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)
c...  pi
      parameter(pi=3.141592653589793d0)

c...  flavours
      dimension iflproc(mxproc,mxpart),iflcross(mxpart)
c...  momenta
      dimension zq(5),p(mxpart,5),q(mxpart,5),qcross(mxpart,5)
      dimension qij(4),qijk(4)

      dimension znlc(4),zvij(2),zym(4,4),zymp1(mxkmax,8,2),
     &          zvijaux(2,2,2)
      double precision colfac
c... cache for Nagy-Soper scheme
      TYPE cache
      integer iseenmap(mxdip),iseenamp(mxdip),iseendip(mxdip)
      double precision ptilde(mxpart,5)
      double complex zampdip(2,mxncc,mxhel),zvij(2,4)
      END TYPE
      TYPE(cache) NScache
      save ndipNS

c...  masses
      double precision mijsq,misq,mjsq,mksq,muijsq,muisq,mujsq,muksq
c...  polarizations
      dimension ipol(mxpart),ipolcross(mxpart)
      dimension rpol(mxpart),rpolcross(mxpart)
      dimension ipolreal(mxhel,mxpart)
c...  process number for a given dipole
      dimension iprocdip(mxdip)
c...  color flows
      dimension jcoldip(0:mxproc,mxncc,mxpart,0:2)
      dimension jcol(mxpart,2),kcol(mxpart,2)
c...  color correlator matrices for real emission and dipoles
      dimension rmatrixdip(0:mxdip,mxncc,mxncc)
c...  Dyson-Schwinger skeletons
      dimension kmaxdip(0:mxproc),nccdip(0:mxproc),imaxdip(0:mxproc)
      dimension list1dip(0:mxproc,mxpart,3)
      dimension listdip(0:mxproc,mxkmax,18,mxncc)
      dimension kcoudip(0:mxproc,mxncc)
      dimension isnew(mxfeyg,8,8) ! pure trash, but needed
c...  averages over color and helicity, and symmetry factor (only the
c...  real emission factors are stored, the second elements are trash)
      dimension avcoldip(0:mxdip),avheldip(0:1),symetdip(0:1)
c...  amplitudes
      dimension zy(mxkmax,8),zamp(mxncc),zampdip(2,mxncc)

c...  random color arrays
      dimension icolorsum(0:3),iflowmap(mxpart)
      dimension icolor(mxpart,2),iflow(mxncc)
      dimension icolordip(mxpart,2),iflowdip(mxncc,0:2),nflowsdip(0:2)
      dimension isoftcolor(2),ipkp(2),ipkplink(2,2)
      dimension coeffdip(mxncc,2)
c...  initial setup
      data nclrnrm/-1/
      save nclrnrm

c...  helicity sampling arrays
      dimension alpha(0:mxhel),channel(0:mxhel),Wi(mxhel),Ni(mxhel)
c...  amplitude for different polarizations. the lower precision (real)
c...  is used to spare some space, and is, of course, sufficient
      real amppol(0:mxdip,mxhel)
c...  perform a fast sum over polarizations separately for the real
c...  emission and dipoles
      logical sumpol
c...  sum over polarizations of the difference of real emission and
c...  corresponding dipoles
      logical optpol
c...  accumulate wights for helicity sampling optimization
      logical accpol
c...  perform helicity sampling
      logical ranpol
c...  start counting events in summation
      logical fixedPS
c...  initial setup
      data nevents/0/,sumpol/.true./,optpol/.false./,accpol/.false./
      data ranpol/.false./,fixedPS/.false./
      save fixedPS

c...  logical setup flags
c...  only the last particle soft/collinear
      logical onlylast
c...  include only dipoles, which compensate divergences
      logical onlydiv
c...  random polarizations for non-partons
      logical hybrid
c...  treatment of contributions with different signs
      integer signmode
c...  type of polarization sum
      integer sumtype
c...  parameters
      double precision kappa
c...  color sampling
      logical colorsampling

c...  setup for self-initialization
      logical firstcall
      save firstcall
      data firstcall/.true./

c...  some local arrays and variables
      character*100 process
      dimension zTijTk(2,2)
      dimension zep(4),zem(4)
      dimension vec(4),zvec(4)
      dimension zekt(2)
      double precision zti,ztj,zp,zm,ztmi,ztmj
      logical match

c...  storage
      save nproc
      save npolreal,ipolreal,iflproc,iprocdip
      save jcoldip,rmatrixdip
      save kmaxdip,nccdip,imaxdip
      save list1dip,listdip,kcoudip
      save avcoldip,avheldip,symetdip
      save nevents,alpha,channel,Wi,Ni
      save amppol,sumtype,nsumpol,nsumpol2,sumpol,noptpol,optpol,nuptpol
      save accpol,ranpol,onlylast,onlydiv,hybrid,signmode,kappa
      save colorsampling
   
#if SINGLEPOINT
      open(99,file='momenta.input',status='old')
      do i=1,n
         read(99,*)(pmom(i,j),j=1,5)
      enddo
      close(99)
#endif     
  
c-----------------------------------------------------------------------
c     INITIALIZATION (TWO PHASES)
c-----------------------------------------------------------------------

      weight=0d0

      if (firstcall) then

c...  read the configuration file
         open(500,file='dipoles.conf',status='old')

         read(500,*)dipoletype
         read(500,*)onlyreal,onlylast,onlydiv,hybrid
         read(500,*)signmode
         read(500,*)sumtype
         read(500,*)nsumpol,noptpol,nuptpol
         read(500,*)alphaMinCut
         read(500,*)alphaMaxII,alphaMaxIF,alphaMaxFI,alphaMaxFF
         read(500,*)kappa
         read(500,*)colorsampling
         read(500,*)ialgorithm
         read(500,*)nbjets
         read(500,*)etamax
         read(500,*)dR
         read(500,*)jetveto,ptveto

         close(500)

         if (nsumpol.lt.1) nsumpol=1

         print*,'--------------------------------------------------'
         print*,'Dipole Subtraction Configuration'
         print*,'--------------------------------------------------'
         if(dipoletype.eq.0) print*,'dipole type:       Catani-Seymour'
         if(dipoletype.eq.1) print*,'dipole type:       Nagy-Soper'
         print*,'only real emission:                      ',onlyreal
         print*,'only last particle soft/collinear:       ',onlylast
         print*,'only divergent dipoles:                  ',onlydiv
         print*,'random polarization for non-partons:     ',hybrid
         print*,'sign mode (0-both,1-positive,2-negative):',signmode

c... random polarization option is for Nagy-Soper scheme only
         if(signmode.eq.3 .AND. dipoletype.ne.1)then
            print*,'dipoles.F ERROR: signmode=3 only allowed in NS'
     .           //' scheme.'
            STOP
         endif

         print*,'helicity sum (0-fast,1-slow,2-flat MC):  ',sumtype
         print*,'events with polarization sum=            ',nsumpol
         print*,'events for sampling optimization=        ',noptpol
         print*,'event increment for sampling update=     ',nuptpol
         print*,'alphaMinCut=                             ',alphaMinCut
         print*,'alphaMaxII=                              ',alphaMaxII
         print*,'alphaMaxIF=                              ',alphaMaxIF
         print*,'alphaMaxFI=                              ',alphaMaxFI
         print*,'alphaMaxFF=                              ',alphaMaxFF
         print*,'kappa=                                   ',kappa
         print*,'color sampling=                         ',colorsampling
         print*,'jet algorithm=                           ',ialgorithm
         print*,'number of b-jets=                        ',nbjets
         print*,'eta_max of clustered partons             ',etamax
         print*,'jet resolution parameter R               ',dR 
         print*,'jet veto included:                       ',jetveto
         print*,'pt of vetoing jet=                       ',ptveto
         print*,'--------------------------------------------------'

c...  polarization sum events are counted from after phase space
c...  optimization, which corresponds to key = 3. we store the
c...  requested original number in nsumpol2
         nsumpol2=nsumpol

c-----------------------------------------------------------------------
c     PHASE ONE                              
c-----------------------------------------------------------------------

c-----------------------------------------------------------------------
c...  dipoles only for QCD particles
c...  QCD particle numbering as in HELAC-PHEGAS:
c...  u d c s t  b  g  
c...  3 4 7 8 11 12 35 
c...  the respective antiparticles have a minus sign
c...  e.g. ubar is -3 
c-----------------------------------------------------------------------

         ndip=0

         if(dipoletype.eq.1)then !... Nagy-Soper
            ndipNS=0
         endif

         nproc=0

c-----------------------------------------------------------------------
c...  1st emitter
c-----------------------------------------------------------------------

         do ip=1,n

c...  only partons
            if (iabs(ifl(ip)).ne.3.and.iabs(ifl(ip)).ne.4.and.
     .          iabs(ifl(ip)).ne.7.and.iabs(ifl(ip)).ne.8.and.
     .          iabs(ifl(ip)).ne.11.and.iabs(ifl(ip)).ne.12.and.
     .           ifl(ip).ne.35) cycle

c...  massless initial states
            if (ip.le.2.and.parmas(ifl(ip)).ne.0d0) then 
               print*,'error: initial states must be massless'
               stop
            endif

c-----------------------------------------------------------------------
c...  2nd emitter with jp > ip
c-----------------------------------------------------------------------

            do jp=ip+1,n

c...  jp emitter always in the final state         
               if (jp.le.2) cycle

c...  only the last particle is soft/collinear
               if (onlylast.and.jp.ne.n) cycle             

c...  only partons
               if (iabs(ifl(jp)).ne.3.and.iabs(ifl(jp)).ne.4.and.
     .             iabs(ifl(jp)).ne.7.and.iabs(ifl(jp)).ne.8.and.
     .             iabs(ifl(jp)).ne.11.and.iabs(ifl(jp)).ne.12.and.
     .              ifl(jp) .ne.35) cycle

c...  gluons split only into a quark-anti-quark pair
               if (ifl(ip).ne.35.and.ifl(jp).ne.35) then
                  if (io(ip)*ifl(ip).ne.-io(jp)*ifl(jp)) cycle
c...  only divergent dipole contributions
                  if (onlydiv.and.parmas(ifl(ip)).ne.0d0) cycle
               endif

c...  initial emitters must be massless
               if (ip.le.2.and.parmas(ifl(jp)).ne.0d0) cycle

               if(dipoletype.eq.1)then !... Nagy-Soper
                  ndipNS=ndipNS+1
               endif

c-----------------------------------------------------------------------
c...  spectator
c-----------------------------------------------------------------------

               do kp=1,n

c...  only partons
                  if (iabs(ifl(kp)).ne.3.and.iabs(ifl(kp)).ne.4.and.
     .                iabs(ifl(kp)).ne.7.and.iabs(ifl(kp)).ne.8.and.
     .                iabs(ifl(kp)).ne.11.and.iabs(ifl(kp)).ne.12.and.
     .                 ifl(kp) .ne.35) cycle
                  if (kp.eq.ip.or.kp.eq.jp) cycle
                  
c...  counting dipoles
                  ndip=ndip+1
                  if (ndip.gt.mxdip) then
                     print*,'maximum number of dipoles is',mxdip
                     print*,'please increase mxdip in dimensions.h'
                     stop
                  endif

c-----------------------------------------------------------------------
c...  storing emitters and spectators
c-----------------------------------------------------------------------

                  ijkdip(ndip,1)=ip
                  ijkdip(ndip,2)=jp
                  ijkdip(ndip,3)=kp

c-----------------------------------------------------------------------
c...  flavour determination
c-----------------------------------------------------------------------

                  ii=0
                  do j=1,n
                     if (j.eq.jp) cycle
                     ii=ii+1
                     ifldip(ndip,ii)=ifl(j)
                  enddo
                  
                  if (io(ip)*ifl(ip).eq.-io(jp)*ifl(jp)) then
                     ifldip(ndip,ip)=35
                  elseif (ifl(ip).eq.35) then
                     ifldip(ndip,ip)=io(ip)*io(jp)*ifl(jp)
                  endif
                  
c-----------------------------------------------------------------------
c...  flavour crossing for process determination
c-----------------------------------------------------------------------

                  iflcross(1:n-1)=ifldip(ndip,1:n-1)
                  if (ip.gt.2) then
                     iflcross(1)=an(ifldip(ndip,ip))
                     iflcross(ip)=an(ifldip(ndip,1))
                  else
                     iflcross(1)=ifldip(ndip,ip)
                     iflcross(ip)=ifldip(ndip,1)
                  endif

c-----------------------------------------------------------------------
c...  check whether crossed process has been seen
c-----------------------------------------------------------------------

                  iseenproc=-1
                  do j=1,nproc
                     match=.true.
                     do k=1,n-1
                        if (iflproc(j,k).ne.iflcross(k)) then
                           match=.false.
                           goto 110
                        endif
                     enddo
 110                 continue
                     if (match) then 
                        iseenproc=j
                        goto 120
                     endif
                  enddo
 120              continue

                  if (iseenproc.ne.-1) then
                     iprocdip(ndip)=iseenproc
                     cycle
                  endif

c...  counting different processes             
                  nproc=nproc+1
                  if (nproc.gt.mxproc) then
                     print*,'maximum number of processes is',mxproc
                     print*,'please increase mxproc in dimensions.h'
                     stop
                  endif
                  iprocdip(ndip)=nproc
                  
c-----------------------------------------------------------------------
c...  storing new process
c-----------------------------------------------------------------------

                  iflproc(nproc,1:n-1)=iflcross(1:n-1)

c-----------------------------------------------------------------------
c...  producing an input file for the underlying process of the dipole
c-----------------------------------------------------------------------

                  call produceinputs(n-1,iflcross,nproc)

               enddo

            enddo

         enddo

c-----------------------------------------------------------------------
c...  input for the real emission process
c-----------------------------------------------------------------------

         call produceinputs(n,ifl,0)

         print*,'--------------------------------------------------'

         if(dipoletype.eq.0)then !...Catani-Seymour
            print*,'Number of Dipoles:  ',ndip
         elseif(dipoletype.eq.1)then !... Nagy-Soper
            print*,'Number of Dipoles:  ',ndipNS
         endif

         print*,'Number of Processes:',nproc
         print*,'--------------------------------------------------'

         if (key.eq.1) return
         
c-----------------------------------------------------------------------
c     PHASE TWO - PROCESSES
c-----------------------------------------------------------------------

c-----------------------------------------------------------------------
c...  reading skeleton from the file for all different processes 
c...  information read once and stored  
c-----------------------------------------------------------------------

         do ipp=0,nproc         ![loop over processes 

c-----------------------------------------------------------------------
c...  finding string for the process
c-----------------------------------------------------------------------

c...  process nr. 0 corresponds to real emission
            if (ipp.eq.0) then
               call getstring(n,ifl,process)
               np=n
            else
               call getstring(n-1,iflproc(ipp,1:n-1),process)
               np=n-1
            endif

            nunit1=123
            open(nunit1,file=
     .           './helac_trees/tree_'//process(2:2*np+1)//'.in'
     .           ,status='old',err=100)
            goto 101

 100        print*,'Could not open tree for ',process(2:2*np+1)
            print*,'Please run "make trees"'
            stop

 101        continue

c-----------------------------------------------------------------------
c...  reading all information from the skeleton
c-----------------------------------------------------------------------

            read(nunit1,*)kmax,ncc,imax
            if (kmax.gt.mxkmax) then
               print*,'maximum number of subamplitudes is',mxkmax
               print*,'please increase mxkmax in dimensions.h'
               stop
            endif
            if (ncc.gt.mxncc) then
               print*,'maximum number of color flows is',mxncc
               print*,'please increase mxncc in dimensions.h'
               stop
            endif

            kmaxdip(ipp)=kmax
            nccdip(ipp)=ncc
            imaxdip(ipp)=imax

            read(nunit1,*)list1dip(ipp,1:np,1:3)
            read(nunit1,*)listdip(ipp,1:kmax,1:18,1:ncc)
            read(nunit1,*)kcoudip(ipp,1:ncc)
            read(nunit1,*)ng  
            if (ng.gt.mxfeyg) then
               print*,'maximum number of graphs is',mxfeyg
               print*,'please increase mxfeyg in dimensions.h'
               stop
            endif

            do ii=1,ng
               do jj=1,8
                  read(nunit1,*)(isnew(ii,jj,kk),kk=1,8)
               enddo
            enddo
            
            if (ipp.eq.0) then
               read(nunit1,*)avheldip(0),avcoldip(0),symetdip(0)
               read(nunit1,*)rmatrixdip(0,1:ncc,1:ncc)
            else
c...  this is trash
               read(nunit1,*)avheldip(1),avcoldip(1),symetdip(1)
               read(nunit1,*)rmatrixdip(1,1:ncc,1:ncc)
            endif
            
            do k=1,ncc
               read(nunit1,*)jcoldip(ipp,k,1:np,0:2)
            enddo

c...  closing skeleton file
            close(nunit1)

         enddo                  !]loop over processes

c-----------------------------------------------------------------------
c     PHASE TWO - COLOR CORRELATORS
c-----------------------------------------------------------------------

c-----------------------------------------------------------------------
c...  loop over dipoles to build a correlator matrix 
c-----------------------------------------------------------------------

         do ipp=1,ndip          ![loop over dipoles

c-----------------------------------------------------------------------
c...  emitters and spectator, process, number of color flows 
c-----------------------------------------------------------------------

            ip=ijkdip(ipp,1)
            jp=ijkdip(ipp,2)
            kp=ijkdip(ipp,3)

            iproc=iprocdip(ipp)

            ncc=nccdip(iproc)
            
c-----------------------------------------------------------------------
c...  Casimir operator of the emitter ij 
c-----------------------------------------------------------------------
            
            if (ifldip(ipp,ip).eq.35) then 
               casimir=3d0 
            else
               casimir=4d0/3d0
            endif
            
c-----------------------------------------------------------------------
c...  color correlator matrix for dipoles
c-----------------------------------------------------------------------
            
            ktp=kp
c...  swap of 1 and ip can induce a change in kp, when kp is 1
            if (kp.eq.1) ktp=ip
c...  removing the line jp shifts kp by 1 if kp > jp
            if (kp.gt.jp) ktp=kp-1

            do i=1,ncc
               do j=1,ncc
                  do k=1,n-1
                     do l=1,2
                        jcol(k,l)=jcoldip(iproc,i,k,l)
                        kcol(k,l)=jcoldip(iproc,j,k,l)
                     enddo
                  enddo
                  rmatrixdip(ipp,i,j) =
     .                 colcor(jcol,kcol,1,ktp,n-1)/casimir
               enddo
            enddo

c-----------------------------------------------------------------------
c     color average
c-----------------------------------------------------------------------

c...  crossing mixes initial and final states, and we need to recompute
c...  this because the value from the skeleton file is wrong
            avcoldip(ipp)=1d0
            do k=1,n-1
               if (io(k).eq. 1) then
                  if (iabs(ifldip(ipp,k)).eq.3.or.
     .                iabs(ifldip(ipp,k)).eq.4.or.
     .                iabs(ifldip(ipp,k)).eq.7.or.
     .                iabs(ifldip(ipp,k)).eq.8.or.
     .                iabs(ifldip(ipp,k)).eq.11.or.
     .                iabs(ifldip(ipp,k)).eq.12)
     .                 avcoldip(ipp)=avcoldip(ipp)/3d0
                  if (ifldip(ipp,k).eq.35)
     .                 avcoldip(ipp)=avcoldip(ipp)/8d0
               endif
            enddo
            
         enddo                  !]loop over dipoles
         
c-----------------------------------------------------------------------
c     end of initialization
c-----------------------------------------------------------------------

      endif

c-----------------------------------------------------------------------
c     PHASE THREE                              
c-----------------------------------------------------------------------

      nevents=nevents+1

c-----------------------------------------------------------------------
c...  phase space points from HELAC via pmom  
c-----------------------------------------------------------------------
c...  p(1:n,1:5) momenta of particles involved in the process
c...  1 - x component
c...  2 - y component
c...  3 - z component
c...  4 - energy
c...  5 - mass
c-----------------------------------------------------------------------

      do i=1,n
         p(i,1:5) = pmom(i,1:5)
      enddo

c-----------------------------------------------------------------------
c...  color setup
c-----------------------------------------------------------------------

      if (.not.firstcall.and.colorsampling) then

c...  generate random color configuration
 200     do i=1,n
            if (iabs(ifl(i)).eq.3.or.iabs(ifl(i)).eq.4.or.
     .          iabs(ifl(i)).eq.7.or.iabs(ifl(i)).eq.8.or.
     .          iabs(ifl(i)).eq.11.or.iabs(ifl(i)).eq.12) then
               if (io(i)*ifl(i).gt.0) then
                  icolor(i,1)=3*rnmy(0)+1
                  icolor(i,2)=0
               else
                  icolor(i,1)=0
                  icolor(i,2)=3*rnmy(0)+1
               endif
            else if (ifl(i).eq.35) then
               icolor(i,1)=3*rnmy(0)+1
               icolor(i,2)=3*rnmy(0)+1
            endif
         enddo

c...  check for color conservation
         icolorsum(0:3)=0
         do i=1,n
            icolorsum(icolor(i,1))=icolorsum(icolor(i,1))+1
            icolorsum(icolor(i,2))=icolorsum(icolor(i,2))-1
         enddo
         do i=1,3
            if (icolorsum(i).ne.0) goto 200
         enddo

c...  map into enumeration elements of the second row of color flows
         iflowmap(1:n)=0
         k=0
         do i=1,n
            if (icolor(i,2).ne.0) then
               k=k+1
               iflowmap(k)=i
            endif
         enddo
         
c...  determine compatible color flows
         nflows=0
         do i=1,nccdip(0)
            do j=1,n
               k=jcoldip(0,i,j,1)
               if (k.eq.0) cycle
               k=iflowmap(k)
               if (icolor(j,1).ne.icolor(k,2)) goto 210
            enddo
            nflows=nflows+1
            iflow(nflows)=i
 210        continue
         enddo
         
c...  determine normalization
         if (nclrnrm.lt.0) nclrnrm=nclrcfns(icolor,n)
         
      endif

c-----------------------------------------------------------------------
c...  helicity setup
c-----------------------------------------------------------------------

c...  polarization sum events are counted from after phase space
c...  optimization, which corresponds to key = 3
      if (key.eq.2) then
         nsumpol=nevents
      else if (.not.fixedPS) then
         print*,''
         print*,'Allowing helicity sampling optimization after',
     .          nsumpol2,'events from now on'
         print*,''
         nsumpol=nsumpol2+nevents
         fixedPS=.true.
      endif

c...  hybrid initialization
      do i=1,n
         if (hybrid.and.
     .        iabs(ifl(i)).ne.3.and.iabs(ifl(i)).ne.4.and.
     .        iabs(ifl(i)).ne.7.and.iabs(ifl(i)).ne.8.and.
     .        iabs(ifl(i)).ne.11.and.iabs(ifl(i)).ne.12.and.
     .        ifl(i).ne.35) then
            rpol(i)=2*pi*rnmy(0)
         else
            rpol(i)=-1
         endif

c...  random polarization (for Nagy-Soper scheme)
         if(dipoletype.eq.1 .AND.sumtype.eq.3)then
            if(.NOT.firstcall)then
               rpol(i)=2*pi*rnmy(0)
               ipol(i)=1
            endif
         endif

      enddo

c...  initialization (helicity configurations, sum type)
      if (nevents.eq.2) then
         i=0
         npolreal=0
         ipol(1:n)=1
         do
            i=i+1
            if (amppol(0,i).gt.1e-15) then
               npolreal=npolreal+1
               ipolreal(npolreal,1:n)=ipol(1:n)
            endif
            if (nextpol(n,ifl,ipol,rpol).eq.0) exit
         enddo
         print*,'--------------------------------------------------'
         print*,'Number of Helicity Configurations:  ',npolreal
         print*,'--------------------------------------------------'
         if (sumtype.eq.1) then
            sumpol=.false.
            optpol=.true.
         else if (sumtype.eq.2) then
            alpha(1:npolreal)=1d0/npolreal
            channel(0)=0d0
            do i=1,npolreal
               channel(i)=channel(i-1)+alpha(i)
            enddo
            sumpol=.false.
            ranpol=.true.
c...  random polarization case (for Nagy-Soper scheme)
         else if(sumtype.eq.3)then
            if(dipoletype.ne.1)then
               print*,'dipoles.F ERROR: option sumtype=3'
               print*,'is allowed for the Nagy-Soper scheme only!'
               STOP
            endif
            sumpol=.false.
            optpol=.false.
            accpol=.false.
            ranpol=.false.
         endif
      endif
      
c...  variance table initialization for helicity sampling optimization
      if (nevents.eq.nsumpol+1) then
         Wi(1:npolreal)=0d0
         Ni(1:npolreal)=0
         sumpol=.false.
         optpol=.true.
         ranpol=.false.
         accpol=.true.
c...  random polarization case (for Nagy-Soper scheme)
         if(dipoletype.eq.1.AND.sumtype.eq.3)then
            sumpol=.false.
            optpol=.false.
            accpol=.false.
            ranpol=.false.
         endif
      endif

c...  from this point on only random helicity
      if (nevents.eq.nsumpol+noptpol+1) then
         optpol=.false.
         ranpol=.true.
c...  random polarization case (for Nagy-Soper scheme)
         if(dipoletype.eq.1.AND.sumtype.eq.3)then
            sumpol=.false.
            optpol=.false.
            accpol=.false.
            ranpol=.false.
         endif
      endif

c...  helicity sampling optimization
      if (nevents.gt.nsumpol+noptpol.and.
     .     mod(nevents-(nsumpol+noptpol+1),nuptpol).eq.0) then
c...  avoid multichannel adaptation in case of random polarization (Nagy-Soper)
         if(.NOT.(dipoletype.eq.1.AND.sumtype.eq.3))then
            alpha(1:npolreal)=dsqrt(Wi(1:npolreal)/Ni(1:npolreal))
            channel(0)=0d0
            do i=1,npolreal
               channel(i)=channel(i-1)+alpha(i)
            enddo
            alpha(1:npolreal)=alpha(1:npolreal)/channel(npolreal)
            channel(1:npolreal)=channel(1:npolreal)/channel(npolreal)
         endif
      endif

c...  random helicity
      alpha(0)=1d0
      if (ranpol) then
         r=rnmy(0)
         do npol=1,npolreal
            if (r.gt.channel(npol-1).and.r.le.channel(npol)) exit
         enddo
         ipol(1:n)=ipolreal(npol,1:n)
         alpha(0)=alpha(npol)
      endif

c-----------------------------------------------------------------------
c...  begin polarization sampling optimization
c-----------------------------------------------------------------------

      if (optpol) npol=0

 300  if (optpol) then
         npol=npol+1
         ipol(1:n)=ipolreal(npol,1:n)
      endif

c-----------------------------------------------------------------------
c...  real emission matrix element squared
c-----------------------------------------------------------------------

      RealSum=0d0

c...  initialize external particle wave functions in m+1 phase space:
      if(dipoletype.eq.1) then  !... Nagy-Soper
c...  with 0
         zymp1(1:n,1:8,1:2)=dcmplx(0d0,0d0)
         if(sumpol) ipol(1:n)=1
 301     call iniqqdip(n,ifl,io,p,ipol,rpol,list1dip,0,zy)
         if(sumpol) then
            do i1=1,n
c...  store external particle wave functions for both helicities
c...  --> use nextpol to get all functions
               zymp1(i1,1:4,ipol(i1))=zy(i1,1:4)
            enddo
         else
            do i1=1,n
c...  store external particle wave functions for given helicity
               zymp1(i1,1:4,1)=zy(i1,1:4)
            enddo
         endif
         if(sumpol) then
            if(nextpol(n,ifl,ipol,rpol).ne.0) then
               goto 301
            endif
         endif
      endif

c...  jet function
      call FJmpo(n,ifl,p,icut)
      if (.not.firstcall.and.icut.eq.0) goto 320

c...  coupling constant
      call couplingjet(n,ifl,p,gqcd,wsf)

c...  initialize the counter and polarizations
      if (sumpol) then
         npol=0
         ipol(1:n)=1
      endif

c...  skip vanishing contributions during summation over polarizations
 310  if (sumpol) then
         npol=npol+1
         if (npol.gt.mxhel) then
            print*,'maximum number of helicity configs is',mxhel
            print*,'please increase mxhel in dimensions.h'
            stop
         endif
         if (.not.firstcall.and.amppol(0,npol).lt.1e-15) then
            if (nextpol(n,ifl,ipol,rpol).ne.0) then
               goto 310
            else
               goto 320
            endif
         endif
      endif

c...  amplitude
      kmax=kmaxdip(0)
      ncc=nccdip(0)
      imax=imaxdip(0)

      call iniqqdip(n,ifl,io,p,ipol,rpol,list1dip,0,zy)

      if (firstcall.or..not.colorsampling) nflows=ncc
      do k=1,nflows
         if (.not.firstcall.and.colorsampling) then
            kk=iflow(k)
         else
            kk=k
         endif
         call nextqdip(n,kk,imax,listdip,kcoudip,0,zy)
         call ampqdip(kk,ifl,zy,listdip,kcoudip,0,zamp)
      enddo
      
      RealME=0d0
      do k1=1,nflows
         do k2=1,nflows
            if (.not.firstcall.and.colorsampling) then
               kk1=iflow(k1)
               kk2=iflow(k2)
               coeff=nclrnrm
            else
               kk1=k1
               kk2=k2
               coeff=rmatrixdip(0,k1,k2)
            endif
            RealME=RealME+coeff*zamp(kk1)*dconjg(zamp(kk2))
         enddo
      enddo
      RealME=RealME*avheldip(0)*avcoldip(0)*symetdip(0)

c-----------------------------------------------------------------------  
c...  pdf for dynamical scale implementation                                 
c-----------------------------------------------------------------------  

      RealME=RealME*wsf

c...  call histogramming routine for real emission
      if (.not.firstcall.and.key.eq.3) then
         if (onlyreal) then
c...  in this case there are no jets defined, so we need the leading
c...  order histogramming routine
            call HISTOm(n,ifl,p,RealME*wmeasure/alpha(0))
         else
            call HISTOmpo(n,ifl,p,RealME*wmeasure/alpha(0))
         endif
      endif

      RealSum = RealSum + RealME

c...  store normalized amplitude values for all polarizations
      if (sumpol) then
         if (firstcall) amppol(0,npol)=RealME
         if (nextpol(n,ifl,ipol,rpol).ne.0) goto 310
         if (firstcall) then
            ampmax=0d0
            do i=1,npol
               if (amppol(0,i).gt.ampmax) ampmax=amppol(0,i)
            enddo
            amppol(0,1:npol)=amppol(0,1:npol)/ampmax
         endif
      endif
      
 320  continue

c-----------------------------------------------------------------------
c...  begin sum over dipoles
c-----------------------------------------------------------------------

      DipSum = 0d0

      if (onlyreal) goto 399

c...  cache for Nagy-Soper scheme
      if(dipoletype.eq.1)then   !...Nagy-Soper
c...  determine number of spectators
         nspectators=ndip/ndipNS
         if(nspectators.le.0)then
            print*,'dipoles.F ERROR: check nspectators'
            STOP
         endif
c...  initialize cache arrays
         NScache%iseenmap(1:mxdip)=0
         NScache%iseenamp(1:mxdip)=0
         NScache%iseendip(1:mxdip)=0
         if(firstcall)then
            NScache%zampdip(1:2,1:mxncc,1:mxhel)=0
            NScache%zvij(1:2,1:4)=0
         endif
      endif


      do ipp=1,ndip             ![loop over dipoles 

         ip=ijkdip(ipp,1)
         jp=ijkdip(ipp,2)
         kp=ijkdip(ipp,3)
         iproc=iprocdip(ipp)

c-----------------------------------------------------------------------
c...  defining mass**2 of i,j emitters and k spectator
c-----------------------------------------------------------------------

         misq=parmas(ifl(ip))**2
         mjsq=parmas(ifl(jp))**2
         mksq=parmas(ifl(kp))**2

c-----------------------------------------------------------------------
c...  defining mass**2 of ij emitter
c-----------------------------------------------------------------------

         mijsq=parmas(ifldip(ipp,ip))**2
         
c-----------------------------------------------------------------------
c...  dimensionless kinematic variable x (xijk or yijk)
c-----------------------------------------------------------------------

         call variable(p,ip,jp,kp,misq,mjsq,mijsq,x)

c...  alpha cutoffs
         pij=2*dot(p,ip,jp)
         pik=2*dot(p,ip,kp)
         pjk=2*dot(p,jp,kp)


         if(dipoletype.eq.0) then !... Catani-Seymour phase-space restriction

            if ((ip .le. 2) .and. (kp .le. 2)) then      
               vtildei=pij/pik
               if (.not.firstcall.and.vtildei.gt.alphaMaxII) cycle
            elseif ((ip .le. 2) .and. (kp .gt. 2)) then      
               ui=pij/(pij+pik)
               if (.not.firstcall.and.ui.gt.alphaMaxIF) cycle
            elseif ((ip .gt. 2) .and. (kp .le. 2)) then
               if (.not.firstcall.and.(1-x).gt.alphaMaxFI) cycle
            elseif ((ip .gt. 2) .and. (kp .gt. 2)) then      
               if (.not.firstcall.and.x.gt.alphaMaxFF) cycle
            endif

         elseif(dipoletype.eq.1) then !... Nagy-Soper phase-space restriction

c...  in the present version, no phase-space restriction is implemented for the
c...  Nagy-Soper scheme
            alphaMaxII=1d0
            alphaMaxIF=1d0
            alphaMaxFI=1d0
            alphaMaxFF=1d0

         endif


c-----------------------------------------------------------------------
c...  ptilde kinematics stored in the array q for each dipole
c-----------------------------------------------------------------------

         if(dipoletype.eq.0)then !... Catani-Seymour

            call ptilde(n-2,io,p,x,ip,jp,kp,mksq,mijsq,q)

         elseif(dipoletype.eq.1)then !...Nagy-Soper

c...  cache for Nagy-Soper scheme
c...  compute NS mapping only if not yet done
            ippNS=((ipp-1)/nspectators)+1
            if(NScache%iseenmap(ippNS).ne.1)then
               call ptildeNS(n-2,io,p,x,ip,jp,kp,mksq,mijsq,znlc,q)
               NScache%iseenmap(ippNS)=1
            endif

         endif

c...  jet function
         call FJm(n-1,ifldip(ipp,1:n-1),q,icut)
         if (.not.firstcall.and.icut.eq.0) cycle

c...  coupling constant
         call coupling(n-1,ifldip(ipp,1:n-1),q,gqcd,wsf)
         
c-----------------------------------------------------------------------
c...  crossing 
c-----------------------------------------------------------------------

c...  flavours
         iflcross(1:n-1)=iflproc(iproc,1:n-1)

c...  momentum crossing
         do i=1,n-1
            qcross(i,1:5)=q(i,1:5)
         enddo

         if (ip.gt.2) then
            qcross(1,1:4)=-q(ip,1:4)
            qcross(ip,1:4)=-q(1,1:4)
         else
            qcross(1,1:4)=q(ip,1:4)
            qcross(ip,1:4)=q(1,1:4)
         endif
         qcross(1,5)=q(ip,5)
         qcross(ip,5)=q(1,5)

c...  helicity crossing
c...  we ignore the helicity of the emitter, since it has to be computed
c...  with two values. notice also that despite crossing, helicity
c...  remains unchanged (because crossing here is basically only a swap
c...  of two particles)
         ii=0
         do j=1,n
            if (j.eq.jp) cycle
            ii=ii+1
            ipolcross(ii)=ipol(j)
            rpolcross(ii)=rpol(j)
         enddo

         ipolcross(ip)=ipolcross(1)
         rpolcross(ip)=rpolcross(1)

c-----------------------------------------------------------------------
c... compatible color flows
c-----------------------------------------------------------------------

         if (firstcall.or..not.colorsampling) goto 327

         icolordip(1,1:2)=icolor(ip,1:2)
         if (ifl(ip).eq.35.or.ifl(jp).eq.35) then
c...  choose soft gluon and copy color
            isoft=jp
            if (ifl(jp).ne.35) isoft=ip
            if (ifl(ip).eq.ifl(jp).and.p(ip,4).lt.p(jp,4)) isoft=ip
            if (isoft.eq.ip) icolordip(1,1:2)=icolor(jp,1:2)
            isoftcolor(1:2)=icolor(isoft,1:2)
         else
c...  copy color from the quarks
            if (icolor(jp,1).ne.0) icolordip(1,1)=icolor(jp,1)
            if (icolor(jp,2).ne.0) icolordip(1,2)=icolor(jp,2)
         endif

c...  map rest of color from n-space to n-1-space (crossing!)
         do i=2,n-1
            if (i.ne.ip) then
               if (i.lt.jp) then
                  icolordip(i,1:2)=icolor(i,1:2)
               else
                  icolordip(i,1:2)=icolor(i+1,1:2)
               endif
            endif
         enddo
         if (ip.ne.1) icolordip(ip,1:2)=icolor(1,1:2)

c...  emitter and spectator in n-1-space (emitter is always 1)
         ipkp(1)=1
         ipkp(2)=kp
         if (kp.eq.1) ipkp(2)=ip
         if (kp.gt.jp) ipkp(2)=ipkp(2)-1

c...  map into enumeration elements of the second row of color flows
         iflowmap(1:n-1)=0
         k=0
         do i=1,n-1
c...  all color flows have the same second raw, so use first flow
            if (jcoldip(iproc,1,i,2).ne.0) then
               k=k+1
               iflowmap(k)=i
            endif
         enddo

c...  determine compatible color flows
         nflowsdip(0:2)=0
         do i=1,nccdip(iproc)

c...  check all colors but those of ip and kp, and at the same time find
c...  the positions of the particles linked to ip and kp by color flow
c...
c...  ipkplink(i,j) - position of the particle linked by color flow
c...  i - 1 for ip, 2 for kp
c...  j - 1 for up, 2 for down color flow
            do j=1,n-1
               k=jcoldip(iproc,i,j,1)
               if (k.eq.0) cycle
               k=iflowmap(k)

               icheckflow=1
               if (j.eq.ipkp(1)) then
                  ipkplink(1,1)=k
                  icheckflow=0
               endif
               if (j.eq.ipkp(2)) then
                  ipkplink(2,1)=k
                  icheckflow=0
               endif
               if (k.eq.ipkp(1)) then
                  ipkplink(1,2)=j
                  icheckflow=0
               endif
               if (k.eq.ipkp(2)) then
                  ipkplink(2,2)=j
                  icheckflow=0
               endif

               if (icheckflow.eq.1.and.icolordip(j,1).ne.icolordip(k,2))
     &              goto 325

            enddo

c...  check the colors of ip and kp

            if (ifl(ip).ne.35.and.ifl(jp).ne.35) then
c...  gluon splitting into quarks

c...  first check for normal color conservation
               do j=1,2
                  do k=1,2
                     l=icolordip(ipkp(j),k)
                     if (l.ne.0) then
                        if (l.ne.icolordip(ipkplink(j,k),3-k)) goto 326
                     endif
                  enddo
               enddo

               nflowsdip(0)=nflowsdip(0)+1
               iflowdip(nflowsdip(0),0)=i

               nflowsdip(1)=nflowsdip(1)+1
               iflowdip(nflowsdip(1),1)=i
               coeffdip(nflowsdip(1),1)=1

 326           continue

c...  check for color conservation after insertion of the correlator
               coeff=0
               do j=1,2
                  do k=1,2
c...  check colors of the remaining lines (if they are not the same as
c...  the transformed lines)
                     i2=icolordip(ipkp(1),3-j)
                     il2=icolordip(ipkplink(1,3-j),j)
                     if (ipkplink(1,3-j).ne.ipkp(1).and.
     &                   ipkplink(1,3-j).ne.ipkp(2).and.
     &                   i2.ne.il2) cycle
                     k2=icolordip(ipkp(2),3-k)
                     if (k2.ne.0) then
                        kl2=icolordip(ipkplink(2,3-k),k)
                        if (ipkplink(2,3-k).ne.ipkp(1).and.
     &                      ipkplink(2,3-k).ne.ipkp(2).and.
     &                      k2.ne.kl2) cycle
                     endif
c...  check colors of the transformed lines
                     i1=icolordip(ipkp(1),j)
                     il1=icolordip(ipkplink(1,j),3-j)
                     k1=icolordip(ipkp(2),k)
                     if (k1.eq.0) cycle
                     kl1=icolordip(ipkplink(2,k),3-k)
                     if (j.eq.k) then
                        if (i1.eq.kl1.and.k1.eq.il1) coeff=coeff+1
                     else if (i1.eq.k1) then
                        if (ipkp(1).eq.ipkplink(2,k)) then
                           coeff=coeff-3
                        else if (il1.eq.kl1) then
                           coeff=coeff-1
                        endif
                     endif
                  enddo
               enddo

               if (coeff.ne.0) then
                  if (nflowsdip(0).eq.0.or.
     &                iflowdip(nflowsdip(0),0).ne.i) then
                     nflowsdip(0)=nflowsdip(0)+1
                     iflowdip(nflowsdip(0),0)=i
                  endif
                  nflowsdip(2)=nflowsdip(2)+1
                  iflowdip(nflowsdip(2),2)=i
                  coeffdip(nflowsdip(2),2)=coeff
               endif

            else
c...  gluon emission from quark or gluon lines

               do j=1,2
                  coeff=0
                  do k=1,2
c...  color of the lines of the emitter (ip or kp depending on j)
                     i1=icolordip(ipkp(j),k)
                     i2=icolordip(ipkp(j),3-k)
                     if (i1.ne.0) il1=icolordip(ipkplink(j,k),3-k)
                     if (i2.ne.0) il2=icolordip(ipkplink(j,3-k),k)
c...  color of the lines of the spectator
                     k1=icolordip(ipkp(3-j),k)
                     k2=icolordip(ipkp(3-j),3-k)
                     if (k1.ne.0) kl1=icolordip(ipkplink(3-j,k),3-k)
                     if (k2.ne.0) kl2=icolordip(ipkplink(3-j,3-k),k)
c...  color conservation on lines not involving the soft gluon
                     if (i1.eq.0.or.
     &                   (i2.ne.0.and.i2.ne.il2).or.
     &                   (k1.ne.0.and.k1.ne.kl1)) cycle
c...  singlet contribution
                     if (isoftcolor(1).eq.isoftcolor(2).and.
     &                   i1.eq.il1.and.(k2.eq.0.or.k2.eq.kl2))
     &                    coeff=coeff-(3-2*k)/3d0
c...  color conservation on the lines, which may be connected to emitter
                     if (k2.ne.0) then
                        if (ipkplink(3-j,3-k).eq.ipkp(j)) then
                           if (k2.ne.isoftcolor(k)) cycle
                        else
                           if (k2.ne.kl2) cycle
                        endif
                     endif
c...  octet contribution
                     if (i1.eq.isoftcolor(3-k).and.il1.eq.isoftcolor(k))
     &                    coeff=coeff+(3-2*k)
                  enddo
c...  update flow lists
                  if (coeff.ne.0) then
                     if (nflowsdip(0).eq.0.or.
     &                   iflowdip(nflowsdip(0),0).ne.i) then
                        nflowsdip(0)=nflowsdip(0)+1
                        iflowdip(nflowsdip(0),0)=i
                     endif
                     nflowsdip(j)=nflowsdip(j)+1
                     iflowdip(nflowsdip(j),j)=i
                     coeffdip(nflowsdip(j),j)=coeff
                  endif
               enddo

            endif

 325        continue
         enddo

 327     continue

c-----------------------------------------------------------------------
c... color correlators
c-----------------------------------------------------------------------

         zTijTk(1:2,1:2)=0d0

c...  initialize the counter and polarizations
         if (sumpol) then
            npol=0
            ipolcross(1:n)=1
         endif

c...  cache for Nagy-Soper scheme
         if(sumpol)then
            npolnonzero=0
         else
            npolnonzero=1
         endif

c...  skip vanishing contributions during summation over polarizations
 330     if (sumpol) then
            npol=npol+1
c...  cache for Nagy-Soper scheme
            if(.not.firstcall.and.amppol(ipp,npol).gt.1e-15) then
               npolnonzero=npolnonzero+1
            endif

            if (.not.firstcall.and.amppol(ipp,npol).lt.1e-15) then
c...  hack to have the first polarization untouched
               rpolcross(1)=+1
               if (nextpol(n-1,iflcross,ipolcross,rpolcross).ne.0) then
                  goto 330
               else
                  goto 340
               endif
            endif
         endif

         kmax=kmaxdip(iproc)
         ncc=nccdip(iproc)
         imax=imaxdip(iproc)

c...  cache for Nagy-Soper scheme
c...  compute matrix elements for the given mapping only if not yet done
         if(dipoletype.eq.1)then !... Nagy-Soper
            ippNS=((ipp-1)/nspectators)+1
            if(NScache%iseenamp(ippNS).eq.1)then
               goto 341
            endif
         endif

         call iniqqdip(n-1,iflcross,io,qcross,ipolcross,rpolcross,
     .        list1dip,iproc,zy)
         
         if (firstcall.or..not.colorsampling) nflowsdip(0:2)=ncc
         do k=1,nflowsdip(0)     ![loop over color flows
            if (.not.firstcall.and.colorsampling) then
               kk=iflowdip(k,0)
            else
               kk=k
            endif
            call nextqdip(n-1,kk,imax,listdip,kcoudip,iproc,zy)
            do ii=1,2           ![loop over helicities of the ij emitter
               ipolcross(1)=ii
               rpolcross(1)=-1
               call iniqqdip(1,iflcross,io,qcross,ipolcross,
     .              rpolcross,list1dip,iproc,zy)

c...  store all m phase space external particle wave functions:
               if(dipoletype.eq.1.and.k.eq.1)then !... Nagy-Soper
                  zym(ii,1:4)=zy(1,1:4)
               endif

               call ampqdip(kk,iflcross,zy,listdip,kcoudip,iproc,zamp)
               zampdip(ii,kk)=zamp(kk)

c...  cache for Nagy-Soper scheme
c...  store matrix element after computation
               if(dipoletype.eq.1)then !... Nagy-Soper
                  NScache%zampdip(ii,kk,npolnonzero)=zampdip(ii,kk)
               endif

            enddo               !]end loop over helicities of ij emitter
         enddo                  !]end loop over color flows


 341     continue
c...  cache for Nagy-Soper scheme
c...  read matrix element from cache if already computed
         if(dipoletype.eq.1 .AND. NScache%iseenamp(ippNS).eq.1)then
            zampdip(1:2,1:mxncc) = 
     .           NScache%zampdip(1:2,1:mxncc,npolnonzero)
         endif


c...  calculating the color correlator
 400     CorrMax=0d0
         do i1=1,2
            do i2=i1,2

               zCorr=0d0
               do k1=1,nflowsdip(1)
                  do k2=1,nflowsdip(2)
                     if (.not.firstcall.and.colorsampling) then
                        kk1=iflowdip(k1,1)
                        kk2=iflowdip(k2,2)
                        coeff=coeffdip(k1,1)*coeffdip(k2,2)*nclrnrm/2d0
                        if (ifldip(ipp,ip).eq.35) then
                           coeff=coeff/3d0
                        else
                           coeff=coeff/(4/3d0)
                        endif
                     else
                        kk1=k1
                        kk2=k2
                        coeff=rmatrixdip(ipp,k1,k2)
                     endif
                     zCorr=zCorr
     .                    +coeff*zampdip(i1,kk1)*dconjg(zampdip(i2,kk2))
                  enddo
               enddo

               zCorr=zCorr*avheldip(0)*avcoldip(ipp)*symetdip(0)
               if (abs(zCorr).gt.CorrMax) CorrMax=abs(zCorr)

               zTijTk(i1,i2)=zTijTk(i1,i2)+zCorr

            enddo
         enddo

c...  store normalized absolute correlator values for all polarizations
         if (sumpol) then
            if (firstcall) amppol(ipp,npol)=CorrMax
c...  hack to have the first polarization untouched
            rpolcross(1)=+1
            if (nextpol(n-1,iflcross,ipolcross,rpolcross).ne.0) goto 330
            if (firstcall) then
               ampmax=0d0
               do i=1,npol
                  if (amppol(ipp,i).gt.ampmax) ampmax=amppol(ipp,i)
               enddo
               if (ampmax.ne.0d0)
     .              amppol(ipp,1:npol)=amppol(ipp,1:npol)/ampmax
            endif
         endif
      
 340     continue

c...  cache for Nagy-Soper scheme
c...  at this point, summation over helicities (if set) is over and
c...  one can safely set the NS cache flag for the amplitude
         if(dipoletype.eq.1 .AND. .NOT.firstcall)then !... Nagy-Soper
            if(.NOT.colorsampling)then
               ippNS=((ipp-1)/nspectators)+1
               NScache%iseenamp(ippNS)=1
            endif
         endif

         zTijTk(2,1)=dconjg(zTijTk(1,2))
         
c-----------------------------------------------------------------------
c...  calculation of the polarisation vectors for the gluon emitter ij.
c...  The kT(mu) vector is also calculated together c...  with the
c...  scalar product epsilon.kT for each case ff, fi, if, ii  
c-----------------------------------------------------------------------

c...  polarization vectors of an emitter gluon
         if (ifldip(ipp,ip).eq.35) then
            pt=dsqrt(q(ip,1)**2+q(ip,2)**2)
            pq=dsqrt(pt**2+q(ip,3)**2)
            
            zq(1)=dcmplx(q(ip,4) +q(ip,3),q(ip,3))
            zq(2)=dcmplx(q(ip,4) -q(ip,3),pt)
            zq(3)=dcmplx(q(ip,1), q(ip,2))
            zq(4)=dcmplx(q(ip,1),-q(ip,2))
            zq(5)=dcmplx(parmas(ifl(ip)), pq)
            
            call eminus(zem,zq)
            call  eplus(zep,zq)

c...  store polarization vectors for the Nagy-Soper scheme:
            if(dipoletype.eq.1) then !... Nagy-Soper
               zym(3,1:4)=zem(1:4)
               zym(4,1:4)=zep(1:4)
            endif

         endif

c-----------------------------------------------------------------------
c...  begin summation over polarizations of the emitter pair
c-----------------------------------------------------------------------
         
         if (sumpol) then
            ipol(ip)=1
            ipol(jp)=1
         endif

c...  cache for Nagy-Soper scheme
         if(dipoletype.eq.1)then !... Nagy-Soper
            if(sumpol)then
               npolnonzero=0
            else
               npolnonzero=1
            endif
         endif

 350     continue


         if(dipoletype.eq.0) then ! Catani-Seymour

c-----------------------------------------------------------------------
c...  initial-initial
c-----------------------------------------------------------------------
  
         if ((ip .le. 2) .and. (kp .le. 2)) then      

c...  emitter is a gluon
            if (ifldip(ipp,ip).eq.35) then

               vec(1:4)=p(jp,1:4)+vtildei*p(kp,1:4)
               
               zvec(1)=dcmplx(vec(4) +vec(3),0d0)
               zvec(2)=dcmplx(vec(4) -vec(3),0d0)
               zvec(3)=dcmplx(vec(1), vec(2))
               zvec(4)=dcmplx(vec(1),-vec(2))
               
               zekt(1)=zprod(zem,zvec)
               zekt(2)=zprod(zep,zvec)
               
               spincor=0d0
               do i1=1,2
                  do i2=1,2
                     spincor=spincor
     .                    +dconjg(zekt(i1))*zekt(i2)*zTijTk(i1,i2)
                  enddo
               enddo

c...  g -> gg
               if (ifl(ip).eq.35) then
                  
                  Dgg=
     .                 +zTijTk(ipol(ip),ipol(ip))*
     .                 (1-2*x+1/(1-x))
     .                 +zTijTk(3-ipol(jp),3-ipol(jp))*
     .                 (-2+x*(1-x)) 
                  
                  if (ipol(ip).eq.ipol(jp))
     .                 Dgg=Dgg+2*(1-x)/(x*vtildei)*spincor/pjk
     .                 +2*x*(zTijTk(1,1)+zTijTk(2,2))
                  
                  Dgg=Dgg*2*gqcd**2*CA
                  Dgg=-Dgg/pij/x
                  DipME=Dgg
#if DEBUG1
                  print*,'Dgg(',ip,',',jp,',',kp,') = ',Dgg
#endif

c...  g -> qq
               else

                  Dqq=0d0

                  if (ipol(ip).eq.ipol(jp)) then
                     Dqq=(zTijTk(1,1)+zTijTk(2,2))*(x-1)
                     Dqq=Dqq+zTijTk(ipol(ip),ipol(ip))*(2-x)
                     Dqq=Dqq+2*(1-x)/(x*vtildei)*spincor/pjk
                  endif

                  Dqq=Dqq*2*gqcd**2*CF
                  Dqq=-Dqq/pij/x
                  DipME=Dqq
#if DEBUG1
                     print*,'Dqq(',ip,',',jp,',',kp,') = ',Dqq
#endif

               endif

c...  emitter is a quark
            else

c...  q -> qg
               if (ifl(jp).eq.35) then
                  
                  Dqg=1/(1-x)-(1+x)
                  if (ipol(ip).eq.ipol(jp)) Dqg = Dqg+1+x
                  Dqg=Dqg*zTijTk(ipol(ip),ipol(ip))

                  Dqg=Dqg*2*gqcd**2*CF
                  Dqg=-Dqg/pij/x
                  DipME=Dqg
#if DEBUG1
                  print*,'Dqg(',ip,',',jp,',',kp,') = ',Dqg
#endif
                  
c...  q -> gq
               elseif (ifl(ip).eq.35) then
                  
                  Dgq=x**2
                  if (ipol(ip).eq.ipol(jp)) Dgq = Dgq+1-2*x
                  Dgq =Dgq*zTijTk(3-ipol(jp),3-ipol(jp))

                  Dgq =Dgq*2*gqcd**2*TF
                  Dgq =-Dgq/pij/x
                  DipME=Dgq
#if DEBUG1
                  print*,'Dgq(',ip,',',jp,',',kp,') = ',Dgq
#endif

               endif

            endif
            
c-----------------------------------------------------------------------
c...  initial-final 
c-----------------------------------------------------------------------

         elseif ((ip .le. 2) .and. (kp .gt. 2)) then      
            
c...  emitter is a gluon
            if (ifldip(ipp,ip).eq.35) then
               
               vec(1:4)=p(jp,1:4)/ui-p(kp,1:4)/(1-ui)
               
               zvec(1)=dcmplx(vec(4) +vec(3),0d0)
               zvec(2)=dcmplx(vec(4) -vec(3),0d0)
               zvec(3)=dcmplx(vec(1), vec(2))
               zvec(4)=dcmplx(vec(1),-vec(2))
               
               zekt(1)=zprod(zem,zvec)
               zekt(2)=zprod(zep,zvec)
               
               spincor=0d0
               do i1=1,2
                  do i2=1,2
                     spincor=spincor
     .                    +dconjg(zekt(i1))*zekt(i2)*zTijTk(i1,i2)
                  enddo
               enddo

c...  g -> gg
               if (ifl(ip).eq.35) then
                  
                  Dgg=
     .                 +zTijTk(ipol(ip),ipol(ip))*
     .                 (1-2*x+1/(1-x+ui))
     .                 +zTijTk(3-ipol(jp),3-ipol(jp))*
     .                 (-2+x*(1-x)) 
                  
                  if (ipol(ip).eq.ipol(jp))
     .                 Dgg=Dgg+2*(1-x)/x*ui*(1-ui)*spincor/pjk
     .                 +2*x*(zTijTk(1,1)+zTijTk(2,2))
                  
                  Dgg=Dgg*2*gqcd**2*CA
                  Dgg=-Dgg/pij/x
                  DipME=Dgg
#if DEBUG1
                  print*,'Dgg(',ip,',',jp,',',kp,') = ',Dgg
#endif

c...  g -> qq
               else

                  Dqq=0d0

                  if (ipol(ip).eq.ipol(jp)) then
                     Dqq=Dqq+(zTijTk(1,1)+zTijTk(2,2))*(x-1)
                     Dqq=Dqq+zTijTk(ipol(ip),ipol(ip))*(2-x)
                     Dqq=Dqq+2*(1-x)/x*ui*(1-ui)*spincor/pjk
                  endif

                  Dqq=Dqq*2*gqcd**2*CF
                  Dqq=-Dqq/pij/x
                  DipME=Dqq
#if DEBUG1
                     print*,'Dqq(',ip,',',jp,',',kp,') = ',Dqq
#endif

               endif

c...  emitter is a quark
            else
               
c...  q -> qg
               if (ifl(jp).eq.35) then
                  
                  Dqg=1/(1-x+ui)-(1+x)
                  if (ipol(ip).eq.ipol(jp)) Dqg = Dqg+1+x
                  Dqg=Dqg*zTijTk(ipol(ip),ipol(ip))

                  Dqg=Dqg*2*gqcd**2*CF
                  Dqg=-Dqg/pij/x
                  DipME=Dqg
#if DEBUG1
                  print*,'Dqg(',ip,',',jp,',',kp,') = ',Dqg
#endif
                  
c...  q -> gq
               elseif (ifl(ip).eq.35) then

                  Dgq=x**2
                  if (ipol(ip).eq.ipol(jp)) Dgq = Dgq+1-2*x
                  Dgq =Dgq*zTijTk(3-ipol(jp),3-ipol(jp))

                  Dgq =Dgq*2*gqcd**2*TF
                  Dgq =-Dgq/pij/x
                  DipME=Dgq
#if DEBUG1
                  print*,'Dgq(',ip,',',jp,',',kp,') = ',Dgq
#endif

               endif

            endif

c-----------------------------------------------------------------------
c...  final-initial
c-----------------------------------------------------------------------

         elseif ((ip .gt. 2) .and. (kp .le. 2)) then
            
            qij(1:4)=p(ip,1:4)+p(jp,1:4)
            qijsq=qij(4)**2-qij(1)**2-qij(2)**2-qij(3)**2

            zti=pik/(pjk+pik)
            ztj=1-zti

c...  emitter is a gluon
            if (ifldip(ipp,ip).eq.35) then

               vec(1:4)=zti*p(ip,1:4)-ztj*p(jp,1:4)

               zvec(1)=dcmplx(vec(4) +vec(3),0d0)
               zvec(2)=dcmplx(vec(4) -vec(3),0d0)
               zvec(3)=dcmplx(vec(1), vec(2))
               zvec(4)=dcmplx(vec(1),-vec(2))
               
               zekt(1)=zprod(zem,zvec)
               zekt(2)=zprod(zep,zvec)
               
               spincor=0d0
               do i1=1,2
                  do i2=1,2
                     spincor=spincor
     .                    +zekt(i1)*dconjg(zekt(i2))*zTijTk(i1,i2)
                  enddo
               enddo
               
c...  g -> gg
               if (ifl(ip).eq.35) then
                  
                  Dgg=
     .                 +zTijTk(ipol(ip),ipol(ip))*
     .                 (1-2*zti+1/(1-zti+(1-x)))
     .                 +zTijTk(ipol(jp),ipol(jp))*
     .                 (1-2*ztj+1/(1-ztj+(1-x))) 
                  
                  if (ipol(ip).ne.ipol(jp))
     .                 Dgg=Dgg+2*spincor/pij-2*zTijTk(1,1)
     .                 -2*zTijTk(2,2)
                  
                  Dgg=Dgg*2*gqcd**2*CA
                  Dgg=-Dgg/pij/x
                  DipME=Dgg
#if DEBUG1
                  print*,'Dgg(',ip,',',jp,',',kp,') = ',Dgg
#endif

c...  g -> qq
               else

                  Dqq=zTijTk(ipol(ip),ipol(ip))/zti*misq/qijsq
     .               +zTijTk(ipol(jp),ipol(jp))/ztj*misq/qijsq
                  
                  if (ipol(ip).ne.ipol(jp)) then
                     Dqq=Dqq+zTijTk(ipol(ip),ipol(ip))*zti
     .                  +zTijTk(ipol(jp),ipol(jp))*ztj
                     Dqq=Dqq-2*spincor/qijsq
                     Dqq=Dqq-(zTijTk(1,1)+zTijTk(2,2))*(1/zti+1/ztj)*
     .                   misq/qijsq
                  endif
                  
                  Dqq=Dqq*2*gqcd**2*TF
                  Dqq=-Dqq/qijsq/x
                  DipME=Dqq
#if DEBUG1
                  print*,'Dqq(',ip,',',jp,',',kp,') = ',Dqq
#endif

               endif

c...  emitter is a quark
            else

c...  q -> qg
               if (ifl(jp).eq.35) then
                  
                  Dqg=1/(1-zti+(1-x))
                  if (ipol(ip).ne.ipol(jp)) Dqg=Dqg-(1+zti*(1+misq/pij))
                  if (ipol(ip).eq.ipol(jp)) Dqg=Dqg-misq/pij/zti
                  Dqg=Dqg*zTijTk(ipol(ip),ipol(ip))
                  
                  if (ipol(ip).ne.ipol(jp))
     .                 Dqg=Dqg-zTijTk(ipol(jp),ipol(jp))*(2-zti-1/zti)*
     .                 misq/pij
                  
                  Dqg=Dqg*2*gqcd**2*CF
                  Dqg=-Dqg/(qijsq-mijsq)/x
                  DipME=Dqg
#if DEBUG1
                  print*,'Dqg(',ip,',',jp,',',kp,') = ',Dqg
#endif
                  
c...  q -> gq
               elseif (ifl(ip).eq.35) then
                  
                  Dgq=1/(1-ztj+(1-x))
                  if (ipol(jp).ne.ipol(ip)) Dgq=Dgq-(1+ztj*(1+mjsq/pij))
                  if (ipol(jp).eq.ipol(ip)) Dgq=Dgq-mjsq/pij/ztj
                  Dgq=Dgq*zTijTk(ipol(jp),ipol(jp))
                  
                  if (ipol(jp).ne.ipol(ip))
     .                 Dgq=Dgq-zTijTk(ipol(ip),ipol(ip))*(2-ztj-1/ztj)*
     .                 mjsq/pij
                  
                  Dgq=Dgq*2*gqcd**2*CF
                  Dgq=-Dgq/(qijsq-mijsq)/x
                  DipME=Dgq
#if DEBUG1
                  print*,'Dgq(',ip,',',jp,',',kp,') = ',Dgq
#endif

               endif

            endif

c-----------------------------------------------------------------------
c...  final-final
c-----------------------------------------------------------------------

         elseif ((ip .gt. 2) .and. (kp .gt. 2)) then      

            qij(1:4)=p(ip,1:4)+p(jp,1:4)
            qijsq=qij(4)**2-qij(1)**2-qij(2)**2-qij(3)**2

            qijk(1:4)=qij(1:4)+p(kp,1:4)
            Qsq=qijk(4)**2-qijk(1)**2-qijk(2)**2-qijk(3)**2

            muisq=misq/Qsq
            mujsq=mjsq/Qsq
            muksq=mksq/Qsq
            muijsq=mijsq/Qsq

            viji=dsqrt(((1-mujsq-muisq-muksq)*x)**2-4*muisq*mujsq)
     .           /((1-mujsq-muisq-muksq)*x+2*muisq)
            
            vijk=dsqrt((2*muksq+(1-muisq-mujsq-muksq)*(1-x))**2-4*muksq)
     .           /((1-mujsq-muisq-muksq)*(1-x))
            
            zp=(2*muisq+(1-muisq-mujsq-muksq)*x)
     .         /(2*(muisq+mujsq+(1-muisq-mujsq-muksq)*x)) 
            zm=zp*(1-viji*vijk)
            zp=zp*(1+viji*vijk)

            zti=pik/(pjk+pik)
            ztj=1-zti

            ztmi=zti-0.5d0+0.5d0*vijk
            ztmj=ztj-0.5d0+0.5d0*vijk

c...  emitter is a gluon
            if (ifldip(ipp,ip).eq.35) then

               vec(1:4)=ztmi*p(ip,1:4)-ztmj*p(jp,1:4)
               
               zvec(1)=dcmplx(vec(4) +vec(3),0d0)
               zvec(2)=dcmplx(vec(4) -vec(3),0d0)
               zvec(3)=dcmplx(vec(1), vec(2))
               zvec(4)=dcmplx(vec(1),-vec(2))

               zekt(1)=zprod(zem,zvec)
               zekt(2)=zprod(zep,zvec)
               
               spincor=0d0
               do i1=1,2
                  do i2=1,2
                     spincor=spincor+
     .                    zekt(i1)*dconjg(zekt(i2))*zTijTk(i1,i2)
                  enddo
               enddo
               
c...  g -> gg
               if (ifl(ip).eq.35) then
                  
                  Dgg=
     .                 +zTijTk(ipol(ip),ipol(ip))*
     .                 ((1-2*zti)/vijk+1/(1-zti*(1-x)))
     .                 +zTijTk(ipol(jp),ipol(jp))*
     .                 ((1-2*ztj)/vijk+1/(1-ztj*(1-x))) 
                  
                  if (ipol(ip).ne.ipol(jp))
     .                 Dgg=Dgg+2*spincor/pij/vijk
     .                 -2*(zTijTk(1,1)+zTijTk(2,2))/vijk

c...  not fixed by the limits, so it's added to all polarizations
                  Dgg=Dgg+(zTijTk(1,1)+zTijTk(2,2))*kappa*zp*zm/vijk/2

                  Dgg=Dgg*2*gqcd**2*CA
                  Dgg=-Dgg/pij
                  DipME=Dgg
#if DEBUG1
                  print*,'Dgg(',ip,',',jp,',',kp,') = ',Dgg
#endif

c...  g -> qq
               else

                  Dqq=zTijTk(ipol(ip),ipol(ip))/zti*misq/qijsq
     .               +zTijTk(ipol(jp),ipol(jp))/ztj*misq/qijsq
                  
                  if (ipol(ip).ne.ipol(jp)) then
                     Dqq=Dqq+zTijTk(ipol(ip),ipol(ip))*zti
     .                  +zTijTk(ipol(jp),ipol(jp))*ztj
                     Dqq=Dqq-2*spincor/qijsq
                     Dqq=Dqq-(zTijTk(1,1)+zTijTk(2,2))*(1/zti+1/ztj)*
     .                   misq/qijsq
                  endif
                  
c...  not fixed by the limits, so it's added to all polarizations
                  Dqq=Dqq-(zTijTk(1,1)+zTijTk(2,2))*kappa*
     .                (zp*zm-misq/qijsq)/2
                  
                  Dqq=Dqq/vijk
                  
                  Dqq=Dqq*2*gqcd**2*TF
                  Dqq=-Dqq/qijsq
                  DipME=Dqq
#if DEBUG1
                  print*,'Dqq(',ip,',',jp,',',kp,') = ',Dqq
#endif

               endif

c...  emitter is a quark
            else
               
               vtijk=sqrt((1-muijsq-muksq)**2-4*muijsq*muksq)
     .              /(1-muijsq-muksq)
               
c...  q -> qg
               if (ifl(jp).eq.35) then
                  
                  Dqg=1/(1-zti*(1-x))*vijk/vtijk
                  if (ipol(ip).ne.ipol(jp)) Dqg=Dqg-(1+zti*(1+misq/pij))
                  if (ipol(ip).eq.ipol(jp)) Dqg=Dqg-misq/pij/zti
                  Dqg=Dqg*zTijTk(ipol(ip),ipol(ip))
                  
                  if (ipol(ip).ne.ipol(jp))
     .                 Dqg=Dqg-zTijTk(ipol(jp),ipol(jp))*(2-zti-1/zti)*
     .                 misq/pij
                  
                  Dqg=Dqg*vtijk/vijk
                  
                  Dqg=Dqg*2*gqcd**2*CF
                  Dqg=-Dqg/(qijsq-mijsq)
                  DipME=Dqg
#if DEBUG1
                  print*,'Dqg(',ip,',',jp,',',kp,') = ',Dqg
#endif
                  
c...  q -> gq
               elseif (ifl(ip).eq.35) then
                  
                  Dgq=1/(1-ztj*(1-x))*vijk/vtijk
                  if (ipol(jp).ne.ipol(ip)) Dgq=Dgq-(1+ztj*(1+mjsq/pij))
                  if (ipol(jp).eq.ipol(ip)) Dgq=Dgq-mjsq/pij/ztj
                  Dgq=Dgq*zTijTk(ipol(jp),ipol(jp))
                  
                  if (ipol(jp).ne.ipol(ip))
     .                 Dgq=Dgq-zTijTk(ipol(ip),ipol(ip))*(2-ztj-1/ztj)*
     .                 mjsq/pij
                  
                  Dgq=Dgq*vtijk/vijk
                  
                  Dgq=Dgq*2*gqcd**2*CF
                  Dgq=-Dgq/(qijsq-mijsq)
                  DipME=Dgq
#if DEBUG1
                  print*,'Dgq(',ip,',',jp,',',kp,') = ',Dgq
#endif

               endif

            endif

         endif

         endif                     ! end: Catani-Seymour



         if(dipoletype.eq.1) then  ! Nagy-Soper Dipoles 

c...  initialize variables with 0
            vsoftsqjp=0d0
            vsoftsqip=0d0

c...  splitting functions for collinear and soft contributions:
c...  1. Diagonal terms

c...  cache for Nagy-Soper scheme
c...  compute diagonal term only if not yet done, otherwise read from cache
            ippNS=((ipp-1)/nspectators)+1
            if(NScache%iseendip(ippNS).ne.1)then
               CALL iniSPLfunc(ip,jp,sumpol,ipol,rpol,ipp,p,q,mijsq,
     &              znlc,zymp1,zym,zvijaux(1,ipol(ip),ipol(jp)))
               NScache%zvij(1:2,npolnonzero)=
     .              zvijaux(1:2,ipol(ip),ipol(jp))
            else
               zvijaux(1:2,ipol(ip),ipol(jp))=
     .              NScache%zvij(1:2,npolnonzero)
            endif

c...  2. Interference terms (soft contributions)
            iAklswitch=2  
            if (ifl(jp).eq.35) CALL iniSPLfuncSoft(ip,jp,kp,sumpol,
     &           ipol,rpol,p,zymp1,iAklswitch,vsoftsqjp) 
            if(ifl(ip).eq.35.and.ip.ge.3) CALL iniSPLfuncSoft(jp,ip,kp,
     &           sumpol,ipol,rpol,p,zymp1,iAklswitch,vsoftsqip)

c...  initialize color factors
            if (ifldip(ipp,ip).eq.35) then
c...  g -> g g
               if (ifl(ip).eq.35) then
                  colfac=gqcd**2*CA
c...  g -> q qbar
               else
                  if (ip.le.2) colfac=gqcd**2*CF 
                  if (ip.gt.2) colfac=gqcd**2*TF 
               endif
            else
c...  q -> q g         
               if (ifl(jp).eq.35) then
                  colfac=gqcd**2*CF
               else
c...  q -> g q
                  if (ip.le.2) colfac=gqcd**2*TF 
                  if (ip.gt.2) colfac=gqcd**2*CF
               endif
            endif
c...  initialize variables with 0
            DipME    =0d0
c... sum over polarizations (helicities) in m_space
            do i1=1,2
               do i2=1,2

c...  add diagonal terms
                  DipME=DipME-DREAL(zvijaux(i1,ipol(ip),ipol(jp))
     &                 *dconjg(zvijaux(i2,ipol(ip),ipol(jp)))
     &                 *zTijTk(i1,i2))

c...  random polarization (for Nagy-Soper scheme) 
c...  if random polarization is on, interference terms must be added in a 
c...  different way (done later)
                  if(dipoletype.eq.1.AND.sumtype.eq.3) cycle

c...  add interference terms
                  if ((i1.eq.i2).and.i1.eq.ipol(ip))
     &                 DipME=DipME+DREAL(vsoftsqjp*zTijTk(i1,i2))
                  if ((i1.eq.i2).and.i1.eq.ipol(jp))
     &                 DipME=DipME+DREAL(vsoftsqip*zTijTk(i1,i2))

               enddo
            enddo

c...  random polarization (for Nagy-Soper scheme)
c...  add interference terms for the random polarization case
            if(dipoletype.eq.1.AND.sumtype.eq.3)then
               zi=dcmplx(0.d0,1.d0)
               phi=2.d0*rpol(ip)
               DipME=DipME+vsoftsqjp*(zTijTk(1,1)+zTijTk(2,2)+
     .              (dcos(phi)+zi*dsin(phi))*zTijTk(1,2)+
     .              (dcos(phi)-zi*dsin(phi))*zTijTk(2,1))
               phi=2.d0*rpol(jp)
               DipME=DipME+vsoftsqip*(zTijTk(1,1)+zTijTk(2,2)+
     .              (dcos(phi)+zi*dsin(phi))*zTijTk(1,2)+
     .              (dcos(phi)-zi*dsin(phi))*zTijTk(2,1))
            endif


            DipME=colfac*DipME

       endif                       ! end: Nagy-Soper
c end +NS         
        
c-----------------------------------------------------------------------
c...  pdf for dynamical scale implementation
c-----------------------------------------------------------------------

c         call pdf_reweight(n-1,ifldip(ipp,1:n-1),q,reweight)
       DipME = DipME*wsf

c-----------------------------------------------------------------------
c...  end summation over polarizations of the emitter pair
c-----------------------------------------------------------------------

c...  call histogramming routine for dipoles
         if (.not.firstcall.and.key.eq.3)
     .     call HISTOm(n-1,ifldip(ipp,1:n-1),q,-DipME*wmeasure/alpha(0))

         DipSum = DipSum + DipME

         if (sumpol) then

c...  cache for Nagy-Soper scheme
            npolnonzero=npolnonzero+1

            if (ipol(jp).eq.2.and.ipol(ip).eq.2) goto 360
            ipol(ip)=3-ipol(ip)
            if (ipol(ip).eq.1) ipol(jp)=3-ipol(jp)
            goto 350

         endif

 360     continue

c...  cache for Nagy-Soper scheme
c...  at this point, summation over helicities (if set) is over and
c...  one can safely set the NS cache flag for the diagonal term
         if(dipoletype.eq.1 .AND. .NOT.firstcall)then !... Nagy-Soper
            ippNS=((ipp-1)/nspectators)+1
            NScache%iseendip(ippNS)=1
         endif

      enddo                     !]loop over dipoles
#if DEBUG2
         print*,''
         print*,ipol(1:n)
         print*,'Real emission |M|^2 =',RealSum
         print*,'Sum over dipoles D  =',DipSum
         print*,'Difference |M|^2-D  =',RealSum-DipSum
         print('(1P,A18,E14.2)'),'(|M|^2-D)/|M|^2 =',
     .        (RealSum-DipSum)/RealSum
#endif


c-----------------------------------------------------------------------
c...  end sum over dipoles
c-----------------------------------------------------------------------

 399  w = RealSum - DipSum

      if (signmode.ne.0) then
         if (signmode.eq.2) w=-w
         if (w.lt.0d0) w=0d0
      endif

      weight = weight + w/alpha(0)
      
c-----------------------------------------------------------------------
c...  end polarization sampling optimization
c-----------------------------------------------------------------------

      if (accpol) then
         Wi(npol)=Wi(npol)+(w*wmeasure)**2
         Ni(npol)=Ni(npol)+1
      endif

      if (optpol.and.npol.lt.npolreal) goto 300

c-----------------------------------------------------------------------
c...  initialization is done only once !
c-----------------------------------------------------------------------

      if (firstcall) firstcall=.false.
      

#if SINGLEPOINT
      print*,'gqcd:',gqcd
      print*,'momenta:x,y,z,E,m'
      do i=1,n
         write(6,'(5d24.16)'), (pmom(i,j),j=1,5)
      enddo
      print*,'value:',RealSum,DipSum,weight
      print*,''
      stop
#endif

      return
      end

c-----------------------------------------------------------------------

      subroutine variable(p,ip,jp,kp,misq,mjsq,mijsq,x)

c-----------------------------------------------------------------------
c...  dimensionless kinematic variables xijk and yijk 
c...  x is xijk for i-i, i-f,f-i and yijk for f-f
c...  smooth interpolation between soft and collinear limits 
c...  ip labels the emitter parton    
c...  jp labels the emitted parton 
c...  kp labels the spectator parton  
c...  misq labels mass**2 of ip emitter
c...  mjsq labels mass**2 of jp emitter
c...  mksq labels mass**2 of kp spectator
c...  mijsq labels mass**2 of ij emitter 
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension p(mxpart,5)
      double precision mijsq, misq, mjsq

c-----------------------------------------------------------------------
c...  all momenta are taken as incoming 
c-----------------------------------------------------------------------

      pij=2*dot(p,ip,jp)
      pik=2*dot(p,ip,kp)
      pjk=2*dot(p,jp,kp)

c-----------------------------------------------------------------------
c...  initial state singularities with initial state partons
c-----------------------------------------------------------------------

      if ((ip .le. 2) .and. (kp .le. 2)) then
         omx=(pij+pjk)/pik
         x=1-omx

c-----------------------------------------------------------------------
c...  initial state singularities with final state partons
c-----------------------------------------------------------------------

      elseif ((ip .le. 2) .and. (kp .gt. 2)) then
         omx=pjk/(pij+pik)
         x=1-omx

c-----------------------------------------------------------------------
c...  final state singularities with initial state partons
c-----------------------------------------------------------------------

      elseif ((ip .gt. 2) .and. (kp .le. 2)) then
         omx=-(mijsq-misq-mjsq-pij)/(pjk+pik)
         x=1-omx

c-----------------------------------------------------------------------
c...  final state singularities with final state partons
c-----------------------------------------------------------------------

      elseif ((ip .gt. 2) .and. (kp .gt. 2)) then
         x=pij/(pij+pjk+pik)

      endif

      return
      end

c-----------------------------------------------------------------------

      subroutine ptilde(n,io,p,x,ip,jp,kp,mksq,mijsq,q)

c-----------------------------------------------------------------------
c...  Taken from MCFM
c...
************************************************************************
*     Author: R.K. Ellis                                               *
*     June, 2002.                                                      *
*     Given p (-p1 + -p2 --> p3 ... px .. p_(npart+2))                 *
*     produce q (-q1 + -q2 --> q3 ... qx .. q_(npart+1))               *
*     by Lorentz transformation with jp denoting the vector            *
*     which is removed (ie all components if q(jp) set to zero)        *
*     ip is the emitter, kp is the spectator                           *
*     Correct branch chosen automatically                              *
*     x is x for ii,if,fi and y for ff                                 *
************************************************************************
c...  
c...  Modified to match the momentum directions (in/out) of HELAC
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension io(mxpart)
      dimension p(mxpart,5),q(mxpart,5),BigQ(4),pij(4)
      double precision kDk,ksDks,mksq,mijsq
      double precision k(4),kt(4),ks(4)
      double precision kDp(3:mxpart),ksDp(3:mxpart)
      dimension pktilde(4)

c...  all momenta outgoing
      do i=1,n+2
         if (io(i).gt.0) p(i,1:4)=-p(i,1:4)
      enddo

      ipart=1
      do j=1,n+2
         if (j.ne.jp) then
            q(ipart,5)=p(j,5)
            ipart=ipart+1
         endif
      enddo
      q(ip,5)=dsqrt(mijsq)

c-----------------------------------------------------------------------
c...  initial-initial
c-----------------------------------------------------------------------

      if ((ip .le. 2) .and. (kp .le. 2)) then      
         do nu=1,4
            q(ip,nu)=x*p(ip,nu)
            q(kp,nu)=p(kp,nu)
            k(nu) =-p(ip,nu)-p(kp,nu)-p(jp,nu)
            kt(nu) =-x*p(ip,nu)-p(kp,nu)
            ks(nu)=k(nu)+kt(nu)
         enddo

         kDk=k(4)**2-k(1)**2-k(2)**2-k(3)**2
         ksDks=ks(4)**2-ks(1)**2-ks(2)**2-ks(3)**2
         ipart=3
         do j=3,n+2
            if (j .eq. jp) then
               go to 18
            else
               kDp(j)=k(4)*p(j,4)-k(1)*p(j,1)-k(2)*p(j,2)-k(3)*p(j,3)
               ksDp(j)=ks(4)*p(j,4)-ks(1)*p(j,1)
     .              -ks(2)*p(j,2)-ks(3)*p(j,3)
               do nu=1,4
                  q(ipart,nu)=p(j,nu)-2*ksDp(j)*ks(nu)/ksDks
     .                 +2*kDp(j)*kt(nu)/kDk
               enddo
               ipart=ipart+1
            endif
 18         continue
         enddo

c-----------------------------------------------------------------------
c...  initial-final 
c-----------------------------------------------------------------------

      elseif ((ip .le. 2) .and. (kp .gt. 2)) then      
         ipart=1
         omx=1-x
         do j=1,n+2
            do nu=1,4
               if (j.eq.ip) then
                  q(ipart,nu)=x*p(ip,nu)
               elseif (j.eq.jp) then
                  goto 19
               elseif (j.eq.kp) then
                  q(ipart,nu)=p(jp,nu)+p(kp,nu)+omx*p(ip,nu)
               else
                  q(ipart,nu)=p(j,nu)
               endif
            enddo
            ipart=ipart+1
 19         continue
         enddo

c-----------------------------------------------------------------------
c...  final-initial
c-----------------------------------------------------------------------

      elseif ((ip .gt. 2) .and. (kp .le. 2)) then        
         ipart=1
         omx=1-x
         do j=1,n+2
            do nu=1,4
               if (j.eq.kp) then
                  q(ipart,nu)=x*p(kp,nu)
               elseif (j.eq.jp) then
                  goto 20
               elseif (j.eq.ip) then
                  q(ipart,nu)=p(ip,nu)+p(jp,nu)+omx*p(kp,nu)
               else
                  q(ipart,nu)=p(j,nu)
               endif
            enddo
            ipart=ipart+1
 20         continue
         enddo

c-----------------------------------------------------------------------
c...  final-final
c-----------------------------------------------------------------------

      elseif ((ip .gt. 2) .and. (kp .gt. 2)) then      
         do nu=1,4
            BigQ(nu)=p(ip,nu)+p(jp,nu)+p(kp,nu)
            pij(nu)=p(ip,nu)+p(jp,nu)
         enddo   
         Qsq=BigQ(4)**2-BigQ(1)**2-BigQ(2)**2-BigQ(3)**2
         QDpk=
     .        +BigQ(4)*p(kp,4)-BigQ(1)*p(kp,1)
     .        -BigQ(2)*p(kp,2)-BigQ(3)*p(kp,3)
         pijsq=pij(4)**2-pij(1)**2-pij(2)**2-pij(3)**2
         rat=dsqrt((Qsq-mijsq-mksq)**2-4d0*mijsq*mksq)
         rat=rat/dsqrt((Qsq-pijsq-mksq)**2-4d0*pijsq*mksq)
         ipart=1
         do j=1,n+2
            do nu=1,4
               pktilde(nu)=rat*(p(kp,nu)-QDpk/Qsq*BigQ(nu))
     .              +(Qsq+mksq-mijsq)/(2d0*Qsq)*BigQ(nu)
               if (j.eq.ip) then
                  q(ipart,nu)=BigQ(nu)-Pktilde(nu)
               elseif (j.eq.jp) then
                  goto 21
               elseif (j.eq.kp) then
                  q(ipart,nu)=pktilde(nu)
               else
                  q(ipart,nu)=p(j,nu)
               endif
            enddo
            ipart=ipart+1
 21         continue
         enddo 
      endif

c...  momenta in/out-going
      do i=1,n+2
         if (io(i).gt.0) p(i,1:4)=-p(i,1:4)
      enddo

      do i=1,n+1
         if (io(i).gt.0) q(i,1:4)=-q(i,1:4)
      enddo
      
      end
      
c-----------------------------------------------------------------------

      double precision function dot(p,i,j)

c-----------------------------------------------------------------------
c...  scalar product of four-vector p(i,1:4), p(j,1:4)
c-----------------------------------------------------------------------
      
      include 'declare.h'
      include 'dimensions.h'

      double precision p(mxpart,4)

      dot=p(i,4)*p(j,4)-p(i,1)*p(j,1)-p(i,2)*p(j,2)-p(i,3)*p(j,3)
      return
      end



************************************************************************
      SUBROUTINE ptildeNS(n,io,p,x,ip,jp,kp,mksq,mijsq,znlc,q)
************************************************************************
c... Nagy-Soper subtraction scheme
c... KINEMATICS
c... Modified to match the momentum directions (in/out) of HELAC
c-----------------------------------------------------------------------
      include 'declare.h'
      include 'dimensions.h'
      include 'common_strf.h'

      dimension io(mxpart)
      dimension p(mxpart,5),q(mxpart,5),BigQ(4),pij(4)
      double precision kDk,ksDks,mksq,mijsq
      double precision k(4),kt(4),ks(4)
      double precision kDp(3:mxpart),ksDp(3:mxpart)
      double precision BigR(4),BigP(4),nvec(4)
      double precision alpha,beta,Psq,QP,Qptil,Rsq
      double precision ptil(4)
      dimension znlc(4)
      double precision eta

c... initialize q=0=nvec, znlc=(0,0)
      do i=1,n+2
         q(i,1:4)=0d0
      enddo
      nvec(1:4)=0d0
      znlc(1:4)=dcmplx(0d0,0d0)
c... all momenta outgoing 
      ipart=1
      do j=1,n+2
         if (j.ne.jp) then
            q(ipart,5)=p(j,5)
            ipart=ipart+1
         endif
      enddo
      q(ip,5)=dsqrt(mijsq)
    
c-----------------------------------------------------------------------
c... INITIAL
c-----------------------------------------------------------------------
      if (ip.le.2) then
         k(1:4)=p(1,1:4)+p(2,1:4)-p(jp,1:4) 
         kDk=k(4)**2-k(1)**2-k(2)**2-k(3)**2
         BigQ(1:4)=(p(1,1:4)+p(2,1:4))
         Qsq=BigQ(4)**2-BigQ(1)**2-BigQ(2)**2-BigQ(3)**2
         eta=kDk/Qsq
         if(ip.eq.1)then
            eta=kDk/(Qsq/xp1)
            q(2,1:4) =p(2,1:4)
            q(ip,1:4)=eta*p(ip,1:4)/xp1
            kt(1:4)  =q(ip,1:4)+q(2,1:4)
         elseif(ip.eq.2)then
            eta=kDk/(Qsq/xp2)
            q(1,1:4) =p(1,1:4)
            q(ip,1:4)=eta*p(ip,1:4)/xp2
            kt(1:4)  =q(ip,1:4)+q(1,1:4)
         endif
         ipart=3
c... lightlike vector n in light-cone coord. used in splitting 
c... functions
         if(ip.eq.1) then
            znlc(1)=dcmplx(p(2,4)+p(2,3),0d0)/xp2
            znlc(2)=dcmplx(p(2,4)-p(2,3),0d0)/xp2
            znlc(3)=dcmplx(p(2,1),p(2,2))/xp2
            znlc(4)=dcmplx(p(2,1),-p(2,2))/xp2
         else
            znlc(1)=dcmplx(p(1,4)+p(1,3),0d0)/xp1
            znlc(2)=dcmplx(p(1,4)-p(1,3),0d0)/xp1
            znlc(3)=dcmplx(p(1,1),p(1,2))/xp1
            znlc(4)=dcmplx(p(1,1),-p(1,2))/xp1
         endif
      endif
c-----------------------------------------------------------------------
c... FINAL
c-----------------------------------------------------------------------
      if (ip .gt. 2) then
         BigP(1:4)=p(ip,1:4)+p(jp,1:4)
         BigQ(1:4)=(p(1,1:4)+p(2,1:4))

         Psq=BigP(4)**2-BigP(1)**2-BigP(2)**2-BigP(3)**2
         Qsq=BigQ(4)**2-BigQ(1)**2-BigQ(2)**2-BigQ(3)**2
         QP =BigQ(4)*BigP(4)-BigQ(1)*BigP(1)-BigQ(2)*BigP(2)
     &      -BigQ(3)*BigP(3)

         BigR(1:4)=BigP(1:4)-QP/Qsq*BigQ(1:4)
         Rsq=BigR(4)**2-BigR(1)**2-BigR(2)**2-BigR(3)**2

         alpha=(Psq-mijsq)/(2d0*QP)
         beta=dsqrt((1d0-alpha)**2+((1d0-alpha)**2*Psq-mijsq)/(-Rsq))
     &        -(1d0-alpha)
         
         ptil(1:4)=(1d0-alpha)*BigP(1:4)+beta*BigR(1:4)

         k(1:4)=BigQ(1:4)-BigP(1:4)
         kt(1:4)=(BigQ(1:4)-ptil(1:4))

         ipart=1
      endif
c-----------------------------------------------------------------------
      ks(1:4)= k(1:4)+kt(1:4)
      kDk=k(4)**2-k(1)**2-k(2)**2-k(3)**2
      ksDks=ks(4)**2-ks(1)**2-ks(2)**2-ks(3)**2
      
      if(ipart.eq.3) then
         do j=3,n+2
            if (j.eq.jp) cycle
            kDp(j)=k(4)*p(j,4)-k(1)*p(j,1)-k(2)*p(j,2)-k(3)*p(j,3)
            ksDp(j)=ks(4)*p(j,4)-ks(1)*p(j,1)-ks(2)*p(j,2)-ks(3)*p(j,3)
            q(ipart,1:4)=p(j,1:4)-2d0*ksDp(j)*ks(1:4)/ksDks+2d0*kDp(j)
     &                  *kt(1:4)/kDk
            ipart=ipart+1
         enddo
      else
         do j=1,n+2
            if (j .eq. jp) cycle
            if (j.le.2) then
               q(ipart,1:4)=p(j,1:4)
            elseif(j.eq.ip) then
               q(ipart,1:4)=ptil(1:4)
            elseif(j.gt.2) then
               if(n+2.eq.5) then
                  q(ipart,1:4)=kt(1:4)
               else
                  kDp(j)=k(4)*p(j,4)-k(1)*p(j,1)-k(2)*p(j,2)-k(3)*p(j,3)
                  ksDp(j)=ks(4)*p(j,4)-ks(1)*p(j,1)-ks(2)*p(j,2)-ks(3)
     &                   *p(j,3)
                  q(ipart,1:4)=p(j,1:4)-2d0*ksDp(j)*ks(1:4)/ksDks+2d0
     &                        *kDp(j)*kt(1:4)/kDk
               endif
            endif
            ipart=ipart+1
         enddo
c... vector n
         Qptil =BigQ(4)*ptil(4)-BigQ(1)*ptil(1)-BigQ(2)*ptil(2)
     &         -BigQ(3)*ptil(3)
         if(mijsq.lt.10d-10) then
            nvec(1:4)=BigQ(1:4)-Qsq/(2d0*Qptil)*ptil(1:4)
         else
            nvec(1:4)=BigQ(1:4)-Qsq/(Qptil+dsqrt(Qptil**2-Qsq*mijsq))
     &               *ptil(1:4)
         endif
c... convert into light-cone coordinates
         znlc(1)=dcmplx(nvec(4)+nvec(3),0d0)
         znlc(2)=dcmplx(nvec(4)-nvec(3),0d0)
         znlc(3)=dcmplx(nvec(1),nvec(2))
         znlc(4)=dcmplx(nvec(1),-nvec(2))
      endif

      END

************************************************************************
      SUBROUTINE effSpinor(mass,znlc,zBigP,zsp,sw,zeffsp)
************************************************************************
c... calculates an effective spinor in light-cone represenation
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      double precision mass
      logical sw
      dimension znlc(4),zBigP(4),zsp(4),zeffsp(4)

c...  U_eff = (P_sl + m) . n_sl . U 
      if(sw) then
         zeffsp(1)=(zBigP(1)*znlc(2)-zBigP(4)*znlc(3))*zsp(1)+(zBigP(4)
     &        *znlc(1)-zBigP(1)*znlc(4))*zsp(2)-mass*znlc(1)*zsp(3)
     &        -mass*znlc(4)*zsp(4)
         zeffsp(2)=(zBigP(3)*znlc(2)-zBigP(2)*znlc(3))*zsp(1)+(zBigP(2)
     &        *znlc(1)-zBigP(3)*znlc(4))*zsp(2)-mass*znlc(3)*zsp(3)
     &        -mass*znlc(2)*zsp(4)
         zeffsp(3)=(zBigP(2)*znlc(1)-zBigP(4)*znlc(3))*zsp(3)+znlc(4)
     &        *(mass*zsp(2)+zBigP(2)*zsp(4))-znlc(2)*(mass*zsp(1)
     &        +zBigP(4)*zsp(4))
         zeffsp(4)=znlc(3)*(mass*zsp(1)+zBigP(1)*zsp(3))-znlc(1)*(mass
     &        *zsp(2)+zBigP(3)*zsp(3))+(zBigP(1)*znlc(2)-zBigP(3)
     &        *znlc(4))*zsp(4)
      else
c... Vbar_eff = Vbar . n_sl . (P_sl - m)
         zeffsp(1)=-(zBigP(3)*(znlc(4)*zsp(1) + znlc(2)*zsp(2)))
     &        +zBigP(2)*(znlc(1)*zsp(1)+znlc(3)*zsp(2))+mass*znlc(2)
     &        *zsp(3)-mass*znlc(3)*zsp(4)
         zeffsp(2)=zBigP(1)*(znlc(4)*zsp(1)+znlc(2)*zsp(2))-zBigP(4)
     &        *(znlc(1)*zsp(1)+znlc(3)*zsp(2))-mass*znlc(4)*zsp(3)
     &        +mass*znlc(1)*zsp(4)
         zeffsp(3)=(zBigP(1)*znlc(2)-zBigP(3)*znlc(4))*zsp(3)+znlc(3)
     &        *(mass*zsp(2)-zBigP(1)*zsp(4))+znlc(1)*(mass*zsp(1)
     &        +zBigP(3)*zsp(4))
         zeffsp(4)=znlc(4)*(mass*zsp(1)-zBigP(2)*zsp(3))+znlc(2)
     &        *(mass*zsp(2)+zBigP(4)*zsp(3))+(zBigP(2)*znlc(1)-zBigP(4)
     &        *znlc(3))*zsp(4)
      endif

      END

************************************************************************
      SUBROUTINE currLC(zspB,zsp,zjLC)
************************************************************************
c... calculates a vector current in light-cone representation      
c-----------------------------------------------------------------------     

      include 'declare.h'
      include 'dimensions.h'

      dimension zspB(4),zsp(4),zjLC(4)

      zjLC(1)=2d0*(-zsp(4)*zspB(2)-zsp(1)*zspB(3))
      zjLC(2)=2d0*(-zsp(3)*zspB(1)-zsp(2)*zspB(4))
      zjLC(3)=2d0*( zsp(4)*zspB(1)-zsp(2)*zspB(3))
      zjLC(4)=2d0*( zsp(3)*zspB(2)-zsp(1)*zspB(4))
      
      END

************************************************************************
      SUBROUTINE map2LC(pmom,zpmomLC)
************************************************************************
c... maps to light-cone coordinates
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension pmom(4),zpmomLC(4)

      zpmomLC(1)=dcmplx(pmom(4)+pmom(3),0d0)
      zpmomLC(2)=dcmplx(pmom(4)-pmom(3),0d0)
      zpmomLC(3)=dcmplx(pmom(1),pmom(2))
      zpmomLC(4)=dcmplx(pmom(1),-pmom(2))

      END

************************************************************************
      SUBROUTINE effpvec(zpolv,zBigQ,zBigP,znlc,zeffpolv)
************************************************************************
c... calculates an effective polarization vector
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension zpolv(4),zBigQ(4),zBigP(4),znlc(4),zeffpolv(4)

c      zeffpolv(1:4)=-zpolv(1:4)-(zprod(zBigP,zpolv)*znlc(1:4)
c     &        +zprod(zBigQ,zpolv)*zBigP(1:4))
c     &        /(zprod(zBigP,znlc))

      zeffpolv(1:4)=-zpolv(1:4)+(zBigP(1:4)*zprod(zpolv,znlc)
     &             +znlc(1:4)*zprod(zpolv,zBigP))
     &             /dreal(zprod(zBigP,znlc))
      END

c-----------------------------------------------------------------------
************************************************************************
      SUBROUTINE vert3g(pmom,zva,zvb,zvc,zres)
************************************************************************
c... calculates a 3-gluon-vertex contracted with polarization vectors
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      
      dimension pmom(3,4),zva(4),zvb(4),zvc(4)
      dimension auxab(4),auxbc(4),auxca(4)
      dimension zpab(4),zpbc(4),zpca(4)

      auxab(1:4)=pmom(1,1:4)
      auxbc(1:4)=pmom(2,1:4)
      auxca(1:4)=pmom(3,1:4)

      CALL map2LC(auxab,zpab)
      CALL map2LC(auxbc,zpbc)
      CALL map2LC(auxca,zpca)

      zvab=zprod(zva,zvb)
      zvbc=zprod(zvb,zvc)
      zvca=zprod(zva,zvc)
      
      zpvabc=zprod(zpab,zvc)
      zpvbca=zprod(zpbc,zva)
      zpvcab=zprod(zpca,zvb)

      zres=zvab*zpvabc + zvbc*zpvbca + zvca*zpvcab
 
      END

************************************************************************
      FUNCTION spinSum(ipn,jp,p,a,b,sw)
************************************************************************
c... spin sum (Eq. 7.4, arXiv:0805.0216)
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_int.h'
      include 'common_masses.h'

      logical sw
      dimension a(4),b(4),BigQ(4)
      dimension p(mxpart,5)

      BigQ(1:4)=p(1,1:4)+p(2,1:4)
      Qsq=BigQ(4)**2-BigQ(1)**2-BigQ(2)**2-BigQ(3)**2
      apj=a(4)*p(jp,4)-a(1)*p(jp,1)-a(2)*p(jp,2)
     &     -a(3)*p(jp,3)
      aBigQ=a(4)*BigQ(4)-a(1)*BigQ(1)-a(2)*BigQ(2)-a(3)*BigQ(3)
      pjBigQ=p(jp,4)*BigQ(4)-p(jp,1)*BigQ(1)-p(jp,2)*BigQ(2)
     &        -p(jp,3)*BigQ(3)

      if(sw) then
         spinSum=-parmas(ifl(ipn))**2+2d0*apj*aBigQ/pjBigQ
     &          -Qsq*apj**2/pjBigQ**2 
      else
         ab=a(4)*b(4)-a(1)*b(1)-a(2)*b(2)-a(3)*b(3)
         bpj=b(4)*p(jp,4)-b(1)*p(jp,1)-b(2)*p(jp,2)
     &        -b(3)*p(jp,3)
         bBigQ=b(4)*BigQ(4)-b(1)*BigQ(1)-b(2)*BigQ(2)-b(3)*BigQ(3)

         spinSum=-ab+(apj*bBigQ+aBigQ*bpj)/pjBigQ-Qsq*apj*bpj/pjBigQ**2
      endif

      RETURN
      END

************************************************************************
      SUBROUTINE iniSPLfuncSoft(ip,jp,kp,sumpol,ipol,rpol,p,zymp1,
     &     iswitch,vsoftsq)
************************************************************************
c... soft splitting function with different definitions of A_lk
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      logical sumpol,noSum,debug
      PARAMETER(debug=.false.)

      dimension zymp1(mxkmax,8,2),zyip(4),zyjp(4),zyjpcc(4)
      dimension zpi(4),zpj(4),zpk(4)
      dimension ipol(mxpart),rpol(mxpart)
      dimension BigQ(4),Plk(4),zPlk(4)
      dimension p(mxpart,5)


cx      noSum=.false.
      noSum=.true.

      A_lkP=0d0
      A_lk =0d0
      vsoftsq=0d0

      pipj=dot(p,ip,jp)
      pjpk=dot(p,jp,kp)

      if(noSum) then
         CALL map2LC (p(ip,1:4),zpi(1:4))
         CALL map2LC (p(jp,1:4),zpj(1:4))
         CALL map2LC (p(kp,1:4),zpk(1:4))
         if(sumpol) then
            zyip(1:4)=zymp1(ip,1:4,ipol(ip))
            zyjp(1:4)=zymp1(jp,1:4,ipol(jp))
         else
            zyip(1:4)=zymp1(ip,1:4,1)
            zyjp(1:4)=zymp1(jp,1:4,1)
         endif
         Plk(1:4)=pipj*p(kp,1:4)-pjpk*p(ip,1:4)
         CALL map2LC (Plk,zPlk)
      endif

c      piDpi=spinSum(ip,jp,p,p(ip,1:4),p(ip,1:4),.true.)
c      pkDpk=spinSum(kp,jp,p,p(kp,1:4),p(kp,1:4),.true.)
      piDpi=spinSum(ip,jp,p,p(ip,1:4),p(ip,1:4),.false.)
      pkDpk=spinSum(kp,jp,p,p(kp,1:4),p(kp,1:4),.false.)
      piDpk=spinSum(ip,jp,p,p(ip,1:4),p(kp,1:4),.false.)

      if(iswitch.eq.1) then
         if(noSum) then
            A_lkP=pjpk**2*piDpi/(pjpk**2*piDpi+pipj**2*pkDpk)
            A_lk=pipj*pjpk/DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpk,zyjp))+(zprod(zpk,zyjp))
     &           *dconjg(zprod(zpi,zyjp)))*(DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpi,zyjp)))/pipj**2-A_lkP/(pipj*pjpk)**2
     &           *DREAL(zprod(zPlk,zyjp)*dconjg(zprod(zPlk,zyjp))))
            vsoftsq=(A_lk*(DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpk,zyjp)))+DREAL(zprod(zpk,zyjp)
     &           *dconjg(zprod(zpi,zyjp))))/(pipj*pjpk))
c            vsoftsq=A_lk*(zprod(zpi,zyjp)
c     &           *dconjg(zprod(zpk,zyjp))/(pipj*pjpk))
         else
            A_lk=pjpk**2*piDpi/(pjpk**2*piDpi+pipj**2*pkDpk)
            vsoftsq=A_lk*piDpk/(pipj*pjpk)
         endif
      elseif(iswitch.gt.1) then
         if(iswitch.eq.2) then
            BigQ(1:4)=(p(1,1:4)+p(2,1:4))
            piQ=BigQ(4)*p(ip,4)-BigQ(1)*p(ip,1)-BigQ(2)*p(ip,2)
     &           -BigQ(3)*p(ip,3)
            pkQ=BigQ(4)*p(kp,4)-BigQ(1)*p(kp,1)-BigQ(2)*p(kp,2)
     &           -BigQ(3)*p(kp,3)
c... arXiv:0805.0216v1, p.22, Eq. 7.12
            A_lkP=pjpk*piQ/(pjpk*piQ+pipj*pkQ)
         endif
c... arXiv:0805.0216v1, p.22, Eq. 7.13
         if(iswitch.eq.3) A_lkP=pjpk/(pjpk+pipj)
         Plk(1:4)=pipj*p(kp,1:4)-pjpk*p(ip,1:4)
         if(nosum) then
c...  arXiv:0805.0216v1, p.23, Eq. 7.16
            A_lk=pipj*pjpk/(DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpk,zyjp)))+DREAL(zprod(zpk,zyjp)
     &           *dconjg(zprod(zpi,zyjp))))*(DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpi,zyjp)))/pipj**2-A_lkP/(pipj*pjpk)**2
     &           *DREAL(zprod(zPlk,zyjp)*dconjg(zprod(zPlk,zyjp))))
            vsoftsq=(A_lk*(DREAL(zprod(zpi,zyjp)
     &           *dconjg(zprod(zpk,zyjp)))+DREAL(zprod(zpk,zyjp)
     &           *dconjg(zprod(zpi,zyjp))))/(pipj*pjpk))
         else
            PlkDPlk=spinSum(ip,jp,p,Plk,Plk,.false.)
            A_lk=pipj*pjpk/(2d0*piDpk)*(piDpi/(pipj**2)-A_lkP*PlkDPlk
     &           /(pipj*pjpk)**2)
            vsoftsq=A_lk*piDpk/(pipj*pjpk)
         endif
      endif

      if(debug) then
c         print*,'Wii',piDpi/pipj**2,piDpi/pipj**2/2d0
         print*,'Wii',piDpi/pipj**2/2d0
c         print*,'Wjj',spinSum(ip,jp,p,p(jp,1:4),p(jp,1:4),.false.)
c     &        /pipj**2,spinSum(ip,jp,p,p(jp,1:4),p(jp,1:4),.false.)
c     &        /pipj**2/2d0
      endif
         
      END

************************************************************************
      SUBROUTINE iniSPLfunc(ip,jp,sumpol,ipol,rpol,ipp,p,q,mijsq,znlc,
     &     zymp1,zym,zvij)
************************************************************************
c... calculates splitting functions (direct terms) for both helicities 
c... of the emitter parton
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles.h'

      dimension zymp1(mxkmax,8,2),zym(4,4),zymB(2,4),zyip(4),zyjp(4)
      dimension p(mxpart,5),q(mxpart,5)
      dimension zq(5),znlc(4)
      dimension ipol(mxpart),rpol(mxpart)
      dimension BigP(4),BigQ(4),p3g(3,4)
      dimension zBigP(4),zBigQ(4)
      dimension zeffsp(2,4),zjLC(2,4),zpol(2,4),zeffpol(2,4)
      double precision mij,mijsq,Qsq,Psq
      dimension zvij(2)
      logical sumpol
*-----------------------------------------------------------------------

      BigQ(1:4)=p(1,1:4)+p(2,1:4)
      Qsq=BigQ(4)**2-BigQ(1)**2-BigQ(2)**2-BigQ(3)**2
      CALL map2LC(BigQ,zBigQ)  
c... external wave function in m+1_space
      if(sumpol) then
         zyip(1:4)=zymp1(ip,1:4,ipol(ip))
         zyjp(1:4)=zymp1(jp,1:4,ipol(jp))
      else
         zyip(1:4)=zymp1(ip,1:4,1)
         zyjp(1:4)=zymp1(jp,1:4,1)
      endif
c... m_space momentum in light cone coordinates
      zq(1)=dcmplx(q(ip,4) +q(ip,3),0d0)
      zq(2)=dcmplx(q(ip,4) -q(ip,3),0d0)
      zq(3)=dcmplx(q(ip,1), q(ip,2))
      zq(4)=dcmplx(q(ip,1),-q(ip,2))

c... complex conjugation of external wave functions in light-cone repr.
      zymB(1,1)=-dconjg(zym(1,3))
      zymB(1,2)=-dconjg(zym(1,4))
      zymB(1,3)=-dconjg(zym(1,1))
      zymB(1,4)=-dconjg(zym(1,2))

      zymB(2,1)=-dconjg(zym(2,3))
      zymB(2,2)=-dconjg(zym(2,4))
      zymB(2,3)=-dconjg(zym(2,1))
      zymB(2,4)=-dconjg(zym(2,2))

c... initial            
      if (ip.le.2) then   
         if (ifldip(ipp,ip).eq.35) then
            zpol(1,1:4)=-zym(4,1:4)
            zpol(2,1:4)=-zym(3,1:4)
c... g -> g g
            if (ifl(ip).eq.35) then
               prop=2d0*dot(p,ip,jp)
               BigP(1:4)=p(ip,1:4)-p(jp,1:4)
               CALL map2LC(BigP,zBigP)
               p3g(1,1:4)= p(jp,1:4)+p(ip,1:4)
               p3g(2,1:4)=-2d0*p(ip,1:4)+p(jp,1:4)
               p3g(3,1:4)= p(ip,1:4)-2d0*p(jp,1:4)
               CALL effpvec(zpol(1,1:4),zBigQ,zBigP,znlc,
     &              zeffpol(1,1:4))
               CALL effpvec(zpol(2,1:4),zBigQ,zBigP,znlc,
     &              zeffpol(2,1:4))
               CALL vert3g(p3g,zyjp,zyip,zeffpol(1,1:4),z3gres1)
               CALL vert3g(p3g,zyjp,zyip,zeffpol(2,1:4),z3gres2)
               zvij(1)=z3gres1/prop
               zvij(2)=z3gres2/prop
#if DEBUG3
               print*,'INITIAL g -> g g'
#endif
            else
c... g -> q qbar  
#if DEBUG3
               print*,'INITIAL g -> q qbar'
#endif
               BigP(1:4)=p(ip,1:4)-p(jp,1:4)
               CALL map2LC(BigP,zBigP)       
               Psq=BigP(4)**2-BigP(1)**2-BigP(2)**2-BigP(3)**2
               prop=Psq
               CALL effpvec(zpol(1,1:4),zBigQ,zBigP,znlc,
     &                      zeffpol(1,1:4))
               CALL effpvec(zpol(2,1:4),zBigQ,zBigP,znlc,
     &              zeffpol(2,1:4))
               if(ifl(ip).gt.0) then 
                  CALL currLC(zyjp,zyip,zjLC)
               else
                  CALL currLC(zyip,zyjp,zjLC)
               endif
               zvij(1)=zprod(zjLC,zeffpol(1,1:4))/prop
               zvij(2)=zprod(zjLC,zeffpol(2,1:4))/prop
            endif
         else
c... emitter is a quark
            BigP(1:4)=p(ip,1:4)-p(jp,1:4)
            Psq=BigP(4)**2-BigP(1)**2-BigP(2)**2-BigP(3)**2
            CALL map2LC(BigP,zBigP)      
            prop=DREAL(2d0*zprod(znlc,zq)*(Psq-mijsq))
            mij=dsqrt(mijsq)
            if(ifldip(ipp,ip).gt.0) then 
               CALL effSpinor(mij,znlc,zBigP,zymB(1,1:4),.false.,
     &              zeffsp(1,1:4))
               CALL effSpinor(mij,znlc,zBigP,zymB(2,1:4),.false.,
     &              zeffsp(2,1:4))
               if(ifl(jp).eq.35) then
#if DEBUG3
                  print*,'INITIAL q -> q g'
#endif
                  CALL currLC(zeffsp(1,1:4),zyip,zjLC(1,1:4))
                  CALL currLC(zeffsp(2,1:4),zyip,zjLC(2,1:4))
                  zvij(1)=zprod(zyjp,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyjp,zjLC(2,1:4))/prop
               elseif(ifl(ip).eq.35) then 
#if DEBUG3
                  print*,'INITIAL q -> g q'
#endif
                  CALL currLC(zeffsp(1,1:4),zyjp,zjLC(1,1:4))
                  CALL currLC(zeffsp(2,1:4),zyjp,zjLC(2,1:4))
                  zvij(1)=zprod(zyip,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyip,zjLC(2,1:4))/prop
               endif
            elseif(ifldip(ipp,ip).lt.0) then 
               CALL effSpinor(mij,znlc,zBigP,zymB(1,1:4),.true.,
     &              zeffsp(1,1:4))
               CALL effSpinor(mij,znlc,zBigP,zymB(2,1:4),.true.,
     &              zeffsp(2,1:4))
               if(ifl(jp).eq.35) then
#if DEBUG3
                  print*,'INITIAL qbar -> qbar g'
#endif
                  CALL currLC(zyip,zeffsp(1,1:4),zjLC(1,1:4)) 
                  CALL currLC(zyip,zeffsp(2,1:4),zjLC(2,1:4)) 
                  zvij(1)=-zprod(zyjp,zjLC(1,1:4))/prop
                  zvij(2)=-zprod(zyjp,zjLC(2,1:4))/prop
               elseif(ifl(ip).eq.35) then
#if DEBUG3
                  print*,'INITIAL qbar -> g qbar'
#endif
                  CALL currLC(zyjp,zeffsp(1,1:4),zjLC(1,1:4)) 
                  CALL currLC(zyjp,zeffsp(2,1:4),zjLC(2,1:4)) 
                  zvij(1)=-zprod(zyip,zjLC(1,1:4))/prop
                  zvij(2)=-zprod(zyip,zjLC(2,1:4))/prop
               endif
            endif
         endif
      endif

c... final         
      if (ip.gt.2) then
c... emitter is a gluon
         if (ifldip(ipp,ip).eq.35) then
            zpol(1,1:4)=zym(3,1:4)
            zpol(2,1:4)=zym(4,1:4)
c... g -> g g
            if (ifl(ip).eq.35) then
#if DEBUG3
               print*,'FINAL g -> g g'
#endif
               prop=2d0*dot(p,ip,jp)
               BigP(1:4)=p(ip,1:4)+p(jp,1:4)
               CALL map2LC(BigP,zBigP)
               p3g(1,1:4)= p(jp,1:4)-p(ip,1:4)
               p3g(2,1:4)= 2d0*p(ip,1:4)+p(jp,1:4)
               p3g(3,1:4)=-2d0*p(jp,1:4)-p(ip,1:4)
               CALL effpvec(zpol(1,1:4),zBigQ,zBigP,znlc,zeffpol(1,1:4))
               CALL effpvec(zpol(2,1:4),zBigQ,zBigP,znlc,zeffpol(2,1:4))
               CALL vert3g(p3g,zyjp,zyip,zeffpol(1,1:4),z3gres1)
               CALL vert3g(p3g,zyjp,zyip,zeffpol(2,1:4),z3gres2)
               zvij(1)=z3gres1/prop
               zvij(2)=z3gres2/prop
            else
c... g -> q qbar  
#if DEBUG3
               print*,'FINAL g -> q qbar'
#endif
               BigP(1:4)=p(ip,1:4)+p(jp,1:4)
               prop=BigP(4)**2-BigP(1)**2-BigP(2)**2-BigP(3)**2
               CALL map2LC(BigP,zBigP)

               CALL effpvec(zpol(1,1:4),zBigQ,zBigP,znlc,zeffpol(1,1:4))
               CALL effpvec(zpol(2,1:4),zBigQ,zBigP,znlc,zeffpol(2,1:4))
               if(ifl(ip).gt.0) then
                  CALL currLC(zyip,zyjp,zjLC) 
               else
                  CALL currLC(zyjp,zyip,zjLC)
               endif
               zvij(1)=-zprod(zjLC,zeffpol(1,1:4))/prop
               zvij(2)=-zprod(zjLC,zeffpol(2,1:4))/prop
            endif
c... emitter is a quark
c... q -> qg  &  qbar -> qbar g
         else
            BigP(1:4)=p(ip,1:4)+p(jp,1:4)
            CALL map2LC(BigP,zBigP)
            Psq=BigP(4)**2-BigP(1)**2-BigP(2)**2-BigP(3)**2
            prop=DREAL(2d0*zprod(znlc,zq)*(Psq-mijsq))
            mij=dsqrt(mijsq)
            if(ifldip(ipp,ip).gt.0) then 
               CALL effSpinor(mij,znlc,zBigP,zymB(1,1:4),.true.,
     &              zeffsp(1,1:4))
               CALL effSpinor(mij,znlc,zBigP,zymB(2,1:4),.true.,
     &              zeffsp(2,1:4))
               if(ifl(jp).eq.35) then
#if DEBUG3
               print*,'FINAL q -> q g'
#endif
                  CALL currLC(zyip,zeffsp(1,1:4),zjLC(1,1:4))
                  CALL currLC(zyip,zeffsp(2,1:4),zjLC(2,1:4))
                  zvij(1)=zprod(zyjp,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyjp,zjLC(2,1:4))/prop
               elseif(ifl(ip).eq.35) then
#if DEBUG3
               print*,'FINAL q -> g q'
#endif
                  CALL currLC(zyjp,zeffsp(1,1:4),zjLC(1,1:4))
                  CALL currLC(zyjp,zeffsp(2,1:4),zjLC(2,1:4))
                  zvij(1)=zprod(zyip,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyip,zjLC(2,1:4))/prop   
               endif
            elseif(ifldip(ipp,ip).lt.0) then 
               CALL effSpinor(mij,znlc,zBigP,zymB(1,1:4),.false.,
     &              zeffsp(1,1:4))
               CALL effSpinor(mij,znlc,zBigP,zymB(2,1:4),.false.,
     &              zeffsp(2,1:4))
               if(ifl(jp).eq.35) then
#if DEBUG3
               print*,'FINAL qbar -> qbar g'
#endif
                  CALL currLC(zeffsp(1,1:4),zyip,zjLC(1,1:4))
                  CALL currLC(zeffsp(2,1:4),zyip,zjLC(2,1:4))
                  zvij(1)=zprod(zyjp,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyjp,zjLC(2,1:4))/prop
               elseif(ifl(ip).eq.35) then
#if DEBUG3
               print*,'FINAL qbar -> g qbar'
#endif
                  CALL currLC(zeffsp(1,1:4),zyjp,zjLC(1,1:4))
                  CALL currLC(zeffsp(2,1:4),zyjp,zjLC(2,1:4))
                  zvij(1)=zprod(zyip,zjLC(1,1:4))/prop
                  zvij(2)=zprod(zyip,zjLC(2,1:4))/prop
               endif
            endif
         endif
      endif
      END

