c-----------------------------------------------------------------------
c...  Copyright (C) 2009 M. Czakon, C. Papadopoulos and M. Worek
c...  Copyright (C) 2013 G. Bevilacqua, M. Czakon, M. Kubocz, M. Worek
c...  For details see: e-Print: arXiv:0905.0883 [hep-ph] 
c...                   e-Print: arXiv:1308.5605 [hep-ph]  
c-----------------------------------------------------------------------
#define SINGLEPOINT 0
#define DEBUG1 0
#define POLECHECK 0

      subroutine dipoles_I(key,weight,wmeasure)

c-----------------------------------------------------------------------
c...  
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'dimensions.h'


      common/helac_dipoles_ext/dipoletype
c...  dipoletype: 0-Catani-Seymour, 1-Nagy-Soper
      integer dipoletype
      logical stable
c...  pi
      parameter(pi=3.141592653589793d0)
c...  8*pi^2
      double precision norm
      parameter(norm=78.9568352087148690d0)

c...  polarizations
      dimension ipol(mxpart),rpol(mxpart)
c...  color flows
      dimension jcoldip(mxncc,mxpart,0:2)
      dimension jcol(mxpart,2),kcol(mxpart,2)
c...  color correlator matrix (dimension for 10 partons !!!)
      dimension rmatrix((10*9)/2,mxncc,mxncc)
c...  Dyson-Schwinger skeletons
      dimension list1(0:mxproc,mxpart,3)
      dimension list(0:mxproc,mxkmax,18,mxncc)
      dimension kcou(0:mxproc,mxncc)
      dimension isnew(mxfeyg,8,8) ! pure trash, but needed
c...  amplitudes
      dimension zy(mxkmax,8),zamp(mxncc)

c...  amplitude for different polarizations. the lower precision (real)
c...  is used to spare some space, and is, of course, sufficient
      real amppol(mxhel)
c...  perform a sum over polarizations
      logical sumpol
c...  evaluate for random polarization phases
      logical ranpol

c...  parameters
      double precision mh(mxpart)
      double precision mu,kappa

c...  setup for self-initialization
      logical firstcall
      save firstcall
      data firstcall/.true./

c...  some local arrays and variables
      character*100 process
      double precision Im2,Im1,Im0

c...  storage
      save jcoldip,rmatrix
      save kmax,ncc,imax
      save list1,list,kcou
      save avcol,avhel,symet
      save amppol,sumpol,ranpol
      save nl,nh,mh,mu,alpha,kappa

#if SINGLEPOINT
      open(99,file='momenta.input',status='old')
      do i=1,n
         read(99,*)(pmom(i,j),j=1,5)
      enddo
      close(99)
#endif      
c-----------------------------------------------------------------------
c...  initialisation
c-----------------------------------------------------------------------

      if (firstcall) then
         
c-----------------------------------------------------------------------
c...  reading the configuration file
c-----------------------------------------------------------------------
         
         open(500,file='dipoles_I.conf',status='old')

         read(500,*)dipoletype
         read(500,*)nl,nh
         read(500,*)mh(1:nh)
         read(500,*)alpha
         read(500,*)kappa
         read(500,*)ranpol

         close(500)

         call qcdscale(mu)

         print*,'--------------------------------------------------'
         print*,'Integrated Dipole I Operator Configuration'
         print*,'--------------------------------------------------'
         if(dipoletype.eq.0) print*,'dipole type:       Catani-Seymour'
         if(dipoletype.eq.1) print*,'dipole type:       Nagy-Soper'
         print*,'renormalization scale:',mu
         print*,'number of light quark flavors:',nl
         print*,'number of heavy quark flavors:',nh
         print*,'heavy quark masses:',(mh(i),i=1,nh)
         print*,'alpha:',alpha
         print*,'kappa:',kappa
         print*,'random polarizations:',ranpol
         print*,'--------------------------------------------------'

         if (nh.ne.0.and.alpha.ne.1) then
            print*,'alpha dependence in the heavy quark splitting'
            print*,'is not supported yet!'
            stop
         endif

         if (ranpol) then
            sumpol=.false.
         else
            sumpol=.true.
         endif

c-----------------------------------------------------------------------
c...  input for the process
c-----------------------------------------------------------------------
         
c...  we need the tree for the process independently of the main HELAC
c...  routines, since the computation uses the same technology as the
c...  unintegrated dipoles from dipoles.F
         call produceinputs(n,ifl,0)
         
c-----------------------------------------------------------------------
c...  reading the skeleton from the tree file
c-----------------------------------------------------------------------
         
         call getstring(n,ifl,process)

         nunit1=123
         open(nunit1,file=
     .        './helac_trees/tree_'//process(2:2*n+1)//'.in',
     .        status='old',err=100)
         goto 101
         
 100     print*,'Could not open tree for ',process(2:2*n+1)
         print*,'Please run "make trees"'
         stop
         
 101     continue
         
         read(nunit1,*)kmax,ncc,imax
         if (kmax.gt.mxkmax) then
            print*,'maximum number of subamplitudes is',mxkmax
            print*,'please increase mxkmax in dimensions.h'
            stop
         endif
         if (ncc.gt.mxncc) then
            print*,'maximum number of color flows is',mxncc
            print*,'please increase mxncc in dimensions.h'
            stop
         endif
         
         read(nunit1,*)list1(0,1:n,1:3)
         read(nunit1,*)list(0,1:kmax,1:18,1:ncc)
         read(nunit1,*)kcou(0,1:ncc)
         read(nunit1,*)ng  
         if (ng.gt.mxfeyg) then
            print*,'maximum number of graphs is',mxfeyg
            print*,'please increase mxfeyg in dimensions.h'
            stop
         endif
         
         do ii=1,ng
            do jj=1,8
               read(nunit1,*)(isnew(ii,jj,kk),kk=1,8)
            enddo
         enddo
         
         read(nunit1,*)avhel,avcol,symet
         read(nunit1,*)rmatrix(1,1:ncc,1:ncc)
         
         do k=1,ncc
            read(nunit1,*)jcoldip(k,1:n,0:2)
         enddo
         
c...  closing tree file
         close(nunit1)

c-----------------------------------------------------------------------
c...  color correlator matrix
c-----------------------------------------------------------------------
            
         do i=1,ncc
            do j=1,ncc
               do k=1,n
                  do l=1,2
                     jcol(k,l)=jcoldip(i,k,l)
                     kcol(k,l)=jcoldip(j,k,l)
                  enddo
               enddo
               icount=0
               do k=1,n
c...  only partons
                  if (iabs(ifl(k)).ne.3.and.iabs(ifl(k)).ne.4.and.
     .                iabs(ifl(k)).ne.7.and.iabs(ifl(k)).ne.8.and.
     .                iabs(ifl(k)).ne.11.and.iabs(ifl(k)).ne.12.and.
     .                ifl(k).ne.35) cycle
                  do l=k+1,n
c...  only partons
                     if (iabs(ifl(l)).ne.3.and.iabs(ifl(l)).ne.4.and.
     .                    iabs(ifl(l)).ne.7.and.iabs(ifl(l)).ne.8.and.
     .                    iabs(ifl(l)).ne.11.and.iabs(ifl(l)).ne.12.and.
     .                    ifl(l).ne.35) cycle
                     icount=icount+1
                     rmatrix(icount,i,j)=colcor(jcol,kcol,k,l,n)
                  enddo
               enddo
            enddo
         enddo

c-----------------------------------------------------------------------
c     end of initialization
c-----------------------------------------------------------------------
         
      endif
      
c-----------------------------------------------------------------------
c...  helicity setup
c-----------------------------------------------------------------------

c...  random polarization initialization
      do i=1,n
         if (ranpol) then
            rpol(i)=2*pi*rnmy(0)
         else
            rpol(i)=-1
         endif
      enddo

c-----------------------------------------------------------------------
c...  I operator contribution
c-----------------------------------------------------------------------

      weight=0d0

c...  coupling constant
      call coupling(n,ifl,pmom,gqcd,wsf)

c...  dynamical scale
      call qcdscale(mu)
      
c...  initialize the counter and polarizations
      if (sumpol) then
         npol=0
         ipol(1:n)=1
      endif

c...  skip vanishing contributions during summation over polarizations
 310  if (sumpol) then
         npol=npol+1
         if (npol.gt.mxhel) then
            print*,'maximum number of helicity configs is',mxhel
            print*,'please increase mxhel in dimensions.h'
            stop
         endif
         if (.not.firstcall.and.amppol(npol).lt.1e-15) then
            if (nextpol(n,ifl,ipol,rpol).ne.0) then
               goto 310
            else
               goto 320
            endif
         endif
      endif

c...  amplitude
      call iniqqdip(n,ifl,io,pmom,ipol,rpol,list1,0,zy)

      do k=1,ncc
         call nextqdip(n,k,imax,list,kcou,0,zy)
         call ampqdip(k,ifl,zy,list,kcou,0,zamp)
      enddo

      w=0d0
      icount=0
      do jp=1,n
c...  only partons
         if (iabs(ifl(jp)).ne.3.and.iabs(ifl(jp)).ne.4.and.
     .        iabs(ifl(jp)).ne.7.and.iabs(ifl(jp)).ne.8.and.
     .        iabs(ifl(jp)).ne.11.and.iabs(ifl(jp)).ne.12.and.
     .        ifl(jp).ne.35) cycle
         do kp=jp+1,n
c...  only partons
            if (iabs(ifl(kp)).ne.3.and.iabs(ifl(kp)).ne.4.and.
     .           iabs(ifl(kp)).ne.7.and.iabs(ifl(kp)).ne.8.and.
     .           iabs(ifl(kp)).ne.11.and.iabs(ifl(kp)).ne.12.and.
     .           ifl(kp).ne.35) cycle
            icount=icount+1

            if(dipoletype.eq.0)then
c...  Catani-Seymour subtraction scheme
               call Ioperator(jp,kp,nl,nh,mh,mu,alpha,kappa,Im2,Im1,Im0)
               coeff=Im0
#if POLECHECK
               coeff1=Im1
               coeff2=Im2
#endif
               call Ioperator(kp,jp,nl,nh,mh,mu,alpha,kappa,Im2,Im1,Im0)
               coeff=coeff+Im0
#if POLECHECK
               coeff1=coeff1+Im1
               coeff2=coeff2+Im2
               print*,'*** CS coeff2  kp,jp',kp,jp,coeff2
               print*,'*** CS coeff1',coeff1
#endif
            elseif(dipoletype.eq.1)then
c...  Nagy-Soper subtraction scheme
c              ioption=0        !...Numerical integ. of the unconstrained parton
               ioption=1        !...Monte Carlo integ. of the unconstrained parton
               call NSIoperator(jp,kp,nl,mu,ioption,alpha,Im2,Im1,Im0,
     -              stable)
               if(.not.stable)then
                  print*,' '
                  print*,'MESSAGE from dipoles_I.F: unstable event'
     -                 //' found, weight set to zero'
                  print*,' '
                  GOTO 320
               endif
               coeff=Im0
#if POLECHECK
               coeff1=Im1
               coeff2=Im2
#endif
               call NSIoperator(kp,jp,nl,mu,ioption,alpha,Im2,Im1,Im0,
     -              stable)
               if(.not.stable)then
                  print*,' '
                  print*,'MESSAGE from dipoles_I.F: unstable event'
     -                 //' found, weight set to zero'
                  print*,' '
                  GOTO 320
               endif
               coeff=coeff+Im0
#if POLECHECK
               coeff1=coeff1+Im1
               coeff2=coeff2+Im2
               print*,'*** NS coeff2  kp,jp',kp,jp,coeff2
               print*,'*** NS coeff1',coeff1
#endif
            else
               print*,'dipoles_I.f ERROR: invalid dipole '//
     .              'scheme setting',dipoletype
               STOP
            endif

            do k1=1,ncc
               do k2=1,ncc
                  w=w+zamp(k1)*dconjg(zamp(k2))
     .               *rmatrix(icount,k1,k2)*coeff
               enddo
            enddo
         enddo
      enddo

#if POLECHECK
      STOP
#endif

      w=w*avhel*avcol*symet
      w=gqcd**2/norm*w

c-----------------------------------------------------------------------  
c...  pdf for dynamical scale implementation                                 
c-----------------------------------------------------------------------  
      w=w*wsf

      weight = weight + w

c...  store normalized amplitude values for all polarizations
      if (sumpol) then
         if (firstcall) amppol(npol)=abs(w)
         if (nextpol(n,ifl,ipol,rpol).ne.0) goto 310
         if (firstcall) then
            ampmax=0d0
            do i=1,npol
               if (amppol(i).gt.ampmax) ampmax=amppol(i)
            enddo
            amppol(1:npol)=amppol(1:npol)/ampmax
         endif
      endif
      
 320  continue

c-----------------------------------------------------------------------
c...  histogramming
c-----------------------------------------------------------------------

      if (.not.firstcall.and.key.eq.2) 
     .     call HISTOm(n,ifl,pmom,weight*wmeasure)

c-----------------------------------------------------------------------
c...  initialization is done only once !
c-----------------------------------------------------------------------

      firstcall=.false.

#if SINGLEPOINT
      print*,'gqcd:',gqcd
      print*,'momenta:x,y,z,E,m'
      do i=1,n
         write(6,'(5d24.16)'), (pmom(i,j),j=1,5)
      enddo
      print*,'value:',weight
      stop
#endif
      return
      end

c-----------------------------------------------------------------------

      subroutine Ioperator(jp,kp,nl,nh,mh,mu,alpha,kappa,Im2,Im1,Im0)

c-----------------------------------------------------------------------
c...  coefficient of the color correlator Tj.Tk in the I operator
c...  the normalization does not include the coupling alphas/2pi
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)
c...  Riemann zeta(2)
      double precision z2
      parameter(z2=1.64493406684822644d0)

      integer jp,kp,nl,nh,nh1
      double precision mh(*),mu,kappa,kappa1

      double precision mj,mjsq,mk,mksq,muj,mujsq,muk,muksq
      double precision Kj
      double precision Im2,Im1,Im0
      double precision Ieik,IgQcoll,Iggcoll,Iqqcoll,ddilog

c-----------------------------------------------------------------------
c...  setup
c-----------------------------------------------------------------------

      nh1=nh
      kappa1=kappa

c...  in the case of initial state emitter, there are never heavy
c...  quark splittings
      if (jp.le.2) then
         nh1=0
         kappa1=2d0/3
      endif

c...  auxilliary scale (arbitrary)
      Qaux=mu

c...  constants
      gq=3d0/2*CF
      gg=11d0/6*CA-2d0/3*TF*nl

c...  masses of the emitter and spectator
      mj=parmas(ifl(jp))
      mk=parmas(ifl(kp))
      mjsq=mj**2
      mksq=mk**2

c...  invariants
      sjk=2*(pmom(jp,4)*pmom(kp,4)-pmom(jp,1)*pmom(kp,1)
     .      -pmom(jp,2)*pmom(kp,2)-pmom(jp,3)*pmom(kp,3))
      Qjksq=sjk+mjsq+mksq
      Qjk=sqrt(Qjksq)

c...  dimensionless ratios
      muj=mj/Qjk
      muk=mk/Qjk
      mujsq=mjsq/Qjksq
      muksq=mksq/Qjksq

c-----------------------------------------------------------------------
c...  singular part of V_j
c-----------------------------------------------------------------------

      if (mj.gt.0.and.mk.gt.0) then

         vjk=sqrt((1-mujsq-muksq)**2-4*mujsq*muksq)/(1-mujsq-muksq)
         rho=sqrt((1-vjk)/(1+vjk))
         rhoj=sqrt((1-vjk+2*mujsq/(1-mujsq-muksq))/
     .             (1+vjk+2*mujsq/(1-mujsq-muksq)))
         rhok=sqrt((1-vjk+2*muksq/(1-mujsq-muksq))/
     .             (1+vjk+2*muksq/(1-mujsq-muksq)))
         VjSm2=0
         VjSm1=log(rho)/vjk
         VjSm0=(log(rho)*log(Qjksq/sjk)-log(rhoj)**2-log(rhok)**2-z2)/
     .        vjk

      else if (mj.gt.0.and.mk.eq.0) then

         VjSm2=0.5d0
         VjSm1=0.5d0*log(mjsq/sjk)
         VjSm0=-.25d0*log(mjsq/sjk)**2-0.5d0*z2-0.5d0*log(mjsq/sjk)
     .        *log(sjk/Qjksq)-0.5d0*log(mjsq/Qjksq)*log(sjk/Qjksq)

      else if (mj.eq.0.and.mk.gt.0) then

         VjSm2=0.5d0
         VjSm1=0.5d0*log(mksq/sjk)
         VjSm0=-.25d0*log(mksq/sjk)**2-0.5d0*z2-0.5d0*log(mksq/sjk)
     .        *log(sjk/Qjksq)-0.5d0*log(mksq/Qjksq)*log(sjk/Qjksq)

      else

         VjSm2=1.0d0
         VjSm1=0
         VjSm0=0

      endif

c-----------------------------------------------------------------------
c...  non-singular part of V_j
c-----------------------------------------------------------------------

      if (ifl(jp).ne.35) then
c...  quark emitter

         if (mj.gt.0.and.mk.gt.0) then

            VjNS=gq/CF*log(sjk/Qjksq)+(log(rho**2)*log(1+rho**2)
     .          +2*ddilog(rho**2)-ddilog(1-rhoj**2)-ddilog(1-rhok**2)
     .          -z2)/vjk+log((Qjk-mk)/Qjk)-2*log(((Qjk-mk)**2-mjsq)/
     .          Qjksq)-2*mjsq/sjk*log(mj/(Qjk-mk))-mk/(Qjk-mk)
     .          +2*mk*(2*mk-Qjk)/sjk+3*z2

         else if (mj.gt.0.and.mk.eq.0) then

            VjNS=gq/CF*log(sjk/Qjksq)+z2-ddilog(sjk/Qjksq)
     .          -2*log(sjk/Qjksq)-mjsq/sjk*log(mjsq/Qjksq)

         else if (mj.eq.0.and.mk.gt.0) then

            VjNS=gq/CF*(log(sjk/Qjksq)-2*log((Qjk-mk)/Qjk)
     .          -2*mk/(Qjk+mk))+z2-ddilog(sjk/Qjksq)

         else

            VjNS=0

         endif

      else
c...  gluon emitter

         if (mk.gt.0) then
            
            VjNS=gg/CA*(log(sjk/Qjksq)-2*log((Qjk-mk)/Qjk)
     .          -2*mk/(Qjk+mk))+z2-ddilog(sjk/Qjksq)
            do i=1,nh1
               if (sjk.le.4*mh(i)*(mh(i)+mk)) cycle
               rho1=sqrt(1-4*mh(i)**2/(Qjk-mk)**2)
               VjNS=VjNS+4d0/3*TF/CA*(log((Qjk-mk)/Qjk)+mk*rho1**3/
     .             (Qjk+mk)+log((1+rho1)/2)-rho1/3*(3+rho1**2)
     .             -0.5d0*log(mh(i)**2/Qjksq))
            enddo
            do i=1,nh1
               VjNS=VjNS+2d0/3*TF/CA*log(mh(i)**2/Qaux**2)
            enddo
            VjNS=VjNS+(kappa1-2d0/3)*mksq/sjk*(2*TF/CA*nl-1)*log(2*mk/
     .          (Qjk+mk))
            do i=1,nh1
               if (sjk.le.4*mh(i)*(mh(i)+mk)) cycle
               rho1=sqrt(1-4*mh(i)**2/(Qjk-mk)**2)
               rho2=sqrt(1-4*mh(i)**2/(Qjksq-mksq))
               VjNS=VjNS+(kappa1-2d0/3)*mksq/sjk*2*TF/CA*(rho2**3*
     .             log((rho2-rho1)/(rho2+rho1))-log((1-rho1)/(1+rho1))
     .             -8*rho1*mh(i)**2/sjk)
            enddo

         else

            VjNS=0
            do i=1,nh1
               if (sjk.le.4*mh(i)*(mh(i)+mk)) cycle
               rho1=sqrt(1-4*mh(i)**2/(Qjk-mk)**2)
               VjNS=VjNS+4d0/3*TF/CA*(log((1+rho1)/2)-rho1/3*(3+rho1**2)
     .             -0.5d0*log(mh(i)**2/Qjksq))
            enddo
            do i=1,nh1
               VjNS=VjNS+2d0/3*TF/CA*log(mh(i)**2/Qaux**2)
            enddo

         endif

      endif
      
c-----------------------------------------------------------------------
c...  Gamma_j, gamma_j, K_j and T_j^2
c-----------------------------------------------------------------------

      if (ifl(jp).ne.35) then
c...  quark emitter

         if (mj.eq.0) then
            Gjm1=gq
            Gjm0=0
         else
            Gjm1=CF
            Gjm0=CF*(0.5d0*log(mjsq/mu**2)-2)
         endif
         gj=gq
         Kj=(7d0/2-z2)*CF
         Tjsq=CF

      else
c...  gluon emitter

         Gjm1=gg
         Gjm0=0
         do i=1,nh1
            Gjm0=Gjm0-2d0/3*TF*log(mh(i)**2/Qaux**2)
         enddo
         gj=gg
         Kj=(67d0/18-z2)*CA-10d0/9*TF*nl
         Tjsq=CA

      endif

c-----------------------------------------------------------------------
c...  modification to include the alpha parameter.
c-----------------------------------------------------------------------

      if (jp.gt.2.and.kp.gt.2) then
c...  final-final integrated dipole alpha dependence
         if (ifl(jp).ne.35) then
c...  quark emitter
            Kj=Kj+CF*(2*Ieik(muj,muk,alpha)+IgQcoll(muj,muk,alpha))
         else
c...  gluon emitter
            Kj=Kj+CA*(2*Ieik(muj,muk,alpha)+Iggcoll(muk,kappa,alpha))
     .           +TF*nl*Iqqcoll(muk,kappa,alpha)
         endif
      else
c...  remaining integrated dipoles
         Kj=Kj-Tjsq*log(alpha)**2+gj*(alpha-1-log(alpha))
      endif

c-----------------------------------------------------------------------
c...  complete operator
c-----------------------------------------------------------------------

      Im2=-VjSm2
      Im1=-(VjSm1+VjSm2*log(mu**2/sjk)+Gjm1/Tjsq)
      Im0=-(VjSm0+VjSm1*log(mu**2/sjk)+0.5d0*VjSm2*log(mu**2/sjk)**2
     .   +VjNS-2*z2)-(Gjm0+gj*log(mu**2/sjk)+gj+Kj)/Tjsq

      return
      end

c-----------------------------------------------------------------------

      double precision function Ieik(muj,muk,alpha)

c-----------------------------------------------------------------------
c...  alpha dependence of the eikonal integral
c-----------------------------------------------------------------------

      include 'declare.h'

      double precision muj,muk,alpha

      if (muk.eq.0) then
      Ieik = ddilog(muj**2/(-1 + muj**2)) - 
     -  ddilog(muj**2/(alpha*(-1 + muj**2))) - 
     -  (log(alpha)*(log(alpha) + 2*log(1 - muj**2)))/2
      return
      endif

      yp = 1 - (2*(1 - muk)*muk)/(1 - muj**2 - muk**2)
      if (alpha.gt.yp) then
         Ieik=0
         return
      endif

      x = 1 - alpha - (2*(1 - muk)*muk)/(1 - muj**2 - muk**2) + 
     -  sqrt((1 - alpha - (2*(1 - muk)*muk)/(1 - muj**2 - muk**2))*
     -    (-alpha + (4*muj**2*muk**2)/
     -       ((muj**2 - (1 - muk)**2)*(1 - muj**2 - muk**2)) + 
     -      1/(1 - (2*(1 - muk)*muk)/(1 - muj**2 - muk**2))))

      if (muj.eq.0) then
      Ieik = -log((1 - x + yp)/(1 + yp))**2 + 
     -  log((1 + 2*x*yp - yp**2)/((1 + x - yp)*(1 - x + yp)))**2/2 + 
     -  2*(-ddilog((1 - yp)/2) + ddilog((1 + x - yp)/2) + 
     -     ddilog((1 - yp)/(1 + yp)) - 
     -     ddilog((1 + 2*x*yp - yp**2)/(1 + yp)**2) + 
     -     log((1 + x - yp)/(1 - yp))*log((1 + yp)/2) + 
     -     log((1 + yp)/(2*yp))*log((1 + 2*x*yp - yp**2)/(1 - yp**2)))
      return
      endif

      a = (2*muk)/(1 - muj**2 - muk**2)
      b = (2*(1 - muk))/(1 - muj**2 - muk**2)
      c = (2*(1 - muk)*muk)/(1 - muj**2 - muk**2)
      d = log((1 - muj**2 - muk**2)/2)
      xm = (-muj**2 + (1 - muk)**2 - 
     -    sqrt(1 - 2*muj**2 + muj**4 - 2*muk**2 - 2*muj**2*muk**2 + 
     -      muk**4))/(1 - muj**2 - muk**2)
      xp = (-muj**2 + (1 - muk)**2 + 
     -    sqrt(1 - 2*muj**2 + muj**4 - 2*muk**2 - 2*muj**2*muk**2 + 
     -      muk**4))/(1 - muj**2 - muk**2)
      vjk = sqrt(1 - 2*muj**2 + muj**4 - 2*muk**2 - 2*muj**2*muk**2 + 
     -    muk**4)/(1 - muj**2 - muk**2)

      Ieik = -ddilog(b/(a + b)) + ddilog(-(c/(a - c))) + 
     -  ddilog((b - x)/(a + b)) - ddilog(-((c + x)/(a - c))) + 
     -  ddilog(b/(b - xm)) - ddilog((b - x)/(b - xm)) - 
     -  ddilog(xm/(a + xm)) + ddilog(xm/(c + xm)) + 
     -  ddilog((-x + xm)/(a + xm)) - ddilog((-x + xm)/(c + xm)) + 
     -  ddilog((x - xp)/(b - xp)) - ddilog(-(xp/(b - xp))) + 
     -  ddilog(a/(a + xp)) - ddilog((a + x)/(a + xp)) - 
     -  ddilog(c/(c + xp)) + ddilog((c + x)/(c + xp)) - log(a)**2/2 + 
     -  log(a + x)**2/2 + log(b - x)*
     -   log(((a + x)*(b - xm))/((a + b)*(x - xm))) - 
     -  log(b)*log((a*(-b + xm))/((a + b)*xm)) + 
     -  (log(x - xm) - log(-xm))*log((c + xm)/(a + xm)) - 
     -  d*log(-(a/(xm*xp))) + log(a*(b - xp))*log(xp) + 
     -  (-log(a) + log(a + x))*log(a + xp) + 
     -  log(c + x)*log(-(((a - c)*(x - xp))/((a + x)*(c + xp)))) - 
     -  log(c)*log(((a - c)*xp)/(a*(c + xp))) + 
     -  d*log((a + x)/((x - xm)*(-x + xp))) - 
     -  log((a + x)*(b - xp))*log(-x + xp)

      Ieik=Ieik/vjk

      end

c-----------------------------------------------------------------------

      double precision function IgQcoll(muj,muk,alpha)

c-----------------------------------------------------------------------
c...  alpha dependence of the eikonal integral
c-----------------------------------------------------------------------

      include 'declare.h'

      double precision muj,muk,alpha

      yp = 1 - (2*(1 - muk)*muk)/(1 - muj**2 - muk**2)
      if (alpha.gt.yp) then
         IgQcoll=0
         return
      endif

      IgQcoll = (3*(1 + alpha))/2 + 1/(1 - muk) - 
     -  (2*(2 - 2*muj**2 - muk))/(1 - muj**2 - muk**2) + 
     -  ((1 - alpha)*muj**2)/
     -   (2*(muj**2 + alpha*(1 - muj**2 - muk**2))) - 
     -  2*log((alpha*(1 - muj**2 - muk**2))/(-muj**2 + (1 - muk)**2)) + 
     -  ((1 + muj**2 - muk**2)*
     -     log((muj**2 + alpha*(1 - muj**2 - muk**2))/(1 - muk)**2))/
     -   (2*(1 - muj**2 - muk**2))

      end

c-----------------------------------------------------------------------

      double precision function Iggcoll(muk,kappa,alpha)

c-----------------------------------------------------------------------
c...  alpha dependence of the eikonal integral
c-----------------------------------------------------------------------

      include 'declare.h'

      double precision muk,kappa,alpha

      yp = (1 - muk)/(1 + muk)
      if (alpha.gt.yp) then
         Iggcoll=0
         return
      endif

      Iggcoll = (11*((-1 + muk + alpha*(1 + muk))/(1 + muk) - 
     -       log((alpha*(1 + muk))/(1 - muk))))/6

      if (muk.ne.0) Iggcoll = Iggcoll + 
     -  ((2 - 3*kappa)*muk**2*log(((1 - alpha)*(1 + muk))/(2*muk)))/
     -   (3*(1 - muk**2))

      end

c-----------------------------------------------------------------------

      double precision function Iqqcoll(muk,kappa,alpha)

c-----------------------------------------------------------------------
c...  alpha dependence of the eikonal integral
c-----------------------------------------------------------------------

      include 'declare.h'

      double precision muk,kappa,alpha

      yp = (1 - muk)/(1 + muk)
      if (alpha.gt.yp) then
         Iqqcoll=0
         return
      endif

      Iqqcoll = (-2*((-1 + muk + alpha*(1 + muk))/(1 + muk) - 
     -       log((alpha*(1 + muk))/(1 - muk))))/3

      if (muk.ne.0) Iqqcoll = Iqqcoll + 
     -  (2*muk**2*(kappa - 2d0/3)*log(((1 - alpha)*(1 + muk))/(2*muk)))/
     -   (1 - muk**2)

      end

c-----------------------------------------------------------------------

      FUNCTION DDILOG(X)

c-----------------------------------------------------------------------
c...  double precision dilogarithm function taken from CERNlib
c-----------------------------------------------------------------------

      IMPLICIT DOUBLE PRECISION (A-H,O-Z)

      DIMENSION C(0:19)

      PARAMETER (Z1 = 1, HF = Z1/2)
      PARAMETER (PI = 3.14159 26535 89793 24D0)
      PARAMETER (PI3 = PI**2/3, PI6 = PI**2/6, PI12 = PI**2/12)

      DATA C( 0) / 0.42996 69356 08136 97D0/
      DATA C( 1) / 0.40975 98753 30771 05D0/
      DATA C( 2) /-0.01858 84366 50145 92D0/
      DATA C( 3) / 0.00145 75108 40622 68D0/
      DATA C( 4) /-0.00014 30418 44423 40D0/
      DATA C( 5) / 0.00001 58841 55418 80D0/
      DATA C( 6) /-0.00000 19078 49593 87D0/
      DATA C( 7) / 0.00000 02419 51808 54D0/
      DATA C( 8) /-0.00000 00319 33412 74D0/
      DATA C( 9) / 0.00000 00043 45450 63D0/
      DATA C(10) /-0.00000 00006 05784 80D0/
      DATA C(11) / 0.00000 00000 86120 98D0/
      DATA C(12) /-0.00000 00000 12443 32D0/
      DATA C(13) / 0.00000 00000 01822 56D0/
      DATA C(14) /-0.00000 00000 00270 07D0/
      DATA C(15) / 0.00000 00000 00040 42D0/
      DATA C(16) /-0.00000 00000 00006 10D0/
      DATA C(17) / 0.00000 00000 00000 93D0/
      DATA C(18) /-0.00000 00000 00000 14D0/
      DATA C(19) /+0.00000 00000 00000 02D0/

      IF(X .EQ. 1) THEN
       H=PI6
      ELSEIF(X .EQ. -1) THEN
       H=-PI12
      ELSE
       T=-X
       IF(T .LE. -2) THEN
        Y=-1/(1+T)
        S=1
        A=-PI3+HF*(LOG(-T)**2-LOG(1+1/T)**2)
       ELSEIF(T .LT. -1) THEN
        Y=-1-T
        S=-1
        A=LOG(-T)
        A=-PI6+A*(A+LOG(1+1/T))
       ELSE IF(T .LE. -HF) THEN
        Y=-(1+T)/T
        S=1
        A=LOG(-T)
        A=-PI6+A*(-HF*A+LOG(1+T))
       ELSE IF(T .LT. 0) THEN
        Y=-T/(1+T)
        S=-1
        A=HF*LOG(1+T)**2
       ELSE IF(T .LE. 1) THEN
        Y=T
        S=1
        A=0
       ELSE
        Y=1/T
        S=-1
        A=PI6+HF*LOG(T)**2
       ENDIF
       H=Y+Y-1
       ALFA=H+H
       B1=0
       B2=0
       DO 1 I = 19,0,-1
       B0=C(I)+ALFA*B1-B2
       B2=B1
    1  B1=B0
       H=-(S*(B0-H*B2)+A)
      ENDIF

      DDILOG=H

      RETURN
      END


c-----------------------------------------------------------------------

      subroutine NSIoperator(jp,kp,nl,muscale,ioption,alpha,Im2,Im1,Im0,
     -     stab)

c-----------------------------------------------------------------------
c...  Coefficient of the color correlator Tj.Tk in the I operator based 
c...  on Nagy-Soper subtraction scheme.
c...  The normalization does not include the coupling alphas/2pi.
c...  The input variable ioption sets the numerical integration mode:
c...   ioption = 0  -->  returns the exact I-operator for a given phase-
c...                     space point. Integration over the kinematics 
c...                     of the unconstrained parton is performed 
c...                     numerically using the Gauss-Legendre algorithm.
c...   ioption = 1  -->  returns the value of the I-operator for given
c...                     values of the integration variables of the 
c...                     unconstrained parton (randomly generated inside
c...                     the routine). Suitable for Monte Carlo summation.
c...  In the present version, only ioption=1 is available.
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'

      integer jp,kp,id_unconstr,ioption
      integer nl            !... nl = number of light quarks
      double precision muscale,xrandom1,xrandom2
      double precision mj,mk,mjsq,mksq
      double precision Im2,Im1,Im0

c...  phase-space restriction parameter (alphaMax)
      double precision alpha
c...  setup for self-initialization
      logical firstcall
      save firstcall
      data firstcall/.true./

      logical stab,stable
      dimension res(0:2)

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)

c-----------------------------------------------------------------------
c...  setup
c-----------------------------------------------------------------------

c...  masses of the emitter and spectator
      mj=parmas(ifl(jp))
      mk=parmas(ifl(kp))
      mjsq=mj**2
      mksq=mk**2

      Im2=0.d0
      Im1=0.d0
      Im0=0.d0

      stab=.true.

c-----------------------------------------------------------------------
c...  Sum all the integrated dipoles contributing in the case  
c...  of the partonic pair {j,k}.
c...  The variable id_unconstr denotes the flavour of the unconstrained
c...  parton. In general there is more than one possible choice for such
c...  flavour given a parton as emitter.
c...  For a given unconstrained parton flavour, the contribution to be 
c...  evaluated has the form (Wjj + Wjk), where Wjj is the diagonal (aka
c...  Collinear) term and Wjk is the interference (aka Soft) term in the
c...  Nagy-Soper formalism.
c-----------------------------------------------------------------------


c...  In the present version, only the Monte Carlo sum option (ioption=1) is 
c...  available
      if(ioption.eq.0)then
         print*,' '
         print*,'WARNING: numerical integration mode'
         print*,'         not available yet. STOP   '
         print*,' '
         STOP
      endif


c...  Generate two random variables using Parni (adaptive)
      if(firstcall)then
         call avh_parni_init(3 ,1 ,1,500,15000) !--> xrandom1
         call avh_parni_init(4 ,1 ,1,500,15000) !--> xrandom2
         firstcall=.false.
      endif
      call avh_parni_generate(3,xrandom1)
      call avh_parni_generate(4,xrandom2)
      weight_parni=1.d0/avh_parni_density(3,xrandom1)
      weight_parni=weight_parni/avh_parni_density(4,xrandom2)


      if(ifl(jp).ne.35)then

c...  quark emitter: q->qg
         id_unconstr=35         !...gluon
c... Wjj contribution
c         print*,'CALLING q->qg diag. jp,jp',jp,jp
         call NSintdipole(n,ifl,pmom,id_unconstr,jp,jp,
     &        xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
         if(.not.stable)then
            stab=.false.
            return
         endif
         Im2=Im2+res(2)
         Im1=Im1+res(1)
         Im0=Im0+res(0)

c... Wjk contribution
c         print*,'CALLING q->qg interf. jp,kp',jp,kp
         call NSintdipole(n,ifl,pmom,id_unconstr,jp,kp,
     &        xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
         if(.not.stable)then
            stab=.false.
            return
         endif
         Im2=Im2+res(2)
         Im1=Im1+res(1)
         Im0=Im0+res(0)

c...  In case of initial-state radiation, Wjj receives an additional contribution
c...  from the q->gq splitting (with quark as unconstrained parton).
         if(jp.le.2)then
c            print*,'CALLING q->gq diag. jp,jp',jp,jp
            id_unconstr=-ifl(jp)
            call NSintdipole(n,ifl,pmom,id_unconstr,jp,jp,
     &           xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            Im2=Im2+res(2)
            Im1=Im1+res(1)
            Im0=Im0+res(0)
         endif

      elseif(ifl(jp).eq.35) then

c...  gluon emitter. Two possible contributions: g->gg and g->qq (the 
c...  latter only if the quark is massless)

c...  1. g->gg
         id_unconstr=35         !...gluon
c... Wjj contribution
c         print*,'CALLING g->gg diag. jp,jp',jp,jp
         call NSintdipole(n,ifl,pmom,id_unconstr,jp,jp,
     &        xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
         if(.not.stable)then
            stab=.false.
            return
         endif
         Im2=Im2+res(2)
         Im1=Im1+res(1)
         Im0=Im0+res(0)
c... Wjk contribution
c         print*,'CALLING g->gg interf. jp,kp',jp,kp
         call NSintdipole(n,ifl,pmom,id_unconstr,jp,kp,
     &        xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
         if(.not.stable)then
            stab=.false.
            return
         endif
         Im2=Im2+res(2)
         Im1=Im1+res(1)
         Im0=Im0+res(0)

c...  2. g->qq
         id_unconstr=3          !...up-quark (any light-quark flavour is ok here)
c... Wjj contribution
c         print*,'CALLING g->qq diag. jp,jp',jp,jp
         call NSintdipole(n,ifl,pmom,id_unconstr,jp,jp,
     &        xrandom1,xrandom2,muscale,nl,alpha,kp,res,stable)
         if(.not.stable)then
            stab=.false.
            return
         endif

c... Results from the g->qq splitting must be rescaled by TF/CA in order the final
c... result is the coefficient of the color correlator Tj.Tk
         Im2=Im2+res(2)*nl*TF/CA
         Im1=Im1+res(1)*nl*TF/CA
         Im0=Im0+res(0)*nl*TF/CA
c... Unconstrained parton is not a gluon, hence no Wjk contribution applies in
c... this case.

      endif                     !...if(ifl(jp).ne.35)then

c... Add minus sign coming from color factor: -1/2[ tl x tk  + tk x tl]
      Im2=-Im2
      Im1=-Im1
      Im0=-Im0

c... Make Parni adapt to the finite result
      Im0=Im0*weight_parni
      wparni=Im0
      if(Im0.lt.0)wparni=-Im0
      call avh_parni_adapt(3,wparni,xrandom1)
      call avh_parni_adapt(4,wparni,xrandom2)

      return
      end




c-----------------------------------------------------------------------

      subroutine NSintdipole(nt,iproct,pmomt,idj,ipos_i,ipos_k,
     &     xrandom1,xrandom2,muscale,nl,alphaMax,kp,res,stable)


c-----------------------------------------------------------------------
c...  Coefficient of the color correlator Ti.Tk in the integrated dipole
c...  term Wik based on Nagy-Soper subtraction scheme.
c...  The normalization does not include the coupling alphas/2pi
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

c... Input variables
      integer nt,iproct,idj,ipos_i,ipos_k,nl,kp
      dimension iproct(mxpart)
      double precision pmomt,muscale,xrandom1,xrandom2,alphaMax
      dimension pmomt(mxpart,1:5)

c... Output: res(k,0)=finite part; res(k,1)=1/eps pole; res(k,2)=1/eps**2 pole
      double precision res
      dimension res(0:2)

      logical stable

c... Internal variables
      integer idit,idkt,mu,i,igo
      double precision Qsq,Ktsq,pxtot,pytot,pztot,Etot
      double precision pmomt_boost,z,ys,bcm,gcm
      dimension pmomt_boost(nt,1:4)
      double precision rmit,rmkt,rMK2
      double precision Pij,pti,ptk,pta,pa,Ktilde,K,pk
      dimension Pij(4),pti(4),ptk(4),pta(4),pa(4),Ktilde(4),K(4),pk(4)
      double precision pijsq,lambda,gamma,varej,varejmax
      double precision ctij,ctj,stj
      double precision Eti,Etk,ptkz,ptky,misq,mksq,Ei,Ej
      double precision Ea,Eta,paz,pay,masq
      double precision Eta0,Etk0
      double precision pimod,ptimod,ptkmod
      double precision ptiQ,Pigreco
      double precision pijdotpj,pidotQ,pjdotQ,pkdotQ,ptidotptk,ptkdotQ,
     -     ptidotQ,pijdotpti,pijdotQ,pijdotpk,pijdotptk,pjdotpti,
     -     pidotpj,pidotpti,padotpj,ptadotptk,padotptk,padotQ,pijdotpa,
     -     ptidotpa
      double precision parA,parB,parF,parL,parM,parN,parO,
     -     parP,parQ,parR,parS,parT,
     -     parAM,parBM,parGM,parFM,parHM,parLM,parMM,parNM,parOM,parPM,
     -     parQM,parRM,parSM,parTM,parUM,parVM,parXM,
     -     parAa,parAj,parAq,parBa,parBj,parBq,parAMj,parAMq,parBMj,
     -     parBMq

      double precision xjac

      logical alphaMaxcutoff


      double precision dot3,dot4
      external dot3,dot4

      parameter(Pigreco=3.141592653589793d0)

c...  color factors
      double precision CA,CF,TF
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)

      stable=.true.

c... Cutoff on input random numbers
      if(xrandom1.lt.1.d-10 .OR. abs(1.d0-xrandom1).lt.1.d-10)then
         res(2)=0.d0
         res(1)=0.d0
         res(0)=0.d0
         return
      endif
      if(xrandom2.lt.1.d-10 .OR. abs(1.d0-xrandom2).lt.1.d-10)then
         res(2)=0.d0
         res(1)=0.d0
         res(0)=0.d0
         return
      endif

c...  Default setup: no phase-space restriction
c...  Note: in the present version, no phase-space restriction is implemented for
c...  the Nagy-Soper scheme
      alphaMaxcutoff=.false.


c---------------------------------------------------------------------------------
c... 1. Identify the dipole contribution.
c---------------------------------------------------------------------------------

c...  Qsq = CM energy squared of the m-parton kinematics
      pxtot=0.d0
      pytot=0.d0
      pztot=0.d0
      Etot=0.d0
      do i=3,nt
         pxtot=pxtot+pmomt(i,1)
         pytot=pytot+pmomt(i,2)
         pztot=pztot+pmomt(i,3)
         Etot=Etot+pmomt(i,4)
      enddo
      Qsq = Etot**2-pxtot**2-pytot**2-pztot**2

c... ipos_i = position of the emitter parton
c... ipos_k = position of the spectator parton
      idit=iproct(ipos_i)
      idkt=iproct(ipos_k)
      rmit=pmomt(ipos_i,5)
      rmkt=pmomt(ipos_k,5)
      do mu=1,4
         pti(mu)=pmomt(ipos_i,mu)
         ptk(mu)=pmomt(kp,mu)
      enddo

c---------------------------------------------------------------------------------
c... Check if the selected particles {i,k} are eligible to give a dipole 
c... contribution, given the input ID (idj) of the kinematically 
c... unconstrained particle.
c---------------------------------------------------------------------------------
      igo=0
      if(idj.eq.35)then         ! unconstrained parton = gluon
         if(idit.eq.35
     &        .OR. abs(idit).eq.3 .OR. abs(idit).eq.4
     &        .OR. abs(idit).eq.7 .OR. abs(idit).eq.8
     &        .OR. abs(idit).eq.11 .OR. abs(idit).eq.12 )then ! i = gluon or q(bar)
            igo=1
         else
            igo=0
         endif
         if(idkt.eq.35
     &        .OR. abs(idkt).eq.3 .OR. abs(idkt).eq.4
     &        .OR. abs(idkt).eq.7 .OR. abs(idkt).eq.8
     &        .OR. abs(idkt).eq.11 .OR. abs(idkt).eq.12 )then ! k = gluon or q(bar)
            igo=igo*1
         else
            igo=igo*0
         endif
      elseif( abs(idj).eq.3 .OR. abs(idj).eq.4 .OR.
     &        abs(idj).eq.7 .OR. abs(idj).eq.8 .OR.
     &        abs(idj).eq.11 .OR. abs(idj).eq.12 )then ! unconstr. parton = q(bar)
         if ( idit.eq.35 .OR.
     &        (ipos_i.gt.2 .AND. idit.eq.idj ) .OR.
     &        (ipos_i.le.2 .AND. idit.eq.-idj ) )then ! i = gluon or q(bar)
            igo=1
         else
            igo=0
         endif
         if( idkt.eq.35 .OR.
     &        (ipos_k.gt.2 .AND. idkt.eq.idj) .OR.
     &        (ipos_k.le.2 .AND. idkt.eq.-idj) )then   ! k = gluon or q(bar)
            igo=igo*1
         else
            igo=igo*0
         endif
      else
         print*,'NSintdipole ERROR: idj is not a parton.'
         STOP
      endif
      do i=0,2
         res(i)=0.d0
      enddo
      if(igo.eq.0)then
         return
      endif

c---------------------------------------------------------------------------------
c... 2. Compute the dipole contribution corresponding to the selected pair {i,k}
c... given j as the unconstrained particle.
c---------------------------------------------------------------------------------

c... If the two indexes ipos_i and ipos_k are equal, the code evaluates the 
c... contribution of the "diagonal" (aka "collinear") integrated subtraction term 
c... (Wii-Wii_eikonal), otherwise the "interference" term (Wii_eikonal-Wik) is 
c... evaluated.

      if(ipos_i.eq.ipos_k)then
c---------------------------------------------------------------------------------
c...  COLLINEAR SUBTRACTION TERMS (Wii-Wii_eikonal): basic kinematics and formulae
c---------------------------------------------------------------------------------

c... Note that the diagonal term can only be collinear-divergent by construction, 
c... hence it gives a contribution to the 1/eps pole only if parton i is massless.
c... If particle i is massive, then (Wii - Wii_eikonal) is finite.


         if(ipos_i.ge.3)then
c---------------------------------------------------------------------------------
c...  FINAL state collinear subtraction 
c---------------------------------------------------------------------------------

c... Basic kinematics.
c... The input variables xrandom1,xrandom2 (random numbers in the range [0,1]) are 
c... used to define the basic quantities ctj and varej (ctj = cosine of the angle 
c... between pj and pti, in the range [-1,1]; varej = rescaled energy of the 
c... unconstrained parton pj, in the range [0,1])
c... These quantities, together with the list of momenta in the m-particle 
c... kinematics ("ptilde"), define unambiguously the (m+1)-particle kinematics.

            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            varej = xrandom2
            misq = rmit**2
            Eti = pti(4)
            ptimod = sqrt(Eti**2-misq)
            ptiQ = Eti*sqrt(Qsq)
            rMK2 = misq - 2.d0*ptiQ + Qsq
            if(rMK2.lt.0.d0) rMK2=0.d0
            varejmax = 
     -           (Sqrt(Qsq)*(-misq + (Sqrt(Qsq) - Sqrt(rMK2))**2))/
     -           (-misq + 2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq) + 
     -           (Sqrt(Qsq) - Sqrt(rMK2))**2)
            pijsq = 
     -           (2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq)*varej*varejmax + 
     -           misq*(Sqrt(Qsq) - varej*varejmax))/(Sqrt(Qsq) - 
     -           varej*varejmax)
            lambda = 
     -           Sqrt(((-misq + pijsq + 2.d0*ptiQ)**2 - 4.d0*pijsq*
     -           Qsq)/(ptiQ**2 - misq*Qsq))/2.d0
            gamma = 
     -           (-misq + pijsq - 2.d0*(-1.d0 + lambda)*ptiQ)/
     -           (2.d0*Qsq)
            Ej = 
     -           ((-misq + pijsq)*Sqrt(Qsq))/(-misq + pijsq + 2.d0*
     -           (Eti - ctj*lambda*ptimod)*Sqrt(Qsq))
            Ei =
     -           Sqrt(Ej**2 + misq - 2.d0*ctj*Ej*lambda*ptimod + 
     -           lambda**2*ptimod**2)
            pimod = sqrt(Ei**2 - misq)
            ctij = 
     -           (-Ej + ctj*lambda*ptimod)/Sqrt(Ej**2 - 2.d0*ctj*Ej*
     -           lambda*ptimod + lambda**2*ptimod**2)
            stj = sqrt(1.d0 - ctj**2)

c... Dot products
            pidotQ = Ei*Sqrt(Qsq)
            pjdotQ = Ej*Sqrt(Qsq)
            pijdotQ = pidotQ+pjdotQ
            ptidotQ = Eti*Sqrt(Qsq)
            pidotpj = Ej*(Ei-pimod*ctij)
            pijdotpti = gamma*ptidotQ+lambda*misq
            pjdotpti = Ej*(Eti-ptimod*ctj)
            pidotpti = pijdotpti-pjdotpti
            pijdotpj = pidotpj

c...  Phase-space restriction (to be implemented)
c...  
c...  if(.CUTOFF.)then
c...      alphaMaxcutoff=.true.
c...  endif


c... Formulae

            if(idit.ne.35 .AND. idj.eq.35)then 

               if(misq.eq.0.d0)then
c---------------------------------------------------------------------------------
c... q -> qg splitting (FINAL-FINAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c               print*,"CALLING q->qg FINAL MASSLESS"

c... The (1/epsilon**2) term
                  res(2) = 
     -                 0.d0
c... The (1/epsilon) term
                  res(1) = 
     -                 -1.d0/2.d0
c... The finite term
                  parA  =  (Eti*Qsq*varej)/(Pigreco*(Eti - Eti*varej
     -                 + Sqrt(Qsq)*varej)**3)
                  res(0) = 
     -                 Log((2.d0*Eti)/muscale)/xjac
     -                 +  (parA*Pigreco)/(-1.d0 + ctj)
c...  Phase-space restriction (alphaMax)
                  if(.NOT.alphaMaxcutoff)then
                     parL  =  ((-1.d0 + ctj)*Ej**2*lambda*(-Eti + ctj*
     -                    ptimod)*(-pijsq + 2.d0*(Ej - Eti)*Sqrt(Qsq))*
     -                    Sqrt(Qsq)*(-2.d0*pjdotQ*(2.d0*pidotQ**2 + 
     -                    pjdotQ**2 + 2.d0*pidotQ*(pjdotQ - ptidotQ))*
     -                    ptidotQ + (pjdotQ*(pjdotpti*(pidotQ + pjdotQ)
     -                    + pidotpti*(2.d0*pidotQ + pjdotQ)) + pidotpj*
     -                    (2.d0*pidotQ + pjdotQ)*ptidotQ - 2.d0*pidotpj*
     -                    ptidotQ**2)*Qsq - pidotpj*pidotpti*Qsq**2)*
     -                    varejmax)/(2.d0*Ei*pidotpj*Pigreco*pijsq*
     -                    pjdotQ**2*ptidotQ**2*(Sqrt(Qsq) - varej*
     -                    varejmax)**2)
                     res(0) = res(0) + 
     -                    (parL*Pigreco)/(1.d0 - ctj)
                  endif
               endif

               if(misq.ne.0.d0)then
c---------------------------------------------------------------------------------
c... q -> qg splitting (FINAL-FINAL, MASSIVE)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c               print*,"CALLING q->qg MASSIVE"

c... The (1/epsilon**2) term
                  res(2) = 
     -                 0.d0
c... The (1/epsilon) term
                  res(1) = 
     -                 0.d0
c... The finite term
c...  Phase-space restriction (alphaMax)
                  if(.NOT.alphaMaxcutoff)then
                     parL  =  
     -                    (Ej**2*lambda*(Eti - ctj*ptimod)*(misq - 
     -                    pijsq + 2.d0*(Ej - Eti)*Sqrt(Qsq))*Sqrt(Qsq)*
     -                    (misq/pidotpj**2 - (2.d0*pidotQ)/(pidotpj*
     -                    pjdotQ) + Qsq/pjdotQ**2 - ((misq*pjdotQ**2 - 
     -                    pidotpj*pjdotQ*(2.d0*pidotQ + pjdotQ) + 
     -                    pidotpj**2*Qsq)*(pidotQ - (pidotpti*Qsq)/(
     -                    ptidotQ + Sqrt(ptidotQ**2 - misq*Qsq))) + 
     -                    pjdotQ*(misq*pjdotQ - pidotpj*(pidotQ + 
     -                    pjdotQ))*(pjdotQ - (pjdotpti*Qsq)/(ptidotQ + 
     -                    Sqrt(ptidotQ**2 - misq*Qsq))) + pidotpj**2*
     -                    pjdotQ*(Qsq - (ptidotQ*Qsq)/(ptidotQ + Sqrt(
     -                    ptidotQ**2 - misq*Qsq))))/(pidotpj**2*
     -                    pjdotQ**2*(ptidotQ - (misq*Qsq)/(ptidotQ + 
     -                    Sqrt(ptidotQ**2 - misq*Qsq)))))*varejmax)/(
     -                    Ei*Pigreco*(misq - pijsq)*(Sqrt(Qsq) - 
     -                    varej*varejmax)**2)
                     res(0) = parL*Pigreco
                  endif
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) 
     -              .OR. res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in q->qg FINAL'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'ptimod',ptimod
                  print*,'misq',misq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parL',parL
                  STOP
#endif
               endif
            endif

            if(idit.ne.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... q -> gq splitting (FINAL-FINAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c               print*,"CALLING q->gq FINAL"

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
     -              -1.d0/2.d0
c... The finite term
               parA  =  (Eti**2*Sqrt(Qsq)*(-1.d0 + varej))/(Pigreco*
     -              (Eti*(-1.d0 + varej) - Sqrt(Qsq)*varej)**3)
               res(0) = 
     -              (-1.d0 + Log((2.d0*Eti)/muscale))/xjac
     -              +  (parA*Pigreco)/(-1.d0 + ctj)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parL  =  -(((1.d0 - ctj)*Ej**2*lambda*(Eti - ctj*
     -                 ptimod)*(-pijsq + 2.d0*(Ej - Eti)*Sqrt(Qsq))*
     -                 Sqrt(Qsq)*((-2.d0*pidotQ*pjdotQ)/pidotpj + Qsq
     -                 - ((pidotpj*pidotQ*Sqrt(ptidotQ**2)*Qsq)/(ptidotQ
     -                 + Sqrt(ptidotQ**2)) + pidotQ*(-pidotQ - pjdotQ)*(
     -                 pidotQ - (pidotpti*Qsq)/(ptidotQ + Sqrt(
     -                 ptidotQ**2))) - (pidotQ**2 + 2.d0*pidotQ*pjdotQ -
     -                 pidotpj*Qsq)*(pjdotQ - (pjdotpti*Qsq)/(ptidotQ +
     -                 Sqrt(ptidotQ**2))))/(pidotpj*ptidotQ))*varejmax)/
     -                 (Ei*pidotQ**2*Pigreco*pijsq*(Sqrt(Qsq)- varej*
     -                 varejmax)**2))
                  res(0) = res(0) + 
     -                 (parL*Pigreco)/(1.d0 - ctj)
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) 
     -              .OR. res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in q->gq FINAL'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'ptimod',ptimod
                  print*,'misq',misq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parL',parL
                  STOP
#endif
               endif
            endif

            if(idit.eq.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... g -> qq splitting (FINAL-FINAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c               print*,"CALLING g->qq FINAL"

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
     -              -2.d0/3.d0
c... The finite term
               parA  =  (Eti*Sqrt(Qsq)*(Eti**2*(-1.d0 + varej)**2 +
     -              Qsq*varej**2))/(Pigreco*(Eti - Eti*varej + 
     -              Sqrt(Qsq)*varej)**4)
               res(0) = 
     -              ((2.d0*(-5.d0 - 6.d0*Log(muscale/(2.d0*Eti))))/
     -              9.d0)/xjac
     -              +  (parA*Pigreco)/(-1.d0 + ctj)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parL  =  -(((-1.d0 + ctj)*Ej**2*lambda*(-Eti + 
     -                 ctj*ptimod)*(-pijsq + 2.d0*(Ej - Eti)*Sqrt(Qsq))
     -                 *Sqrt(Qsq)*(4.d0*(pidotQ**2 + pjdotQ**2)*
     -                 ptidotQ**2 - 4.d0*(pidotpti*pidotQ + pjdotpti*
     -                 pjdotQ)*ptidotQ*Qsq + (pidotpti**2 + pjdotpti**2)
     -                 *Qsq**2)*varejmax)/(Ei*pidotpj*Pigreco*pijsq*(
     -                 -2.d0*(pidotQ + pjdotQ)*ptidotQ + (pidotpti + 
     -                 pjdotpti)*Qsq)**2*(Sqrt(Qsq) - varej*varejmax)**2
     -                 ))
                  res(0) = res(0) + 
     -                 (parL*Pigreco)/(1.d0 - ctj)
               endif


c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) 
     -              .OR. res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in g->qq FINAL'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'ptimod',ptimod
                  print*,'misq',misq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parL',parL
                  STOP
#endif
               endif
            endif

            if(idit.eq.35 .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c... g -> gg splitting (FINAL-FINAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c               print*,"CALLING g->gg FINAL"

c... The (1/epsilon**2) term
               res(2) = 0.d0
     -              
c... The (1/epsilon) term
               res(1) = 
     -              -1.d0/6.d0
c... The finite term
               parA  =  -((Eti**2*Qsq*(-1.d0 + varej)*varej)/(Pigreco*
     -              (Eti - Eti*varej + Sqrt(Qsq)*varej)**4))
               res(0) = 
     -              (-5.d0 - 6.d0*Log(muscale/(2.d0*Eti)))/18.d0/xjac
     -              +  (parA*Pigreco)/(-1.d0 + ctj)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parL  =  -((-1.d0 + ctj)*Ej**2*lambda*(-Eti + ctj*
     -                 ptimod)*(-pijsq + 2.d0*(Ej - Eti)*Sqrt(Qsq))*
     -                 Sqrt(Qsq)*(8.d0*pidotQ**3*pjdotQ**3*ptidotQ**2 +
     -                 2.d0*pidotQ*pjdotQ*ptidotQ*(3.d0*pidotQ**3*
     -                 pjdotpti + pidotQ**2*((-3.d0*pidotpti + 2.d0*
     -                 pjdotpti)*pjdotQ - 3.d0*pidotpj*ptidotQ) + 
     -                 pidotQ*pjdotQ*((-6.d0*pidotpti + pjdotpti)*pjdotQ
     -                 - 2.d0*pidotpj*ptidotQ) + pjdotQ**2*(-(pidotpti*
     -                 pjdotQ) + pidotpj*ptidotQ))*Qsq + pidotQ*(pjdotQ*
     -                 (-3.d0*pidotQ**2*pjdotpti*(pidotpti + pjdotpti) +
     -                 pidotQ*(3.d0*pidotpti**2 + 4.d0*pidotpti*pjdotpti
     -                 - pjdotpti**2)*pjdotQ + pidotpti*(pidotpti + 
     -                 pjdotpti)*pjdotQ**2) + pidotpj*(-4.d0*pidotQ**2*
     -                 pjdotpti + 5.d0*pidotQ*(pidotpti - pjdotpti)*
     -                 pjdotQ + (7.d0*pidotpti - 3.d0*pjdotpti)*
     -                 pjdotQ**2)*ptidotQ + 2.d0*pidotpj**2*(2.d0*pidotQ
     -                 + pjdotQ)*ptidotQ**2)*Qsq**2 + pidotpj*(pidotQ*(
     -                 2.d0*pidotQ*pjdotpti*(pidotpti + pjdotpti) + (
     -                 -pidotpti**2 - 2.d0*pidotpti*pjdotpti + 
     -                 pjdotpti**2)*pjdotQ) - 2.d0*pidotpj*pidotpti*
     -                 (pidotQ + pjdotQ)*ptidotQ)*Qsq**3 + pidotpj**2*
     -                 pidotpti*pjdotpti*Qsq**4)*varejmax)/(2.d0*Ei*
     -                 pidotpj*pidotQ**2*Pigreco*pijsq*pjdotQ**2*(-2.d0*
     -                 (pidotQ + pjdotQ)*ptidotQ + (pidotpti + pjdotpti)
     -                 *Qsq)**2*(Sqrt(Qsq) - varej*varejmax)**2)
                  res(0) = res(0) + 
     -                 (parL*Pigreco)/(1.d0 - ctj)
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) 
     -              .OR. res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in g->gg FINAL'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'ptimod',ptimod
                  print*,'misq',misq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parL',parL
                  STOP
#endif
               endif
            endif


         elseif(ipos_i.eq.1 .OR. ipos_i.eq.2)then
c---------------------------------------------------------------------------------
c...  INITIAL state collinear subtraction 
c---------------------------------------------------------------------------------

c... Formulae

            if(idit.ne.35 .AND. idj.eq.35)then 
c---------------------------------------------------------------------------------
c... q -> qg splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c              print*,"CALLING q->qg INITIAL"
               xjac = 2.d0      !... jacobian of the transf. xrandom1 -> ctj
               Ktsq = Qsq
               if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
                  print*,'NSintdipole: nonzero initial-state parton'//
     -                 ' mass detected',rmit
                  print*,'             Initial state partons must be'//
     -                 ' massless.'
                  STOP
               endif

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
c... O(1/epsilon) Dirac delta contribution from the collinear counterterm
     -              3.d0/2.d0
c... The finite term
c... Note: the O(eps**0) Dirac delta contribution from the Altarelli-Parisi 
c... splitting kernel Pqq, being mu_F (factorization scale) dependent, is not 
c... included in the I operator but in the KP operator.
c... This delta(1-x) term is required for consistency with the 
c... KP-operator implementation (delta contribution)
               res(0) = 
     -              3.d0/2.d0*Log(muscale**2/Ktsq)/xjac
            endif

            if(idit.ne.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... q -> gq splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c              print*,"CALLING q->gq INITIAL"

               if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
                  print*,'NSintdipole: nonzero initial-state parton'//
     -                 ' mass detected',rmit
                  print*,'             Initial state partons must be'//
     -                 ' massless.'
                  STOP
               endif

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
     -              0.d0
c... The finite term is considered in the KP operator
               res(0) = 
     -              0.d0
            endif

            if(idit.eq.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... g -> qq splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c              print*,"CALLING g->qq INITIAL"

               if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
                  print*,'NSintdipole: nonzero initial-state parton'//
     -                 ' mass detected',rmit
                  print*,'             Initial state partons must be'//
     -                 ' massless.'
                  STOP
               endif

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
     -              0.d0
c... The finite term is considered in the KP operator
               res(0) = 
     -              0.d0
            endif

            if(idit.eq.35 .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c... g -> gg splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c              print*,"CALLING g->gg INITIAL"
               xjac = 2.d0      !... jacobian of the transf. xrandom1 -> ctj
               Ktsq=Qsq
               if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
                  print*,'NSintdipole: nonzero initial-state parton'//
     -                 ' mass detected',rmit
                  print*,'             Initial state partons must be'//
     -                 ' massless.'
                  STOP
               endif

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
               res(1) = 
c... O(1/epsilon) Dirac delta contribution from the collinear counterterm
c... (Altarelli-Parisi splitting kernel Pgg divided by overall color factor)
     -              (11d0/6*CA-2d0/3*TF*nl)/CA
c... The finite term
c... Note: the O(eps**0) Dirac delta contribution from the Altarelli-Parisi 
c... splitting kernel Pgg, being mu_F (factorization scale) dependent, is not 
c... included in the I operator but in the KP operator.
c... This delta(1-x) term is required for consistency with the 
c... KP-operator implementation (delta contribution)
               res(0) = 
     -              (11d0/6*CA-2d0/3*TF*nl)/CA*Log(muscale**2/Ktsq)/xjac
            endif

         else                   !...something went wrong with ipos_i

            print*,'NSintdipole ERROR: check ipos_i',ipos_i
            STOP

         endif                  !if(ipos_i.ge.3)then



      elseif(ipos_i.ne.ipos_k .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c... INTERFERENCE TERMS (Wii_eikonal-Wik): basic kinematics and formulae
c---------------------------------------------------------------------------------


         if(ipos_i.ge.3 .AND. ipos_k.ge.3)then
c---------------------------------------------------------------------------------
c...  FINAL-FINAL state interference
c---------------------------------------------------------------------------------

c... Basic kinematics
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            varej = xrandom2
            misq = rmit**2
            mksq = rmkt**2
            if(nt.eq.4)then
c... Using simplified formulae in case of two-particle final states (in the 
c... m-particle kinematics) to minimize instabilities due to limited numerical 
c... precision
               Eti = (Qsq + misq - mksq)/(2.d0*sqrt(Qsq))
               Etk = (Qsq + mksq - misq)/(2.d0*sqrt(Qsq))
               ptimod = sqrt(Eti**2-misq)
               ptkmod = sqrt(Etk**2-mksq)
               ptkz = -ptimod
               ptky = 0.d0
               ptiQ = Eti*sqrt(Qsq)
               rMKt = Sqrt(mksq)
            elseif(nt.gt.4)then
               Eti = pti(4)
               Etk = ptk(4)
               ptimod = sqrt(Eti**2-misq)
               ptkmod = sqrt(Etk**2-mksq)
               ptkz = dot3(ptk,pti)/ptimod
               ptky = sqrt(Etk**2 -mksq - ptkz**2)
               ptiQ = Eti*sqrt(Qsq)
               rMKt = Sqrt(misq - 2.d0*ptiQ + Qsq)
            endif
            varejmax = 
     -           (Sqrt(Qsq)*(-misq + (Sqrt(Qsq) - 
     -           rMKt )**2))/
     -           (-misq + 2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq) + 
     -           (Sqrt(Qsq) - rMKt )**2)
            pijsq = 
     -           (2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq)*varej*varejmax + 
     -           misq*(Sqrt(Qsq) - varej*varejmax))/(Sqrt(Qsq) - 
     -           varej*varejmax)
            lambda = 
     -           Sqrt(((-misq + pijsq + 2.d0*ptiQ)**2 - 4.d0*pijsq*
     -           Qsq)/(ptiQ**2 - misq*Qsq))/2.d0
            gamma = 
     -           (-misq + pijsq - 2.d0*(-1.d0 + lambda)*ptiQ)/
     -           (2.d0*Qsq)
            Ej = 
     -           ((-misq + pijsq)*Sqrt(Qsq))/(-misq + pijsq + 2.d0*
     -           (Eti - ctj*lambda*ptimod)*Sqrt(Qsq))
            Ei =
     -           Sqrt(Ej**2 + misq - 2.d0*ctj*Ej*lambda*ptimod + 
     -           lambda**2*ptimod**2)
            pimod = sqrt(Ei**2 - misq)
            ctij = 
     -           (-Ej + ctj*lambda*ptimod)/Sqrt(Ej**2 - 2.d0*ctj*Ej*
     -           lambda*ptimod + lambda**2*ptimod**2)
            stj = sqrt(1.d0 - ctj**2)

c... Dot products
            pidotQ = Ei*Sqrt(Qsq)
            pjdotQ = Ej*Sqrt(Qsq)
            pijdotQ = pidotQ+pjdotQ
            ptidotQ = Eti*Sqrt(Qsq)
            pidotpj = Ej*(Ei-pimod*ctij)
            pijdotpti = gamma*ptidotQ+lambda*misq
            pjdotpti = Ej*(Eti-ptimod*ctj)
            pidotpti = pijdotpti-pjdotpti
            pijdotpj = pidotpj
            do i=1,3
               pij(i)=lambda*pti(i)
               K(i)=-pij(i)
               Ktilde(i)=-pti(i)
            enddo
            pij(4)=gamma*sqrt(Qsq)+lambda*pti(4)
            K(4)=sqrt(Qsq)-pij(4)
            Ktilde(4)=sqrt(Qsq)-pti(4)
            call Lorentz(nt,K,Ktilde,ptk,pk)
            ptkdotQ = Etk*Sqrt(Qsq)
            ptidotptk = dot4(pti,ptk)
            pijdotptk=lambda*dot4(pti,ptk)+gamma*ptkdotQ
            pkdotQ = pk(4)*Sqrt(Qsq)
            pijdotpk=lambda*dot4(pti,pk)+gamma*pkdotQ

c...  Phase-space restriction (to be implemented)
c...  
c...  if(.CUTOFF.)then
c...      alphaMaxcutoff=.true.
c...  endif


            if(misq.eq.0.d0)then
c---------------------------------------------------------------------------------
c... FINAL-FINAL state interference -- MASSLESS CASE
c---------------------------------------------------------------------------------
c               print*,'CALLING FINAL-FINAL INTERFERENCE MASSLESS'

c... The (1/epsilon**2) term
               res(2) = 
     -              1.d0
c... The (1/epsilon) term
c... DeltaW contribution to the 1/epsilon pole
               res(1) = 
     -              2.d0 + Log((Eti*(Eti*mksq + 2.d0*Etk*ptidotptk))/
     -              ptidotptk**2) + Log(muscale**2/Qsq) + 
     -              Log(Qsq/(4.d0*Eti**2))
c... Mass term contribution (proportional to mksq) to the 1/epsilon pole
               if(mksq.ne.0.d0)then
                  res(1) = res(1) +
     -                 (Eti*mksq*Log(1.d0 + (2.d0*Etk*ptidotptk)/(
     -                 Eti*mksq)))/(2.d0*Etk*ptidotptk)
               endif

c... The finite term
c... DeltaW contribution to the finite term
               parN  =  2.d0*(Etk - ptkz)*(Log(4.d0) + Log(1.d0 + ctj) -
     -              Log(muscale**2/Qsq) + 2.d0*Log((Eti - Sqrt(Qsq) + 
     -              Sqrt(-2.d0*Eti*Sqrt(Qsq) + Qsq))/(ctj*Eti - 
     -              Sqrt(Qsq) + Sqrt(-2.d0*Eti*Sqrt(Qsq) + Qsq))))
               parO  =  2.d0*(Etk - ptkz)
               parP  =  (-2.d0 + ctj)*Etk*Pigreco + ctj*Pigreco*ptkz
               parQ  =  Sqrt(1.d0 - ctj**2)*Pigreco*ptky
               res(0) = 
     -              (-Log(4.d0) + Log(muscale**2/Qsq))**2/2.d0/xjac
     -              +  (-(((-2.d0 + (2.d0*(Etk - ptkz))/Sqrt((
     -              -1.d0 + ctj**2)*ptky**2 + ((-2.d0 + ctj)*Etk + 
     -              ctj*ptkz)**2))*Log(1.d0 - ctj))/(-1.d0 + ctj)) - 
     -              (Log(4.d0) - (Pigreco*(parN + parO*Log(((
     -              parP - parQ)*(parP + parQ))/(parP - Sqrt(
     -              parP**2 - parQ**2))**2)))/Sqrt((parP - parQ)*(
     -              parP + parQ)) - 2.d0*Log(muscale**2/Qsq))/(
     -              1.d0 - ctj))/2.d0
     -              +  (8.d0 - 4.d0*Log(4.d0) + 4.d0*Log(
     -              muscale**2/Qsq) + (4.d0 - 2.d0*Log(4.d0) + 
     -              2.d0*Log(muscale**2/Qsq))*Log(Qsq/Eti**2) + 
     -              Log(Qsq/Eti**2)**2)/2.d0/xjac
     -              +  (2.d0 - (2.d0*(Etk - ptkz))/Sqrt((-1.d0 + ctj**2)
     -              *ptky**2 + ((-2.d0 + ctj)*Etk + ctj*ptkz)**2) + 
     -              (2.d0*Eti**2*(-1.d0 + varej))/(Eti - Eti*varej + 
     -              Sqrt(Qsq)*varej)**2)/(varej - ctj*varej)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  if(nt.gt.4)then
c... The final state in the m-particle kinematics has more than two particles
                     parA  = Ej*Etk - ctj*Ej*ptkz + (2.d0*(pijdotpj - 
     -                    pjdotQ)*(ptidotptk - ptkdotQ))/(-2.d0*ptidotQ
     -                    + Qsq) - (2.d0*(pijdotpj + pjdotpti - 2.d0*
     -                    pjdotQ)*(pijdotptk + ptidotptk - 2.d0*ptkdotQ)
     -                    )/(2.d0*pijdotpti - 4.d0*pijdotQ + pijsq - 
     -                    4.d0*ptidotQ + 4.d0*Qsq)
                     parB  = -(Sqrt(1.d0 - ctj**2)*Ej*ptky)
                  elseif(nt.eq.4)then
c... The final state in the m-particle kinematics has only two particles. In this
c... case the standard formula for the Lorentz boost is singular for Kt**2 -> 0 and
c... one needs an alternative representation for the Lorentz transformation which 
c... reflects in a different definition of the parameter parA.
                     parA = -pijdotpj + pjdotpti + Ej*(Etk - ctj*ptkz)
                     parB = -(Sqrt(1.d0 - ctj**2)*Ej*ptky)
                  else
                     print*,'NSintdipole.f: check nt value'
                     STOP
                  endif
                  parF = (2.d0*(-1.d0 + ctj)**2*Ej**2*Eti*lambda*(
     -                 pijdotQ - pjdotQ)*(-pijsq + 2.d0*(Ej - Eti)*
     -                 Sqrt(Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*Pigreco*
     -                 pijdotpj*pijsq*(Sqrt(Qsq) - varej*varejmax)**2)
                  parL  =  parA*pijdotQ - parA*pjdotQ + pijdotpj*pkdotQ
                  res(0) = res(0) + 
     -                 (-((parF*Pigreco*(-parL + pidotQ*(parA - 
     -                 pijdotpk) + Sqrt(parL**2 + (-1.d0 + ctj**2)*
     -                 Ej**2*pidotQ**2*ptky**2)))/((-1.d0 + ctj)*pidotQ*
     -                 Sqrt(parL**2 + (-1.d0 + ctj**2)*Ej**2*pidotQ**2*
     -                 ptky**2)*varej)))
               endif

c... Mass term contribution (proportional to mksq) to the finite term
               if(mksq.ne.0.d0)then
                  parRM = Log(4.d0 - 4.d0*ctj**2) - Log(muscale**2/Qsq)
     -                 + 2.d0*Log((Eti - Sqrt(Qsq) + Sqrt(-2.d0*Eti*
     -                 Sqrt(Qsq) + Qsq))/(ctj*Eti - Sqrt(Qsq) + 
     -                 Sqrt(-2.d0*Eti*Sqrt(Qsq) + Qsq)))
                  parTM = -Etk + ctj*ptkz
                  parVM = (-2.d0 + ctj)*Etk + ctj*ptkz
                  res(0) = res(0) + 
     -                 (mksq*(parRM*Sqrt(parTM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2) - parRM*Sqrt(parVM**2 + (-1.d0 + ctj**2)
     -                 *ptky**2) - Sqrt(parVM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2)*Log((parTM**2 + (-1.d0 + ctj**2)*ptky**2
     -                 )/(parTM - Sqrt(parTM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2))**2) + Sqrt(parTM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2)*Log((parVM**2 + (-1.d0 + ctj**2)*ptky**2
     -                 )/(parVM - Sqrt(parVM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2))**2)))/(2.d0*(parTM - parVM)*Sqrt((
     -                 parTM**2 + (-1.d0 + ctj**2)*ptky**2)*(parVM**2 +
     -                 (-1.d0 + ctj**2)*ptky**2)))
     -                 +  (-((mksq*(-(1.d0/Sqrt((-2.d0 + ctj)**2*Etk**2
     -                 + (-1.d0 + ctj**2)*ptky**2 + 2.d0*(-2.d0 + ctj)*
     -                 ctj*Etk*ptkz + ctj**2*ptkz**2)) + 1.d0/Sqrt((
     -                 -1.d0 + ctj**2)*ptky**2 + (Etk - ctj*ptkz)**2)))/
     -                 ((-1.d0 + ctj)*Etk*varej)))
c...  Phase-space restriction (alphaMax)
                  if(.NOT.alphaMaxcutoff)then
                     parAM = parA
                     parBM = parB
                     parFM = ((-1.d0 + ctj)*Ej**2*Eti*lambda*mksq*pidotQ
     -                    *(pijsq + 2.d0*(-Ej + Eti)*Sqrt(Qsq))*Sqrt(
     -                    Qsq)*varej*varejmax)/(Ei*Pigreco*pijsq*(Sqrt(
     -                    Qsq) - varej*varejmax)**2)
                     parGM = parAM*pidotQ + pidotpj*pkdotQ
                     parHM = parBM*pidotQ
                     if(nt.gt.4)then
                        res(0) = res(0) + 
     -                       (parFM*(parBM/Sqrt(parAM**2 - parBM**2) - 
     -                       parHM/Sqrt(parGM**2 - parHM**2))*Pigreco)/
     -                       (parBM*parGM*varej - parAM*parHM*varej)
                     elseif(nt.eq.4)then
c... Expand around ptky=0
                        res(0) = res(0) + 
     -                       (parFM*Pigreco*(-(Sqrt(parAM**2)*
     -                       parGM**2) + parAM**2*Sqrt(parGM**2)*
     -                       pidotQ))/(parAM**2*parGM**2*varej*
     -                       (-parGM + parAM*pidotQ))
                     endif
                  endif
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR.
     -              res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in FIN-FIN int.massles'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'Etk',Etk
                  print*,'ptimod',ptimod
                  print*,'ptkmod',ptkmod
                  print*,'ptkz',ptkz
                  print*,'ptky',ptky
                  print*,'misq',misq
                  print*,'mksq',mksq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parB',parB
                  print*,'parF',parF
                  print*,'parL',parL
                  print*,'parM',parM
                  print*,'parN',parN
                  print*,'parO',parO
                  print*,'parP',parP
                  print*,'parQ',parQ
                  if(mksq.ne.0.d0)then
                     print*,'parAM',parAM
                     print*,'parBM',parBM
                     print*,'parFM',parFM
                     print*,'parGM',parGM
                     print*,'parHM',parHM
                     print*,'parRM',parRM
                     print*,'parTM',parTM
                     print*,'parVM',parVM
                  endif
                  STOP
#endif
               endif
            endif


            if(misq.ne.0.d0)then
c---------------------------------------------------------------------------------
c... FINAL-FINAL state interference -- MASSIVE CASE
c---------------------------------------------------------------------------------
c               print*,'CALLING FINAL-FINAL INTERFERENCE MASSIVE'

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0
c... The (1/epsilon) term
c... DeltaW contribution to the 1/epsilon pole
               res(1) = 
     -              (ptidotptk*Log(((Eti - ptimod)*(Eti**3*Etk**2 + 
     -              Eti**2*(Etk**2 - mksq)*ptimod - Etk**2*ptimod**3 - 
     -              Sqrt((-(misq*mksq) + ptidotptk**2)*(Eti**2*Etk - 
     -              Eti*ptidotptk + 2.d0*Eti*Etk*ptimod + Etk*ptimod**2
     -              )**2) - Eti*(ptidotptk**2 + Etk**2*ptimod**2 + 
     -              mksq*ptimod**2)))/((Eti + ptimod)*(Eti**3*Etk**2 + 
     -              Eti**2*(-Etk**2 + mksq)*ptimod + Etk**2*ptimod**3 - 
     -              Eti*(ptidotptk**2 + Etk**2*ptimod**2 + mksq*
     -              ptimod**2) - Sqrt((-(misq*mksq) + ptidotptk**2)*
     -              (Eti**2*Etk + Etk*ptimod**2 - Eti*(ptidotptk + 
     -              2.d0*Etk*ptimod))**2)))))/Sqrt(-(misq*mksq) + 
     -              ptidotptk**2)
c... Mass term contribution (misq) to the 1/epsilon pole
               res(1) = res(1) +
     -              (2.d0 + (Etk*misq*Log(((Eti - ptimod)*(-(Etk**2*
     -              ptimod*(2.d0*Eti + ptimod)) + 2.d0*Eti**2*Etk*ptkz + 
     -              Eti*((Eti + ptimod)*ptky**2 + Eti*ptkz**2) - Sqrt((
     -              Etk*ptimod + Eti*(2.d0*Etk + ptkz))**2*(Eti**2*
     -              ptky**2 + (Etk*ptimod - ptimod*ptky - Eti*ptkz)*
     -              (ptimod*(Etk + ptky) - Eti*ptkz)))))/((Eti + 
     -              ptimod)*(Etk**2*ptimod**2 + Eti*ptimod*(-2.d0*Etk**2
     -              + ptky**2) - Eti**2*(ptky**2 - 2.d0*Etk*ptkz + 
     -              ptkz**2) - Sqrt((-2.d0*Eti*Etk + Etk*ptimod + 
     -              Eti*ptkz)**2*(Eti**2*ptky**2 + (Etk*ptimod - 
     -              ptimod*ptky - Eti*ptkz)*(ptimod*(Etk + ptky) - 
     -              Eti*ptkz)))))))/(Eti*Sqrt(Eti**2*ptky**2 + 
     -              (Etk*ptimod - ptimod*ptky - Eti*ptkz)*(ptimod*(
     -              Etk + ptky) - Eti*ptkz))))/2.d0

c... Mass term contribution (mksq) to the 1/epsilon pole
               if(mksq.ne.0.d0)then
                  res(1) = res(1) + 
     -                 (Eti*mksq*Log(((-(Etk**2*ptimod) + ptimod*ptky**2
     -                 + Etk*(Eti + ptimod)*ptkz - Eti*(ptky**2 + 
     -                 ptkz**2) - Sqrt((Etk - ptkz)**2*(Eti**2*ptky**2 + 
     -                 (Etk*ptimod - ptimod*ptky - Eti*ptkz)*(ptimod*(
     -                 Etk + ptky) - Eti*ptkz))))*(-(Etk**2*ptimod**2) + 
     -                 Eti*ptimod*(-2.d0*Etk**2 + ptky**2) + Eti**2*(
     -                 ptky**2 + 2.d0*Etk*ptkz + ptkz**2) - Sqrt((
     -                 2.d0*Eti*Etk + Etk*ptimod + Eti*ptkz)**2*(
     -                 Eti**2*ptky**2 + (Etk*ptimod - ptimod*ptky - 
     -                 Eti*ptkz)*(ptimod*(Etk + ptky) - Eti*ptkz)))))/
     -                 ((-(Etk**2*ptimod) + (Eti + ptimod)*ptky**2 + 
     -                 Etk*(Eti - ptimod)*ptkz + Eti*ptkz**2 - Sqrt((
     -                 Etk + ptkz)**2*(Eti**2*ptky**2 + (Etk*ptimod - 
     -                 ptimod*ptky - Eti*ptkz)*(ptimod*(Etk + ptky) - 
     -                 Eti*ptkz))))*(Etk**2*ptimod**2 + Eti*ptimod*(
     -                 -2.d0*Etk**2 + ptky**2) - Eti**2*(ptky**2 - 
     -                 2.d0*Etk*ptkz + ptkz**2) - Sqrt((-2.d0*Eti*Etk +
     -                 Etk*ptimod + Eti*ptkz)**2*(Eti**2*ptky**2 + 
     -                 (Etk*ptimod - ptimod*ptky - Eti*ptkz)*(ptimod*(
     -                 Etk + ptky) - Eti*ptkz)))))))/(2.d0*Etk*Sqrt(
     -                 Eti**2*ptky**2 + (Etk*ptimod - ptimod*ptky - 
     -                 Eti*ptkz)*(ptimod*(Etk + ptky) - Eti*ptkz)))
               endif

c... The finite term
c... DeltaW contribution to the finite term
               parN  =  (2.d0*Eti*(-(Eti*Etk) + ptimod*ptkz))/(Pigreco*(
     -              Eti - ctj*ptimod))
               parO = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
               parP = Sqrt(1.d0 - ctj**2)*Eti*ptky
               parQ = (2.d0*Eti*(Eti*Etk - ptimod*ptkz))/(Pigreco*(
     -              Eti - ctj*ptimod))
               parR = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
               parT = Log(4.d0) + Log(1.d0 - ctj**2) - 
     -              Log(muscale**2/Qsq) + 2.d0*Log((Eti - Sqrt(Qsq) + 
     -              rMKt )/(ctj*ptimod - 
     -              Sqrt(Qsq) + rMKt ))
               res(0) = 
     -              (parQ*Pigreco*(parT + Log(parR**2 + (-1.d0 + ctj**2)
     -              *Eti**2*ptky**2) - 2.d0*Log(-parR + Sqrt(parR**2 + (
     -              -1.d0 + ctj**2)*Eti**2*ptky**2))))/(2.d0*Sqrt(
     -              parR**2 + (-1.d0 + ctj**2)*Eti**2*ptky**2))
     -              +  (parN*Pigreco)/(Sqrt(parO**2 - parP**2)*varej)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  if(nt.gt.4)then
c... The final state in the m-particle kinematics has more than two particles
                     parA  = Ej*(Etk - ctj*ptkz) + 
     -                    (2.d0*(pijdotpj - pjdotQ)*(ptidotptk - ptkdotQ
     -                    ))/(misq - 2.d0*ptidotQ + Qsq) - (2.d0*(
     -                    pijdotpj + pjdotpti - 2.d0*pjdotQ)*(pijdotptk
     -                    + ptidotptk - 2.d0*ptkdotQ))/(misq + 2.d0*
     -                    pijdotpti - 4.d0*pijdotQ + pijsq - 4.d0*
     -                    ptidotQ + 4.d0*Qsq)
                     parB  = -(Sqrt(1.d0 - ctj**2)*Ej*ptky)
                  elseif(nt.eq.4)then
c... The final state in the m-particle kinematics has only two particles. In this
c... case the standard formula for the Lorentz boost is singular for Kt**2 -> 0 and
c... one needs an alternative representation for the Lorentz transformation which 
c... reflects in a different definition of the parameter parA.
                     parA = -pijdotpj + pjdotpti + Ej*(Etk - ctj*ptkz)
                     parB = -(Sqrt(1.d0 - ctj**2)*Ej*ptky)
                  else
                     print*,'NSintdipole.f: check nt value'
                     STOP
                  endif
                  parF = (2.d0*Ej**2*lambda*(pijdotQ - pjdotQ)*(Eti - 
     -                 ctj*ptimod)*(misq - pijsq + 2.d0*(Ej - Eti)*
     -                 Sqrt(Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*Pigreco*
     -                 pijdotpj*(misq - pijsq)*(Sqrt(Qsq) - varej*
     -                 varejmax)**2)
                  parL = parA*pijdotQ - parA*pjdotQ + pijdotpj*pkdotQ
                  res(0) = res(0) + 
     -                 (-((parF*Pigreco*(-parL + parA*pidotQ - pidotQ*
     -                 pijdotpk + Sqrt(parL**2 + (-1.d0 + ctj**2)*
     -                 Ej**2*pidotQ**2*ptky**2)))/(pidotQ*Sqrt(parL**2 + 
     -                 (-1.d0 + ctj**2)*Ej**2*pidotQ**2*ptky**2)*varej)
     -                 ))
               endif

c... Mass term contribution (misq) to the finite term
               parLM = (Eti*misq)/(Pigreco*(Eti - ctj*ptimod)**2)
               parMM = Etk - ctj*ptkz
               parOM = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
               parQM = (Eti*misq)/(Pigreco*(Eti - ctj*ptimod)**2)
               parRM = -Etk + ctj*ptkz
               parTM = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
               parVM = Log(4.d0 - 4.d0*ctj**2) - Log(muscale**2/Qsq) + 
     -              2.d0*Log((Eti - Sqrt(Qsq) + rMKt
     -              )/(ctj*ptimod - Sqrt(Qsq) + 
     -              rMKt ))
               res(0) = res(0) + 
     -              (-(parQM*Pigreco*(parTM*parVM*Sqrt(parTM**2 + (
     -              -1.d0 + ctj**2)*Eti**2*ptky**2) + Eti*(-(parRM*
     -              parVM*Sqrt(parTM**2 + (-1.d0 + ctj**2)*Eti**2*
     -              ptky**2)) + (-1.d0 + ctj**2)*Eti*ptky**2*(parVM -
     -              Log(4.d0))) + parTM**2*(parVM - Log(4.d0)) - 2.d0*(
     -              -(Eti*parRM) + parTM)*Sqrt(parTM**2 + (-1.d0 + 
     -              ctj**2)*Eti**2*ptky**2)*Log(1.d0 - parTM/Sqrt(
     -              parTM**2 + (-1.d0 + ctj**2)*Eti**2*ptky**2))))/
     -              (2.d0*(Eti*parTM**2 + (-1.d0 + ctj**2)*Eti**3*
     -              ptky**2)))
     -              +  (parLM*Pigreco*(Eti*parMM + parOM + Sqrt(parOM**2
     -              + (-1.d0 + ctj**2)*Eti**2*ptky**2)))/(Eti*Sqrt(
     -              parOM**2 + (-1.d0 + ctj**2)*Eti**2*ptky**2)*varej)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parAM = parA
                  parBM = parB
                  parFM = -((Ej**2*lambda*misq*(pijdotQ - pjdotQ)*(Eti - 
     -                 ctj*ptimod)*(misq - pijsq + 2.d0*(Ej - Eti)*
     -                 Sqrt(Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*Pigreco*
     -                 pijdotpj**2*(misq - pijsq)*(Sqrt(Qsq) - varej*
     -                 varejmax)**2))
                  parGM = parAM*pijdotQ - parAM*pjdotQ + pijdotpj*pkdotQ
                  res(0) = res(0) + 
     -                 (parFM*Pigreco*(-parGM + parAM*pidotQ + Sqrt(
     -                 parGM**2 + (-1.d0 + ctj**2)*Ej**2*pidotQ**2*
     -                 ptky**2)))/(pidotQ*Sqrt(parGM**2 + (-1.d0 + 
     -                 ctj**2)*Ej**2*pidotQ**2*ptky**2)*varej)
               endif

c... Mass term contribution (mksq) to the finite term
               if(mksq.ne.0.d0)then
                  parMM = -Etk + ctj*ptkz
                  parOM = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
                  parRM = -Etk + ctj*ptkz
                  parTM = -2.d0*Eti*Etk + ctj*Etk*ptimod + ctj*Eti*ptkz
                  parVM = Log(4.d0 - 4.d0*ctj**2) - Log(muscale**2/Qsq)
     -                 + 2.d0*Log((Eti - Sqrt(Qsq) + Sqrt(misq - 2.d0*
     -                 Eti*Sqrt(Qsq) + Qsq))/(ctj*ptimod - Sqrt(Qsq) + 
     -                 Sqrt(misq - 2.d0*Eti*Sqrt(Qsq) + Qsq)))
                  res(0) = res(0) + 
     -                 (-(Eti*mksq*parVM)/(2.d0*(Eti*parRM - parTM)*
     -                 Sqrt(parRM**2 + (-1.d0 + ctj**2)*ptky**2)) - 
     -                 (Eti**2*mksq*parVM)/(2.d0*(-(Eti*parRM) + parTM)*
     -                 Sqrt(parTM**2 + (-1.d0 + ctj**2)*Eti**2*ptky**2))
     -                 - (Eti*mksq*Log((parRM**2 + (-1.d0 + ctj**2)*
     -                 ptky**2)/(parRM - Sqrt(parRM**2 + (-1.d0 + ctj**2
     -                 )*ptky**2))**2))/(2.d0*(Eti*parRM - parTM)*Sqrt(
     -                 parRM**2 + (-1.d0 + ctj**2)*ptky**2)) + (Eti**2*
     -                 mksq*Log((parTM - Sqrt(parTM**2 + (-1.d0 + ctj**2
     -                 )*Eti**2*ptky**2))**2/(parTM**2 + (-1.d0 + ctj**2
     -                 )*Eti**2*ptky**2)))/(2.d0*(-(Eti*parRM) + parTM)*
     -                 Sqrt(parTM**2 + (-1.d0 + ctj**2)*Eti**2*ptky**2)
     -                 ))
     -                 +  (Eti*mksq*(1.d0/Sqrt(parMM**2 + (-1.d0 + 
     -                 ctj**2)*ptky**2) - Eti/Sqrt(parOM**2 + (-1.d0 + 
     -                 ctj**2)*Eti**2*ptky**2)))/((Eti*parMM - parOM)*
     -                 varej)
c...  Phase-space restriction (alphaMax)
                  if(.NOT.alphaMaxcutoff)then
                     parAM = parA
                     parBM = parB
                     parFM = (Ej**2*lambda*mksq*(pijdotQ - pjdotQ)*(Eti
     -                    - ctj*ptimod)*(-misq + pijsq + 2.d0*(-Ej + 
     -                    Eti)*Sqrt(Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*
     -                    Pigreco*(misq - pijsq)*(Sqrt(Qsq) - varej*
     -                    varejmax)**2)
                     parGM = parAM*pijdotQ - parAM*pjdotQ + pijdotpj*
     -                    pkdotQ
                     res(0) = res(0) + 
     -                    (parFM*(1.d0/Sqrt(parAM**2 - parBM**2) - 
     -                    pidotQ/Sqrt(parGM**2 - parBM**2*pidotQ**2))*
     -                    Pigreco)/((parGM - parAM*pidotQ)*varej)
                  endif

               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR.
     -              res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in FIN-FIN int.massive'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'Etk',Etk
                  print*,'ptimod',ptimod
                  print*,'ptkmod',ptkmod
                  print*,'ptkz',ptkz
                  print*,'ptky',ptky
                  print*,'misq',misq
                  print*,'mksq',mksq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'pidotQ',pidotQ
                  print*,'pjdotQ',pjdotQ
                  print*,'pijdotQ',pijdotQ
                  print*,'ptidotQ',ptidotQ
                  print*,'pidotpj',pidotpj
                  print*,'pijdotpti',pijdotpti
                  print*,'pjdotpti',pjdotpti
                  print*,'pidotpti',pidotpti
                  print*,'pijdotpj',pijdotpj
                  print*,'ptkdotQ',ptkdotQ
                  print*,'pkdotQ',pkdotQ
                  print*,'ptidotptk',ptidotptk
                  print*,'pijdotptk',pijdotptk
                  print*,'pijdotpk',pijdotpk
                  print*,'parA',parA
                  print*,'parB',parB
                  print*,'parF',parF
                  print*,'parL',parL
                  print*,'parM',parM
                  print*,'parN',parN
                  print*,'parO',parO
                  print*,'parP',parP
                  print*,'parQ',parQ
                  print*,'parR',parR
                  print*,'parT',parT
                  print*,'parFM',parFM
                  print*,'parGM',parGM
                  print*,'parHM',parHM
                  print*,'parLM',parLM
                  print*,'parMM',parMM
                  print*,'parNM',parNM
                  print*,'parOM',parOM
                  print*,'parPM',parPM
                  print*,'parQM',parQM
                  print*,'parRM',parRM
                  print*,'parTM',parTM
                  print*,'parVM',parVM
                  STOP
#endif
               endif
            endif


         elseif(ipos_i.ge.3 .AND. (ipos_k.eq.1 .OR. ipos_k.eq.2))then
c---------------------------------------------------------------------------------
c...  FINAL-INITIAL state interference
c---------------------------------------------------------------------------------

c... Basic kinematics
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            varej = xrandom2
            misq = rmit**2
            masq = rmkt**2
            if(masq.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSintdipole: nonzero initial-state parton mass'//
     -              ' detected',sqrt(masq)
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif
            do mu=1,4
               pa(mu)=ptk(mu)
            enddo
            Eti = pti(4)
            Ea = pa(4)
            ptimod = sqrt(Eti**2-misq)
            paz = dot3(pa,pti)/ptimod
            pay = sqrt(Ea**2 -masq - paz**2)
            ptiQ = Eti*sqrt(Qsq)
            rMK2 = misq - 2.d0*ptiQ + Qsq
            if(rMK2.lt.0.d0) rMK2 = 0.d0
            varejmax = 
     -           (Sqrt(Qsq)*(-misq + (Sqrt(Qsq) - 
     -           Sqrt(rMK2))**2))/
     -           (-misq + 2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq) + 
     -           (Sqrt(Qsq) - Sqrt(rMK2))**2)
            pijsq = 
     -           (2.d0*(Eti - ctj*ptimod)*Sqrt(Qsq)*varej*varejmax + 
     -           misq*(Sqrt(Qsq) - varej*varejmax))/(Sqrt(Qsq) - 
     -           varej*varejmax)
            lambda = 
     -           Sqrt(((-misq + pijsq + 2.d0*ptiQ)**2 - 4.d0*pijsq*
     -           Qsq)/(ptiQ**2 - misq*Qsq))/2.d0
            gamma = 
     -           (-misq + pijsq - 2.d0*(-1.d0 + lambda)*ptiQ)/
     -           (2.d0*Qsq)
            Ej = 
     -           ((-misq + pijsq)*Sqrt(Qsq))/(-misq + pijsq + 2.d0*
     -           (Eti - ctj*lambda*ptimod)*Sqrt(Qsq))
            Ei =
     -           Sqrt(Ej**2 + misq - 2.d0*ctj*Ej*lambda*ptimod + 
     -           lambda**2*ptimod**2)
            pimod = sqrt(Ei**2 - misq)
            ctij = 
     -           (-Ej + ctj*lambda*ptimod)/Sqrt(Ej**2 - 2.d0*ctj*Ej*
     -           lambda*ptimod + lambda**2*ptimod**2)
            stj = sqrt(1.d0 - ctj**2)

c... Dot products
            pidotQ = Ei*Sqrt(Qsq)
            pjdotQ = Ej*Sqrt(Qsq)
            padotQ = Ea*Sqrt(Qsq)
            pijdotQ = pidotQ+pjdotQ
            pidotpj = Ej*(Ei-pimod*ctij)
            pijdotpj = pidotpj
            pijdotpa = lambda*dot4(pti,pa)+gamma*padotQ
            ptidotpa = dot4(pti,pa)

c...  Phase-space restriction (to be implemented)
c...  
c...  if(.CUTOFF.)then
c...      alphaMaxcutoff=.true.
c...  endif


            if(misq.eq.0.d0)then
c---------------------------------------------------------------------------------
c... FINAL-INITIAL state interference -- MASSLESS CASE
c---------------------------------------------------------------------------------
c               print*,'CALLING FINAL-INITIAL INTERFERENCE MASSLESS'

c... The (1/epsilon**2) term
               res(2) = 
     -              1.d0

c... The (1/epsilon) term
               res(1) = 
     -              2.d0 + Log((Ea*Eti)/(2.d0*ptidotpa)) + 
     -              Log(muscale**2/Qsq) + Log(Qsq/Eti**2)

c... The finite term
               rMK2 = -2.d0*Eti*Sqrt(Qsq) + Qsq
               if(rMK2.lt.0.d0) rMK2=0.d0
               parN  =  2.d0*(Ea - paz)*(Log(4.d0) + Log(1.d0 + ctj) - 
     -              Log(muscale**2/Qsq) + 2.d0*Log((Eti - Sqrt(Qsq) + 
     -              Sqrt(rMK2))/(ctj*Eti - Sqrt(Qsq) + Sqrt(rMK2))))
               parO  =  2.d0*(Ea - paz)
               parP  =  (-2.d0 + ctj)*Ea*Pigreco + ctj*paz*Pigreco
               parQ  =  Sqrt(1.d0 - ctj**2)*pay*Pigreco
               res(0) = 
     -              (-Log(4.d0) + Log(muscale**2/Qsq))**2/2.d0/xjac
     -              +  (-(((-2.d0 + (2.d0*(Ea - paz))/Sqrt((
     -              -1.d0 + ctj**2)*pay**2 + ((-2.d0 + ctj)*Ea + 
     -              ctj*paz)**2))*Log(1.d0 - ctj))/(-1.d0 + ctj)) - 
     -              (-((Pigreco*(parN + parO*Log(parP**2 - parQ**2) - 
     -              2.d0*parO*Log(-parP + Sqrt(parP**2 - parQ**2))))/
     -              Sqrt(parP**2 - parQ**2)) - 2.d0*Log(muscale**2/(
     -              2.d0*Qsq)))/(1.d0 - ctj))/2.d0
     -              +  (8.d0 - 4.d0*Log(4.d0) + 4.d0*Log(muscale**2/Qsq)
     -              + (4.d0 - 2.d0*Log(4.d0) + 2.d0*Log(muscale**2/Qsq))
     -              *Log(Qsq/Eti**2) + Log(Qsq/Eti**2)**2)/2.d0/xjac
     -              +  (2.d0 - (2.d0*(Ea - paz))/Sqrt((-1.d0 + ctj**2)*
     -              pay**2 + ((-2.d0 + ctj)*Ea + ctj*paz)**2) + (2.d0*
     -              Eti**2*(-1.d0 + varej))/(Eti - Eti*varej + Sqrt(Qsq)
     -              *varej)**2)/(varej - ctj*varej)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parA  =  Ej*(Ea - ctj*paz)
                  parB  =  -(Sqrt(1.d0 - ctj**2)*Ej*pay)
                  parF  =  (2.d0*(-1.d0 + ctj)**2*Ej**2*Eti*lambda*(
     -                 pijdotQ - pjdotQ)*(-pijsq + 2.d0*(Ej - Eti)*Sqrt(
     -                 Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*Pigreco*
     -                 pijdotpj*pijsq*(Sqrt(Qsq) - varej*varejmax)**2)
                  parL  =  padotQ*pijdotpj + parA*(pijdotQ - pjdotQ)
                  parM  =  parB*(pijdotQ - pjdotQ)
                  res(0) = res(0) + 
     -                 (-((parF*Pigreco*(parA*parM + parB*(-parL + 
     -                 Sqrt(parL**2 - parM**2)) - parM*pijdotpa))/
     -                 ((-1.d0 + ctj)*parM*Sqrt(parL**2 - parM**2)*
     -                 varej)))
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR.
     -              res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in FINAL-INITIAL'//
     -                 ' interference MASSLESS'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'Ea',Ea
                  print*,'ptimod',ptimod
                  print*,'paz',paz
                  print*,'pay',pay
                  print*,'misq',misq
                  print*,'masq',masq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parB',parB
                  print*,'parF',parF
                  print*,'parL',parL
                  print*,'parM',parM
                  print*,'parN',parN
                  print*,'parO',parO
                  print*,'parP',parP
                  print*,'parQ',parQ
                  STOP
#endif
               endif
            endif


            if(misq.ne.0.d0)then
c---------------------------------------------------------------------------------
c... FINAL-INITIAL state interference -- MASSIVE CASE
c---------------------------------------------------------------------------------
c               print*,'CALLING FINAL-INITIAL INTERFERENCE MASSIVE'

c... The (1/epsilon**2) term
               res(2) = 
     -              0.d0

c... The (1/epsilon) term
c... DeltaW contribution to the 1/epsilon pole
               res(1) = 
     -              Log((Ea*misq)/(Ea*misq + 2.d0*Eti*ptidotpa))
c... Mass term contribution (proportional to misq) to the 1/epsilon pole
               res(1) = res(1) +
     -              (2.d0 + (Ea*misq*Log((Ea*misq)/
     -              (Ea*misq + 2.d0*Eti*ptidotpa)))/
     -              (Eti*ptidotpa))/2.d0

c... The finite term
c... DeltaW contribution to the finite term
               parN  =  (2.d0*Eti*(-(Ea*Eti) + paz*ptimod))/(
     -              Pigreco*(Eti - ctj*ptimod))
               parO  =  -2.d0*Ea*Eti + ctj*Eti*paz + ctj*Ea*ptimod
               parP  =  Sqrt(1.d0 - ctj**2)*Eti*pay
               parQ  =  (2.d0*Eti*(Ea*Eti - paz*ptimod))/(Pigreco*(Eti
     -              - ctj*ptimod))
               parR  =  -2.d0*Ea*Eti + ctj*Eti*paz + ctj*Ea*ptimod
               parS  =  Sqrt(1.d0 - ctj**2)*Eti*pay
               rMK2 = misq - 2.d0*Eti*Sqrt(Qsq) + Qsq
               if(rMK2.lt.0.d0) rMK2=0.d0
               parT  =  Log(4.d0) + Log(1.d0 - ctj**2) - 
     -              Log(muscale**2/Qsq) + 2.d0*Log((Eti - Sqrt(Qsq) + 
     -              Sqrt(rMK2))/(ctj*ptimod - Sqrt(Qsq) + Sqrt(rMK2)))
               res(0) = 
     -              (Pigreco*(parQ*parT + parQ*Log(parR**2 - parS**2) - 
     -              2.d0*parQ*Log(-parR + Sqrt(parR**2 - parS**2))))/
     -              (2.d0*Sqrt(parR**2 - parS**2))
     -              +  (parN*Pigreco)/(Sqrt(parO**2 - parP**2)*varej)
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parA  =  Ej*(Ea - ctj*paz)
                  parB  =  -(Sqrt(1.d0 - ctj**2)*Ej*pay)
                  parF  =  (2.d0*Ej**2*lambda*(pijdotQ - pjdotQ)*(Eti - 
     -                 ctj*ptimod)*(misq - pijsq + 2.d0*(Ej -Eti)*
     -                 Sqrt(Qsq))*Sqrt(Qsq)*varej*varejmax)/(Ei*Pigreco*
     -                 pijdotpj*(misq - pijsq)*(Sqrt(Qsq) - 
     -                 varej*varejmax)**2)
                  parL  =  padotQ*pijdotpj + parA*(pijdotQ - pjdotQ)
                  parM  =  parB*(pijdotQ - pjdotQ)
                  res(0) = res(0) + 
     -                 (-((parF*Pigreco*(parA*parM + parB*(-parL + 
     -                 Sqrt(parL**2 - parM**2)) - parM*pijdotpa))/
     -                 (parM*Sqrt(parL**2 - parM**2)*varej)))
               endif

c... Mass term contribution (proportional to misq) to the finite term
               parNM  =  Eti*misq*(Ea - ctj*paz)
               parOM  =  -(Sqrt(1.d0 - ctj**2)*Eti*misq*pay)
               parPM  =  Pigreco*(Eti - ctj*ptimod)**2*(-2.d0*Ea*Eti + 
     -              ctj*Eti*paz + ctj*Ea*ptimod)
               parQM  =  Sqrt(1.d0 - ctj**2)*Eti*pay*Pigreco*(Eti - 
     -              ctj*ptimod)**2
               parRM = (Eti*misq)/(Pigreco*(Eti - ctj*ptimod)**2)
               parSM  =  -Ea + ctj*paz
               rMK2 = misq - 2.d0*Eti*Sqrt(Qsq) + Qsq
               if(rMK2.lt.0.d0) rMK2=0.d0
               parUM  =  Log(4.d0 - 4.d0*ctj**2) - Log(muscale**2/Qsq) + 
     -              2.d0*Log((Eti - Sqrt(Qsq) + Sqrt(rMK2))/(ctj*ptimod
     -              - Sqrt(Qsq) + Sqrt(rMK2)))
               parXM  =  -2.d0*Ea*Eti + ctj*Eti*paz + ctj*Ea*ptimod
               res(0) = res(0) + 
     -              (-(parRM*Pigreco*(-(Eti*parSM*parUM*Sqrt(parXM**2 +
     -              (-1.d0 + ctj**2)*Eti**2*pay**2)) + parUM*parXM*
     -              Sqrt(parXM**2 + (-1.d0 + ctj**2)*Eti**2*pay**2)
     -              + parXM**2*(parUM - Log(4.d0)) + (-1.d0 + ctj**2)*
     -              Eti**2*pay**2*(parUM - Log(4.d0)) - (Eti*parSM - 
     -              parXM)*Sqrt(parXM**2 + (-1.d0 + ctj**2)*Eti**2*
     -              pay**2)*(Log(parXM**2 + (-1.d0 + ctj**2)*Eti**2*
     -              pay**2) - 2.d0*Log(-parXM + Sqrt(parXM**2 + 
     -              (-1.d0 + ctj**2)*Eti**2*pay**2)))))/(2.d0*(Eti*
     -              parXM**2 + (-1.d0 + ctj**2)*Eti**3*pay**2)))
     -              +  (-(((parOM + (parOM*parPM - parNM*parQM)/
     -              Sqrt((parPM - parQM)*(parPM + parQM)))*Pigreco)/
     -              (parQM*varej)))
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parAM  =  Ej*(Ea - ctj*paz)
                  parBM  =  -(Sqrt(1.d0 - ctj**2)*Ej*pay)
                  parFM  =  -((Ej**2*lambda*misq*pidotQ*(Eti - ctj*
     -                 ptimod)*(misq - pijsq + 2.d0*(Ej -Eti)*Sqrt(Qsq)
     -                 )*Sqrt(Qsq)*varej*varejmax)/(Ei*pidotpj**2*
     -                 Pigreco*(misq - pijsq)*(Sqrt(Qsq) - varej*
     -                 varejmax)**2))
                  parLM  =  padotQ*pidotpj + parAM*pidotQ
                  parMM  =  parBM*pidotQ
                  res(0) = res(0) + 
     -                 (parFM*(parAM*parMM + parBM*(-parLM + Sqrt(
     -                 parLM**2 - parMM**2)))*Pigreco)/(parMM*Sqrt(
     -                 parLM**2 - parMM**2)*varej)
               endif

c... Numerical exception (NaN) trap
               if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR.
     -              res(2).ne.res(2))then
                  stable=.false.
#if DEBUG1
                  print*,'NSintdipole: NaN found in FINAL-INITIAL'//
     -                 ' interference MASSIVE'
                  print*,'x(1)',xrandom1
                  print*,'x(2)',xrandom2
                  print*,'res(2)',res(2)
                  print*,'res(1)',res(1)
                  print*,'res(0)',res(0)
                  print*,'ctj',ctj
                  print*,'varej',varej
                  print*,'Eti',Eti
                  print*,'Ea',Ea
                  print*,'ptimod',ptimod
                  print*,'paz',paz
                  print*,'pay',pay
                  print*,'misq',misq
                  print*,'masq',masq
                  print*,'ptiQ',ptiQ
                  print*,'varejmax',varejmax
                  print*,'pijsq',pijsq
                  print*,'lambda',lambda
                  print*,'gamma',gamma
                  print*,'Ej',Ej
                  print*,'Ei',Ei
                  print*,'pimod',pimod
                  print*,'ctij',ctij
                  print*,'stj',stj
                  print*,'parA',parA
                  print*,'parB',parB
                  print*,'parF',parF
                  print*,'parL',parL
                  print*,'parM',parM
                  print*,'parN',parN
                  print*,'parO',parO
                  print*,'parP',parP
                  print*,'parQ',parQ
                  print*,'parR',parR
                  print*,'parS',parS
                  print*,'parT',parT
                  print*,'parAM',parAM
                  print*,'parBM',parBM
                  print*,'parFM',parFM
                  print*,'parLM',parLM
                  print*,'parMM',parMM
                  print*,'parOM',parOM
                  print*,'parPM',parPM
                  print*,'parQM',parQM
                  print*,'parRM',parRM
                  print*,'parSM',parSM
                  print*,'parUM',parUM
                  print*,'parXM',parXM
                  STOP
#endif
               endif
            endif


         elseif((ipos_i.eq.1 .OR. ipos_i.eq.2) .AND. ipos_k.ge.3)then
c---------------------------------------------------------------------------------
c...  INITIAL-FINAL state interference
c---------------------------------------------------------------------------------

c... Basic kinematics
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            x = xrandom2
            masq = rmit**2
            mksq = rmkt**2
            if(masq.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSintdipole: nonzero initial-state parton mass'//
     -              ' detected',sqrt(masq)
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

c... Note: all formulae of the integrated dipoles are simpler if derived in
c... the frame where (pa + pb) = Q = (sqrt(Qsq),0,0,0) is at rest, where 
c... {pa,pb,Q} belong to the (m+1)-parton kinematics.
c... On the other hand, the momenta of the m-particle ('ptilde') kinematics 
c... provided by the Helac framework are expressed in the frame where 
c... (pta + ptb) = Kt = (sqrt(Ktsq),0,0,0) is at rest. Hence, in case of
c... non-zero initial state radiation (i.e. x<1) the two frames do not coincide.
c... In this case, in order to get the same frame of the formulae, the 'ptilde' 
c... momenta must be properly boosted. The invariance of the integrated
c... dipoles (Lorentz scalars) under boosts ensures that this operation leaves 
c... the final result unchanged.
c... The following rule has to be followed.
c... In the finite terms, one has to take either the original or the boosted
c... ptilde's on a case by case basis:
c...  - for all finite terms arising from soft-limit contributions 
c...    (i.e. x->1), one has to consider the original ptilde's;
c...  - for all finite terms arising from contributions away from the 
c...    soft limit (i.e. with x<1), one has to consider the boosted ptilde's.
c... NOTE: here, the dependence of the Dirac delta contribution upon x is justified
c... by the fact that we are adopting Monte Carlo integration. Dirac delta terms 
c... would obviously have no x dependence if analytically integrated, and the two
c... reference frames would coincide (since x=1).

            Ktsq = Qsq
            Qsq = Ktsq/x        !...CM energy squared of the (m+1)-parton kinematics
            z = 1.d0/x
            ys = -0.5d0*log(x)  !...rapidity
c...  If emitter is parton b (coming from right), rapidity has opposite sign
            if(ipos_i.eq.2) ys=-ys
            bcm = (exp(ys)-exp(-ys))/(exp(ys)+exp(-ys)) !...beta
            gcm = (exp(ys)+exp(-ys))/2.d0 !...gamma
c... Boost ptilde momenta
            do i=1,nt
               pmomt_boost(i,4) = gcm*(pmomt(i,4)-bcm*pmomt(i,3))
               pmomt_boost(i,1) = pmomt(i,1)
               pmomt_boost(i,2) = pmomt(i,2)
               pmomt_boost(i,3) = pmomt(i,3)/gcm-bcm*pmomt_boost(i,4)
            enddo

            do mu=1,4
               pti(mu)=pmomt_boost(ipos_i,mu)
               ptk(mu)=pmomt_boost(ipos_k,mu)
            enddo
            do mu=1,4
               pta(mu)=pti(mu)
               pa(mu)=pta(mu)/x
            enddo

c... Boosted quantities (x<1)
            Ea = pa(4)
            Eta = pta(4)
            Etk = ptk(4)
            ptkz = ptk(3)
            if(ipos_i.eq.2) ptkz=-ptkz
            ptky = sqrt(Etk**2 -mksq - ptkz**2)

c... Non-boosted quantities (x=1)
            Eta0 = pmomt(ipos_i,4)
            Etk0 = pmomt(ipos_k,4)
            ptkz0 = pmomt(ipos_k,3)
            if(ipos_i.eq.2) ptkz0=-ptkz0
            ptky0 = sqrt(Etk0**2 -mksq - ptkz0**2)

            Ej = -(Sqrt(Qsq)*(-1.d0 + x))/2.d0

c... Dot products
            padotQ = Ea*Sqrt(Qsq)
            ptkdotQ = Etk*Sqrt(Qsq)
            pjdotQ = Ej*Sqrt(Qsq)
            padotptk = dot4(pa,ptk)
            ptadotptk = dot4(pta,ptk)
            padotpj = Ej*Ea*(1.d0 - ctj)

c...  Phase-space restriction (to be implemented)
c...  
c...  if(.CUTOFF.)then
c...      alphaMaxcutoff=.true.
c...  endif


c... The (1/epsilon**2) term
            res(2) = 
     -           1.d0

c... The (1/epsilon) term
c... DeltaW contribution to the 1/epsilon pole
            res(1) = 
     -           Log(muscale**2/Ktsq) + 
     -           Log((Eta0*(Eta0*mksq + 2*Etk0*ptadotptk))/ptadotptk**2)

c... Mass term contribution (proportional to mksq) to the 1/epsilon pole
            if(mksq.ne.0.d0)then
               res(1) = res(1) +
     -              (Eta0*mksq*Log(1.d0 + (2.d0*Etk0*ptadotptk)/(
     -              Eta0*mksq)))/(2.d0*Etk0*ptadotptk)
            endif

c... The finite term
c... DeltaW contribution to the finite term
            parN  =  2.d0*(Etk0 - ptkz0)*(Log(1.d0 + ctj) - 
     -           Log(muscale**2/Ktsq))
            parO  =  2.d0*(Etk0 - ptkz0)
            parP  =  (-2.d0 + ctj)*Etk0*Pigreco + ctj*Pigreco*ptkz0
            parQ  =  Sqrt(1.d0 - ctj**2)*Pigreco*ptky0
            res(0) = 
     -           Log(muscale**2/Ktsq)**2/2.d0/xjac
     -           +  (-(((-2.d0 + (2.d0*(Etk0 - ptkz0))/Sqrt((
     -           -1.d0 + ctj**2)*ptky0**2 + ((-2.d0 + ctj)*Etk0 +
     -           ctj*ptkz0)**2))*Log(1.d0 - ctj))/(-1.d0 + ctj)) - 
     -           (-2.d0*Log((2.d0*muscale**2)/Ktsq) - (Pigreco*(
     -           parN + parO*Log(parP**2 - parQ**2) - 2.d0*parO*
     -           Log(-parP + Sqrt(parP**2 - parQ**2))))/
     -           Sqrt(parP**2 - parQ**2))/(1.d0 - ctj))/2.d0
     -           +  Pigreco**2/3.d0/xjac
     -           +  (-2.d0*(Etk0 - ptkz0))/((-1.d0 + ctj)*Sqrt((
     -           -1.d0 + ctj**2)*ptky0**2 + ((-2.d0 + ctj)*Etk0 + 
     -           ctj*ptkz0)**2)*(-1.d0 + x))
c...  Phase-space restriction (alphaMax)
            if(.NOT.alphaMaxcutoff)then
               parAa  =  (2.d0*(padotQ*(-2.d0*ptkdotQ*(pjdotQ + padotQ*(
     -              -1.d0 + x)) + Ej*(Etk - ctj*ptkz)*(Qsq + 2.d0*padotQ
     -              *(-1.d0 + x))) + padotptk*(Qsq*(-pjdotQ + Qsq) + 
     -              4.d0*padotQ*(-pjdotQ + Qsq)*(-1.d0 + x) + 2.d0*
     -              padotQ**2*(-1.d0 + x)**2)) + padotpj*(-2.d0*ptkdotQ*
     -              (-2.d0*pjdotQ + Qsq + padotQ*(-1.d0 + x)) - Ej*(
     -              Etk - ctj*ptkz)*(Qsq + 2.d0*padotQ*(-1.d0 + x)) - 
     -              2.d0*padotptk*(2.d0*(-pjdotQ + Qsq) + 3.d0*padotQ*(
     -              -1.d0 + x))*(-1.d0 + x)) + 2.d0*padotpj**2*(ptkdotQ
     -              + padotptk*(-1.d0 + x))*(-1.d0 + x))/((2.d0*(-pjdotQ
     -              + Qsq + padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))*(
     -              Qsq + 2.d0*padotQ*(-1.d0 + x)))
               parBa  =  -((Sqrt(1.d0 - ctj**2)*Ej*(padotpj - 2.d0*
     -              padotQ)*ptky)/(2.d0*(pjdotQ - Qsq - padotQ*(
     -              -1.d0 + x)) + padotpj*(-1.d0 + x)))
               parAj  =  (2.d0*(Ej*(Etk - ctj*ptkz)*Qsq**2 - 2.d0*pjdotQ
     -              *ptkdotQ*(pjdotQ + padotQ*(-1.d0 + x)) + (-2.d0*
     -              padotptk*pjdotQ**2 + padotptk*pjdotQ*Qsq + Ej*padotQ
     -              *(Etk - ctj*ptkz)*(3.d0*Qsq + 2.d0*padotQ*(-1.d0 + x
     -              )))*(-1.d0 + x)) - padotpj*(2.d0*ptkdotQ*(pjdotQ + 
     -              Qsq + 2.d0*padotQ*(-1.d0 + x)) + padotptk*(2.d0*
     -              pjdotQ + Qsq + 2.d0*padotQ*(-1.d0 + x))*(-1.d0 + x))
     -              *(-1.d0 + x))/((2.d0*(-pjdotQ + Qsq + padotQ*(
     -              -1.d0 + x)) - padotpj*(-1.d0 + x))*(Qsq + 2.d0*
     -              padotQ*(-1.d0 + x)))
               parBj  =  (-2.d0*Sqrt(1.d0 - ctj**2)*Ej*ptky*(Qsq + 
     -              padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + padotQ*(
     -              -1.d0 + x)) - padotpj*(-1.d0 + x))
               parAq  =  (Ej*(Etk - ctj*ptkz)*(-pjdotQ + 2.d0*Qsq + 
     -              padotQ*(-1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x))
     -              + ptkdotQ*(2.d0*Qsq**2 + 4.d0*pjdotQ*(pjdotQ - 2.d0*
     -              Qsq - padotQ*(-1.d0 + x)) - padotpj*(-2.d0*pjdotQ + 
     -              3.d0*Qsq + 2.d0*padotQ*(-1.d0 + x))*(-1.d0 + x)) + 
     -              padotptk*(4.d0*pjdotQ**2 - 7.d0*pjdotQ*Qsq + 2.d0*
     -              Qsq**2 - 2.d0*padotpj*(-pjdotQ + Qsq)*(-1.d0 + x) -
     -              padotQ*(2.d0*pjdotQ + Qsq)*(-1.d0 + x) - 2.d0*
     -              padotQ**2*(-1.d0 + x)**2)*(-1.d0 + x))/((2.d0*(-
     -              pjdotQ + Qsq + padotQ*(-1.d0 + x)) - padotpj*(
     -              -1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x)))
               parBq  =  (Sqrt(1.d0 - ctj**2)*Ej*ptky*(pjdotQ - 2.d0*Qsq
     -              - padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + padotQ
     -              *(-1.d0 + x)) - padotpj*(-1.d0 + x))
               parF  =  -((-1.d0 + ctj)*padotQ*Qsq*(-1.d0 + x)**2)/(
     -              2.d0*padotpj*Pigreco)
               parL  =  padotQ*parAj + padotpj*parAq
               parM  =  padotQ*parBj + padotpj*parBq
               res(0) = res(0) + 
     -              (parF*(parAa*parM + parBa*(-parL + Sqrt(parL**2 -
     -              parM**2)))*Pigreco)/((-1.d0 + ctj)*parM*Sqrt(parL**2
     -              - parM**2)*(-1.d0 + x))
            endif


c... Mass term contribution (proportional to mksq) to the finite term
            if(mksq.ne.0.d0)then
               parQM  =  (-2.d0 + ctj)*Etk0 + ctj*ptkz0
               parTM  =  Log(1.d0 - ctj**2) - Log(muscale**2/Ktsq)
               parXM  =  (-2.d0 + ctj)*Etk0 + ctj*ptkz0
               res(0) = res(0) + 
     -              (mksq*(parTM*(Sqrt(parXM**2 + (-1.d0 + ctj**2)*
     -              ptky0**2) - Sqrt((-1.d0 + ctj**2)*ptky0**2 + (Etk0 -
     -              ctj*ptkz0)**2)) - Sqrt((-1.d0 + ctj**2)*ptky0**2 + 
     -              (Etk0 - ctj*ptkz0)**2)*Log((parXM**2 + (-1.d0 +
     -              ctj**2)*ptky0**2)/(parXM - Sqrt(parXM**2 + (-1.d0 +
     -              ctj**2)*ptky0**2))**2) + Sqrt(parXM**2 + (-1.d0 +
     -              ctj**2)*ptky0**2)*Log(((-1.d0 + ctj**2)*ptky0**2 + 
     -              (Etk0 - ctj*ptkz0)**2)/(Etk0 - ctj*ptkz0 + Sqrt((
     -              -1.d0 + ctj**2)*ptky0**2 + (Etk0 - ctj*ptkz0)**2)
     -              )**2)))/(2.d0*(Etk0 + parXM - ctj*ptkz0)*Sqrt(
     -              (parXM**2 + (-1.d0 + ctj**2)*ptky0**2)*((-1.d0 + 
     -              ctj**2)*ptky0**2 + (Etk0 - ctj*ptkz0)**2)))
     -              +  (mksq*(-(1.d0/Sqrt(parQM**2 + (-1.d0 + ctj**2)*
     -              ptky0**2)) + 1.d0/Sqrt((-1.d0 + ctj**2)*ptky0**2 + 
     -              (Etk0 - ctj*ptkz0)**2)))/((Etk0 + parQM - ctj*ptkz0
     -              )*(-1.d0 + x))
c...  Phase-space restriction (alphaMax)
               if(.NOT.alphaMaxcutoff)then
                  parAMj  =  (2.d0*(Ej*(Etk - ctj*ptkz)*Qsq**2 - 2.d0*
     -                 pjdotQ*ptkdotQ*(pjdotQ + padotQ*(-1.d0 + x)) + (
     -                 -2.d0*padotptk*pjdotQ**2 + padotptk*pjdotQ*Qsq + 
     -                 Ej*padotQ*(Etk - ctj*ptkz)*(3.d0*Qsq + 2.d0*
     -                 padotQ*(-1.d0 + x)))*(-1.d0 + x)) - padotpj*(2.d0
     -                 *ptkdotQ*(pjdotQ + Qsq + 2.d0*padotQ*(-1.d0 + x))
     -                 + padotptk*(2.d0*pjdotQ + Qsq + 2.d0*padotQ*(
     -                 -1.d0 + x))*(-1.d0 + x))*(-1.d0 + x))/((2.d0*(
     -                 -pjdotQ + Qsq + padotQ*(-1.d0 + x)) - padotpj*(
     -                 -1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x)))
                  parBMj  =  (-2.d0*Sqrt(1.d0 - ctj**2)*Ej*ptky*(Qsq + 
     -                 padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + 
     -                 padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))
                  parAMq  =  (Ej*(Etk - ctj*ptkz)*(-pjdotQ + 2.d0*Qsq + 
     -                 padotQ*(-1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x
     -                 )) + ptkdotQ*(2.d0*Qsq**2 + 4.d0*pjdotQ*(pjdotQ -
     -                 2.d0*Qsq - padotQ*(-1.d0 + x)) - padotpj*(-2.d0*
     -                 pjdotQ + 3.d0*Qsq + 2.d0*padotQ*(-1.d0 + x))*(
     -                 -1.d0 + x)) + padotptk*(4.d0*pjdotQ**2 - 7.d0*
     -                 pjdotQ*Qsq + 2.d0*Qsq**2 - 2.d0*padotpj*(-pjdotQ
     -                 + Qsq)*(-1.d0 + x) - padotQ*(2.d0*pjdotQ + Qsq)*(
     -                 -1.d0 + x) - 2.d0*padotQ**2*(-1.d0 + x)**2)*(
     -                 -1.d0 + x))/((2.d0*(-pjdotQ + Qsq + padotQ*(
     -                 -1.d0 + x)) - padotpj*(-1.d0 + x))*(Qsq + 2.d0*
     -                 padotQ*(-1.d0 + x)))
                  parBMq  =  (Sqrt(1.d0 - ctj**2)*Ej*ptky*(pjdotQ - 2.d0
     -                 *Qsq - padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq
     -                 + padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))
                  parFM  =  -(mksq*padotQ*Qsq*(-1.d0 + x)**2)/(
     -                 4.d0*Pigreco)
                  parLM  =  padotQ*parAMj + padotpj*parAMq
                  parMM  =  padotQ*parBMj + padotpj*parBMq
                  res(0) = res(0) + 
     -                 (parFM*(parBMj/Sqrt(parAMj**2 - parBMj**2) - 
     -                 parMM/Sqrt(parLM**2 - parMM**2))*Pigreco)/
     -                 ((-(parBMj*parLM) + parAMj*parMM)*(-1.d0 + x))
               endif
            endif


c... Numerical exception (NaN) trap
            if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR. 
     -           res(2).ne.res(2))then
               stable=.false.
#if DEBUG1
               print*,'NSintdipole: NaN found in INITIAL-FINAL'//
     -              ' interference'
               print*,'x(1)',xrandom1
               print*,'x(2)',xrandom2
               print*,'res(2)',res(2)
               print*,'res(1)',res(1)
               print*,'Eta0',Eta0
               print*,'Etk0',Etk0
               print*,'masq',masq
               print*,'mksq',mksq
               print*,'padotQ',padotQ
               print*,'ptkdotQ',ptkdotQ
               print*,'pjdotQ',pjdotQ
               print*,'padotptk',padotptk
               print*,'ptadotptk',ptadotptk
               print*,'padotpj',padotpj
               STOP
#endif
            endif



         elseif( (ipos_i.eq.1 .AND. ipos_k.eq.2) .OR.
     -           (ipos_i.eq.2 .AND. ipos_k.eq.1) )then
c---------------------------------------------------------------------------------
c...  INITIAL-INITIAL state interference
c---------------------------------------------------------------------------------
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            x = xrandom2
            masq = rmit**2
            mksq = rmkt**2
            if(masq.ne.0.d0 .OR. mksq.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSintdipole: nonzero initial-state parton mass'//
     -              ' detected',sqrt(masq),sqrt(mksq)
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

c... Note: all formulae of the integrated dipoles are simpler if derived in
c... the frame where (pa + pb) = Q = (sqrt(Qsq),0,0,0) is at rest, where 
c... {pa,pb,Q} belong to the (m+1)-parton kinematics.
c... On the other hand, the momenta of the m-particle ('ptilde') kinematics 
c... provided by the Helac framework are expressed in the frame where 
c... (pta + ptb) = Kt = (sqrt(Ktsq),0,0,0) is at rest. Hence, in case of
c... non-zero initial state radiation (i.e. x<1) the two frames do not coincide.
c... In this case, in order to get the same frame of the formulae, the 'ptilde' 
c... momenta must be properly boosted. The invariance of the integrated
c... dipoles (Lorentz scalars) under boosts ensures that this operation leaves 
c... the final result unchanged.
c... The following rule has to be followed.
c... In the finite terms, one has to take either the original or the boosted
c... ptilde's on a case by case basis:
c...  - for all finite terms arising from soft-limit contributions 
c...    (i.e. x->1), one has to consider the original ptilde's;
c...  - for all finite terms arising from contributions away from the 
c...    soft limit (i.e. with x<1), one has to consider the boosted ptilde's.
c... NOTE: here, the dependence of the Dirac delta contribution upon x is justified
c... by the fact that we are adopting Monte Carlo integration. Dirac delta terms 
c... would obviously have no x dependence if analytically integrated, and the two
c... reference frames would coincide (since x=1).

            Ktsq = Qsq
            Qsq = Ktsq/x        !...CM energy squared of the (m+1)-parton kinematics
            z = 1.d0/x
            ys = -0.5d0*log(x)  !...rapidity
c...  If emitter is parton b (coming from right), rapidity has opposite sign
            if(ipos_i.eq.2) ys=-ys
            bcm = (exp(ys)-exp(-ys))/(exp(ys)+exp(-ys)) !...beta
            gcm = (exp(ys)+exp(-ys))/2.d0 !...gamma
c... Boost ptilde momenta
            do i=1,nt
               pmomt_boost(i,4) = gcm*(pmomt(i,4)-bcm*pmomt(i,3))
               pmomt_boost(i,1) = pmomt(i,1)
               pmomt_boost(i,2) = pmomt(i,2)
               pmomt_boost(i,3) = pmomt(i,3)/gcm-bcm*pmomt_boost(i,4)
            enddo
            do mu=1,4
               pti(mu)=pmomt_boost(ipos_i,mu)
               pta(mu)=pti(mu)
               pa(mu)=pta(mu)/x
            enddo

c... Boosted quantities (x<1)
            Ea = pa(4)
            Ej = -(Sqrt(Qsq)*(-1.d0 + x))/2.d0

c... Dot products
            padotQ = Ea*Sqrt(Qsq)
            padotpj = Ej*Ea*(1.d0 - ctj)

c...  Phase-space restriction (to be implemented)
c...  
c...  if(.CUTOFF.)then
c...      alphaMaxcutoff=.true.
c...  endif


c... The (1/epsilon**2) term
            res(2) = 
     -           1.d0
c... The (1/epsilon) term
            res(1) = 
     -           Log(muscale**2/Ktsq)
c... The finite term
            res(0) = 
     -           Log(muscale**2/Ktsq)**2/2.d0/xjac
     -           +  Pigreco**2/6.d0/xjac
     -           +  (-2.d0/((-1.d0 + ctj)*(-1.d0 + x)))
c...  Phase-space restriction (alphaMax)
            if(.NOT.alphaMaxcutoff)then
               res(0) = res(0) + 
     -              2.d0/((-1.d0 + ctj)*(-1.d0 + x))
            endif


c... Numerical exception (NaN) trap
            if(res(0).ne.res(0) .OR. res(1).ne.res(1) .OR.
     -           res(2).ne.res(2))then
               stable=.false.
#if DEBUG1
               print*,'NSintdipole: NaN found in INITIAL-INITIAL'//
     -              ' interference'
               print*,'res(2)',res(2)
               print*,'res(1)',res(1)
               print*,'res(0)',res(0)
               STOP
#endif
            endif


         else                   !...something went wrong with ipos_i or ipos_k


            print*,'NSintdipole ERROR: check ipos_i(k)',ipos_i,ipos_k
            STOP


         endif


      endif                     !if(ipos_i.eq.ipos_k)then


c... Include jacobian contribution for the transformation xrandom1 -> ctj
         res(0)=res(0)*xjac


      return
      end                       ! subroutine NSIntdipole


c-----------------------------------------------------------------------

      double precision function dot4(pi,pj)

      double precision pi,pj
      dimension pi(4),pj(4)

      dot4=pi(4)*pj(4)-pi(1)*pj(1)-pi(2)*pj(2)-pi(3)*pj(3)

      return
      end

c-----------------------------------------------------------------------

      double precision function dot3(pi,pj)

      double precision pi,pj
      dimension pi(4),pj(4)

      dot3=pi(1)*pj(1)+pi(2)*pj(2)+pi(3)*pj(3)

      return
      end

c-----------------------------------------------------------------------

      subroutine Lorentz(nt,K,Ktilde,ptk,pk)

      integer nt,mu
      double precision K,Ktilde,ptk,pk
      dimension K(4),Ktilde(4),ptk(4),pk(4)
      double precision dot4
      external dot4

c... In case of two-particle final states (in the m-particle kinematics), then
c... pk is simply K=Q-pti
      if(nt.eq.4)then
         do mu=1,4
            pk(mu) = K(mu)
         enddo
      elseif(nt.gt.4)then
         do mu=1,4
            pk(mu) = ptk(mu)
     &           -2.d0*(dot4(K,ptk)+dot4(Ktilde,ptk))*(K(mu)+Ktilde(mu))
     &           /(dot4(K,K)+dot4(Ktilde,Ktilde)+2.d0*dot4(K,Ktilde))
     &           + 2.d0*dot4(Ktilde,ptk)*K(mu)/dot4(Ktilde,Ktilde)
         enddo
      else
         print*,'NSintdipole.f: ERROR in LORENTZ subroutine'
         STOP
      endif

      return
      end

