c-----------------------------------------------------------------------
c...  Copyright (C) 2009 M. Czakon, C. Papadopoulos and M. Worek
c...  Copyright (C) 2013 G. Bevilacqua, M. Czakon, M. Kubocz, M. Worek
c...  For details see: e-Print: arXiv:0905.0883 [hep-ph] 
c...                   e-Print: arXiv:1308.5605 [hep-ph]
c-----------------------------------------------------------------------
#define SINGLEPOINT 0
#define DEBUG1 0


      subroutine dipoles_KP(key,weight,wmeasure)

c-----------------------------------------------------------------------
c...  
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles_KP.h'
      include 'dimensions.h'

      common/helac_dipoles_ext/dipoletype
c...  dipoletype: 0-Catani-Seymour, 1-Nagy-Soper
      integer dipoletype
c...  pi
      parameter(pi=3.141592653589793d0)
c...  8*pi^2
      double precision norm
      parameter(norm=78.9568352087148690d0)

c...  polarizations
      dimension ipol(mxpart),rpol(mxpart)
c...  color flows
      dimension jcoldip(mxncc,mxpart,0:2)
      dimension jcol(mxpart,2),kcol(mxpart,2)
c...  color correlator matrix (dimension for 10 partons !!!)
      dimension rmatrix(2,10,mxncc,mxncc)
c...  Dyson-Schwinger skeletons
      dimension list1(0:mxproc,mxpart,3)
      dimension list(0:mxproc,mxkmax,18,mxncc)
      dimension kcou(0:mxproc,mxncc)
      dimension isnew(mxfeyg,8,8) ! pure trash, but needed
c...  amplitudes
      dimension zy(mxkmax,8),zamp(mxncc)

c...  amplitude for different polarizations. the lower precision (real)
c...  is used to spare some space, and is, of course, sufficient
      real amppol(mxhel)
c...  perform a sum over polarizations
      logical sumpol
c...  evaluate for random polarization phases
      logical ranpol

c...  setup for self-initialization
      logical firstcall
      save firstcall
      data firstcall/.true./

c...  some local arrays and variables
      character*100 process
      double precision KP(2,10),KPoperator

      double precision NSKPoperator
      logical stable,stab

      integer app,jpp

c...  storage
      save jcoldip,rmatrix
      save kmax,ncc,imax
      save list1,list,kcou
      save avcol,avhel,symet
      save amppol,sumpol,ranpol

#if SINGLEPOINT
      open(99,file='momenta.input',status='old')
      do i=1,n
         read(99,*)(pmom(i,j),j=1,5)
      enddo
      close(99)
#endif

c-----------------------------------------------------------------------
c...  initialisation
c-----------------------------------------------------------------------

      if (firstcall) then
         
c-----------------------------------------------------------------------
c...  reading the configuration file
c-----------------------------------------------------------------------
         
         open(500,file='dipoles_KP.conf',status='old')

         read(500,*)dipoletype
         read(500,*)nl,nh
         read(500,*)mh(1:nh)
         read(500,*)nflsum(1)
         read(500,*)iflsum(1,1:nflsum(1))
         read(500,*)nflsum(2)
         read(500,*)iflsum(2,1:nflsum(2))
         read(500,*)alpha
         read(500,*)ranpol

         close(500)

         call qcdscale(muF)

         print*,'--------------------------------------------------'
         print*,'Integrated Dipole K+P Operator Configuration'
         print*,'--------------------------------------------------'
         if(dipoletype.eq.0) print*,'dipole type:       Catani-Seymour'
         if(dipoletype.eq.1) print*,'dipole type:       Nagy-Soper'
         print*,'factorization scale:',muF
         print*,'number of light quark flavors:',nl
         print*,'number of heavy quark flavors:',nh
         print*,'heavy quark masses:',(mh(i),i=1,nh)
         print*,'flavors in the summation over the 1st initial state:'
         print*,(iflsum(1,i),i=1,nflsum(1))
         print*,'flavors in the summation over the 2nd initial state:'
         print*,(iflsum(2,i),i=1,nflsum(2))
         print*,'alpha:',alpha
         print*,'random polarizations:',ranpol
         print*,'--------------------------------------------------'

         if (nh.ne.0.and.alpha.ne.1) then
            print*,'alpha dependence in the heavy quark splitting'
            print*,'is not supported yet!'
            stop
         endif

         if (ranpol) then
            sumpol=.false.
         else
            sumpol=.true.
         endif

c-----------------------------------------------------------------------
c...  input for the process
c-----------------------------------------------------------------------
         
c...  we need the tree for the process independently of the main HELAC
c...  routines, since the computation uses the same technology as the
c...  unintegrated dipoles from dipoles.F
         call produceinputs(n,ifl,0)
         
c-----------------------------------------------------------------------
c...  reading the skeleton from the tree file
c-----------------------------------------------------------------------
         
         call getstring(n,ifl,process)

         nunit1=123
         open(nunit1,file=
     .        './helac_trees/tree_'//process(2:2*n+1)//'.in',
     .        status='old',err=100)
         goto 101
         
 100     print*,'Could not open tree for ',process(2:2*n+1)
         print*,'Please run "make trees"'
         stop
         
 101     continue
         
         read(nunit1,*)kmax,ncc,imax
         if (kmax.gt.mxkmax) then
            print*,'maximum number of subamplitudes is',mxkmax
            print*,'please increase mxkmax in dimensions.h'
            stop
         endif
         if (ncc.gt.mxncc) then
            print*,'maximum number of color flows is',mxncc
            print*,'please increase mxncc in dimensions.h'
            stop
         endif
         
         read(nunit1,*)list1(0,1:n,1:3)
         read(nunit1,*)list(0,1:kmax,1:18,1:ncc)
         read(nunit1,*)kcou(0,1:ncc)
         read(nunit1,*)ng  
         if (ng.gt.mxfeyg) then
            print*,'maximum number of graphs is',mxfeyg
            print*,'please increase mxfeyg in dimensions.h'
            stop
         endif
         
         do ii=1,ng
            do jj=1,8
               read(nunit1,*)(isnew(ii,jj,kk),kk=1,8)
            enddo
         enddo
         
         read(nunit1,*)avhel,avcol,symet
         read(nunit1,*)rmatrix(1,1,1:ncc,1:ncc)
         
         do k=1,ncc
            read(nunit1,*)jcoldip(k,1:n,0:2)
         enddo
         
c...  closing tree file
         close(nunit1)

c-----------------------------------------------------------------------
c...  color correlator matrix
c-----------------------------------------------------------------------
            
         do i=1,ncc
            do j=1,ncc
               do k=1,n
                  do l=1,2
                     jcol(k,l)=jcoldip(i,k,l)
                     kcol(k,l)=jcoldip(j,k,l)
                  enddo
               enddo
               do k=1,2
c...  only partons
                  if (iabs(ifl(k)).ne.3.and.iabs(ifl(k)).ne.4.and.
     .                iabs(ifl(k)).ne.7.and.iabs(ifl(k)).ne.8.and.
     .                iabs(ifl(k)).ne.11.and.iabs(ifl(k)).ne.12.and.
     .                ifl(k).ne.35) cycle
                  icount=0
                  do l=1,n
c...  no diagonal correlators
                     if (k.eq.l) cycle
c...  only partons
                     if (iabs(ifl(l)).ne.3.and.iabs(ifl(l)).ne.4.and.
     .                    iabs(ifl(l)).ne.7.and.iabs(ifl(l)).ne.8.and.
     .                    iabs(ifl(l)).ne.11.and.iabs(ifl(l)).ne.12.and.
     .                    ifl(l).ne.35) cycle
                     icount=icount+1
                     rmatrix(k,icount,i,j)=colcor(jcol,kcol,k,l,n)
                  enddo
               enddo
            enddo
         enddo

c-----------------------------------------------------------------------
c     end of initialization
c-----------------------------------------------------------------------
         
      endif

c-----------------------------------------------------------------------
c...  helicity setup
c-----------------------------------------------------------------------

c...  random polarization initialization
      do i=1,n
         if (ranpol) then
            rpol(i)=2*pi*rnmy(0)
         else
            rpol(i)=-1
         endif
      enddo

c-----------------------------------------------------------------------
c...  KP operator contribution
c-----------------------------------------------------------------------

      if(dipoletype.eq.1)then   !...Nagy-Soper scheme
         stab=.true.
      endif

      weight=0d0

c...  coupling constant
      call coupling(n,ifl,pmom,gqcd,wsf)

c...  dynamical scale
      call qcdscale(muF)

c...  precompute the integrals
      do app=1,2
c...  only partons
         if (iabs(ifl(app)).ne.3.and.iabs(ifl(app)).ne.4.and.
     .        iabs(ifl(app)).ne.7.and.iabs(ifl(app)).ne.8.and.
     .        iabs(ifl(app)).ne.11.and.iabs(ifl(app)).ne.12.and.
     .        ifl(app).ne.35) cycle
         icount=0
         do jpp=1,n
c...  no diagonal correlators
            if (app.eq.jpp) cycle
c...  only partons
            if (iabs(ifl(jpp)).ne.3.and.iabs(ifl(jpp)).ne.4.and.
     .           iabs(ifl(jpp)).ne.7.and.iabs(ifl(jpp)).ne.8.and.
     .           iabs(ifl(jpp)).ne.11.and.iabs(ifl(jpp)).ne.12.and.
     .           ifl(jpp).ne.35) cycle
            icount=icount+1

            if(dipoletype.eq.0)then
c...  Catani-Seymour subtraction scheme
               KP(app,icount)=KPoperator(app,jpp)
            elseif(dipoletype.eq.1)then
c...  Nagy-Soper subtraction scheme
c              ioption=0        !...Numerical integ. of the unconstrained parton
               ioption=1        !...Monte Carlo integ. of the unconstrained parton
               KP(app,icount)=NSKPoperator(app,jpp,ioption,stable)
               if(.not.stable)then
                  stab=.false.
                  print*,' '
                  print*,'MESSAGE from dipoles_KP.F: unstable event'
     -                 //' found, weight set to zero'
                  print*,' '
                  GOTO 320
               endif
            else
               print*,'dipoles_KP.f ERROR: invalid dipole '//
     .              'scheme setting',dipoletype
               STOP
            endif

         enddo
      enddo

c...  initialize the counter and polarizations
      if (sumpol) then
         npol=0
         ipol(1:n)=1
      endif

c...  skip vanishing contributions during summation over polarizations
 310  if (sumpol) then
         npol=npol+1
         if (npol.gt.mxhel) then
            print*,'maximum number of helicity configs is',mxhel
            print*,'please increase mxhel in dimensions.h'
            stop
         endif
         if (.not.firstcall.and.amppol(npol).lt.1e-15) then
            if (nextpol(n,ifl,ipol,rpol).ne.0) then
               goto 310
            else
               goto 320
            endif
         endif
      endif

c...  amplitude
      call iniqqdip(n,ifl,io,pmom,ipol,rpol,list1,0,zy)

      do k=1,ncc
         call nextqdip(n,k,imax,list,kcou,0,zy)
         call ampqdip(k,ifl,zy,list,kcou,0,zamp)
      enddo
      
      w=0d0
      do app=1,2
c...  only partons
         if (iabs(ifl(app)).ne.3.and.iabs(ifl(app)).ne.4.and.
     .        iabs(ifl(app)).ne.7.and.iabs(ifl(app)).ne.8.and.
     .        iabs(ifl(app)).ne.11.and.iabs(ifl(app)).ne.12.and.
     .        ifl(app).ne.35) cycle
         icount=0
         do jpp=1,n
c...  no diagonal correlators
            if (app.eq.jpp) cycle
c...  only partons
            if (iabs(ifl(jpp)).ne.3.and.iabs(ifl(jpp)).ne.4.and.
     .           iabs(ifl(jpp)).ne.7.and.iabs(ifl(jpp)).ne.8.and.
     .           iabs(ifl(jpp)).ne.11.and.iabs(ifl(jpp)).ne.12.and.
     .           ifl(jpp).ne.35) cycle
            icount=icount+1
            do k1=1,ncc
               do k2=1,ncc
                  w=w+zamp(k1)*dconjg(zamp(k2))
     .               *rmatrix(app,icount,k1,k2)*KP(app,icount)
               enddo
            enddo
         enddo
      enddo
      w=w*avhel*avcol*symet
      w=gqcd**2/norm*w
c-----------------------------------------------------------------------  
c...  pdf for dynamical scale implementation                                 
c----------------------------------------------------------------------- 
      w=w*wsf

      weight = weight + w

c...  store normalized amplitude values for all polarizations
      if (sumpol) then
         if (firstcall) amppol(npol)=abs(w)
         if (nextpol(n,ifl,ipol,rpol).ne.0) goto 310
         if (firstcall) then
            ampmax=0d0
            do i=1,npol
               if (amppol(i).gt.ampmax) ampmax=amppol(i)
            enddo
            amppol(1:npol)=amppol(1:npol)/ampmax
         endif
      endif
      
 320  continue

c-----------------------------------------------------------------------
c...  histogramming
c-----------------------------------------------------------------------

      if (.not.firstcall.and.key.eq.2) 
     .     call HISTOm(n,ifl,pmom,weight*wmeasure)

c-----------------------------------------------------------------------
c...  initialization is done only once !
c-----------------------------------------------------------------------

      firstcall=.false.

#if SINGLEPOINT
      print*,'gqcd:',gqcd
      print*,'momenta:'
      do i=1,n
         print*,(pmom(i,j),j=1,5)
      enddo
      print*,'value:',weight
      stop
#endif

      return
      end

c-----------------------------------------------------------------------

      double precision function KPoperator(app,jpp)

c-----------------------------------------------------------------------
c...  coefficient of the color correlator Tj.Ta in the K+P operator
c...  the normalization does not include the coupling alphas/2pi
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_strf.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)
c...  Riemann zeta(2)
      double precision z2
      parameter(z2=1.64493406684822644d0)

      integer app,jpp

      double precision mj,mjsq,Ka
      double precision integral
      double precision JQQNS

      external KPkrn1sum1,KPkrn1sum2,KPkrn2,KPkrn3

c-----------------------------------------------------------------------
c...  setup
c-----------------------------------------------------------------------
      
      ap=app
      jp=jpp

      if (ap.eq.1) then
         eta=xp1
      else
         eta=xp2
      endif

      ifla=ifl(ap)
      sja=2*(pmom(jp,4)*pmom(ap,4)-pmom(jp,1)*pmom(ap,1)
     .      -pmom(jp,2)*pmom(ap,2)-pmom(jp,3)*pmom(ap,3))
      mj=parmas(ifl(jp))
      mjsq=mj**2

      if (ifla.eq.35) then
c...  gluon emitter
         ga=11d0/6*CA-2d0/3*TF*nl
         Ka=(67d0/18-z2)*CA-10d0/9*TF*nl
         Tasq=CA
      else
c...  quark emitter
         ga=3d0/2*CF
         Ka=(7d0/2-z2)*CF
         Tasq=CF
      endif

c...  modification to include the alpha parameter.
      Ka=Ka-Tasq*log(alpha)**2+ga*(alpha-1-log(alpha))

c...  pdf
      if (ipp.eq.1.and.ap.eq.2.and.ifla.ne.35) then
         f1=fpdf(-ifla,eta,muF)
      else
         f1=fpdf(ifla,eta,muF)
      endif
      
c-----------------------------------------------------------------------
c...  Dirac delta contributions
c-----------------------------------------------------------------------

c...  check that there exists a flavor diagonal element in the sum
      do i=1,nflsum(ap)
         if (iflsum(ap,i).eq.ifla) goto 1
      enddo

      KPoperator=0
      goto 2

c...  P contribution
 1    KPoperator=ga*log(muF**2/sja)/Tasq

c...  Kbar contribution with factorized color correlator
      KPoperator=KPoperator+(ga+Ka-5*z2*Tasq)/Tasq-log(alpha)**2

      if (jp.gt.2) then
c...  Kj contribution
         if (ifl(jp).eq.35) then
c...  gluon emitter
            gj=11d0/6*CA-2d0/3*TF*nl
            Tjsq=CA
         else
c...  quark emitter
            gj=3d0/2*CF
            Tjsq=CF
         endif
         KPoperator=KPoperator+gj/Tjsq*alpha
c...  purely massive contribution including the effect of the alpha
c...  dependence of the final-initial integrated dipoles in the quark
c...  case. the latter is written in such a way that there is no smooth
c...  massless limit in the expression below!!!
         if (ifl(jp).ne.35.and.mjsq.ne.0) KPoperator=KPoperator
     .        -(mjsq/sja*log(mjsq/(sja+mjsq))+0.5d0*mjsq/(sja+mjsq))
     .        +log(alpha)*(0.5d0-2*log((mjsq+sja)/mjsq))
c...  additional purely massive contribution
         KPoperator=KPoperator-ga*(log((sja-2*mj*sqrt(sja+mjsq)
     .        +2*mjsq)/sja)+2*mj/(sqrt(sja+mjsq)+mj))/Tasq
      else
c...  Ktilde contribution
         KPoperator=KPoperator+2*z2
      endif

c...  include the pdf
      KPoperator=KPoperator*f1

c-----------------------------------------------------------------------
c...  contributions from gluon splitting into massive quarks
c-----------------------------------------------------------------------

      if (jp.gt.2.and.ifl(jp).eq.35) then
         do i=1,nh
            mp=mh(i)**2/sja
c...  delta(1-x) contribution
            if (sja.gt.4*mh(i)**2) KPoperator=KPoperator
     .           -TF/CA*(2d0/3*(log(mp)+5d0/3)-JQQNS(mp)
     .           +2d0/3*sqrt(1-4*mp)**3)*f1
c...  the change in the definition of xp is due to the change of
c...  variables to keep the cross section at fixed momentum, just as
c...  everywhere else
            xp=1/(1+4*mp)
            if (xp.gt.eta) then
c...  delta(1-xp) contribution
               if (ipp.eq.1.and.ap.eq.2.and.ifla.ne.35) then
                  fxp=fpdf(-ifla,eta/xp,muF)
               else
                  fxp=fpdf(ifla,eta/xp,muF)
               endif
               KPoperator=KPoperator+TF/CA*(2d0/3*(log(mp*xp)+5d0/3)
     .              -JQQNS(mp*xp))*fxp
c...  x+ distribution contribution
               KPoperator=KPoperator-TF/CA*integral(KPkrn2,eta,xp)
               KPoperator=KPoperator-TF/CA*integral(KPkrn3,0d0,eta)
            endif
         enddo
      endif

c-----------------------------------------------------------------------
c...  integral of the plus distribution and regular part over [eta,1]
c-----------------------------------------------------------------------

 2    if (1-alpha.lt.eta) then
         KPoperator=KPoperator+integral(KPkrn1sum1,eta,1d0)
      else
         KPoperator=KPoperator+integral(KPkrn1sum1,eta,1-alpha)
         KPoperator=KPoperator+integral(KPkrn1sum1,1-alpha,1d0)
      endif

c-----------------------------------------------------------------------
c...  integral of the plus distribution end point over [0,eta]
c-----------------------------------------------------------------------

      if (1-alpha.gt.eta) then
         KPoperator=KPoperator+integral(KPkrn1sum2,0d0,eta)
      else
         KPoperator=KPoperator+integral(KPkrn1sum2,0d0,1-alpha)
         KPoperator=KPoperator+integral(KPkrn1sum2,1-alpha,eta)
      endif

c-----------------------------------------------------------------------
c...  include the other pdf. WORKS ONLY IN HADRON-HADRON COLLISIONS!
c-----------------------------------------------------------------------

      iflb=ifl(3-ap)
      if (ap.eq.1) then
         eta=xp2
      else
         eta=xp1
      endif
      if (ipp.eq.1.and.ap.eq.1.and.iflb.ne.35) then
         KPoperator=KPoperator*fpdf(-iflb,eta,muF)
      else
         KPoperator=KPoperator*fpdf(iflb,eta,muF)
      endif
      
      return
      end

c-----------------------------------------------------------------------

      double precision function KPkrn1sum1(x)

c-----------------------------------------------------------------------
c...  KPkrn1 summed over the flavors for integration over [eta,1]
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

      double precision KPkrn1

      KPkrn1sum1=0
      do i=1,nflsum(ap)
         iflb=iflsum(ap,i)
         if (ifla.eq.iflb.or.ifla.eq.35.or.iflb.eq.35) then
            if (ipp.eq.1.and.ap.eq.2.and.iflb.ne.35) then
               fx=fpdf(-iflb,eta/x,muF)/x
               f1=fpdf(-iflb,eta,muF)
            else
               fx=fpdf(iflb,eta/x,muF)/x
               f1=fpdf(iflb,eta,muF)
            endif
            KPkrn1sum1=KPkrn1sum1+KPkrn1(x,fx,f1)
         endif
      enddo

      end

c-----------------------------------------------------------------------

      double precision function KPkrn1sum2(x)

c-----------------------------------------------------------------------
c...  KPkrn1 summed over the flavors for integration over [0,eta]
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

      double precision KPkrn1

      KPkrn1sum2=0
      do i=1,nflsum(ap)
         iflb=iflsum(ap,i)
         if (ifla.eq.iflb.or.ifla.eq.35.or.iflb.eq.35) then
            if (ipp.eq.1.and.ap.eq.2.and.iflb.ne.35) then
               fx=0
               f1=fpdf(-iflb,eta,muF)
            else
               fx=0
               f1=fpdf(iflb,eta,muF)
            endif
            KPkrn1sum2=KPkrn1sum2+KPkrn1(x,fx,f1)
         endif
      enddo

      end

c-----------------------------------------------------------------------

      double precision function KPkrn1(x,fx,f1)

c-----------------------------------------------------------------------
c...  contains all plus distributions and regular functions
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_dipoles_KP.h'
      include 'dimensions.h'

      double precision mjsq,zp
      double precision Kbar,Kj,Ktilde

c-----------------------------------------------------------------------
c...  setup
c-----------------------------------------------------------------------

c...  sjax corresponds to sja expressed through the original initial
c...  state momentum before the x boost
      sjax=sja/x

c...  emitter mass (spectator in the case of initial-final dipoles)
      mjsq=parmas(ifl(jp))**2

c...  Casimir operator
      if (ifla.eq.35) then
         Tasq=3d0
      else
         Tasq=4d0/3
      endif

c...  splitting functions
      if (iflb.eq.35.and.ifla.eq.35) then
         Pregx=Pggreg(x)
         Phatx=Pgghat(x)
         Pavgx=Pggavg(x)
      else if (iflb.eq.35.and.ifla.ne.35) then
         Pregx=Pgqreg(x)
         Phatx=Pgqhat(x)
         Pavgx=Pgqavg(x)
      else if (iflb.ne.35.and.ifla.eq.35) then
         Pregx=Pqgreg(x)
         Phatx=Pqghat(x)
         Pavgx=Pqgavg(x)
      else
         Pregx=Pqqreg(x)
         Phatx=Pqqhat(x)
         Pavgx=Pqqavg(x)
      endif

c...  difference of the test functions as in the definition of the
c...  plus distributions
      df=fx-f1

c-----------------------------------------------------------------------
c...  P contribution
c-----------------------------------------------------------------------

      P=Pregx*fx
      if (iflb.eq.ifla) P=P+2*Tasq/(1-x)*df
      P=P/Tasq*log(muF**2/sja)

c-----------------------------------------------------------------------
c...  Kbar contribution
c-----------------------------------------------------------------------

      Kbar=(Pregx*log((1-x)/x)+Phatx)*fx
      if (iflb.eq.ifla) then
         Kbar=Kbar+2*Tasq*log((1-x)/x)/(1-x)*df
      endif

c...  initial-final integrated dipole alpha dependence. the initial-
c...  initial case is also included as given by Nagy.
c...
c...  notice that in the massive case, there is a dependence on the
c...  spectator mass and therefore the Kbar operator acquires a "j"
c...  index. it's against the concept, but we don't want to do major
c...  design changes, like shifting things between the Kbar and Kj
c...  operators.
      zp=(1-x)/(1-x+mjsq/sjax)
      if (zp.gt.alpha) then
         Kbar=Kbar+Pregx*log(alpha/zp)*fx
         if (iflb.eq.ifla) Kbar=Kbar+2*Tasq
     .        *log(alpha*(1+zp-x)/(zp*(1+alpha-x)))/(1-x)*fx
         if (ifla.eq.35.and.mjsq.ne.0) then
            if (iflb.eq.35) then
               Kbar=Kbar-6*mjsq/(x*sjax)*log((1-zp)/(1-alpha))*fx
            else
               Kbar=Kbar-8d0/3*mjsq/(x*sjax)*log((1-zp)/(1-alpha))*fx
            endif
         endif
      endif

c...  first part of the final-initial integrated dipole alpha dependence
c...  in the massless case. the initial-initial case is also included
c...  as given by Nagy.
      if (iflb.eq.ifla.and.mjsq.eq.0) then
         if (x.lt.1-alpha) Kbar=Kbar-2*Tasq*log((2-x)/(1-x))/(1-x)*fx
      endif

c-----------------------------------------------------------------------
c...  initial state emitter
c-----------------------------------------------------------------------

      if (jp.le.2) then

         Ktilde=Pregx*log((1-x)/alpha)*fx
         if (alpha.lt.1-x) then
            Ktilde=Ktilde-Pavgx*log((1-x)/alpha)*fx
         endif
         if (iflb.eq.ifla) then
            if (x.gt.1-alpha) Ktilde=Ktilde+2*Tasq*log(1-x)/(1-x)*df
            Ktilde=Ktilde+2*Tasq*log((1+alpha-x)/alpha)/(1-x)*fx
            if (x.gt.1-alpha) Ktilde=Ktilde-2*Tasq*log(2-x)/(1-x)*fx
         endif
         KPkrn1=P-Kbar/Tasq-Ktilde/Tasq
         return

c-----------------------------------------------------------------------
c...  final state emitter
c-----------------------------------------------------------------------

      else
      
c-----------------------------------------------------------------------
c...  Kj contribution
c-----------------------------------------------------------------------

         if (ifl(jp).eq.35) then
c...  gluon emitter

            if (iflb.eq.ifla.and.x.gt.1-alpha) then
               Kj=-(11d0/6-1d0/9*nl)/(1-x)*df
            else
               Kj=0
            endif

         else
c...  quark emitter

            if (ifla.eq.35.and.mjsq.ne.0) then
               Kj=2*mjsq/(x*sjax)*log(mjsq/((1-x)*sjax+mjsq))*fx
               if (iflb.ne.35) Kj=4d0/9*Kj
            else
               Kj=0
            endif

            if (iflb.eq.ifla) then
               Kj=Kj+2*(log(1-x)/(1-x)*df-log(2-x)/(1-x)*fx)
               Kj=Kj+((1-x)/(2*(1-x+mjsq/sjax)**2)-2/(1-x)*(1
     .              +log(1-x+mjsq/sjax))+2/(1-x)*(log(2+mjsq/sjax-x)))
     .              *fx-((1-x)/(2*(1-x+mjsq/sja)**2)-2/(1-x)*(1
     .              +log(1-x+mjsq/sja))+2/(1-x)*(log(1+mjsq/sja)))*f1
               Kj=Kj+2/(1-x)*(log((2-x)*sjax/((2-x)*sjax+mjsq))*fx
     .              -log(sja/(sja+mjsq))*f1)

c...  second part of the final-initial integrated dipole alpha
c...  dependence
               if (x.lt.1-alpha) then
                  if (mjsq.eq.0) then
c...  massless case
                     Kj=Kj+1.5d0/(1-x)*df
                  else
c...  massive case
                     Kj=Kj-2*(log((mjsq+sja)/mjsq)-1)/(1-x)*df
                     Kj=Kj-((1-x)/(2*(1-x+mjsq/sjax)**2)+2/(1-x)
     .                    *log((2-x+mjsq/sjax)*mjsq/((mjsq+sja)
     .                    *(1-x+mjsq/sjax))))*fx
                  endif
               endif
               
            endif

         endif

      endif

c-----------------------------------------------------------------------
c...  complete contribution
c-----------------------------------------------------------------------

      KPkrn1=P-Kbar/Tasq-Kj
     .     -Pregx*log((1-x)*sjax/((1-x)*sjax+mjsq))/Tasq*fx

      return
      end

c-----------------------------------------------------------------------

      double precision function KPkrn2(x)

c-----------------------------------------------------------------------
c...  integration kernel for the x+ distibution JQQ
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

      double precision JQQxp
      
      if (ipp.eq.1.and.ap.eq.2.and.ifla.ne.35) then
         fx=fpdf(-ifla,eta/x,muF)/x
         fxp=fpdf(-ifla,eta/xp,muF)
      else
         fx=fpdf(ifla,eta/x,muF)/x
         fxp=fpdf(ifla,eta/xp,muF)
      endif
      
      KPkrn2=JQQxp(x,mp*x)*fx-JQQxp(x,mp*xp)*fxp

      end

c-----------------------------------------------------------------------

      double precision function KPkrn3(x)

c-----------------------------------------------------------------------
c...  integration kernel for the x+ distibution JQQ
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

      double precision JQQxp

      if (ipp.eq.1.and.ap.eq.2.and.ifla.ne.35) then
         fxp=fpdf(-ifla,eta/xp,muF)
      else
         fxp=fpdf(ifla,eta/xp,muF)
      endif
     
      KPkrn3=-JQQxp(x,mp*xp)*fxp

      end

c-----------------------------------------------------------------------

      double precision function JQQNS(muQsq)
      include 'declare.h'
      double precision muQsq
      vmuQ=sqrt(1-4*muQsq)
      JQQNS=10d0/9*(1-vmuQ)-8d0/9*muQsq*vmuQ+4d0/3*log((1+vmuQ)/2)

      end

c-----------------------------------------------------------------------

      double precision function JQQxp(x,muQsq)
      include 'declare.h'
      double precision muQsq
      JQQxp=2d0/3*(1-x+2*muQsq)/(1-x)**2*sqrt(1-4*muQsq/(1-x))
      end

c-----------------------------------------------------------------------
c...  spin averaged Altarelli-Parisi splitting functions
c-----------------------------------------------------------------------

      double precision function Pqqavg(x)
      include 'declare.h'
      Pqqavg=4d0/3*((1+x**2)/(1-x))
      end

      double precision function Pqgavg(x)
      include 'declare.h'
      Pqgavg=4d0/3*(1+(1-x)**2)/x
      end

      double precision function Pgqavg(x)
      include 'declare.h'
      Pgqavg=0.5d0*(x**2+(1-x)**2)
      end

      double precision function Pggavg(x)
      include 'declare.h'
      Pggavg=6*(x/(1-x)+(1-x)/x+x*(1-x))
      end

c-----------------------------------------------------------------------
c...  regular parts of the Altarelli-Parisi splitting functions
c-----------------------------------------------------------------------

      double precision function Pqqreg(x)
      include 'declare.h'
      Pqqreg=-4d0/3*(1+x)
      end

      double precision function Pqgreg(x)
      include 'declare.h'
      Pqgreg=4d0/3*(1+(1-x)**2)/x
      end

      double precision function Pgqreg(x)
      include 'declare.h'
      Pgqreg=0.5d0*(x**2+(1-x)**2)
      end

      double precision function Pggreg(x)
      include 'declare.h'
      Pggreg=6*((1-x)/x-1+x*(1-x))
      end

c-----------------------------------------------------------------------
c...  epsilon parts of the Altarelli-Parisi splitting functions
c-----------------------------------------------------------------------

      double precision function Pqqhat(x)
      include 'declare.h'
      Pqqhat=4d0/3*(1-x)
      end

      double precision function Pqghat(x)
      include 'declare.h'
      Pqghat=4d0/3*x
      end

      double precision function Pgqhat(x)
      include 'declare.h'
      Pgqhat=x*(1-x)
      end

      double precision function Pgghat(x)
      include 'declare.h'
      Pgghat=0
      end

c-----------------------------------------------------------------------

      double precision function integral(f,a,b)

c-----------------------------------------------------------------------
c...  simplified integrator interface
c-----------------------------------------------------------------------

      include 'declare.h'
c...  MC integration
      if (b.gt.a) then
         r1=rnmy(0)*(b-a)+a
         integral=(b-a)*f(r1)
c...  deterministic Gaussian integration
c     call dadapt(f,a,b,1,1d-4,0d0,integral,err)
c     print*,'dadapt',integral,err
      else
         integral=0
      endif

      end

c-----------------------------------------------------------------------

      SUBROUTINE DADAPT(F,A,B,NSEG,RELTOL,ABSTOL,RES,ERR)

c-----------------------------------------------------------------------
c...  double precision adaptive gaussian quadrature taken from CERNlib
c-----------------------------------------------------------------------

      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
 
C     RES = Estimated Integral of F from A to B,
C     ERR = Estimated absolute error on RES.
C     NSEG  specifies how the adaptation is to be done:
C        =0   means use previous binning,
C        =1   means fully automatic, adapt until tolerance attained.
C        =n>1 means first split interval into n equal segments,
C             then adapt as necessary to attain tolerance.
C     The specified tolerances are:
C            relative: RELTOL ;  absolute: ABSTOL.
C        It stops when one OR the other is satisfied, or number of
C        segments exceeds NDIM.  Either TOLA or TOLR (but not both!)
C        can be set to zero, in which case only the other is used.
 
      EXTERNAL F
 
      PARAMETER (NDIM=100)
      PARAMETER (R1 = 1, HF = R1/2)
 
      DIMENSION XLO(NDIM),XHI(NDIM),TVAL(NDIM),TERS(NDIM)
      SAVE XLO,XHI,TVAL,TERS,NTER
      DATA NTER /0/
 
      IF(NSEG .LE. 0)  THEN
       IF(NTER .EQ. 0) THEN
        NSEGD=1
        GO TO 2
       ENDIF
       TVALS=0
       TERSS=0
       DO 1 I = 1,NTER
       CALL DGS56P(F,XLO(I),XHI(I),TVAL(I),TE)
       TERS(I)=TE**2
       TVALS=TVALS+TVAL(I)
       TERSS=TERSS+TERS(I)
    1  CONTINUE
       ROOT= SQRT(2*TERSS)
       GO TO 9
      ENDIF
      NSEGD=MIN(NSEG,NDIM)
    2 XHIB=A
      BIN=(B-A)/NSEGD
      DO 3 I = 1,NSEGD
      XLO(I)=XHIB
      XLOB=XLO(I)
      XHI(I)=XHIB+BIN
      IF(I .EQ. NSEGD) XHI(I)=B
      XHIB=XHI(I)
      CALL DGS56P(F,XLOB,XHIB,TVAL(I),TE)
      TERS(I)=TE**2
    3 CONTINUE
      NTER=NSEGD
      DO 4 ITER = 1,NDIM
      TVALS=TVAL(1)
      TERSS=TERS(1)
      DO 5 I = 2,NTER
      TVALS=TVALS+TVAL(I)
      TERSS=TERSS+TERS(I)
    5 CONTINUE
      ROOT= SQRT(2*TERSS)
      IF(ROOT .LE. ABSTOL .OR. ROOT .LE. RELTOL*ABS(TVALS)) GO TO 9
      IF(NTER .EQ. NDIM) GO TO 9
      BIGE=TERS(1)
      IBIG=1
      DO 6 I = 2,NTER
      IF(TERS(I) .GT. BIGE) THEN
       BIGE=TERS(I)
       IBIG=I
      ENDIF
    6 CONTINUE
      NTER=NTER+1
      XHI(NTER)=XHI(IBIG)
      XNEW=HF*(XLO(IBIG)+XHI(IBIG))
      XHI(IBIG)=XNEW
      XLO(NTER)=XNEW
      CALL DGS56P(F,XLO(IBIG),XHI(IBIG),TVAL(IBIG),TE)
      TERS(IBIG)=TE**2
      CALL DGS56P(F,XLO(NTER),XHI(NTER),TVAL(NTER),TE)
      TERS(NTER)=TE**2
    4 CONTINUE
    9 RES=TVALS
      ERR=ROOT
      RETURN
      END

c-----------------------------------------------------------------------

      SUBROUTINE DGS56P(F,A,B,RES,ERR)

c-----------------------------------------------------------------------
c...  double precision gaussian quadrature points taken from CERNlib
c-----------------------------------------------------------------------

      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
 
      PARAMETER (R1 = 1, HF = R1/2)
      DIMENSION X5(5),W5(5),X6(6),W6(6)
 
      DATA (X5(I),W5(I),I=1,5)
     1/4.6910077030668004D-02, 1.1846344252809454D-01,
     2 2.3076534494715846D-01, 2.3931433524968324D-01,
     3 5.0000000000000000D-01, 2.8444444444444444D-01,
     4 7.6923465505284154D-01, 2.3931433524968324D-01,
     5 9.5308992296933200D-01, 1.1846344252809454D-01/
 
      DATA (X6(I),W6(I),I=1,6)
     1/3.3765242898423989D-02, 8.5662246189585178D-02,
     2 1.6939530676686775D-01, 1.8038078652406930D-01,
     3 3.8069040695840155D-01, 2.3395696728634552D-01,
     4 6.1930959304159845D-01, 2.3395696728634552D-01,
     5 8.3060469323313225D-01, 1.8038078652406930D-01,
     6 9.6623475710157601D-01, 8.5662246189585178D-02/
 
      RANG=B-A
      E5=0
      E6=0
      DO 1 I = 1,5
      E5=E5+W5(I)*F(A+RANG*X5(I))
      E6=E6+W6(I)*F(A+RANG*X6(I))
    1 CONTINUE
      E6=E6+W6(6)*F(A+RANG*X6(6))
      RES=HF*(E6+E5)*RANG
      ERR=ABS((E6-E5)*RANG)
      RETURN
      END


c-------------------------------------------------------------------------

      double precision function NSKPoperator(app,jpp,ioption,stab)

c-------------------------------------------------------------------------
c...  coefficient of the color correlator Tj.Ta in the K+P operator based 
c...  on Nagy-Soper subtraction scheme.
c...  The normalization does not include the coupling alphas/2pi.
c...  The input variable ioption sets the numerical integration mode:
c...   ioption = 0  -->  returns the exact KP-operator for a given phase-
c...                     space point. Integration over the kinematics 
c...                     of the unconstrained parton is performed 
c...                     numerically using the Gauss-Legendre algorithm.
c...   ioption = 1  -->  returns the value of the KP-operator for given
c...                     values of the integration variables of the 
c...                     unconstrained parton (randomly generated inside
c...                     the routine). Suitable for Monte Carlo summation.
c...  In the present version, only ioption=1 is available.
c-------------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_strf.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles_KP.h'
      include 'parameter_pp.h'          
      include 'dimensions.h'

      integer app,jpp,id_unconstr,ioption
      integer nl,i,k,ifla,iflb
      double precision muscale,xrandom1,xrandom2,xmin,xmax
      double precision eta,mj,ma,mjsq,masq
      double precision KPoperator,res_reg,res_plus,res_delta,x
      double precision res_Kqq,res_Kqg,res_Kgq,res_Kgg
      double precision res_Gqq,res_Ggg
      double precision res_Hqq,res_Hgg
      double precision fpdfx_q,fpdf1_q,fpdfx_g,fpdf1_g
      integer ipdf_q,ipdf_g
      double precision Koperator,Poperator
      logical stable,stab

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)


c...  In the present version, only the Monte Carlo sum option (ioption=1) is 
c...  available
      if(ioption.eq.0)then
         print*,' '
         print*,'WARNING: numerical integration mode'
         print*,'         not available yet. STOP'
         print*,' '
         STOP
      endif



c-----------------------------------------------------------------------
c...  setup
c-----------------------------------------------------------------------

c...  app = position of the emitter parton
c...  jpp = position of the spectator parton
      mj=parmas(ifl(jpp))
      ma=parmas(ifl(app))
      mjsq=mj**2
      masq=ma**2

c... momentum fraction
      if (app.eq.1) then
         eta=xp1
      else
         eta=xp2
      endif

      fpdfx_q=0.d0
      fpdf1_q=0.d0
      ipdf_q=0
      fpdfx_g=0.d0
      fpdf1_g=0.d0
      ipdf_g=0

      Koperator=0.d0
      Poperator=0.d0
      NSKPoperator = 0.d0
      stab=.true.

c... Note: we introduce here an arbitrary renormalization scale (muscale)
c... because intermediate results need it (also useful for cross-checks). 
c... In the end, combining K+P, it must cancel leaving the final result 
c... dependent upon the factorization scale only (muF).

      muscale = 2.d0*muF
      xrandom1=rnmy(0)
      xrandom2=rnmy(0)

c---------------------------------------------------------------------------
c...  KP operator.
c...  The contributing terms for both K and P induce dependences on x which 
c...  combined yield result in the structure
c...
c...                [G(x)]_plus + K(x) + delta(1-x)*H(x)
c...
c...  where K(x) is a regular function of x, G(x) is a plus distribution
c...  and H(x) is the coefficient of a Dirac delta(1-x).
c---------------------------------------------------------------------------

c---------------------------------------------------------------------------
c...  Dirac delta contribution + integral of the 'plus' distribution and 
c...  regular part over [eta,1]
c---------------------------------------------------------------------------

      xmin = eta
      xmax = 1.d0
      x = xmin+(xmax-xmin)*xrandom2

c... pdf's
c... Note: in case of MASSLESS initial-state partons, the integrated dipole formulae
c... are independent on the quark flavour and therefore the pdf contributions can be
c... summed up
      ifla=ifl(app)
      do i=1,nflsum(app)
         iflb=iflsum(app,i)
         if(iflb.eq.35)then
            ipdf_g=1
            fpdfx_g = fpdf(35,eta/x,muF)/x
            fpdf1_g = fpdf(35,eta,muF)
         endif
         if(iflb.ne.35 .AND. (ifla.eq.iflb .OR. ifla.eq.35))then
            ipdf_q=1
            if(ipp.eq.1.and.app.eq.2)then !p-pbar collider
               fpdfx_q = fpdfx_q + fpdf(-iflb,eta/x,muF)/x
               fpdf1_q = fpdf1_q + fpdf(-iflb,eta,muF)
            else
               fpdfx_q = fpdfx_q + fpdf(iflb,eta/x,muF)/x
               fpdf1_q = fpdf1_q + fpdf(iflb,eta,muF)
            endif
         endif
      enddo


      if(ifla.ne.35)then
c... Quark emitter. Two possible contributions: q->qg and q->gq

c...  1. Q->QG splitting:
         if(ipdf_q.ne.0)then
            id_unconstr=35
c...  K operator: regular part from diagonal contribution (Waa)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,app,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Kqq = res_reg*x*fpdfx_q
            res_Hqq = res_delta*fpdf1_q
            Koperator = Koperator + res_Kqq + res_Hqq
c...  K operator: 'plus' distribution and Dirac delta part from interference 
c...  contribution (Waj)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,jpp,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Gqq = res_plus*(x*fpdfx_q - fpdf1_q)
            res_Hqq = res_delta*fpdf1_q
            Koperator = Koperator + res_Gqq + res_Hqq
c...  P operator (Pqq)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Kqq = res_reg*fpdfx_q
            res_Gqq = res_plus*(fpdfx_q - fpdf1_q)
            res_Hqq = res_delta*fpdf1_q
            Poperator = Poperator + res_Kqq + res_Gqq + res_Hqq
         endif

c... 2. Q->GQ splitting:
         if(ipdf_g.ne.0)then
            id_unconstr=-ifla
c...  K operator: regular part from diagonal contribution (Waa)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,app,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Kgq = (res_reg*TF/CF)*x*fpdfx_g
            Koperator = Koperator + res_Kgq
c... P operator (Pgq)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Kgq = (res_reg*TF/CF)*fpdfx_g
            Poperator = Poperator + res_Kgq
         endif

      elseif(ifla.eq.35) then
c... Gluon emitter. Two possible contributions: g->gg and g->qq

c... 1. G->GG splitting:
         if(ipdf_g.ne.0)then
            id_unconstr=35
c... K operator: regular part from diagonal contribution (Waa)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,app,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Kgg = res_reg*x*fpdfx_g
            res_Hgg = res_delta*fpdf1_g
            Koperator = Koperator + res_Kgg + res_Hgg
c... K operator: 'plus' distribution and Dirac delta part from interference 
c... contribution (Waj) 
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,jpp,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Ggg = res_plus*(x*fpdfx_g - fpdf1_g)
            res_Hgg = res_delta*fpdf1_g
            Koperator = Koperator + res_Ggg + res_Hgg
c... P operator (Pgg)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Kgg = res_reg*fpdfx_g
            res_Ggg = res_plus*(fpdfx_g - fpdf1_g)
            res_Hgg = res_delta*fpdf1_g
            Poperator = Poperator + res_Kgg + res_Ggg + res_Hgg
         endif

c... 2. G->QQ splitting:
         if(ipdf_q.ne.0)then
            id_unconstr=3       !...light quark flavours are degenerate
c... K operator: regular part from diagonal contribution (Waa)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,app,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Kqg = (res_reg*CF/CA)*(x*fpdfx_q)
            Koperator = Koperator + res_Kqg
c... P operator (Pqg)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Kqg = (res_reg*CF/CA)*fpdfx_q
            Poperator = Poperator + res_Kqg
         endif

      endif                     ! if(ifla.ne.35)then


c---------------------------------------------------------------------------
c...  Integral of the 'plus' distribution end point over [0,eta]
c---------------------------------------------------------------------------

      xmin = 0.d0
      xmax = eta
      x = xmin+(xmax-xmin)*xrandom2

      if(ifla.ne.35)then

c... Q->QG splitting
         if(ipdf_q.ne.0)then
            id_unconstr=35
c... K operator: 'plus' ditribution part from interference contribution (Waj)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,jpp,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Gqq = res_plus*(-fpdf1_q)
            Koperator = Koperator + res_Gqq
c... P operator (Pqq)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Gqq = res_plus*(-fpdf1_q)
            Poperator = Poperator + res_Gqq
         endif

      elseif(ifla.eq.35)then

c... G->GG splitting
         if(ipdf_g.ne.0)then
            id_unconstr=35
c... K operator: 'plus' distribution from interference contribution (Waj)
            call NSxintdipole(n,ifl,pmom,id_unconstr,app,jpp,
     -           xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -           res_reg,res_plus,res_delta,stable)
            if(.not.stable)then
               stab=.false.
               return
            endif
            res_Ggg = res_plus*(-fpdf1_g)
            Koperator = Koperator + res_Ggg
c... P operator (Pgg)
            call Pkrn(n,ifl,id_unconstr,app,xrandom2,xmin,xmax,
     -           muscale,muF,nl,res_reg,res_plus,res_delta)
            res_Ggg = res_plus*(-fpdf1_g)
            Poperator = Poperator + res_Ggg
         endif

      endif                     ! if(ifla.ne.35)then

c... Add minus sign coming from color factor:  -1/2[ tl x tk  + tk x tl]
      Koperator=-Koperator
      Poperator=-Poperator

      NSKPoperator = Koperator + Poperator

c-----------------------------------------------------------------------
c...  include the other pdf. WORKS ONLY IN HADRON-HADRON COLLISIONS!
c-----------------------------------------------------------------------

      iflb=ifl(3-app)
      if (app.eq.1) then
         eta=xp2
      else
         eta=xp1
      endif
      if (ipp.eq.1.and.app.eq.1.and.iflb.ne.35) then
         NSKPoperator=NSKPoperator*fpdf(-iflb,eta,muF)
      else
         NSKPoperator=NSKPoperator*fpdf(iflb,eta,muF)
      endif


      return
      end


c-------------------------------------------------------------------------

      subroutine Pkrn(nt,iproct,idj,ipos_i,xrandom2,xmin,xmax,
     -     muscale,muF,nl,res_reg,res_plus,res_delta)

c-----------------------------------------------------------------------
c...  P operator, based on the regularized Altarelli-Parisi splitting 
c...  kernels.
c...  The normalization does not include the overall color factor.
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

c... Input variables
      integer nt,iproct,idj,ipos_i,nl
      dimension iproct(mxpart)
      double precision muscale,muF,xrandom2,xmin,xmax

c... Output: 
c... res_reg = contribution from the regular part;
c... res_plus = contribution from 'plus' distribution;
c... res_delta = contribution from Dirac delta(1-x).
      double precision res_reg,res_plus,res_delta

c... Internal variables
      integer idit
      double precision x,xjac_ext

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)


      if(ipos_i.gt.2)then
         print*,'Poperator ERROR: the emitter can only be an'//
     -        'initial-state parton'
         STOP
      endif


c... Cutoff on input random numbers
      if(xrandom2.lt.1.d-10 .OR. abs(1.d0-xrandom2).lt.1.d-10)then
         res_reg=0.d0
         res_plus=0.d0
         res_delta=0.d0
         return
      endif

      idit = iproct(ipos_i)
      x = xmin+(xmax-xmin)*xrandom2
      xjac_ext = (xmax-xmin)

      if(idit.ne.35 .AND. idj.eq.35)then
c...  Pqq splitting kernel

         res_reg = -(1.d0+x)
         res_plus = 2.d0/(1.d0-x)
         res_delta = 3.d0/2.d0/xjac_ext

      elseif(idit.eq.35 .AND. idj.ne.35)then
c...  Pqg splitting kernel

         res_reg = (1.d0+(1.d0-x)**2)/x
         res_plus = 0.d0
         res_delta = 0.d0

      elseif(idit.ne.35 .AND. idj.ne.35)then
c...  Pgq splitting kernel

         res_reg = x**2+(1.d0-x)**2
         res_plus = 0.d0
         res_delta = 0.d0

      elseif(idit.eq.35 .AND. idj.eq.35)then
c...  Pgg splitting kernel

         res_reg = 2.d0*((1.d0-x)/x-1.d0+x*(1.d0-x))
         res_plus = 2.d0/(1.d0-x)
         res_delta = (11d0/6*CA-2d0/3*TF*nl)/CA/xjac_ext

      else
         print*,'Poperator ERROR: check input',ipos_i,idj
         STOP
      endif

c... Include jacobian and scale log contribution
      res_reg=res_reg*xjac_ext*log(muscale**2/muF**2)
      res_plus=res_plus*xjac_ext*log(muscale**2/muF**2)
      res_delta=res_delta*xjac_ext*log(muscale**2/muF**2)

      return
      end

c-------------------------------------------------------------------------

      subroutine NSxintdipole(nt,iproct,pmomt,idj,ipos_i,ipos_k,
     -     xrandom1,xrandom2,muscale,nl,xmin,xmax,
     -     res_reg,res_plus,res_delta,stable)

c-----------------------------------------------------------------------
c...  Contribution of the integrated dipole term Wik based on 
c...  initial-state radiation in the Nagy-Soper subtraction scheme.
c...  The normalization does not include the coupling alphas/2pi
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

c... Input variables
      integer nt,iproct,idj,ipos_i,ipos_k,nl
      dimension iproct(mxpart)
      double precision pmomt,muscale,xrandom1,xrandom2,xmin,xmax
      dimension pmomt(mxpart,1:5)

c... Output: 
c... res_reg = contribution from the regular part;
c... res_plus = contribution from 'plus' distribution;
c... res_delta = contribution from Dirac delta(1-x).
      double precision res_reg,res_plus,res_delta
      logical stable

c... Internal variables
      integer idit,idkt,mu,i,j,igo
      double precision Ktsq,Qsq,x,ctj,Ej,Etk,ptkz,ptky
      double precision Ea,Eta
      double precision pti,ptk,pta,pa
      dimension pti(4),ptk(4),pta(4),pa(4)
      double precision rmit,rmkt,masq,mksq
      double precision pxtot,pytot,pztot,Etot
      double precision pmomt_boost,z,ys,bcm,gcm
      dimension pmomt_boost(nt,1:4)
      double precision Pigreco,xjac,xjac_ext
      double precision padotQ,ptkdotQ,pjdotQ,padotptk,ptadotptk,padotpj
      double precision huge,sqrtL2mM2,BaFL,AaFM,BjL,AjM

      double precision dot4
      external dot4

      parameter(Pigreco=3.141592653589793d0)
      parameter(huge=1.d+12)

c...  color factors
      parameter(CA=3d0,CF=4d0/3,TF=0.5d0)

      stable=.true.

c... Cutoff on input random numbers
      if(xrandom1.lt.1.d-10 .OR. abs(1.d0-xrandom1).lt.1.d-10)then
         res_reg=0.d0
         res_plus=0.d0
         res_delta=0.d0
         return
      endif
      if(xrandom2.lt.1.d-10 .OR. abs(1.d0-xrandom2).lt.1.d-10)then
         res_reg=0.d0
         res_plus=0.d0
         res_delta=0.d0
         return
      endif

c---------------------------------------------------------------------------------
c... 1. Identify the dipole contribution.
c---------------------------------------------------------------------------------

c...  Ktsq = center-of-mass energy squared of the m-parton kinematics
      pxtot=0.d0
      pytot=0.d0
      pztot=0.d0
      Etot=0.d0
      do i=3,nt
         pxtot=pxtot+pmomt(i,1)
         pytot=pytot+pmomt(i,2)
         pztot=pztot+pmomt(i,3)
         Etot=Etot+pmomt(i,4)
      enddo
      Ktsq = Etot**2-pxtot**2-pytot**2-pztot**2

c... ipos_i = position of the emitter parton
c... ipos_k = position of the spectator parton
      idit = iproct(ipos_i)
      idkt = iproct(ipos_k)
      rmit = pmomt(ipos_i,5)
      rmkt = pmomt(ipos_k,5)

c---------------------------------------------------------------------------------
c... Check if the selected particles {i,k} are eligible to give a dipole 
c... contribution, given the input ID (idj) of the kinematically 
c... unconstrained particle.
c---------------------------------------------------------------------------------
      igo=0
      if(idj.eq.35)then         ! unconstrained parton = gluon
         if(idit.eq.35
     -        .OR. abs(idit).eq.3 .OR. abs(idit).eq.4
     -        .OR. abs(idit).eq.7 .OR. abs(idit).eq.8
     -        .OR. abs(idit).eq.11 .OR. abs(idit).eq.12 )then ! i = gluon or q(bar)
            igo=1
         else
            igo=0
         endif
         if(idkt.eq.35
     -        .OR. abs(idkt).eq.3 .OR. abs(idkt).eq.4
     -        .OR. abs(idkt).eq.7 .OR. abs(idkt).eq.8
     -        .OR. abs(idkt).eq.11 .OR. abs(idkt).eq.12 )then ! k = gluon or q(bar)
            igo=igo*1
         else
            igo=igo*0
         endif
      elseif( abs(idj).eq.3 .OR. abs(idj).eq.4 .OR.
     -        abs(idj).eq.7 .OR. abs(idj).eq.8 .OR.
     -        abs(idj).eq.11 .OR. abs(idj).eq.12 )then ! unconstr. parton = q(bar)
         if ( idit.eq.35 .OR.
     -        (ipos_i.gt.2 .AND. idit.eq.idj ) .OR.
     -        (ipos_i.le.2 .AND. idit.eq.-idj ) )then ! i = gluon or q(bar)
            igo=1
         else
            igo=0
         endif
         if( idkt.eq.35 .OR.
     -        (ipos_k.gt.2 .AND. idkt.eq.idj) .OR.
     -        (ipos_k.le.2 .AND. idkt.eq.-idj) )then ! k = gluon or q(bar)
            igo=igo*1
         else
            igo=igo*0
         endif
      else
         print*,'NSxintdipole ERROR: idj is not a parton.'
         STOP
      endif
      res_reg=0.d0
      res_plus=0.d0
      res_delta=0.d0
      if(igo.eq.0)then
         return
      endif

c---------------------------------------------------------------------------------
c... 2. Compute the dipole contribution corresponding to the selected pair {i,k} 
c... given j as the unconstrained particle, for the given input value of x.
c---------------------------------------------------------------------------------

c... If ipos_i .eq. ipos_k, the code evaluates the x-dependent part of the 
c... "diagonal" (aka "collinear") integrated subtraction terms (Wii-Wii_eikonal),
c... whereas if ipos_i .ne. ipos_k the x-dependent part of the "interference" 
c... terms (Wii_eikonal-Wik) is evaluated.
c... All terms proportional to Dirac delta, together with the pole parts, are
c... taken into account into the I-operator; the remaining x-dependent pole parts 
c... are cancelled by the Altarelli-Parisi collinear counterterms.
c... Note: we introduce here an arbitrary renormalization scale (muscale) because
c... intermediate results need it. The inclusion of the Altarelli-Parisi 
c... contributions will cancel the dependence on muscale in the end, leaving the 
c... final results dependent only upon the factorization scale muF.

      if(ipos_i.gt.2)then
         print*,'NSxintdipole ERROR: the emitter can only be an'//
     -        'initial-state parton'
         STOP
      endif

      if(ipos_i.eq.ipos_k)then
c---------------------------------------------------------------------------------
c...  x-dependent part of the COLLINEAR SUBTRACTION TERM (Wii-Wii_eikonal): 
c...  basic kinematics and formulae
c---------------------------------------------------------------------------------

c... Formulae

         if(idit.ne.35 .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c... q -> qg splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c               print*,"CALLING q->qg INITIAL"

c... Basic kinematics
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',rmit
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

            res_reg  =  
     -           -(((-1.d0 + x)*(-Log(muscale**2/Ktsq) + 
     -           2.d0*Log(1.d0 - x) - Log(x)))/x)
     -           /xjac
            res_plus = 
     -           0.d0
            res_delta = 
     -           -3.d0/2.d0*Log(muscale**2/Ktsq)
     -           /xjac/xjac_ext

#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_reg.ne.res_reg)then
               print*,'NSxintdipole: NaN found in q->qg INITIAL'
               print*,'res_reg',res_reg
               print*,'x',x
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               STOP
            endif
#endif

         endif

         if(idit.ne.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... q -> gq splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c               print*,"CALLING q->gq INITIAL"

c... Basic kinematics
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',rmit
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

            res_reg  =
     -           (-1.d0 + (6.d0 - 5.d0*x)*x + 
     -           (2.d0 + 4.d0*(-1.d0 + x)*x)*Log(1.d0 - x) + 
     -           (-1.d0 - 2.d0*(-1.d0 + x)*x)*Log(x) + 
     -           (-1.d0 - 2.d0*(-1.d0 + x)*x)*Log(muscale**2/Ktsq))/x
     -           /xjac
            res_plus = 
     -           0.d0
            res_delta = 
     -           0.d0

#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_reg.ne.res_reg)then
               print*,'NSxintdipole: NaN found in q->gq INITIAL'
               print*,'res_reg',res_reg
               print*,'x',x
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               STOP
            endif
#endif

         endif

         if(idit.eq.35 .AND. idj.ne.35)then
c---------------------------------------------------------------------------------
c... g -> qq splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = q(bar)
c---------------------------------------------------------------------------------
c               print*,"CALLING g->qq INITIAL"

c... Basic kinematics
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',rmit
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

            res_reg  =
     -           (-2.d0 + x*(2.d0 + x) - 
     -           (2.d0 + (-2.d0 + x)*x)*Log(muscale**2/Ktsq) + 
     -           2.d0*(2.d0 + (-2.d0 + x)*x)*Log(1.d0 - x) - 
     -           x**2*Log(x))/x**2
     -           /xjac
            res_plus = 
     -           0.d0
            res_delta = 
     -           0.d0


#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_reg.ne.res_reg)then
               print*,'NSxintdipole: NaN found in g->qq INITIAL'
               print*,'res_reg',res_reg
               print*,'x',x
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               STOP
            endif
#endif

         endif

         if(idit.eq.35 .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c... g -> gg splitting (INITIAL-INITIAL, MASSLESS)
c... Unconstrained parton = gluon
c---------------------------------------------------------------------------------
c               print*,"CALLING g->gg INITIAL"

c... Basic kinematics
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            if(rmit.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',rmit
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

            res_reg  =
     -           (2.d0*((-1.d0 + x)*(-1.d0 + x + x**3 + 
     -           (-1.d0 + x)*(1.d0 + x**2)*Log(muscale**2/Ktsq)) - 
     -           2.d0*(-1.d0 + x)**2*(1.d0 + x**2)*Log(1.d0 - x) + 
     -           (-2.d0 + x)*x**3*Log(x)))/((-1.d0 + x)*x**2)
     -           /xjac
            res_plus = 
     -           0.d0
            res_delta = 
     -           -(11d0/6*CA-2d0/3*TF*nl)/CA*Log(muscale**2/Ktsq)
     -           /xjac/xjac_ext

#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_reg.ne.res_reg)then
               print*,'NSxintdipole: NaN found in g->gg INITIAL'
               print*,'res_reg',res_reg
               print*,'x',x
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               STOP
            endif
#endif

         endif


      elseif(ipos_i.ne.ipos_k .AND. idj.eq.35)then
c---------------------------------------------------------------------------------
c...  x-dependent part of the INTERFERENCE TERM (Wii_eikonal-Wik): 
c...  basic kinematics and formulae
c---------------------------------------------------------------------------------

         if(ipos_k.le.2)then
c---------------------------------------------------------------------------------
c...  INITIAL-INITIAL state interference (x-dependent part)
c---------------------------------------------------------------------------------

c... Basic kinematics
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            if(rmit.ne.0.d0 .OR. rmkt.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',rmit,rmkt
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

            res_reg =
     -           0.d0
            res_plus  =
     -           (2.d0*(Log(muscale**2/Ktsq) - 2.d0*Log(1.d0 - x) + 
     -           Log(x)))/(-1.d0 + x)
     -           /xjac
            res_delta = 
     -           0.d0

#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_plus.ne.res_plus .OR. res_delta.ne.res_delta)then
               print*,'NSxintdipole: NaN found in INITIAL-INITIAL int.'
               print*,'res_plus',res_plus
               print*,'res_delta',res_delta
               print*,'x',x
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               STOP
            endif
#endif


         elseif(ipos_k.gt.2 .AND. ipos_k.le.nt)then
c---------------------------------------------------------------------------------
c...  INITIAL-FINAL state interference (x-dependent part)
c---------------------------------------------------------------------------------

c... Basic kinematics
            xjac = 2.d0         !... jacobian of the transf. xrandom1 -> ctj
            ctj = -1.d0 + 2.d0*xrandom1
            x = xmin + (xmax-xmin)*xrandom2
            xjac_ext = xmax-xmin !... jacobian of the transf. xrandom2 -> x
            masq = rmit**2
            mksq = rmkt**2
            if(masq.ne.0.d0)then
!... initial-state partons are assumed to be massless
               print*,'NSxintdipole: nonzero initial-state parton mass'
     -              //' detected',sqrt(masq)
               print*,'             Initial state partons must be'//
     -              ' massless.'
               STOP
            endif

c... Note: all formulae of the integrated dipoles are simpler if derived in
c... the frame where (pa + pb) = Q = (sqrt(Qsq),0,0,0) is at rest, where 
c... {pa,pb,Q} belong to the (m+1)-parton kinematics.
c... On the other hand, the momenta of the m-particle ('ptilde') kinematics 
c... provided by the Helac framework are expressed in the frame where 
c... (pta + ptb) = Kt = (sqrt(Ktsq),0,0,0) is at rest. Hence, in case of
c... non-zero initial state radiation (i.e. x<1) the two frames do not coincide.
c... In this case, in order to get the same frame of the formulae, the 'ptilde' 
c... momenta must be properly boosted. The invariance of the integrated
c... dipoles (Lorentz scalars) under boosts ensures that this operation leaves 
c... the final result unchanged.
c... The following rule has to be followed.
c... In the finite terms, one has to take either the original or the boosted
c... ptilde's on a case by case basis:
c...  - for all finite terms arising from soft-limit contributions 
c...    (i.e. x->1), one has to consider the original ptilde's;
c...  - for all finite terms arising from contributions away from the 
c...    soft limit (i.e. with x<1), one has to consider the boosted ptilde's.
c... NOTE: here, the dependence of the Dirac delta contribution upon x is justified
c... by the fact that we are adopting Monte Carlo integration. The analytically
c... integrated Dirac delta terms obviously would have no x dependence, and the two
c... reference frames would coincide for them (since x=1).

            Qsq = Ktsq/x        !...CM energy squared of the (m+1)-parton kinematics
            z = 1.d0/x
            ys = -0.5d0*log(x)  !...rapidity
c...  If emitter is parton b (coming from right), rapidity has opposite sign
            if(ipos_i.eq.2) ys=-ys
            bcm = (exp(ys)-exp(-ys))/(exp(ys)+exp(-ys)) !...beta
            gcm = (exp(ys)+exp(-ys))/2.d0 !...gamma
c... Boost ptilde momenta
            do i=1,nt
               pmomt_boost(i,4) = gcm*(pmomt(i,4)-bcm*pmomt(i,3))
               pmomt_boost(i,1) = pmomt(i,1)
               pmomt_boost(i,2) = pmomt(i,2)
               pmomt_boost(i,3) = pmomt(i,3)/gcm-bcm*pmomt_boost(i,4)
            enddo

            do mu=1,4
               pti(mu)=pmomt_boost(ipos_i,mu)
               ptk(mu)=pmomt_boost(ipos_k,mu)
            enddo
            do mu=1,4
               pta(mu)=pti(mu)
               pa(mu)=pta(mu)/x
            enddo

c... Boosted quantities (x<1)
            Ea = pa(4)
            Eta = pta(4)
            Etk = ptk(4)
            ptkz = ptk(3)
            if(ipos_i.eq.2) ptkz=-ptkz
            ptky = sqrt(Etk**2 -mksq - ptkz**2)

            Ej = -(Sqrt(Qsq)*(-1.d0 + x))/2.d0

c... Dot products
            padotQ = Ea*Sqrt(Qsq)
            ptkdotQ = Etk*Sqrt(Qsq)
            pjdotQ = Ej*Sqrt(Qsq)
            padotptk = dot4(pa,ptk)
            ptadotptk = dot4(pta,ptk)
            padotpj = Ej*Ea*(1.d0 - ctj)


c... DeltaW contribution to the finite term
            parAa  =  (2.d0*(padotQ*(-2.d0*ptkdotQ*(pjdotQ + padotQ*(
     -           -1.d0 + x)) + Ej*(Etk - ctj*ptkz)*(Qsq + 2.d0*padotQ*(
     -           -1.d0 + x))) + padotptk*(Qsq*(-pjdotQ + Qsq) + 
     -           4.d0*padotQ*(-pjdotQ + Qsq)*(-1.d0 + x) + 2.d0*
     -           padotQ**2*(-1.d0 + x)**2)) + padotpj*(-2.d0*ptkdotQ*(
     -           -2.d0*pjdotQ + Qsq + padotQ*(-1.d0 + x)) - Ej*(
     -           Etk - ctj*ptkz)*(Qsq + 2.d0*padotQ*(-1.d0 + x)) - 
     -           2.d0*padotptk*(2.d0*(-pjdotQ + Qsq) + 3.d0*padotQ*(
     -           -1.d0 + x))*(-1.d0 + x)) + 2.d0*padotpj**2*(ptkdotQ + 
     -           padotptk*(-1.d0 + x))*(-1.d0 + x))/((2.d0*(-pjdotQ + 
     -           Qsq + padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))*(Qsq +
     -           2.d0*padotQ*(-1.d0 + x)))
            parBa  =  -((Sqrt(1.d0 - ctj**2)*Ej*(padotpj - 2.d0*
     -           padotQ)*ptky)/(2.d0*(pjdotQ - Qsq - padotQ*(
     -           -1.d0 + x)) + padotpj*(-1.d0 + x)))
            parAj  =  (2.d0*(Ej*(Etk - ctj*ptkz)*Qsq**2 - 2.d0*pjdotQ*
     -           ptkdotQ*(pjdotQ + padotQ*(-1.d0 + x)) + (-2.d0*padotptk
     -           *pjdotQ**2 + padotptk*pjdotQ*Qsq + Ej*padotQ*(
     -           Etk - ctj*ptkz)*(3.d0*Qsq + 2.d0*padotQ*(-1.d0 + x)))*(
     -           -1.d0 + x)) - padotpj*(2.d0*ptkdotQ*(pjdotQ + Qsq + 
     -           2.d0*padotQ*(-1.d0 + x)) + padotptk*(2.d0*pjdotQ + Qsq
     -           + 2.d0*padotQ*(-1.d0 + x))*(-1.d0 + x))*(-1.d0 + x))/((
     -           2.d0*(-pjdotQ + Qsq + padotQ*(-1.d0 + x)) - padotpj*(
     -           -1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x)))
            parBj  =  (-2.d0*Sqrt(1.d0 - ctj**2)*Ej*ptky*(Qsq + padotQ*(
     -           -1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + padotQ*(-1.d0 + x))
     -           - padotpj*(-1.d0 + x))
            parAq  =  (Ej*(Etk - ctj*ptkz)*(-pjdotQ + 2.d0*Qsq + padotQ*
     -           (-1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x)) + ptkdotQ*
     -           (2.d0*Qsq**2 + 4.d0*pjdotQ*(pjdotQ - 2.d0*Qsq - padotQ*
     -           (-1.d0 + x)) - padotpj*(-2.d0*pjdotQ + 3.d0*Qsq + 2.d0*
     -           padotQ*(-1.d0 + x))*(-1.d0 + x)) + padotptk*(4.d0*
     -           pjdotQ**2 - 7.d0*pjdotQ*Qsq + 2.d0*Qsq**2 - 2.d0*
     -           padotpj*(-pjdotQ + Qsq)*(-1.d0 + x) - padotQ*(2.d0*
     -           pjdotQ + Qsq)*(-1.d0 + x) - 2.d0*padotQ**2*(
     -           -1.d0 + x)**2)*(-1.d0 + x))/((2.d0*(-pjdotQ + Qsq + 
     -           padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))*(Qsq + 
     -           2.d0*padotQ*(-1.d0 + x)))
            parBq  =  (Sqrt(1.d0 - ctj**2)*Ej*ptky*(pjdotQ - 2.d0*Qsq -
     -           padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + padotQ*(
     -           -1.d0 + x)) - padotpj*(-1.d0 + x))
            parF  =  -((-1.d0 + ctj)*padotQ*Qsq*(-1.d0 + x)**2)/(
     -           2.d0*padotpj*Pigreco)
            parL  =  padotQ*parAj + padotpj*parAq
            parM  =  padotQ*parBj + padotpj*parBq
            if(abs(parL).lt.huge .AND. abs(parM).lt.huge)then
               sqrtL2mM2 = Sqrt(parL**2 - parM**2)
               BaFL = parBa*parF*parL
               AaFM = parAa*parF*parM
            else
               sqrtL2mM2 = Sqrt( ((parL+parM)/sqrt(huge))*
     -              ((parL-parM)/sqrt(huge))*huge )
               BaFL = parBa*parF*(parL/huge)*huge
               AaFM = parAa*parF*(parM/huge)*huge
            endif
            res_reg = 
     -           0.d0
            res_plus  = 
     -           (-4.d0*Log(1.d0 - x) + 2.d0*(Log(muscale**2/Ktsq) + 
     -           Log(x)))/(-1.d0 + x)/xjac
     -           + (-2.d0 + ((parBa*parF + (-BaFL + AaFM)/sqrtL2mM2)*
     -           Pigreco)/parM)/((-1.d0 + ctj)*(-1.d0 + x))
            res_delta = 
     -           0.d0

            if(abs(parAj).lt.abs(parBj) .OR.
     -           abs(parAq).lt.abs(parBq) .OR.
     -           abs(parL).lt.abs(parM))then
               stable=.false.
            endif


c... mass term contribution (proportional to mksq)
            if(mksq.ne.0.d0)then
               parAMj  =  (2.d0*(Ej*(Etk - ctj*ptkz)*Qsq**2 - 2.d0*
     -              pjdotQ*ptkdotQ*(pjdotQ + padotQ*(-1.d0 + x)) + (
     -              -2.d0*padotptk*pjdotQ**2 + padotptk*pjdotQ*Qsq + 
     -              Ej*padotQ*(Etk - ctj*ptkz)*(3.d0*Qsq + 2.d0*padotQ*(
     -              -1.d0 + x)))*(-1.d0 + x)) - padotpj*(2.d0*ptkdotQ*(
     -              pjdotQ + Qsq + 2.d0*padotQ*(-1.d0 + x)) + padotptk*(
     -              2.d0*pjdotQ + Qsq + 2.d0*padotQ*(-1.d0 + x))*(
     -              -1.d0 + x))*(-1.d0 + x))/((2.d0*(-pjdotQ + Qsq + 
     -              padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))*(Qsq + 
     -              2.d0*padotQ*(-1.d0 + x)))
               parBMj  =  (-2.d0*Sqrt(1.d0 - ctj**2)*Ej*ptky*(Qsq + 
     -              padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + padotQ*(
     -              -1.d0 + x)) - padotpj*(-1.d0 + x))
               parAMq  =  (Ej*(Etk - ctj*ptkz)*(-pjdotQ + 2.d0*Qsq + 
     -              padotQ*(-1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x))
     -              + ptkdotQ*(2.d0*Qsq**2 + 4.d0*pjdotQ*(pjdotQ - 2.d0*
     -              Qsq - padotQ*(-1.d0 + x)) - padotpj*(-2.d0*pjdotQ + 
     -              3.d0*Qsq + 2.d0*padotQ*(-1.d0 + x))*(-1.d0 + x)) + 
     -              padotptk*(4.d0*pjdotQ**2 - 7.d0*pjdotQ*Qsq + 2.d0*
     -              Qsq**2 - 2.d0*padotpj*(-pjdotQ + Qsq)*(-1.d0 + x) -
     -              padotQ*(2.d0*pjdotQ + Qsq)*(-1.d0 + x) - 2.d0*
     -              padotQ**2*(-1.d0 + x)**2)*(-1.d0 + x))/((2.d0*(
     -              -pjdotQ + Qsq + padotQ*(-1.d0 + x)) - padotpj*(
     -              -1.d0 + x))*(Qsq + 2.d0*padotQ*(-1.d0 + x)))
               parBMq  =  (Sqrt(1.d0 - ctj**2)*Ej*ptky*(pjdotQ - 2.d0*
     -              Qsq - padotQ*(-1.d0 + x)))/(2.d0*(-pjdotQ + Qsq + 
     -              padotQ*(-1.d0 + x)) - padotpj*(-1.d0 + x))
               parFM  =  -(mksq*padotQ*Qsq*(-1.d0 + x)**2)/(
     -              4.d0*Pigreco)
               parLM  =  padotQ*parAMj + padotpj*parAMq
               parMM  =  padotQ*parBMj + padotpj*parBMq
               if(abs(parLM).lt.huge .AND. abs(parMM).lt.huge)then
                  sqrtL2mM2 = Sqrt(parLM**2 - parMM**2)
                  BjL = parBMj*parLM
                  AjM = parAMj*parMM
               else
                  sqrtL2mM2 = Sqrt( ((parLM+parMM)/sqrt(huge))*
     -                 ((parLM-parMM)/sqrt(huge))*huge )
                  BjL = (parBMj/sqrt(huge))*(parLM/sqrt(huge))*huge
                  AjM = (parAMj/sqrt(huge))*(parMM/sqrt(huge))*huge
               endif
               res_plus = res_plus + 
     -              (-((parFM*(parBMj/Sqrt(parAMj**2 - parBMj**2) - 
     -              parMM/sqrtL2mM2)*Pigreco)/
     -              ((BjL - AjM)*(-1.d0 + x))))
            endif

#if DEBUG1
c... Numerical exception (NaN) trap
            if(res_plus.ne.res_plus .OR. res_delta.ne.res_delta)then
               print*,'NSxintdipole: NaN found in INITIAL-FINAL int.'
               print*,'res_plus',res_plus
               print*,'res_delta',res_delta
               print*,'Ktsq',Ktsq
               print*,'muscale',muscale
               print*,'x(1)',xrandom1
               print*,'x(2)',xrandom2
               print*,'ctj',ctj
               print*,'Etk',Etk
               print*,'ptkz',ptkz
               print*,'ptky',ptky
               print*,'masq',masq
               print*,'mksq',mksq
               print*,'Ej',Ej
               print*,'padotQ',padotQ
               print*,'ptkdotQ',ptkdotQ
               print*,'pjdotQ',pjdotQ
               print*,'padotptk',padotptk
               print*,'padotpj',padotpj
               print*,'parAa',parAa
               print*,'parBa',parBa
               print*,'parAj',parAj
               print*,'parBj',parBj
               print*,'parAq',parAq
               print*,'parBq',parBq
               print*,'parF',parF
               print*,'parL',parL
               print*,'parM',parM
               if(mksq.ne.0.d0)then
                  print*,'parAMj',parAMj
                  print*,'parBMj',parBMj
                  print*,'parAMq',parAMq
                  print*,'parBMq',parBMq
                  print*,'parFM',parFM
                  print*,'parLM',parLM
                  print*,'parMM',parMM
               endif
               STOP
            endif
#endif

         else                   !...something went wrong with ipos_k

            print*,'NSxintdipole ERROR: check ipos_k',ipos_k
            STOP

         endif                  !if(ipos_k.le.2)then


      else                      !...something went wrong with ipos_i or ipos_k

         print*,'NSxintdipole ERROR: check ipos_i,ipos_k,idj',
     -        ipos_i,ipos_k,idj
         STOP

      endif                     !if(ipos_i.eq.ipos_k)then


c... Include jacobian contributions
      res_reg = res_reg*xjac*xjac_ext
      res_plus = res_plus*xjac*xjac_ext
      res_delta = res_delta*xjac*xjac_ext


      return
      end


c-----------------------------------------------------------------------

      double precision function dot4(pi,pj)

      double precision pi,pj
      dimension pi(4),pj(4)

      dot4=pi(4)*pj(4)-pi(1)*pj(1)-pi(2)*pj(2)-pi(3)*pj(3)

      return
      end
