      subroutine HISTOmpo(n,ifl,p,weight)

c-----------------------------------------------------------------------
c...  histogramming routine with m+1 partons (for real radiation)
c...  
c...  n=m+1
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

c...  jet momenta and coresponding flavors
c...  this information is filled in by the jet function FJmpo
      common/jets/pjet(mxpart,4),ifljet(mxpart),njet

      dimension ifl(mxpart),p(mxpart,5)

c...  uncomment for trivial histogramming
c      call ttbb(n,ifl,p,weight)
c      return

c...  pp -> ttbb + X
c      call bbbb(njet,ifljet,pjet,weight)

c...  pp -> ttbb + X
c      call ttbb(njet,ifljet,pjet,weight)

c...  pp -> ttjj + X
c      call ttjj(njet,ifljet,pjet,weight)

c...  pp -> wwbb + X
c      call wwbb2(njet,ifljet,pjet,weight)

c...  pp -> ttbb + X
c      call ttbb_tev(njet,ifljet,pjet,weight)
 
c... pp -> tttt +X     
c      call tttt(njet,ifljet,pjet,weight)

      return
      end

c-----------------------------------------------------------------------

      subroutine HISTOm(n,ifl,p,weight)

c-----------------------------------------------------------------------
c...  histogramming routine with m partons (for dipoles)
c...  
c...  n=m
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension ifl(mxpart),p(mxpart,5)

c       call bbbb(n,ifl,p,weight)
c      call ttbb(n,ifl,p,weight)
c      call ttjj(n,ifl,p,weight)
c      call wwbb2(n,ifl,p,weight)
c      call ttbb_tev(n,ifl,p,weight)
c      call tttt(n,ifl,p,weight)

      return
      end

c-----------------------------------------------------------------------

      subroutine ttbb(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> ttbb + X process
c...  invariant mass of two b-jets
c...  pT distribution of two b-jets
c...  rapidity distribution of two b-jet
c...  pT distribution of b- or antib-jets
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
      dimension p1(4),p2(4),p12(4)
      dimension p3(4),p4(4),p5(4)
      dimension plab(mxpart,4)

      call fromcmstolab(njet,pjet,plab)

      do i=3,njet
c...  b-jet
         if (ifljet(i).eq.12) then
            p1(1:4)=plab(i,1:4)
         endif
c...  anti-b-jet
         if (ifljet(i).eq.-12) then
            p2(1:4)=plab(i,1:4)
         endif
c...  additional jet (gluon or quark)
         if (ifljet(i).eq.35.or.ifljet(i).eq.3) then
            p3(1:4)=plab(i,1:4)
         endif
c...  t-jet
         if (ifljet(i).eq.11) then
            p4(1:4)=plab(i,1:4)
         endif
c...  anti-t-jet
         if (ifljet(i).eq.-11) then
            p5(1:4)=plab(i,1:4)
         endif         
      enddo

      p12(1:4)=p1(1:4)+p2(1:4)

c...  invariant mass distribution of two b-jets
      w12=dsqrt(scalar_product(p12,p12))
      call histo1(1,40,0d0,400d0,w12,w)

c...  pT distribution of two b-jets
      ptb=transverse(p12)
      call histo1(2,40,0d0,400d0,ptb,w)
     
c...  rapidity distribution of two b-jets
      etab=rapidity_jf(p12)
      call histo1(3,40,-2.5d0,2.5d0,etab,w)

c...  pT distribution of b- or antib-jets
      ptb1=transverse(p1)
      ptb2=transverse(p2)
      call histo1(4,40,0d0,400d0,ptb1,w)
      call histo1(4,40,0d0,400d0,ptb2,w)

c...  pT distribution of b-jet
      call histo1(5,40,0d0,400d0,ptb1,w)

c...  pT distribution of anti-b-jet
      call histo1(6,40,0d0,400d0,ptb2,w)

c...  pT distribution of additional jet
      ptjet=transverse(p3)
      call histo1(7,40,0d0,200d0,ptjet,w)

c...  pT distribution of t-jet
      ptt1=transverse(p4)
      call histo1(8,40,0d0,400d0,ptt1,w)

c...  pT distribution of anti-t jet
      ptt2=transverse(p5)
      call histo1(9,40,0d0,400d0,ptt2,w)

c...  rapidity distribution of b-jet
      etab1=rapidity_jf(p1)
      call histo1(10,40,-2.5d0,2.5d0,etab1,w)

c...  rapidity distribution of anti-b-jet
      etab2=rapidity_jf(p2)
      call histo1(11,40,-2.5d0,2.5d0,etab2,w)

c...  rapidity distribution of additional jet
      etajet=rapidity_jf(p3)
      call histo1(12,40,-2.5d0,2.5d0,etajet,w)

c...  rapidity distribution of t-jet
      etat1=rapidity_jf(p4)
      call histo1(13,40,-5d0,5d0,etat1,w)

c...  rapidity distribution of anti-t-jet
      etat2=rapidity_jf(p5)
      call histo1(14,40,-5d0,5d0,etat2,w)

      return
      end

c-----------------------------------------------------------------------

      subroutine ttjj(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> ttjj + X process
c...  13 different distributions
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_masses.h'
      include 'parameter_pp.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
c...  maximal number of light jets njet = 3
      dimension ptjet(3),ksort(3),pjj(4)
      dimension plab(mxpart,4),ptable(3,4)
      dimension ptop(4),ptopbar(4),ptt(4)
      double precision mTtop1,mTtop2

      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)

      iput=0
      ptable(1:3,1:4)=0d0
      ptjet(1:3)=0d0
  
      do i=3,njet
c...  (anti)top-jets and light-jets
         if (iabs(ifljet(i)).ne.11) then
            iput=iput+1
            ptable(iput,1:4)=plab(i,1:4)
            ptjet(iput)=transverse(plab(i,1:4))
         elseif(ifljet(i).eq.11) then
            ptop(1:4)=plab(i,1:4)
         elseif(ifljet(i).eq.-11) then
            ptopbar(1:4)=plab(i,1:4)
         endif
      enddo
      
c...  ordering jets according to pT, the hardest, the second hardest, etc. 
      call SOR(ptjet,iput,ksort,2)

c...  test
c      print*,(ptjet(i),i=1,iput)
c      print*,(ksort(i),i=1,iput)
c      print*,(ptjet(ksort(i)),i=1,iput)

c...  --------------------------------------------------------
c...  (antit)top quark kinematic
c...  --------------------------------------------------------
c...  pT distribution of top 

      
      pTtop1=transverse(ptop)
      if(ipp.eq.1) call histo1(1,40,0d0,400d0,pTtop1,w*0.5d0)
      if(ipp.eq.2) call histo1(1,40,0d0,500d0,pTtop1,w*0.5d0)
c...  pT distribution of anti-top 
      pTtop2=transverse(ptopbar)
      if(ipp.eq.1) call histo1(1,40,0d0,400d0,pTtop2,w*0.5d0)
      if(ipp.eq.2) call histo1(1,40,0d0,500d0,pTtop2,w*0.5d0)
c...  rapidity distribution of top
      etatop1=rapidity_jf(ptop)
      if(ipp.eq.1) call histo1(2,40,-1.5d0,1.5d0,etatop1,w)
      if(ipp.eq.2) call histo1(2,40,-2.5d0,2.5d0,etatop1,w)
c...  bigger range
      if(ipp.eq.1) call histo1(4,40,-2.5d0,2.5d0,etatop1,w)
      if(ipp.eq.2) call histo1(4,40,-4.5d0,4.5d0,etatop1,w)
c...  rapidity distribution of anti-top
      etatop2=rapidity_jf(ptopbar)
      if(ipp.eq.1) call histo1(3,40,-1.5d0,1.5d0,etatop2,w)
      if(ipp.eq.2) call histo1(3,40,-2.5d0,2.5d0,etatop2,w)
c...  bigger range
      if(ipp.eq.1) call histo1(5,40,-2.5d0,2.5d0,etatop2,w)
      if(ipp.eq.2) call histo1(5,40,-4.5d0,4.5d0,etatop2,w)
 
c...  --------------------------------------------------------
c...  tT pair kinematic
c...  --------------------------------------------------------
      ptt(1:4)=ptop(1:4)+ptopbar(1:4)
c...  pT distribution of tT pair
      pTtt=transverse(ptt)
      if(ipp.eq.1) call histo1(6,40,0d0,250d0,pTtt,w)
      if(ipp.eq.2) call histo1(6,40,0d0,500d0,pTtt,w)
c...  rapidity distribution of tT pair
      etatt=rapidity_jf(ptt)
      if(ipp.eq.1) call histo1(7,40,-1d0,1d0,etatt,w)
      if(ipp.eq.2) call histo1(7,40,-2d0,2d0,etatt,w)
c...  invariant mass distribution of tT system
      wtt=dsqrt(scalar_product(ptt,ptt))
      if(ipp.eq.1) call histo1(8,40,300d0,800d0,wtt,w)
      if(ipp.eq.2) call histo1(8,40,300d0,1000d0,wtt,w)

c...  --------------------------------------------------------
c...  the 1st hardest jet kinematic
c...  --------------------------------------------------------
c...  pT distribution of the 1st hardest jet
      pTj1=transverse(ptable(ksort(1),1:4))
      if(ipp.eq.1) call histo1(9,40,0d0,200d0,pTj1,w)
      if(ipp.eq.2) call histo1(9,40,0d0,500d0,pTj1,w)
c...  rapidity distribution of the 1st hardest jet
      etaj1=rapidity_jf(ptable(ksort(1),1:4))
      if(ipp.eq.1) call histo1(10,40,-2d0,2d0,etaj1,w)
      if(ipp.eq.2) call histo1(10,40,-2.5d0,2.5d0,etaj1,w)

c...  --------------------------------------------------------
c...  the 2nd hardest jet kinematic
c...  --------------------------------------------------------
c...  pT distribution of the 2nd hardest jet
      pTj2=transverse(ptable(ksort(2),1:4))
      if(ipp.eq.1) call histo1(11,40,0d0,80d0,pTj2,w)
      if(ipp.eq.2) call histo1(11,40,0d0,150d0,pTj2,w)
c...  rapidity distribution of the 2nd hardest jet
      etaj2=rapidity_jf(ptable(ksort(2),1:4))
      if(ipp.eq.1) call histo1(12,40,-2d0,2d0,etaj2,w)
      if(ipp.eq.2) call histo1(12,40,-2.5d0,2.5d0,etaj2,w)

c...  --------------------------------------------------------
c...  jj pair kinematic
c...  --------------------------------------------------------
c...  sum of the hardest and the second hardest jet
      pjj(1:4)=ptable(ksort(1),1:4)+ptable(ksort(2),1:4)
c...  invariant mass distribution of two light-jets
      wjj=dsqrt(scalar_product(pjj,pjj))
      if(ipp.eq.1) call histo1(13,40,0d0,250d0,wjj,w)
      if(ipp.eq.2) call histo1(13,40,0d0,800d0,wjj,w)
c...  delta_R separation 
c...  between the hardest and the second hardest light jet
      d1=etaj1-etaj2
      d2=ph4(ptable(ksort(1),1),ptable(ksort(1),2),ptable(ksort(1),3))-
     &   ph4(ptable(ksort(2),1),ptable(ksort(2),2),ptable(ksort(2),3))
      d2=min(dabs(d2),2*pi-dabs(d2))
      if(d2/pi.gt.1.d0)print*,d2/pi
      drjj=sqrt(d1**2+d2**2)
      if(ipp.eq.1) call histo1(14,40,0d0,5d0,drjj,w)
      if(ipp.eq.2) call histo1(14,40,0d0,6d0,drjj,w)

c...  --------------------------------------------------------
c...  full system kinematics
c...  --------------------------------------------------------
c...  total transverse energy of the system

      mTtop1 = dsqrt(parmas(11)**2+pTtop1**2)
      mTtop2 = dsqrt(parmas(11)**2+pTtop2**2)
      Ht = pTj1 + pTj2 + mTtop1 + mTtop2
      if(ipp.eq.1) call histo1(15,40,300d0,1000d0,Ht,w)
      if(ipp.eq.2) call histo1(15,40,300d0,1400d0,Ht,w)

      return
      end

c-----------------------------------------------------------------------

      subroutine wwbb(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> ttbb + X process
c...  invariant mass of two b-jets
c...  pT distribution of two b-jets
c...  rapidity distribution of two b-jet
c...  pT distribution of b- or antib-jets
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'parameter_pp.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
c...  b-quarks
      dimension p1(4),p2(4)
c...  additional jet (only from subtracted real emission)
      dimension p3(4)
c...  electron, muon and neutrinous 
      dimension p4(4),p5(4),p6(4),p7(4)
c...  bb pair, tt pair
      dimension pbb(4),ptt(4),pw1(4),pw2(4),pmiss(4)
c...  top and anti-top 
      dimension pt1(4),pt2(4)
      dimension plab(mxpart,4)

      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)
      
      do i=3,njet
c...  b-jet
         if (ifljet(i).eq.12) then
            p1(1:4)=plab(i,1:4)
         endif
c...  anti-b-jet
         if (ifljet(i).eq.-12) then
            p2(1:4)=plab(i,1:4)
         endif
c...  electron
         if (iabs(ifljet(i)).eq.2) then
            p4(1:4)=plab(i,1:4)
         endif
c...  muon
         if (iabs(ifljet(i)).eq.6) then
            p5(1:4)=plab(i,1:4)
         endif         
c...  electron neutrino
         if (iabs(ifljet(i)).eq.1) then
            p6(1:4)=plab(i,1:4)
         endif
c...  muon neutrino
         if (iabs(ifljet(i)).eq.5) then
            p7(1:4)=plab(i,1:4)
         endif     
      enddo

c...  additional jet (gluon or quark)
      if(njet.eq.9)then       
         p3(1:4)=plab(njet,1:4)
      endif

      ptj=0d0
      etaj=0d0

c....................................................

c...  b-quark pair kinematics

c....................................................

      pbb(1:4)=p1(1:4)+p2(1:4)
      
c...  invariant mass distribution of two b-jets
      wbb=dsqrt(scalar_product(pbb,pbb))
      if(ipp.eq.1) call histo1(1,40,0d0,400d0,wbb,w)
      if(ipp.eq.2) call histo1(1,40,0d0,550d0,wbb,w)

c...  pT distribution of two b-jets
      ptbb=transverse(pbb)
      if(ipp.eq.1) call histo1(2,40,0d0,200d0,ptbb,w)
      if(ipp.eq.2) call histo1(2,40,0d0,250d0,ptbb,w)
     
c...  rapidity distribution of two b-jets
      etabb=rapidity_jf(pbb)
      call histo1(3,40,-3d0,3d0,etabb,w)

c....................................................

c...  b-quark kinematics 

c....................................................

c...  pT distribution of b-jet
      ptb1=transverse(p1)
      ptb2=transverse(p2)
      if(ipp.eq.1)then
         call histo1(4,40,0d0,200d0,ptb1,w*0.5d0)
         call histo1(4,40,0d0,200d0,ptb2,w*0.5d0)
      endif
      if(ipp.eq.2)then
         call histo1(4,40,0d0,300d0,ptb1,w*0.5d0)
         call histo1(4,40,0d0,300d0,ptb2,w*0.5d0)
      endif

c...  rapidity distribution of b-jet
      etab1=rapidity_jf(p1)
      etab2=rapidity_jf(p2)
      call histo1(5,40,-3d0,3d0,etab1,w*0.5d0)
      call histo1(5,40,-3d0,3d0,etab2,w*0.5d0)

c...  delta_R separation between b jets
      d1=etab1-etab2
      d2=ph4(p1(1),p1(2),p1(3))-
     &   ph4(p2(1),p2(2),p2(3))
      d2=min(dabs(d2),2*pi-dabs(d2))
      if(d2/pi.gt.1.d0)print*,d2/pi
      drbb=sqrt(d1**2+d2**2)
      call histo1(6,40,0d0,5d0,drbb,w)

c....................................................

c...  t-anti-t system and single top kinematics

c....................................................

      ptt(1:4)=p1(1:4)+p2(1:4)+p4(1:4)+p5(1:4)+p6(1:4)+p7(1:4)
      pt1(1:4)=p1(1:4)+p4(1:4)+p6(1:4)
      pt2(1:4)=p2(1:4)+p5(1:4)+p7(1:4)

c...  invariant mass distribution of t anti-t system
      wtt=dsqrt(scalar_product(ptt,ptt))
      if(ipp.eq.1) call histo1(7,40,300d0,800d0,wtt,w)
      if(ipp.eq.2) call histo1(7,40,300d0,1000d0,wtt,w)

c...  rapidity distribution t anti-t system
      etatt=rapidity_jf(ptt)
      call histo1(8,40,-2.5d0,2.5d0,etatt,w)

c...  pT distribution of top
      ptt1=transverse(pt1)
      ptt2=transverse(pt2)
      if(ipp.eq.1)then
         call histo1(9,40,0d0,350d0,ptt1,w*0.5d0)
         call histo1(9,40,0d0,350d0,ptt2,w*0.5d0)
      endif
      if(ipp.eq.2)then
         call histo1(9,40,0d0,450d0,ptt1,w*0.5d0)
         call histo1(9,40,0d0,450d0,ptt2,w*0.5d0)
      endif

c...  rapidity distribution of top
      etat1=rapidity_jf(pt1)
      etat2=rapidity_jf(pt2)
      call histo1(10,40,-2.5d0,2.5d0,etat1,w*0.5d0)
      call histo1(10,40,-2.5d0,2.5d0,etat2,w*0.5d0)

c....................................................

c...  W kinematics

c....................................................

      pw1(1:4)=p4(1:4)+p6(1:4)
      pw2(1:4)=p5(1:4)+p7(1:4)

c...  pT distribution of W
      ptw1=transverse(pw1)
      ptw2=transverse(pw2)
      if(ipp.eq.1)then
         call histo1(11,40,0d0,250d0,ptw1,w*0.5d0)
         call histo1(11,40,0d0,250d0,ptw2,w*0.5d0)
      endif
      if(ipp.eq.2)then
         call histo1(11,40,0d0,350d0,ptw1,w*0.5d0)
         call histo1(11,40,0d0,350d0,ptw2,w*0.5d0)
      endif

c...  rapidity distribution of W
      etaw1=rapidity_jf(pw1)
      etaw2=rapidity_jf(pw2)
      call histo1(12,40,-3d0,3d0,etaw1,w*0.5d0)
      call histo1(12,40,-3d0,3d0,etaw2,w*0.5d0)

c....................................................

c...  charged lepton kinematics 

c....................................................

c...  pT distribution of charged lepton
      ptele=transverse(p4)
      ptmu =transverse(p5)
      if(ipp.eq.1)then
         call histo1(13,40,0d0,200d0,ptele,w*0.5d0)
         call histo1(13,40,0d0,200d0,ptmu,w*0.5d0)
      endif
      if(ipp.eq.2)then
         call histo1(13,40,0d0,250d0,ptele,w*0.5d0)
         call histo1(13,40,0d0,250d0,ptmu,w*0.5d0)
      endif

c...  rapidity distribution of charged lepton
      etaele=rapidity_jf(p4)
      etamu =rapidity_jf(p5)
      call histo1(14,40,-2.5d0,2.5d0,etaele,w*0.5d0)
      call histo1(14,40,-2.5d0,2.5d0,etamu,w*0.5d0)

c...  delta_R separation between charged leptons
      d1=etaele-etamu
      d2=ph4(p4(1),p4(2),p4(3))-
     &   ph4(p5(1),p5(2),p5(3))
      d2=min(dabs(d2),2*pi-dabs(d2))
      if(d2/pi.gt.1.d0)print*,d2/pi
      drelemu=sqrt(d1**2+d2**2)
      call histo1(15,40,0d0,5d0,drelemu,w)

c....................................................

c...  neutrino kinematics

c....................................................

      pmiss(1:4)=p6(1:4)+p7(1:4)
c...  pT miss distribution
      ptmiss=transverse(pmiss)
      if(ipp.eq.1) call histo1(16,40,0d0,200d0,ptmiss,w)
      if(ipp.eq.2) call histo1(16,40,0d0,250d0,ptmiss,w)

c....................................................

c...  full system kinematics

c....................................................

c...  total transverse energy of the system
      Ht=0d0
      Ht= ptb1+ptb2+ptele+ptmu+ptmiss
      if(njet.eq.9)then
         ptj=transverse(p3)
         Ht=Ht+ptj
      endif
      
      if(ipp.eq.1) call histo1(17,40,0d0,600d0,Ht,w)
      if(ipp.eq.2) call histo1(17,40,0d0,800d0,Ht,w)

c...  total transverse energy of the system
c...  without additional jet
      Ht=0d0
      Ht= ptb1+ptb2+ptele+ptmu+ptmiss
      if(ipp.eq.1) call histo1(18,40,0d0,600d0,Ht,w)
      if(ipp.eq.2) call histo1(18,40,0d0,800d0,Ht,w)

c....................................................                  

c...  additional jet kinematics                                        
  
c....................................................                    

      if(njet.eq.9)then

c...  pT distribution of jet                                       
         ptj=transverse(p3)
         if(ipp.eq.1) call histo1(19,40,0d0,100d0,ptj,w)
         if(ipp.eq.2) call histo1(19,40,0d0,150d0,ptj,w)

c...  rapidity distribution of jet         
         etaj=rapidity_jf(p3)
         call histo1(20,40,-4d0,4d0,etaj,w)
      endif

      return
      end

      subroutine wwbb2(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> ttbb + X process
c...  invariant mass of two b-jets
c...  pT distribution of two b-jets
c...  rapidity distribution of two b-jet
c...  pT distribution of b- or antib-jets
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'parameter_pp.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
c...  b-quarks
      dimension p1(4),p2(4)
c...  electron, muon and neutrinous 
      dimension p4(4),p5(4),p6(4),p7(4)
c...  top and anti-top 
      dimension pt1(4),pt2(4)
      dimension plab(mxpart,4)

      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)
      
      do i=3,njet
c...  b-jet
         if (ifljet(i).eq.12) then
            p1(1:4)=plab(i,1:4)
         endif
c...  anti-b-jet
         if (ifljet(i).eq.-12) then
            p2(1:4)=plab(i,1:4)
         endif
c...  electron
         if (iabs(ifljet(i)).eq.2) then
            p4(1:4)=plab(i,1:4)
         endif
c...  muon
         if (iabs(ifljet(i)).eq.6) then
            p5(1:4)=plab(i,1:4)
         endif         
c...  electron neutrino
         if (iabs(ifljet(i)).eq.1) then
            p6(1:4)=plab(i,1:4)
         endif
c...  muon neutrino
         if (iabs(ifljet(i)).eq.5) then
            p7(1:4)=plab(i,1:4)
         endif     
      enddo

c...  top 
      pt1(1:4)=p1(1:4)+p4(1:4)+p6(1:4)
c...  anti-top
      pt2(1:4)=p2(1:4)+p5(1:4)+p7(1:4)

c...  invariant mass distribution of t-jets
      wt1=dsqrt(scalar_product(pt1,pt1))
      wt2=dsqrt(scalar_product(pt2,pt2))
      call histo1(1,40,120d0,220d0,wt1,w*0.5d0)      
      call histo1(1,40,120d0,220d0,wt2,w*0.5d0)

c...  rapidity distribution of b-jet
c...  bottom
      etab1=rapidity_jf(p1)
      call histo1(2,40,-3d0,3d0,etab1,w)
c...  anti-bottom
      etab2=rapidity_jf(p2)
      call histo1(3,40,-3d0,3d0,etab2,w)

c...  rapidity distribution of top
c...  top
      etat1=rapidity_jf(pt1)
      call histo1(4,40,-2.5d0,2.5d0,etat1,w)
c...  anti-top
      etat2=rapidity_jf(pt2)
      call histo1(5,40,-2.5d0,2.5d0,etat2,w)

c...  rapidity distribution of charged lepton
c...  e+
      etaele=rapidity_jf(p4)
      call histo1(6,40,-2.5d0,2.5d0,etaele,w)
c...  mu-
      etamu =rapidity_jf(p5)
      call histo1(7,40,-2.5d0,2.5d0,etamu,w)

      return
      end


c-----------------------------------------------------------------------

      subroutine ttbb_tev(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> ttbb + X process at the TeVatron
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
      dimension p1(4),p2(4),p12(4)
      dimension p3(4),p4(4)
      dimension plab(mxpart,4)

      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)


      do i=3,njet
c...  b-jet
         if (ifljet(i).eq.12) then
            p1(1:4)=plab(i,1:4)
         endif
c...  anti-b-jet
         if (ifljet(i).eq.-12) then
            p2(1:4)=plab(i,1:4)
         endif
c...  t-jet
         if (ifljet(i).eq.11) then
            p3(1:4)=plab(i,1:4)
         endif
c...  anti-t-jet
         if (ifljet(i).eq.-11) then
            p4(1:4)=plab(i,1:4)
         endif         
      enddo

c...  -----------------------------------------
c...  b-quark pair kinematics
c...  -----------------------------------------

      p12(1:4)=p1(1:4)+p2(1:4)

c...  invariant mass distribution of two b-jets
      w12=dsqrt(scalar_product(p12,p12))
      call histo1(1,40,0d0,200d0,w12,w)

c...  pT distribution of two b-jets
      ptb=transverse(p12)
      call histo1(2,40,0d0,200d0,ptb,w)
     
c...  rapidity distribution of two b-jets
      etab=rapidity_jf(p12)
      call histo1(3,40,-2.5d0,2.5d0,etab,w)

c...  -----------------------------------------
c...  b-quark kinematics
c...  -----------------------------------------

c...  pT distribution of b-jets
      ptb1=transverse(p1)
      call histo1(4,40,0d0,150d0,ptb1,w*0.5d0)
c...  pT distribution of antib-jets
      ptb2=transverse(p2)
      call histo1(4,40,0d0,150d0,ptb2,w*0.5d0)

c...  rapidity distribution of  b-jets
      etab1=rapidity_jf(p1)
      call histo1(5,40,-2.5d0,2.5d0,etab1,w*0.5d0)
c...  rapidity distribution of antib-jets
      etab2=rapidity_jf(p2)
      call histo1(5,40,-2.5d0,2.5d0,etab2,w*0.5d0)

c...  delta_R separation between b jets
      d1=etab1-etab2
      d2=ph4(p1(1),p1(2),p1(3))-ph4(p2(1),p2(2),p2(3))
      d2=min(dabs(d2),2*pi-dabs(d2))
      if(d2/pi.gt.1.d0)print*,d2/pi
      drbb=sqrt(d1**2+d2**2)
      call histo1(6,40,0d0,4d0,drbb,w)

c...  -----------------------------------------
c...  t-quark kinematics
c...  -----------------------------------------

c...  pT distribution of t-jet
      ptt1=transverse(p3)
      call histo1(7,40,0d0,400d0,ptt1,w*0.5d0)
c...  pT distribution of antit-jet
      ptt2=transverse(p4)
      call histo1(7,40,0d0,400d0,ptt2,w*0.5d0)

c...  rapidity distribution of t-jet
      etat1=rapidity_jf(p3)
      call histo1(8,40,-1.5d0,1.5d0,etat1,w)
c...  rapidity distribution of antit-jet
      etat2=rapidity_jf(p4)
      call histo1(9,40,-1.5d0,1.5d0,etat2,w)

c...  -----------------------------------------
c...  t-quark pair kinematics
c...  -----------------------------------------

      p12(1:4)=p3(1:4)+p4(1:4)

c...  invariant mass distribution of two t-jets
      w12=dsqrt(scalar_product(p12,p12))
      call histo1(10,40,300d0,800d0,w12,w)

c...  pT distribution of two t-jets
      pttt=transverse(p12)
      call histo1(11,40,0d0,200d0,pttt,w)
     
c...  rapidity distribution of two t-jets
      etatt=rapidity_jf(p12)
      call histo1(12,40,-1d0,1d0,etatt,w)

      return
      end

c-----------------------------------------------------------------------

      subroutine tttt(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> tttt + X process at the TeVatron
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_masses.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
      dimension plab(mxpart,4)
      dimension ptable(4,4),ptjet(4),ksort(4),iflavour(4)
      dimension p12(4)
      dimension p5(4)
 
      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)

      iput=0
      ptable(1:4,1:4)=0d0
      ptjet(1:4)=0d0
      iflavour(1:4)=0
      ksort(1:4)=0d0

      do i=3,njet
         if (iabs(ifljet(i)).eq.11) then
            iput=iput+1
            iflavour(iput)=ifljet(i)
            ptable(iput,1:4)=plab(i,1:4)
            ptjet(iput)=transverse(plab(i,1:4))
         endif
      enddo
   
c...  ordering jets according to pT, the hardest, the second hardest, etc. 
      call SOR(ptjet,iput,ksort,2)

c...  test of the ordering
c      print*,'TEST'
c      print*,(ptjet(i),i=1,iput)
c      print*,(ksort(i),i=1,iput)
c      print*,(ptjet(ksort(i)),i=1,iput)
c      print*,(iflavour(ksort(i)),i=1,iput)


c...  -----------------------------------------    
c...  top anti-top pair kinematics         
c...  -----------------------------------------        

      do i=1,iput
         k=i+1
         do j=k,iput
            if(iflavour(i).eq.iflavour(j))cycle
c            print*,i,j,iflavour(i),iflavour(j)
            p12(1:4)=ptable(i,1:4)+ptable(j,1:4)
c...  invariant mass distribution, transverse momentum distribution, 
c...  rapidity distribution  of top anti-top pairs, 
c...  all possible combinations, 4 in total
            w12=dsqrt(scalar_product(p12,p12))
            call histo1(1,40,340d0,1500d0,w12,w*0.25d0)
            pt12=transverse(p12)
            call histo1(3,40,0d0,1000d0,pt12,w*0.25d0)
            y12=rapidity_jf(p12)
            call histo1(5,40,-2.5d0,2.5d0,y12,w*0.25d0)
         enddo
      enddo

      do i=1,iput
         k=i+1
         do j=k,iput
            if(iflavour(ksort(i)).eq.iflavour(ksort(j)))cycle
c            print*,ptjet(ksort(i)),ptjet(ksort(j))
c            print*,iflavour(ksort(i)),iflavour(ksort(j))
c...  invariant mass distribution, transverse momentum distribution 
c...  and rapidity distribution of top anti-top pairs; 
c...  top and anti-top with the highest pT
            p12(1:4)=ptable(ksort(i),1:4)+ptable(ksort(j),1:4)
            w12=dsqrt(scalar_product(p12,p12))
            call histo1(2,40,340d0,1500d0,w12,w)
            pt12=transverse(p12)
            call histo1(4,40,0d0,1000d0,pt12,w)
            y12=rapidity_jf(p12)
            call histo1(6,40,-2.5d0,2.5d0,y12,w)
            goto 111
         enddo
      enddo
 111  continue
 
c...  -----------------------------------------
c...  full system kinematics                                                   
c...  -----------------------------------------                                

c...  total transverse energy of the system                                    
      pmT1 = dsqrt(parmas(11)**2+ptjet(1)**2)
      pmT2 = dsqrt(parmas(11)**2+ptjet(2)**2)
      pmT3 = dsqrt(parmas(11)**2+ptjet(3)**2)
      pmT4 = dsqrt(parmas(11)**2+ptjet(4)**2)
      hT=pmT1+pmT2+pmT3+pmT4
      call histo1(7,40,690d0,2500d0,hT,w)

c...  invariant mass of the top anti-top top anti-top system
      p12(1:4)=ptable(1,1:4)+ptable(2,1:4)+
     .         ptable(3,1:4)+ptable(4,1:4)
      w12=dsqrt(scalar_product(p12,p12))
      call histo1(8,40,690d0,4000d0,w12,w)

c...  -----------------------------------------
c...  single top kinematics                                                   
c...  -----------------------------------------    

c...  transverse momentum distribution of the top quark
      call histo1(9,40,0d0,800d0,ptjet(ksort(1)),w)
      call histo1(10,40,0d0,800d0,ptjet(ksort(2)),w)
      call histo1(11,40,0d0,800d0,ptjet(ksort(3)),w)
      call histo1(12,40,0d0,800d0,ptjet(ksort(4)),w)

c...  rapidity distribution of the top quark
      p12(1:4)=ptable(ksort(1),1:4)
      y1=rapidity_jf(p12)
      call histo1(13,40,-3d0,3d0,y1,w)
      p12(1:4)=ptable(ksort(2),1:4)
      y1=rapidity_jf(p12)
      call histo1(14,40,-3d0,3d0,y1,w)
      p12(1:4)=ptable(ksort(3),1:4)
      y1=rapidity_jf(p12)
      call histo1(15,40,-3d0,3d0,y1,w)
      p12(1:4)=ptable(ksort(4),1:4)
      y1=rapidity_jf(p12)
      call histo1(16,40,-3d0,3d0,y1,w)

c...  -----------------------------------------  

      return
      end


      subroutine bbbb(njet,ifljet,pjet,w)

c-----------------------------------------------------------------------
c...  pp -> bbbb + X process at the TeVatron
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_masses.h'

      dimension pjet(mxpart,4),ifljet(mxpart)
      dimension plab(mxpart,4)
      dimension ptable(4,4),ptjet(4),ksort(4),iflavour(4)
      dimension p12(4)
      dimension p5(4)
 
      pi=dacos(-dnou(1))
      call fromcmstolab(njet,pjet,plab)

      iput=0
      ptable(1:4,1:4)=0d0
      ptjet(1:4)=0d0
      iflavour(1:4)=0
      ksort(1:4)=0d0

      do i=3,njet
         if (iabs(ifljet(i)).eq.12) then
            iput=iput+1
            iflavour(iput)=ifljet(i)
            ptable(iput,1:4)=plab(i,1:4)
            ptjet(iput)=transverse(plab(i,1:4))
         endif
      enddo
   
c...  ordering jets according to pT, the hardest, the second hardest, etc. 
      call SOR(ptjet,iput,ksort,2)

c...  -----------------------------------------    
c...  top anti-top pair kinematics         
c...  -----------------------------------------        

      do i=1,iput
         k=i+1
         do j=k,iput
            if(iflavour(i).eq.iflavour(j))cycle
c            print*,i,j,iflavour(i),iflavour(j)
            p12(1:4)=ptable(i,1:4)+ptable(j,1:4)
c...  invariant mass distribution, transverse momentum distribution, 
c...  rapidity distribution  of bb pairs, 
c...  all possible combinations, 4 in total
            w12=dsqrt(scalar_product(p12,p12))
            call histo1(1,40,0d0,500d0,w12,w*0.25d0)
            pt12=transverse(p12)
            call histo1(3,40,0d0,250d0,pt12,w*0.25d0)
            y12=rapidity_jf(p12)
            call histo1(5,40,-2.5d0,2.5d0,y12,w*0.25d0)
         enddo
      enddo

      do i=1,iput
         k=i+1
         do j=k,iput
            if(iflavour(ksort(i)).eq.iflavour(ksort(j)))cycle
c            print*,ptjet(ksort(i)),ptjet(ksort(j))
c            print*,iflavour(ksort(i)),iflavour(ksort(j))
c...  invariant mass distribution, transverse momentum distribution 
c...  and rapidity distribution of  the bb pairs; 
c...  b-jet and anti-b-jet with the highest pT
            p12(1:4)=ptable(ksort(i),1:4)+ptable(ksort(j),1:4)
            w12=dsqrt(scalar_product(p12,p12))
            call histo1(2,40,0d0,500d0,w12,w)
            pt12=transverse(p12)
            call histo1(4,40,0d0,250d0,pt12,w)
            y12=rapidity_jf(p12)
            call histo1(6,40,-2.5d0,2.5d0,y12,w)
            goto 111
         enddo
      enddo
 111  continue
 
c...  -----------------------------------------
c...  full system kinematics                                                   
c...  -----------------------------------------                                

c...  total transverse energy of the bbbb system                               
      hT=ptjet(1)+ptjet(2)+ptjet(3)+ptjet(4)
      call histo1(7,40,0d0,500d0,hT,w)

c...  invariant mass of the bbbb system
      p12(1:4)=ptable(1,1:4)+ptable(2,1:4)+
     .         ptable(3,1:4)+ptable(4,1:4)
      w12=dsqrt(scalar_product(p12,p12))
      call histo1(8,40,0d0,800d0,w12,w)

c...  -----------------------------------------
c...  single top kinematics                                                   
c...  -----------------------------------------    

c...  transverse momentum distribution of the b-jet
      call histo1( 9,40,0d0,250d0,ptjet(ksort(1)),w)
      call histo1(10,40,0d0,200d0,ptjet(ksort(2)),w)
      call histo1(11,40,0d0,150d0,ptjet(ksort(3)),w)
      call histo1(12,40,0d0,100d0,ptjet(ksort(4)),w)

      call histo1(13,40,0d0,150d0,ptjet(ksort(1)),w*0.25d0)
      call histo1(13,40,0d0,150d0,ptjet(ksort(2)),w*0.25d0)
      call histo1(13,40,0d0,150d0,ptjet(ksort(3)),w*0.25d0)
      call histo1(13,40,0d0,150d0,ptjet(ksort(4)),w*0.25d0)

c...  rapidity distribution of the b-jet
      p12(1:4)=ptable(ksort(1),1:4)
      y1=rapidity_jf(p12)
      call histo1(14,40,-2.5d0,2.5d0,y1,w)
      call histo1(18,40,-2.5d0,2.5d0,y1,w*0.25d0)
      p12(1:4)=ptable(ksort(2),1:4)
      y1=rapidity_jf(p12)
      call histo1(15,40,-2.5d0,2.5d0,y1,w)
      call histo1(18,40,-2.5d0,2.5d0,y1,w*0.25d0)
      p12(1:4)=ptable(ksort(3),1:4)
      y1=rapidity_jf(p12)
      call histo1(16,40,-2.5d0,2.5d0,y1,w)
      call histo1(18,40,-2.5d0,2.5d0,y1,w*0.25d0)
      p12(1:4)=ptable(ksort(4),1:4)
      y1=rapidity_jf(p12)
      call histo1(17,40,-2.5d0,2.5d0,y1,w)
      call histo1(18,40,-2.5d0,2.5d0,y1,w*0.25d0)
c...  -----------------------------------------  

      return
      end
