      subroutine FJmpo(n,ifl,p,icut)

c-----------------------------------------------------------------------
c...  jet function with m+1 partons (for real radiation)
c...  
c...  n=m+1
c...  icut=1 - the point is accepted
c...  icut=0 - the point is rejected
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_dipoles.h'
      include 'common_kt.h'

c...  clustered momenta and coresponding flavors    
      common/clusters/pcl(0:mxpart,4),iflcl(0:mxpart)
c...  jet momenta and coresponding flavors         
      common/jets/pjet(mxpart,4),ifljet(mxpart),njet

      dimension ifl(mxpart),p(mxpart,5)
      dimension plab(mxpart,4)

      dimension pjetnew(mxpart,4),ifljetnew(mxpart)
      dimension inun(20),inun0(20)

c...  uncomment to remove the jet function
c      icut=1
c      return

c...  if only real contribution then use normal cuts
      if (onlyreal) then
         call setcuts(n,ifl)
         call checkcuts(n,ifl,p,icut)
         return
      endif

c... uncomment for naive jet function  
c      call setcuts(n-1,ifl)
c      call checkcuts(n,p,icut)
c      return

c... kT algorithm based jet function  

      call fromcmstolab(n,p,plab)

      ncl=0
      njet=2
      ifljet(1:2)=ifl(1:2)
      pjet(1:2,1:4)=p(1:2,1:4)

      do i=3,n
         if ((abs(ifl(i)).eq.3.or.abs(ifl(i)).eq.4.or.
     .        abs(ifl(i)).eq.7.or.abs(ifl(i)).eq.8.or.
     .        abs(ifl(i)).eq.12.or.abs(ifl(i)).eq.35).and.
     .        abs(rapidity(plab(i,1:3))).lt.etamax) then
            ncl=ncl+1
            iflcl(ncl)=ifl(i)
            pcl(ncl,1:4)=p(i,1:4)
         else
            njet=njet+1
            ifljet(njet)=ifl(i)
            pjet(njet,1:4)=p(i,1:4)
         endif
      enddo

      if(ncl.eq.0) goto 2

 1    continue

      i0=0
      j0=0
      dmin=1d13

      do i1=0,ncl-1
         do j1=i1+1,ncl
            d=dmeasure(i1,j1,dR)
            if (d.lt.dmin) then
               i0=i1
               j0=j1
               dmin=d
            endif
         enddo
      enddo

      if (i0.gt.0.and.j0.gt.0) then
         call combineclusters(ncl,i0,j0)
      else
         njet=njet+1
         ifljet(njet)=iflcl(j0)
         pjet(njet,1:4)=pcl(j0,1:4)
         iflcl(j0:ncl-1)=iflcl(j0+1:ncl)
         pcl(j0:ncl-1,1:4)=pcl(j0+1:ncl,1:4)
      endif
      ncl=ncl-1

 2    continue

      if (ncl.gt.0) goto 1

c...  general NLO condition
      if (njet.lt.n-1) then
         icut=0
         return
      endif

c...  keep ordering if nothing done
      if (njet.eq.n) then
         ifljet(1:njet)=ifl(1:n)
         pjet(1:njet,1:4)=p(1:n,1:4)
      endif

c...  check number of requested b-jets
      ibjet=0
      do i=3,njet
         if (iabs(ifljet(i)).eq.12) ibjet=ibjet+1
      enddo

      if (ibjet.ne.nbjets) then
         icut=0
         return
      endif

      call setcuts(n-1,ifljet)

      if(nbjets.gt.0) then
         call checkcuts(njet,ifljet,pjet,icut)
         if(njet.eq.n.and.icut.eq.1.and.jetveto)then
            ptjet=transverse(pjet(njet,1:4))
            if (ptjet.gt.ptveto) icut=0
         endif
         return
      endif
         
      if(njet.eq.n-1)then
       call checkcuts(njet,ifljet,pjet,icut)
       return
      endif

      if(njet.eq.n)then
         icut=0
         nun=0
         pjetnew(1:njet,1:4)=pjet(1:njet,1:4)
         ifljetnew(1:njet)=ifljet(1:njet)

         do i=3,n 
            if (abs(ifl(i)).eq.3.or.abs(ifl(i)).eq.4.or.
     .          abs(ifl(i)).eq.7.or.abs(ifl(i)).eq.8.or.
     .          abs(ifl(i)).eq.35)then
               nun=nun+1
               inun(nun)=i
            endif
         enddo

c...  one additional jet from real emission only
c...  i.e. QCD corrections to electroweak processes  
         if(nun.eq.1)then
            call checkcuts(njet,ifljet,pjet,icut)
c...  jet veto only for events which have been accepted                   
            if (icut.eq.1.and.jetveto) then
                ptjet=transverse(pjet(inun(1),1:4))
                if (ptjet.gt.ptveto) icut=0
            endif
            return
         endif

         inun0(1:nun)=inun(1:nun)
         do k=1,nun
            inunnew=inun(1) 
            inun(1:nun-1)=inun(2:nun)
            inun(nun)=inunnew
            do i=1,nun
               pjetnew(inun0(i),1:4)=pjet(inun(i),1:4)
               ifljetnew(inun0(i))=ifljet(inun(i))
            enddo
            call setcuts(njet-1,ifljetnew)
            call checkcuts(njet,ifljetnew,pjetnew,icut1)
            icut=max(icut,icut1)
         enddo

c...  jet veto only for events which have been accepted            
         if (icut.eq.1.and.jetveto) then
c...  boost from cms to the lab frame for pT not needed
            ptmin=-1
            do i=n-nun+1,n
               ptjet=transverse(pjet(i,1:4))
               if (ptmin.lt.0.or.ptjet.lt.ptmin) ptmin=ptjet
            enddo
            if (ptmin.gt.ptveto) icut=0
         endif
      endif
      
      return
      end

c-----------------------------------------------------------------------

      subroutine FJm(n,ifl,p,icut)

c-----------------------------------------------------------------------
c...  jet function with m partons (for dipoles)
c...  
c...  n=m
c...  icut=1 - the point is accepted
c...  icut=0 - the point is rejected
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_kt.h'

      dimension ifl(mxpart),p(mxpart,5)

c...  uncomment to remove the jet function
c      icut=1
c      return

c...  uncomment to remove all dipoles
c      icut=0
c      return

c...  check number of requested b-jets
      ibjet=0
      do i=3,n
         if (iabs(ifl(i)).eq.12) ibjet=ibjet+1
      enddo

      if (ibjet.ne.nbjets) then
         icut=0
         return
      endif

      call setcuts(n,ifl)
      call checkcuts(n,ifl,p,icut)

      return
      end

c-----------------------------------------------------------------------

      subroutine checkpoint(icut)

c-----------------------------------------------------------------------
c...  check whether the phase space point can give any contribution
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_masses.h'
      include 'common_phegas_mom.h'
      include 'common_dipoles.h'
      
      dimension q(mxpart,5)

      dimension znlc(4)

      double precision mijsq,misq,mjsq,mksq

c...  is there a real emission contribution?
      call FJmpo(n,ifl,pmom,icutreal)
      
c... if only real contribtion then use normal cuts
      if (onlyreal) then
         icut=icutreal
         return
      endif

c...  are there dipole contributions?
      icutdip=0
      alphaMin=1d0
      do ipp=1,ndip
         ip=ijkdip(ipp,1)
         jp=ijkdip(ipp,2)
         kp=ijkdip(ipp,3)

         pij=2*dot(pmom,ip,jp)
         pik=2*dot(pmom,ip,kp)
         pjk=2*dot(pmom,jp,kp)

         misq=parmas(ifl(ip))**2
         mjsq=parmas(ifl(jp))**2
         mksq=parmas(ifl(kp))**2
         mijsq=parmas(ifldip(ipp,ip))**2

         call variable(pmom,ip,jp,kp,misq,mjsq,mijsq,x)

         if(dipoletype.eq.0)then !... Catani-Seymour
            call ptilde(n-2,io,pmom,x,ip,jp,kp,mksq,mijsq,q)
         endif
         if(dipoletype.eq.1)then !... Nagy-Soper
            call ptildeNS(n-2,io,pmom,x,ip,jp,kp,mksq,mijsq,znlc,q)
         endif

         call FJm(n-1,ifldip(ipp,1:n-1),q,icut)
         if (icut.eq.0) cycle


         if(dipoletype.eq.0) then !... Catani-Seymour phase-space restriction

            if ((ip .le. 2) .and. (kp .le. 2)) then      
               vtildei=pij/pik
               if (vtildei.lt.alphaMin) alphaMin=vtildei
               if (vtildei.gt.alphaMaxII) cycle
            elseif ((ip .le. 2) .and. (kp .gt. 2)) then      
               ui=pij/(pij+pik)
               if (ui.lt.alphaMin) alphaMin=ui
               if (ui.gt.alphaMaxIF) cycle
            elseif ((ip .gt. 2) .and. (kp .le. 2)) then
               if ((1-x).lt.alphaMin) alphaMin=(1-x)
               if ((1-x).gt.alphaMaxFI) cycle
            elseif ((ip .gt. 2) .and. (kp .gt. 2)) then      
               if (x.lt.alphaMin) alphaMin=x
               if (x.gt.alphaMaxFF) cycle
            endif

         elseif(dipoletype.eq.1) then !... Nagy-Soper phase-space restriction

c...  In the present version, no alphaMax cutoff is implemented for the Nagy-Soper
c...  subtraction. Only alphaMin cutoff is considered.
            if ((ip .le. 2) .and. (kp .le. 2)) then      
               vtildei=pij/pik
               if (vtildei.lt.alphaMin) alphaMin=vtildei
            elseif ((ip .le. 2) .and. (kp .gt. 2)) then      
               ui=pij/(pij+pik)
               if (ui.lt.alphaMin) alphaMin=ui
            elseif ((ip .gt. 2) .and. (kp .le. 2)) then
               if ((1-x).lt.alphaMin) alphaMin=(1-x)
            elseif ((ip .gt. 2) .and. (kp .gt. 2)) then      
               if (x.lt.alphaMin) alphaMin=x
            endif

         endif

         icutdip=1
      enddo
      
      if (alphaMin.lt.alphaMinCut) then
         icut=0
         return
      endif

      if (icutreal.eq.1.or.icutdip.eq.1) then
         icut=1
      else
         icut=0
      endif

      return
      end

c-----------------------------------------------------------------------

      subroutine setcuts(n,ifl)

c-----------------------------------------------------------------------
c...  set cut tables in parton numbering using flavor information
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_masses.h'
      include 'dimensions.h'

c...  cut parameters
      common/userc_jf/ec(3:20),c1(3:20),c2(3:20),cc(3:20,3:20),
     .      gmas(3:20,3:20)
      common/usercc_jf/ptc(3:20),etac(3:20),drc(3:20,3:20)

      dimension ifl(mxpart)

c...  user predefined cuts by flavor
      include 'cuts.h'

      ptc(3:20)=0
      drc(3:20,3:20)=0
      etac(3:20)=100
      ec(3:20)=0
      c1(3:20)=1
      c2(3:20)=1
      cc(3:20,3:20)=1
      gmas(3:20,3:20)=0

      do i=3,n
         if(ifl(i).lt.13)then
            i1=iabs(mod(ifl(i),4))
         else
            i1=ifl(i)
         endif
         if(i1.eq.35)i1=0
c     - bottom quarks
         if(iabs(ifl(i)).eq.12)i1=12
c     - top quarks
         if(iabs(ifl(i)).eq.11)i1=11
         goto 111
c     -------
c     special for mc4lhc 2003
c     [
c     if(ifl(i).eq.11)i1=-1
c     if(ifl(i).eq.-11)i1=-1
c     if(ifl(i).eq.12)i1=-1
c     if(ifl(i).eq.-12)i1=-1
c     ]

 111     continue
c     photons
         if(i1.eq.31)then
            ptc(i)=ptg
            etac(i)=etag
         endif
c     neutrinos
         if(i1.eq.1)then
         endif
c     ch. leptons
         if(i1.eq.2)then
c     if(i1.le.2)then
            ptc(i)=ptl
            etac(i)=etal
         endif
c     quarks
         if(i1.eq.3.or.i1.eq.0)then
            ptc(i)=ptq
            etac(i)=etaq
         endif
c     bottom 
         if(i1.eq.12)then
            ptc(i)=ptbottom
            etac(i)=etabottom
         endif
c     top
         if(i1.eq.11)then
            ptc(i)=pttop
            etac(i)=etatop
         endif
      enddo
      
      do i=3,n-1
         if(ifl(i).lt.13)then
            i1=iabs(mod(ifl(i),4))
         else
            i1=ifl(i)
         endif
         if(i1.eq.35)i1=0
c     - bottom quarks                                           
         if(iabs(ifl(i)).eq.12)i1=12
c     - top quarks 
         if(iabs(ifl(i)).eq.11)i1=11
         goto 222
c     -------
c     special for mc4lhc 2003
c     [
c     if(ifl(i).eq.11)i1=-1
c     if(ifl(i).eq.-11)i1=-1
c     if(ifl(i).eq.12)i1=-1
c     if(ifl(i).eq.-12)i1=-1
c     ]

 222     continue
         do j=i+1,n
            if(ifl(j).lt.13)then
               j1=iabs(mod(ifl(j),4))
            else
               j1=ifl(j)
            endif
            if(j1.eq.35)j1=0
c     - bottom quarks 
            if(iabs(ifl(j)).eq.12)j1=12
c     - top quarks
            if(iabs(ifl(j)).eq.11)j1=11
            goto 333

c     -------
c     special for mc4lhc 2003
c     [
c     if(ifl(j).eq.11)j1=-1
c     if(ifl(j).eq.-11)j1=-1
c     if(ifl(j).eq.12)j1=-1
c     if(ifl(j).eq.-12)j1=-1
c     ]
 333        continue

c     n-n
            if(i1.eq.1.or.j1.eq.1)then
c     l-q
            elseif(i1.eq.2.and.(j1.eq.3.or.j1.eq.0.or.j1.eq.12))then
               drc(i,j)=drlq
c     l-l
            elseif(i1.eq.2.and.j1.eq.2)then
               drc(i,j)=drll
c     q-l
            elseif(j1.eq.2.and.(i1.eq.3.or.i1.eq.0.or.i1.eq.12))then
               drc(i,j)=drlq
c     g-x  photon-charged
            elseif(i1.eq.31.and.j1.ne.1)then
               drc(i,j)=drgx
c     x-g  charged-photon
            elseif(j1.eq.31.and.i1.ne.1)then
               drc(i,j)=drgx
c     q-q
            elseif((i1.eq.3.or.i1.eq.0).and.
     .              (j1.eq.3.or.j1.eq.0))then
               drc(i,j)=drqq
               gmas(i,j)=max(gqq,gmas(i,j))
c     q-b
            elseif((i1.eq.3.or.i1.eq.0).and.
     .              (j1.eq.12))then
               drc(i,j)=drqb
               gmas(i,j)=max(gqb,gmas(i,j))
c     b-q
            elseif((j1.eq.3.or.j1.eq.0).and.
     .              (i1.eq.12))then
               drc(i,j)=drqb
               gmas(i,j)=max(gqb,gmas(i,j))
c     b-b
            elseif((i1.eq.12).and.
     .              (j1.eq.12))then
               drc(i,j)=drbb
               gmas(i,j)=max(gbb,gmas(i,j))
c     
            endif
         enddo
      enddo
      
      do i=3,n
c     ec(i)=max(ptc(i),parmas(ifl(i)))
         if(ptc(i).gt.0)then
            ec(i)=max(ec(i),ptc(i))
         else
            ec(i)=max(ec(i),parmas(ifl(i)))
         endif
         r=exp(2*etac(i))
         c1(i)=(r-1)/(r+1)
         c2(i)=c1(i)
      enddo 
      
c      do i=3,n-1
c         do j=i+1,n
c     ri=sqrt(1-c1(i)*c1(i))
c     rj=sqrt(1-c1(j)*c1(j))
c     cc(i,j)=c1(i)*c1(j)+ri*rj*cos(drc(i,j))
c            cc(i,j)=cos(drc(i,j))
c         enddo
c      enddo

      do i=3,n-1
         do j=i+1,n
            gmas(i,j)=max(gmas(i,j),
     &           dsqrt( 2*ptc(i)*ptc(j)*(1-cos(drc(i,j))) ) )
         enddo
      enddo

      do l=4,20
         do k=3,l-1
            drc(l,k)= drc(k,l)
            cc(l,k)=  cc(k,l)
            gmas(l,k)=  gmas(k,l)
         enddo
      enddo

      return
      end

c-----------------------------------------------------------------------

      subroutine checkcuts(n,ifl,p,icut)

c-----------------------------------------------------------------------
c...  check for cuts defined through the common block
c...  
c...  icut=1 - the point is accepted
c...  icut=0 - the point is rejected
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

c...  cut parameters
      common/userc_jf/ec(3:20),c1(3:20),c2(3:20),cc(3:20,3:20),
     .      gmas(3:20,3:20)
      common/usercc_jf/ptc(3:20),etac(3:20),drc(3:20,3:20)

c...  pi
      parameter(pi=3.141592653589793d0)

      dimension p(mxpart,5)
      dimension plab(mxpart,4),p1(4)

      dimension ifl(mxpart)

      logical VBF
      data VBF/.false./   
      dimension ptjet(mxpart), ptable(mxpart,4), ksort(mxpart)
      dimension etalep(mxpart)

      
      icut=0
      call fromcmstolab(n,p,plab)

      do l=3,n 
         pt=sqrt(plab(l,1)**2+plab(l,2)**2)
         if(pt.lt.ptc(l)) then
            goto 1000
         endif
         eta=rapidity_jf(plab(l,1:4))
         if(abs(eta).gt.etac(l)) then 
            goto 1000 
         endif
      enddo

      do l1=3,n
         do l2=3,n
            if(l1.eq.l2)cycle
            d1=rapidity_jf(plab(l1,1:4))-rapidity_jf(plab(l2,1:4))
            d2=ph4(plab(l1,1),plab(l1,2),plab(l1,3))
     .           -ph4(plab(l2,1),plab(l2,2),plab(l2,3))
            d2=min(dabs(d2),2*pi-dabs(d2))
            if(d2/pi.gt.1.d0)print*,d2/pi
            dr=sqrt(d1**2+d2**2)
            if(dr.lt.drc(l1,l2))  then 
               goto 1000
            endif
         enddo
      enddo

c...  ptmiss
c      p1(1:4)=plab(6,1:4)+plab(4,1:4)
c      pTmiss=transverse(p1)
c      if(pTmiss.lt.25d0) goto 1000

c...  invariant mass of two b-jets
c       p1(1:4)=plab(5,1:4)+plab(6,1:4)
c       w12=dsqrt(scalar_product(p1,p1))
c       if(w12.lt.100d0) goto 1000

      if(VBF)then
c...  VBF specific cuts (very exclusive)
c...  finding tagging jets (light quarks, gluons and b-quarks)
c...  ordering according to pT

         ijet=0
         ilep=0
         ptable(1:mxpart,1:4)=0d0
         ptjet(1:mxpart)=0d0
         etalep(1:mxpart)=0d0

         do l1=3,n
c...  only partons
            if (iabs(ifl(l1)).ne.3.and.iabs(ifl(l1)).ne.4.and.
     .          iabs(ifl(l1)).ne.7.and.iabs(ifl(l1)).ne.8.and.
     .          iabs(ifl(l1)).ne.11.and.iabs(ifl(l1)).ne.12.and.
     .               ifl(l1) .ne.35) cycle
            ijet=ijet+1
            ptable(ijet,1:4)=plab(l1,1:4)
            ptjet(ijet)=transverse(plab(l1,1:4))
         enddo
C$$$         print*,'-- NEW ---',ijet
c...  ordering jets according to pT
         call SOR(ptjet,ijet,ksort,2)
         
C$$$         print*,(ptjet(i1),i1=1,ijet)
C$$$         print*,(ptjet(ksort(i1)),i1=1,ijet)
C$$$         print*,'-- END ---',(ksort(i1),i1=1,ijet)
C$$$         print*,''

c...  jet-jet rapidity separation
         etaj1=rapidity_jf(ptable(ksort(1),1:4))
         etaj2=rapidity_jf(ptable(ksort(2),1:4))
         if(abs(etaj1-etaj2).lt.4d0) goto 1000
         
c...  opposite sign jets rapidity
         if((etaj1*etaj2).gt.0d0) goto 1000
         
c...  finding leptons (e, mu, tau)
         do l1=3,n
c...  only charged leptons
            if (iabs(ifl(l1)).ne.2.and.iabs(ifl(l1)).ne.6.and.
     .          iabs(ifl(l1)) .ne.10) cycle
            ilep=ilep+1
            etalep(ilep)=rapidity_jf(plab(l1,1:4))
         enddo

         ylepmax = max(etaj1,etaj2)
         ylepmin = min(etaj1,etaj2)

c...  leptons fall inside rapidity gap         
         do l1=1,ilep
            if(etalep(l1).gt.ylepmax.or.etalep(l1).lt.ylepmin) then
               goto 1000
            endif
         enddo

c...  tagging jets invariant dijet mass
         p1(1:4)=ptable(ksort(1),1:4)+ ptable(ksort(2),1:4)
         wjj=dsqrt(scalar_product(p1,p1))
         if(wjj.lt.600d0) goto 1000

      endif

      icut=1
 1000 continue

      return
      end

c-----------------------------------------------------------------------

      subroutine readcuts

c-----------------------------------------------------------------------
c...  for phase space
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_int.h'
      include 'common_cuts.h'
      include 'common_masses.h'

c...  needed for phase space generator sanity
      parameter(cutoff=1d-3)
      parameter(cutoff1=1d-6)

c...  user predefined cuts by flavor
      include 'cuts.h'

      escale=0d0

c...  sum over pTs of the partons
      do i=3,n
         if(ifl(i).lt.13)then
            i1=iabs(mod(ifl(i),4))
         else
            i1=ifl(i)
         endif
         if(i1.eq.35)i1=0
c     - bottom quarks
         if(iabs(ifl(i)).eq.12)i1=12
c     - top quarks
         if(iabs(ifl(i)).eq.11)i1=11
c     photons
         if(i1.eq.31)then
            escale=escale+ptg
         endif
c     neutrinos
         if(i1.eq.1)then
         endif
c     ch. leptons
         if(i1.eq.2)then
c     if(i1.le.2)then
            escale=escale+ptl
         endif
c     quarks
         if(i1.eq.3.or.i1.eq.0)then
            escale=escale+ptq
         endif
c     bottom 
         if(i1.eq.12)then
            escale=escale+ptbottom
         endif
c     top
         if(i1.eq.11)then
            escale=escale+pttop
         endif
      enddo

c...  sum over masses of all final states
      do i=3,n
         escale=escale+parmas(ifl(i))
      enddo

      if (escale.eq.0d0) escale=1d0

      print*,'Energy scale set to:',escale

      ptc(3:20)=cutoff*escale
      drc(3:20,3:20)=0
      etac(3:20)=100
      ec(3:20)=cutoff*escale
      c1(3:20)=1-cutoff1
      c2(3:20)=1-cutoff1
      cc(3:20,3:20)=1 !-cutoff
      gmas(3:20,3:20)=cutoff*escale
      
      do i=3,n
         if(parmas(ifl(i)).gt.0)then
        c1(i)=1
        c2(i)=1
        endif
      enddo

      return
      end

c-----------------------------------------------------------------------

      function dmeasure(i,j,dmin)

c-----------------------------------------------------------------------
c...  kT algorithm measure  
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'
      include 'common_kt.h'

      common/clusters/pcl(0:mxpart,4),iflcl(0:mxpart)

c...  pi   
      parameter(pi=3.141592653589793d0)

c...  initial-final 
      if (i.le.0.and.j.gt.0) then
         pt2=pcl(j,1)**2+pcl(j,2)**2
         dmeasure=pt2**(ialgorithm)
c...  final-initial    
      elseif (i.gt.0.and.j.le.0) then
         pt1=pcl(i,1)**2+pcl(i,2)**2
         dmeasure=pt1**(ialgorithm)
c...  final-final   
      elseif (i.gt.0.and.j.gt.0) then
         pt1=pcl(i,1)**2+pcl(i,2)**2
         pt2=pcl(j,1)**2+pcl(j,2)**2

         y1=0.5d0*dlog((pcl(i,4)+pcl(i,3))/(pcl(i,4)-pcl(i,3)))
         y2=0.5d0*dlog((pcl(j,4)+pcl(j,3))/(pcl(j,4)-pcl(j,3)))
         dphi=ph4(pcl(i,1),pcl(i,2),pcl(i,3))
     .        -ph4(pcl(j,1),pcl(j,2),pcl(j,3))
         dphi=min(dabs(dphi),2*pi-dabs(dphi))

         dmeasure=min(pt1**(ialgorithm),pt2**(ialgorithm))
     .              *((y1-y2)**2+dphi**2)/dmin**2
      endif

      return
      end

c-----------------------------------------------------------------------

      subroutine combineclusters(n,i,j)

c-----------------------------------------------------------------------
c...  combine two clusters into one jet       
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      common/clusters/pcl(0:mxpart,4),iflcl(0:mxpart)

      i1=min(i,j)
      j1=max(i,j)

      pcl(i1,1:4)=pcl(i1,1:4)+pcl(j1,1:4)
      pcl(j1:n-1,1:4)=pcl(j1+1:n,1:4)

      if1=iflcl(i1)
      if2=iflcl(j1)

      if (if1+if2.eq.0) then
         if0=35
      elseif (iabs(if1).ne.12) then
         if0=if2
      elseif (iabs(if2).ne.12) then
         if0=if1
      else
         if0=if1
      endif

      iflcl(i1)=if0
      iflcl(j1:n-1)=iflcl(j1+1:n)

      return
      end
