      function nextpol(n,ifl,ipol,rpol)

c-----------------------------------------------------------------------
c...  next polarization
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension ifl(mxpart),ipol(n),rpol(n)

c...  array of numbers of helicity states for each flavour
      dimension nhel(-12:44)
      data nhel/2,2,2,2,2,2,2,2,2,2,2,2,1, ! -12,...,0
     .          2,2,2,2,2,2,2,2,2,2,2,2,   !  1,...,12
     .          1,1,1,1,1,1,1,1,1,1,1,1,   ! 13,...,24
     .          1,1,1,1,1,1,2,3,3,3,2,1,   ! 25,...,36
     .          1,1,1,1,1,1,1,1/           ! 37,...,44

      do k=n,1,-1
         if (rpol(k).lt.0d0.and.ipol(k).lt.nhel(ifl(k))) exit
      enddo
      
      if (k.eq.0) then
         nextpol=0
         return
      endif

      ipol(k)=ipol(k)+1
      ipol(k+1:n)=1
      
      nextpol=1
      end

c-----------------------------------------------------------------------

      subroutine getstring(n,ifl,process)

c-----------------------------------------------------------------------
c...  finding string for the process
c...  n - number of particles
c...  ifl - flavour array
c...  process - string
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'dimensions.h'

      dimension ifl(mxpart)
      character*100 process
      character*2 char(20)

      do i=1,n
         if (ifl(i).eq. 2)char(i)='E-'
         if (ifl(i).eq.-2)char(i)='E+'
         if (ifl(i).eq. 1)char(i)='Ne'
         if (ifl(i).eq.-1)char(i)='N1'
         if (ifl(i).eq. 3)char(i)='Uq'
         if (ifl(i).eq.-3)char(i)='Ua'
         if (ifl(i).eq. 4)char(i)='Dq'
         if (ifl(i).eq.-4)char(i)='Da'
         if (ifl(i).eq. 6)char(i)='M-'
         if (ifl(i).eq.-6)char(i)='M+'
         if (ifl(i).eq. 5)char(i)='Nm'
         if (ifl(i).eq.-5)char(i)='N2'
         if (ifl(i).eq. 7)char(i)='Cq'
         if (ifl(i).eq.-7)char(i)='Ca'
         if (ifl(i).eq. 8)char(i)='Sq'
         if (ifl(i).eq.-8)char(i)='Sa'
         if (ifl(i).eq. 10)char(i)='T-'
         if (ifl(i).eq.-10)char(i)='T+'
         if (ifl(i).eq. 9)char(i)='Nt'
         if (ifl(i).eq.-9)char(i)='N3'
         if (ifl(i).eq. 11)char(i)='Tq'
         if (ifl(i).eq.-11)char(i)='Ta'
         if (ifl(i).eq. 12)char(i)='Bq'
         if (ifl(i).eq.-12)char(i)='Ba'
         if (ifl(i).eq.31)char(i)='A0'
         if (ifl(i).eq.32)char(i)='Z0'
         if (ifl(i).eq.33)char(i)='W+'
         if (ifl(i).eq.34)char(i)='W-'
         if (ifl(i).eq.35)char(i)='G0'
         if (ifl(i).eq.41)char(i)='H0'
      enddo

      write(process,*)(char(i),i=1,n)

      return
      end

c-----------------------------------------------------------------------

      subroutine produceinputs(n,ifl,iproc)

c-----------------------------------------------------------------------
c...  produce input files for HELAC to obtain Dyson-Schwinger 
c...  skeletons (phase one from HELAC)
c...  n - number of particles in the process                      
c...  ifl - flavour array for the process 
c...  iproc - number of the underlying processes
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_flags.h'
      include 'common_qcdrun.h'
      include 'dimensions.h'

      dimension ifl(mxpart)

      character*1 a1
      character*2 a2
      character*3 a3

      if (iproc.eq.0) then
         write(a1,'(i1)')iproc                  
         open(400+iproc,file='input'//a1)
      elseif (iproc.lt.10) then
         write(a1,'(i1)')iproc                  
         open(400+iproc,file='input'//a1)
      elseif (iproc.ge.10.and.iproc.le.99) then
         write(a2,'(i2)')iproc
         open(400+iproc,file='input'//a2)
      elseif (iproc.ge.100.and.iproc.le.999) then
         write(a3,'(i3)')iproc
         open(400+iproc,file='input'//a3)
      endif

c-----------------------------------------------------------------------
c...  creating input file to run HELAC 
c-----------------------------------------------------------------------

c...  hi_file,err_file -> histo and error files
      write(400+iproc,*)'hi_file  err_file'

c...  repeat,ranhel
      write(400+iproc,*)1,0

c...  number of particles
      write(400+iproc,*)n

c...  flavour of particles
      write(400+iproc,*)ifl(1:n) 

c...  iflag,iunitary,ihiggs,iwidth
      write(400+iproc,*)iflag,iunitary,ihiggs,iwidth

c...  onlyqcd,withqcd,irun
      write(400+iproc,*)onlyqcd,withqcd,0
      
c...  kaleu: ikaleu nbatch nstep nbatch_grid nbatch_cut
      write(400+iproc,*)1, 20000, 10, 1000, 1000

      close(400+iproc)
      
      return
      end

c-----------------------------------------------------------------------

      subroutine fromcmstolab(n,pin,pout)

c-----------------------------------------------------------------------
c...  fromcmstolab
c-----------------------------------------------------------------------

      include 'declare.h'
      include 'common_strf.h'
      include 'dimensions.h'

      dimension pboo(4)
      dimension pin(mxpart,4)
      dimension pout(mxpart,4)

      pout(3:n,1:4)=pin(3:n,1:4)

      if(istruc.eq.1)then
         q=ehat
         e=q/sqrt(xp1*xp2)
         pboo(4)=(xp1+xp2)*e/2
         pboo(3)=(xp1-xp2)*e/2
         pboo(1:2)=0
         do i=3,n
            call boostl(q,pboo,pout(i,1:4))
         enddo
      endif

      end

c-----------------------------------------------------------------------

      function rapidity_jf(p)

c-----------------------------------------------------------------------
c...  rapidity    
c-----------------------------------------------------------------------

      include 'declare.h'
      dimension p(4)

      rapidity_jf=0.5d0*dlog((p(4)+p(3))/(p(4)-p(3)))

      return
      end

