      SUBROUTINE RAMBO(N,ET,XM,P,WT,IDIR)
C------------------------------------------------------
C
C                       RAMBO
C
C    RA(NDOM)  M(OMENTA)  B(EAUTIFULLY)  O(RGANIZED)
C
C    A DEMOCRATIC MULTI-PARTICLE PHASE SPACE GENERATOR
C    AUTHORS:  S.D. ELLIS,  R. KLEISS,  W.J. STIRLING
C    THIS IS VERSION 1.0 -  WRITTEN BY R. KLEISS
C
C    N  = NUMBER OF PARTICLES (>1, IN THIS VERSION <101)
C    ET = TOTAL CENTRE-OF-MASS ENERGY
C    XM = PARTICLE MASSES ( DIM=100 )
C    P  = PARTICLE MOMENTA ( DIM=(4,100) )
C    WT = WEIGHT OF THE EVENT
C
C------------------------------------------------------
      include 'declare.h'
      DIMENSION XM(100),P(4,100),Q(4,100),RZ(100),R(4),
     .   B(3),P2(100),XM2(100),E(100),V(100),IWARN(5)
     .  ,RN(4)
      SAVE ACC,ITMAX,IBEGIN,TWOPI,PI2LOG,IWARN,RZ
      DATA ITMAX/10/,IBEGIN/0/,IWARN/5*0/
C------------------------------------------------------
C       VERSION DEFINITION
C------------------------------------------------------
      DATA IVERSION/0/
c     IF(IVERSION.EQ.0)PRINT*,'VERSION OF 13/12/94'
      IVERSION=1
C------------------------------------------------------
C         IBEGIN=0
C INITIALIZATION STEP: FACTORIALS FOR THE PHASE SPACE WEIGHT
      IF(IBEGIN.NE.0) GOTO 103
      acc=eps(1)
      IBEGIN=1
      TWOPI=dnou(8)*DATAN(dnou(1))
      PI2LOG=DLOG(TWOPI/dnou(4))
      RZ(2)=PI2LOG
      DO 101 K=3,100
  101 RZ(K)=RZ(K-1)+PI2LOG-dnou(2)*DLOG(DBLE(K-2))
      DO 102 K=3,100
  102 RZ(K)=(RZ(K)-DLOG(DBLE(K-1)))
C
C CHECK ON THE NUMBER OF PARTICLES
  103 IF(N.GT.1.AND.N.LT.101) GOTO 104
      PRINT 1001,N
      STOP
C
C CHECK WHETHER TOTAL ENERGY IS SUFFICIENT; COUNT NONZERO MASSES
  104 XMT=dnou(0)
      NM=0
      DO 105 I=1,N
       IF((XM(I).GT.dnou(0)).OR.(XM(I).LT.dnou(0))) NM=NM+1
  105 XMT=XMT+DABS(XM(I))
      IF(XMT.LE.ET) GOTO 201
      PRINT 1002,XMT,ET
      STOP
C
C THE PARAMETER VALUES ARE NOW ACCEPTED
C
C GENERATE N MASSLESS MOMENTA IN INFINITE PHASE SPACE
 201  CONTINUE
C
      IF(IDIR.EQ.1) GOTO 401 !idir
C
      DO 202 I=1,N
        DO J=1,4
        RN(J)=RNMY(0)
        ENDDO
      C=dnou(2)*RN(1)-dnou(1)
      S=DSQRT(dnou(1)-C*C)
      F=TWOPI*RN(2)
      Q(4,I)=-DLOG(RN(3)*RN(4))
      Q(3,I)=Q(4,I)*C
      Q(2,I)=Q(4,I)*S*DSIN(F)
  202 Q(1,I)=Q(4,I)*S*DCOS(F)
C
C CALCULATE THE PARAMETERS OF THE CONFORMAL TRANSFORMATION
      DO 203 I=1,4
  203 R(I)=dnou(0)
      DO 204 I=1,N
      DO 204 K=1,4
  204 R(K)=R(K)+Q(K,I)
      RMAS=DSQRT(R(4)**2-R(3)**2-R(2)**2-R(1)**2)
      DO 205 K=1,3
  205 B(K)=-R(K)/RMAS
      G=R(4)/RMAS
      A=dnou(1)/(dnou(1)+G)
      X=ET/RMAS
C
C TRANSFORM THE Q'S CONFORMALLY INTO THE P'S
      DO 207 I=1,N
      BQ=B(1)*Q(1,I)+B(2)*Q(2,I)+B(3)*Q(3,I)
      DO 206 K=1,3
  206 P(K,I)=X*(Q(K,I)+B(K)*(Q(4,I)+A*BQ))
  207 P(4,I)=X*(G*Q(4,I)+BQ)
C
  401 CONTINUE !idir
C
C CALCULATE WEIGHT AND POSSIBLE WARNINGS
      WT=PI2LOG
      IF(N.NE.2) WT=(dnou(2)*N-dnou(4))*DLOG(ET)+RZ(N)
      IF(WT.GE.-dnou(18)) GOTO 208
      IF(IWARN(1).LE.5) PRINT 1004,WT
      IWARN(1)=IWARN(1)+1
  208 IF(WT.LE. dnou(174)) GOTO 209
      IF(IWARN(2).LE.5) PRINT 1005,WT
      IWARN(2)=IWARN(2)+1
C
C RETURN FOR WEIGHTED MASSLESS MOMENTA
  209 IF(NM.NE.0) GOTO 210
      WT=DEXP(WT)
      RETURN
C
C MASSIVE PARTICLES: RESCALE THE MOMENTA BY A FACTOR X
  210 XMAX=DSQRT(dnou(1)-(XMT/ET)**2)
C
      IF(IDIR.EQ.1) GOTO 402  !idir
C
      DO 301 I=1,N
      XM2(I)=XM(I)**2
  301 P2(I)=P(4,I)**2
      ITER=0
      X=XMAX
      ACCU=ET*ACC
  302 F0=-ET
      G0=dnou(0)
      X2=X*X
      DO 303 I=1,N
      E(I)=DSQRT(XM2(I)+X2*P2(I))
      F0=F0+E(I)
  303 G0=G0+P2(I)/E(I)
      IF(DABS(F0).LE.ACCU) GOTO 305
      ITER=ITER+1
      IF(ITER.LE.ITMAX) GOTO 304
      PRINT 1006,ITMAX
      GOTO 305
  304 X=X-F0/(X*G0)
      GOTO 302
  305 DO 307 I=1,N
      V(I)=X*P(4,I)
      DO 306 K=1,3
  306 P(K,I)=X*P(K,I)
  307 P(4,I)=E(I)
c     print*,'----------------'
c     print*,X,V(1:N),E(1:N)
C
  402 CONTINUE  ! idir
C
      DO I=1,N
       V(I)=SQRT(P(1,I)**2+P(2,I)**2+P(3,I)**2)
       E(I)=P(4,I)
      ENDDO
      X=0
      DO I=1,N
       X=X+V(I)/ET
      ENDDO
c     print*,X,V(1:N),E(1:N)
C
C CALCULATE THE MASS-EFFECT WEIGHT FACTOR
      WT2=dnou(1)
      WT3=dnou(0)
      DO 308 I=1,N
      WT2=WT2*V(I)/E(I)
  308 WT3=WT3+V(I)**2/E(I)
      WTM=(dnou(2)*N-dnou(3))*DLOG(X)+DLOG(WT2/WT3*ET)
C
C RETURN FOR  WEIGHTED MASSIVE MOMENTA
      WT=WT+WTM
      IF(WT.GE.-dnou(18)) GOTO 309
      IF(IWARN(3).LE.5) PRINT 1004,WT,EXP(WT)
      IWARN(3)=IWARN(3)+1
  309 IF(WT.LE. dnou(174)) GOTO 310
      IF(IWARN(4).LE.5) PRINT 1005,WT,EXP(WT)
      IWARN(4)=IWARN(4)+1
  310 WT=DEXP(WT)
c     print*,WT,X
      RETURN
C
 1001 FORMAT(' RAMBO FAILS: # OF PARTICLES =',I5,' IS NOT ALLOWED')
 1002 FORMAT(' RAMBO FAILS: TOTAL MASS =',D15.6,' IS NOT',
     . ' SMALLER THAN TOTAL ENERGY =',D15.6)
 1004 FORMAT('RAMBO WARNS: WEIGHT = EXP(',F20.9,') MAY UNDERFLOW',D20.9)
 1005 FORMAT('RAMBO WARNS: WEIGHT = EXP(',F20.9,') MAY  OVERFLOW',D20.9)
 1006 FORMAT('RAMBO WARNS:',I3,' ITERATIONS DID NOT GIVE THE',
     . ' DESIRED ACCURACY =',D15.6)
      END
C
